/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * EIO.h
 *
 * PROGRAMMER : Gregg A. Tavares 
 *    VERSION : 00.000
 *    CREATED : 01/29/90 
 *   MODIFIED : 10/02/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w-----
 *	 E C H I D N A
 *	--------------
 *
 * DESCRIPTION
 *		Types and Prototype for EIO (Echidna I/O) routines.
 *
 * HISTORY
 *
*/

#ifndef ECHIDNA_EIO_H
#define ECHIDNA_EIO_H

#include <echidna\platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <echidna\etypes.h>
#include <fcntl.h>

#if __TURBOC__
#include <dos.h>
#include <dir.h>
#include <io.h>
#include <sys\stat.h>
#endif

#if (__ZTC__ && __MSDOS__)
#include <dos.h>
#include <io.h>
#include <sys\stat.h>
#endif

#if __WATCOMC__
#include <dos.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <fcntl.h>
#endif

#if __AMIGAOS__
#include <libraries\dos.h>
#include <libraries\arpbase.h>
#endif

#if __MACSC__
#include <files.h>
#endif

#if __EGGS_SGI__
#include <sys\types.h>
#include <sys\dir.h>
#include <sys\stat.h>
#include <fcntl.h>
#include <utime.h>
#include <unistd.h>
#endif

#include <echidna\listfunc.h>

/**************************** C O N S T A N T S ***************************/

#if __MSDOS__

#define EIO_MAXPATH		80
#define EIO_MAXDIR		66
#define EIO_MAXFILE		20
#define EIO_MAXEXT		20

#define	DIRCURRENT		".\\"
#define	DIRSEPS			":/\\"
#define	DIRMATCHALL		"*.*"
#define DIRSEP			'\\'
#define OPENEXTRA		,S_IREAD|S_IWRITE

#elif __AMIGAOS__

#define EIO_MAXPATH		(LONG_FSIZE * 2 + 2)
#define EIO_MAXDIR		EIO_MAXPATH
#define EIO_MAXFILE		EIO_MAXPATH
#define EIO_MAXEXT		32

#define	DIRCURRENT		""
#define	DIRSEPS			":/\\"
#define DIRMATCHALL		"#?"
#define DIRSEP			'/'
#define O_BINARY		0
#define OPENEXTRA	

#elif __MACOS__

#define EIO_MAXPATH		256
#define EIO_MAXDIR		EIO_MAXPATH
#define EIO_MAXFILE		40
#define EIO_MAXEXT		32

#define	DIRCURRENT		":"
#define	DIRSEPS			":/\\"
#define DIRMATCHALL		""
#define DIRSEP			':'
#define O_BINARY		0
#define OPENEXTRA	

#elif __EGGS_SGI__

#define EIO_MAXPATH		256
#define EIO_MAXDIR		EIO_MAXPATH
#define EIO_MAXFILE		256
#define EIO_MAXEXT		128

#define	DIRCURRENT		"./"
#define	DIRSEPS			":/\\"
#define DIRMATCHALL		"*"
#define DIRSEP			'/'
#define O_BINARY		0
#define OPENEXTRA	

#else
#error need some code
#endif

/*
 * Constants returned EIO_fnsplit()
 */
#define EIO_DIRECTORY	(1 << 0)
#define EIO_FILENAME	(1 << 1)
#define EIO_EXTENSION	(1 << 2)

/*
 * Constants returned EIO_FileType()
 */
#define EIO_TYPE_FILE		1
#define EIO_TYPE_DIRECTORY	2

/*
 * Constants used by EIO_Seek()
 */
#define EIO_SEEK_BEGINNING	0
#define EIO_SEEK_CURRENT	1
#define EIO_SEEK_END		2

/***************************** T Y P E D E F S ****************************/

#if __TURBOC__

typedef unsigned rwtype;

typedef struct {
	short			 DirFlag;
	struct ffblk	 FFBlk;
	char far32		*PathPtr;
	char			 TotalPath[256];
	
	short			 Status;		// public
	short			 IsDir;			// public
	char far32		*EnvString;		// public
	char			 Path[256];		// public
} DirTracker;

typedef struct {
	struct ftime	FTime;
} FileDateType;

typedef struct {
	int		Attrib;
} FileAttribType;

typedef struct {
	char	Comment[1];
} FileCommentType;

#elif (__WATCOMC__ && __MSDOS__)

typedef unsigned rwtype;

typedef struct {
	short			 DirFlag;
	short			 GotFirst;
	struct find_t 	 FFBlk;
	char far32		*PathPtr;
	char			 TotalPath[256];
	
	short			 Status;		// public
	short			 IsDir;			// public
	char far32		*EnvString;		// public
	char			 Path[256];		// public
} DirTracker;

typedef struct {
	unsigned short	 Date;
	unsigned short	 Time;
} FileDateType;

typedef struct {
	unsigned	 Attrib;
} FileAttribType;

typedef struct {
	char	Comment[1];
} FileCommentType;

#elif (__ZTC__ && __MSDOS__)

typedef unsigned rwtype;

typedef struct {
	short			 DirFlag;
	short			 GotFirst;
	struct FIND	 	*ZFind;
	char far32		*PathPtr;
	char			 TotalPath[256];
	
	short			 Status;		// public
	short			 IsDir;			// public
	char far32		*EnvString;		// public
	char			 Path[256];		// public
} DirTracker;

typedef struct {
	unsigned	 Date;
	unsigned	 Time;
} FileDateType;

typedef struct {
	unsigned	 Attrib;
} FileAttribType;

typedef struct {
	char	Comment[1];
} FileCommentType;

#elif __AMIGAOS__

typedef size_t rwtype;

struct Path {
	LONG	path_Next;
	LONG	path_Lock;
};

typedef struct DirTracker {
	short				 DirFlag;
	struct DirTracker	*SubDir;
	struct DirTracker	*ParentDir;
	struct DirTracker	*PrivateDir;
	struct Path			*APath;
	struct AnchorPath	*Anchor;
	char				 TotalPath[EIO_MAXPATH];
	
	short				 IsDir;				// public
	short				 Status;			// public
	char				*EnvString;			// public
	char				 Path[EIO_MAXPATH];	// public
} DirTracker;

typedef struct {
	struct DateStamp	DateStamp;
} FileDateType;

typedef struct {
	LONG		Attrib;
} FileAttribType;

typedef struct {
	char	Comment[81];
} FileCommentType;

#elif __MACOS__

typedef size_t rwtype;

typedef struct {
	char			*PathPtr;		// private
	char			 Pattern[256];	// private
	short			 fileNdx;		// private
	short			 DirFlag;		// private
	FSSpec			 parentfss;		// private
	
	short			 Status;		// public
	short			 IsDir;			// public
	char			*EnvString;		// public
	char			 Path[256];		// public
} DirTracker;

typedef struct {
	CInfoPBRec		cipb;
} FileDateType;

typedef struct {
	FileDateType	fdt;
} FileAttribType;

typedef struct {
	FileDateType	fdt;
} FileCommentType;

#elif __EGGS_SGI__

typedef size_t rwtype;

typedef struct {
	char			*PathPtr;		// private
	char			 Pattern[256];	// private
	short			 fileNdx;		// private
	short			 DirFlag;		// private
	DIR				*dirp;			// private
	char			 FullPath[256];	// private
	short			 Status;		// public
	short			 IsDir;			// public
	char			*EnvString;		// public
	char			 Path[256];		// public
} DirTracker;

typedef struct {
	struct		stat		fileStat;
} FileDateType;

typedef FileDateType FileAttribType;

typedef FileDateType FileCommentType;

#else
#error Need Stuff
#endif

/******************************* M A C R O S ******************************/

#if (__TURBOC__ || __MSDOS32X__)

#define EIO_Open(name,mode)		open((name),(mode),(S_IREAD|S_IWRITE))
#define EIO_Close(fh)			close(fh)
#define EIO_Seek(fh,off,org)	lseek((fh),(off),(org))
#define EIO_Tell(fh)			tell(fh)

#elif AZTEC_C

#define EIO_Open(name,mode)		open((name),(mode))
#define EIO_Read(fh,buf,size)	read((fh),(buf),(size))
#define EIO_Write(fh,buf,size)	write((fh),(buf),(size))
#define EIO_Close(fh)			close(fh)
#define EIO_Seek(fh,off,org)	lseek((fh),(off),(org))
#define EIO_Tell(fh)			tell(fh)

#elif __MACSC__

#define EIO_Open(name,mode)		open((name),(mode))
#define EIO_Read(fh,buf,size)	read((fh),(buf),(size))
#define EIO_Write(fh,buf,size)	write((fh),(buf),(size))
#define EIO_Close(fh)			close(fh)
#define EIO_Seek(fh,off,org)	lseek((fh),(off),(org))
#define EIO_Tell(fh)			tell(fh)

#elif __EGGS_SGI__

#define EIO_Open(name,mode)		open((name),(mode),(S_IREAD|S_IWRITE|S_IRGRP|S_IROTH))
#define EIO_Close(fh)			close(fh)
#define EIO_Read(fh,buf,size)	read((fh),(buf),(size))
#define EIO_Write(fh,buf,size)	write((fh),(buf),(size))
#define EIO_Seek(fh,off,org)	lseek((fh),(off),(org))
#define EIO_Tell(fh)			tell(fh)

#elif LATTICE

#define EIO_Open(name,mode)		open((name),(mode))
#define EIO_Close(fh)			close(fh)
#define EIO_Seek(fh,off,org)	lseek((fh),(off),(org))
#define EIO_Tell(fh)			tell(fh)

#else
#error Need EIO_Open,EIO_Close...
#endif

#define EIO_ReadOpen(name)		EIO_Open((name), O_RDONLY | O_BINARY)
#define EIO_OpenRead(name)		EIO_Open((name), O_RDONLY | O_BINARY)
#define EIO_WriteOpen(name)		EIO_Open((name), O_WRONLY | O_BINARY | O_TRUNC | O_CREAT)
#define EIO_OpenWrite(name)		EIO_Open((name), O_WRONLY | O_BINARY | O_TRUNC | O_CREAT)
#define EIO_UpdateOpen(name)	EIO_Open((name), O_RDWR | O_BINARY | O_CREAT)
#define EIO_OpenUpdate(name)	EIO_Open((name), O_RDWR | O_BINARY | O_CREAT)
#define EIO_AppendOpen(name)	EIO_Open((name), O_RDWR | O_BINARY | O_CREAT | O_APPEND)
#define EIO_OpenAppend(name)	EIO_Open((name), O_RDWR | O_BINARY | O_CREAT | O_APPEND)

#define	EIO_FileListType(nd)	((nd)->Type)

/****************************** G L O B A L S *****************************/


/****************** F U N C T I O N   P R O T O T Y P E S *****************/
#if __NOT_ANSI__		/** Non-ANSI Prototypes **/



#else					/** ANSI Prototypes **/

extern short		 EIO_FileExists (char *filename);
extern char			*EIO_ExpandEVars (char *new, char *old, size_t maxlen);
extern char			*EIO_FixDirSeps (char *str);
extern short		 EIO_InsureEndSlash (char *str);
extern void			 EIO_GetTempFileName (char *temp);
extern void			 EIO_fnmerge (char *path, char *dir, char *name, char *ext);
extern UWORD		 EIO_fnsplit (char *path, char *dir, char *name, char *ext);
extern char			*EIO_Path (char *filespec);
extern char			*EIO_Name (char *filespec);
extern char			*EIO_Ext (char *filespec);
extern char			*EIO_Filename (char *filespec);
extern char			*EIO_CurrentDir (char *path);
extern void			 EIO_FreeDirTracker (DirTracker *dt);
extern short		 EIO_GetNextPath (DirTracker *dt);
extern DirTracker	*EIO_GetFirstPath (void);
extern int			 EIO_SpawnProgram (char *program, char **argv);
extern long			 EIO_FileLength (int fd);
extern DirTracker 	*EIO_GetFirstFile (char *path, short dirflag);
extern short		 EIO_GetNextFile (DirTracker *dt);
extern ListType		*EIO_GetFileList (char *filespec, short recursive, short directories);
extern short		 EIO_FindFile (char *filename, char *guidename, char *pathbuff);
extern short		 EIO_PrintEscString (FILE *fp, char *s);
extern DirTracker 	*EIO_GetFirstEnv (void);
extern short		 EIO_GetNextEnv (DirTracker *dt);
extern short 		 EIO_FileType (char *filename);
extern short 		 EIO_GetFileDate (char *filename, FileDateType *fdt);
extern short 		 EIO_SetFileDate (char *filename, FileDateType *fdt);
extern short 		 EIO_GetFileAttrib (char *filename, FileAttribType *fat);
extern short 		 EIO_SetFileAttrib (char *filename, FileAttribType *fat);
extern short 		 EIO_GetFileComment (char *filename, FileCommentType *fct);
extern short 		 EIO_SetFileComment (char *filename, FileCommentType *fct);
extern short 		 EIO_CopyFile (char *from, char *to);
extern short 		 EIO_MakeDir (char *filename);
extern short		 EIO_CmpDates (FileDateType *fdt1, FileDateType *fdt2);

#if __MSDOS32X__

#define EIO_Read(fh,buf,size)	read((fh),(buf),(size))
#define EIO_Write(fh,buf,size)	write((fh),(buf),(size))

#elif (__TURBOC__ || LATTICE || (__ZTC__  && __MSDOS__))

extern long			 EIO_Read (int fh, void *buf, long size);
extern long			 EIO_Write (int fh, void *buf, long size);

#endif

#endif /** Function Prototypes **/

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_EIO_H */

