/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ELVTGADG.H
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 02/07/92
 *   MODIFIED : 07/26/94
 *       TABS : 04 08
 *
 * DESCRIPTION
 *		Externs and types for elevator gadget.
 *
 * HISTORY
 *
*/

#ifndef ELVTGADG_H
#define ELVTGADG_H

#include <echidna/platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <echidna/etypes.h>
#include <echidna/windows.h>
#include <echidna/hitareas.h>
#include <echidna/naming.h>
#include <echidna/scic.h>
#include <echidna/gadgets.h>

/**************************** C O N S T A N T S ***************************/

/* Turn USE_H_ELEVATORS on or off depending on if you program uses horizontal
 * elevators.  Turning it off if you don't use them saves code space.
 */
#if USE_H_ARROWS
	#define USE_H_ELEVATORS	1
#else
	#define USE_H_ELEVATORS	0
#endif

typedef enum {VERTICAL, HORIZONTAL} AXIS;

#define EGAD_REALTIME_SLIDER	(1 << 0)

#define cpelEGADButtonMin  (cyVArrowMin + cyPushBTNMin)
#define cpelEGADWdthMin	cpelEGADButtonMin
#define cpelEGADLnthMin	(cpelEGADButtonMin * 2)

/******************************** T Y P E S *******************************/

typedef void (*PFNEGAD)(void *pvData);

typedef struct {

	/* PRIVATE VARIABLES: Use macros to look at them */
	short x, y;
	short cx, cy;
	short fDrawn;	/* True if drawn since init */
	short fRealTimeSlider;
	short EUITicksTilRepeat;
	
	short FloorsMax;
	short iFloorCrnt;	/* 0..iFloorsMax-1 */

	AXIS	axis;
	union {
		struct {
			short xLft;		/* Base Horizontal Position */
			short y;		/* Constant Vertical position */
			short dx;		/* Offset from base position */
			short cxSlack;	/* If (slack < 0) then slider will not appear */
			short xRgtBtn;	/* coordinate for right arrow button */
		} H; /* Horizontal */

		struct {
			short yTop;		/* Base Vertical Position */
			short x;		/* Constant Horizontal position */
			short dy;		/* Offset from base position */
			short cySlack;	/* If (slack < 0) then slider will not appear */
			short yBotBtn;	/* coordinate for down arrow button */
		} V; /* Vertical */


	} S; /* Slider */

	HitArea  rgha[4];	/* 2 arrows and 1 slider area and 1 end area */

	SCIC	rgscic[2];	/* Scaled icon structs for 2 arrows */	
	BTN		rgbtn[3];	/* Button structs for 2 arrow buttons and 1 slider */
	GHI		rgghi[2];	/* Button area info structs for 2 arrows */

	/* Gets colors from active window: GetActiveWindow() */

	PFNEGAD	 pfnegad;	/* User func to call when floor changes */
	void   	*pvUserData;	/* User data to pass to user function */

} EGAD;	/* Elevator gadget */

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/

#define iFloorCrntEGAD(pegad)		((pegad)->iFloorCrnt)
#define FloorsMaxEGAD(pegad)		((pegad)->FloorsMax)
#define cxEGAD(pegad)		((pegad)->cx)
#define cyEGAD(pegad)		((pegad)->cy)

#define phaEGAD(pegad)		((pegad)->rgha)

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

extern short InitEGAD (
					EGAD	*pegad,
					AXIS	 axis,			/* VERTICAL or HORIZONTAL */
					short	 cpelLnth,	/* length along axis */
					short	 cpelWdth,	/* width perpendicular to axis */
					short	 iFloorsMax,		/* Number of floors serviced */
					UWORD	 Flags,			
					short	 c60hzTicksTilRepeat, /* < 0 means no repeat */ 
					PFNEGAD	 pfnegad,	/* User func for floor changes */ 
					void	*pvUserData	/* User data to pass to user func */
);

extern short	SetFloorPositionEGAD (
					EGAD *pegad, 
					short FloorsMax,
					short iFloorCrnt
					);

extern void 	DrawEGAD (EGAD *pegad, short x, short y);

#ifdef __cplusplus
}
#endif

#endif /* ELVTGADG_H */

