/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * EMS.H
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/16/91
 *   MODIFIED : 10/06/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Public header file for EMS.C.
 *
 * HISTORY
 *		09/20/94 (dcc) - add support for __MSDOS32X__, __AMIGAOS__,
 *						 __MACOS__ and __EGGS_SGI__
 *
*/

#ifndef ECHIDNA_EMS_H
#define ECHIDNA_EMS_H

#include <echidna/platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <echidna/etypes.h>
#include "stdlib.h"

/**************************** C O N S T A N T S ***************************/

#define EMS_PAGES 4

/******************************** T Y P E S *******************************/

typedef struct {
	UWORD		 EMMHandle;
	UBYTE		 Page;
	UWORD		 Offset;
} EMMTracker;

typedef struct {
	UBYTE		 Type;
	UBYTE		 Pages;
	UWORD		 Size;
	EMMTracker	 EMMs[1];
} EMSPntr16384;

typedef struct {
	UBYTE		 Type;
	UBYTE		 Pages;
	UWORD		 Size;
	EMMTracker	 EMMs[2];
} EMSPntr32768;

typedef struct {
	UBYTE		 Type;
	UBYTE		 Pages;
	UWORD		 Size;
	EMMTracker	 EMMs[3];
} EMSPntr49152;

typedef struct {
	UBYTE		 Type;
	UBYTE		 Pages;
	UWORD		 Size;
	EMMTracker	 EMMs[EMS_PAGES];
} EMSPntr;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/

#define ActivateEMS(ep)	ActivateEMSatWin ((ep), 0)

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

#if (__MSDOS32X__ || __AMIGAOS__ || __MACOS__ || __EGGS_SGI__)

extern	void	*__LastXTRA;

#define SetEMSMemoryPoolSize(cEMSKBytes)
#define SetLargestEMS(size)		
#define ActivateEMSatWin(ep,win)	(__LastXTRA = (void *)(ep))
#define FreeEMS						free
#define AllocEMS					malloc
#define	ReadEMS(ep,off,len,mem)		(memcpy((mem),((char *)(ep))+(off),(len)))
#define	WriteEMS(mem,len,ep,off)	(memcpy(((char *)(ep))+(off),(mem),(len)))
#define	AvailEMS(xmt)				(12345678L)
#define PopEMS()					
#define PushEMS()					

#elif __MSDOS__

extern void		 SetEMSMemoryPoolSize(UWORD cEMSKBytes);
extern void		 SetLargestEMS (size_t size);
extern void	far	*ActivateEMSatWin (EMSPntr *ep, short window);
extern void		 FreeEMS (EMSPntr *ep);
extern EMSPntr	*AllocEMS (size_t size);
extern void		 ReadEMS (EMSPntr *ep, UWORD uwOffset, UWORD uwLength, void far *pMem);
extern void		 WriteEMS (void far *pMem, UWORD uwLength, EMSPntr *ep, UWORD uwOffset);
extern long		 AvailEMS(void);
extern void		 PopEMS (void);
extern short	 PushEMS (void);

#else
#error Need Macros
#endif

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_EMS_H */

