/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * EUI.H
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 07/31/89
 *   MODIFIED : 10/12/93
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Easy User Interface routines.
 *
 * HISTORY
 *
*/

#ifndef ECHIDNA_EUI_H
#define ECHIDNA_EUI_H

#include <echidna/platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#if __AMIGAOS__
#include <libraries/arpbase.h>
#endif

#include <echidna/eio.h>
#include <echidna/windows.h>
#include <echidna/keyboard.h>
#include <echidna/readini.h>

/**************************** C O N S T A N T S ***************************/

#define MAX_EDIT_STRING	80

#define EDIT_NUM_ONLY	(1 << 0)	/* Allow digits only				*/

#define EUI_TIMER_HERTZ	30

/******************************** T Y P E S *******************************/

typedef struct {
	Window	*Window;
	char	 Path[EIO_MAXPATH];
	char	 Name[EIO_MAXPATH];
	char	 FileName[EIO_MAXPATH];
	char	 Title[40];

	/* --- */

	short	 Used;
	short	 FileCount;
	short	 Position;
	void	*Top;
} EasyFileReq;

typedef struct {
	char	*Title;
	char	*Path;
	char	*FileName;
} FileReqInt;

typedef struct EditString {
	char				*String;
	size_t				 MaxLen;
	short				 X;
	short				 Y;
	struct EditString	*Prev;
	struct EditString	*Next;
} EditString;

/****************************** G L O B A L S *****************************/

extern short	 ExitProgram;
extern Font		 SmallFont;
extern Font		 MainFont;
extern Window	*MainWindow;
extern UWORD	 EUICountDownTimer;
extern short	 EUIDBuf;

/******************************* M A C R O S ******************************/

#define SetActiveWindow(pwin)	(MainWindow = (pwin))
#define GetActiveWindow()		(MainWindow)

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

/********************************* EDIT.c *********************************/
extern short		 CheckEditKeys (short *outkey);
extern short		 HighlightString (HitInfo *hi);
extern void			 SelectString (EditString *es);
extern void			 UpdateEditString (EditString *es);
extern void			 NewUpdateStrings (
							Window		*win,
							EditString	*es,
							short		 count,
							UWORD		 flags
						);
extern void			 UpdateStrings (
							Window		*win,
							EditString	*es,
							short		 count,
							UWORD		 flags
						);
extern void			 InitStrings (Window *win);

/******************************** FILEREQ.c *******************************/
extern char			*EasyGetFileName (Window *win, char *title);
extern void			 InitFileReq (EasyFileReq *fr);
extern EasyFileReq	**InitFileReqs (FileReqInt *fri);
extern char			*GetFileName (Window *win, EasyFileReq *fr);
extern char			*GetPathName (Window *win, EasyFileReq *fr);
extern void			 SetFreqTitle (EasyFileReq *ezfrq, char *title);
extern void			 SetDirSortType (short type);
extern void			 RestoreOriginalDrive(void);

/******************************** MESSAGE.c *******************************/
extern short	 Message (Window *window, char *title, char *message, char *options);

/********************************* KEYEVENT.c *********************************/
extern short	 InitKeys (IniList *inilist);
extern void		 UninitKeys (void);
extern short	 DoKeyCommand (USHORT keyvalue);
extern short	 HandleKeys (Window *win);

/******************************** 1FLOPPY.c *******************************/

extern BOOL SingleDrive(void);
extern void SetFloppy(char *filename);

/********************************** EUI.C *********************************/
extern void		 EUIMessageLoop (void);
extern void		 CloseEUI (void);
extern short	 OpenEUI (char *inifile);

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_EUI_H */

