/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * GADGETS.H
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 03/06/92
 *   MODIFIED : 07/26/94
 *       TABS : 04 08
 *
 * DESCRIPTION
 *		Typedefs for gadget structs.
 *
 * HISTORY
 *
*/

#ifndef GADGETS_H
#define GADGETS_H

#include <echidna/platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <echidna/etypes.h>
#include <echidna/events.h>
#include <echidna/naming.h>

/**************************** C O N S T A N T S ***************************/

#define cxPushBTNMin	8
#define cyPushBTNMin	8
#define cxCheckBoxBTNMin	5	
#define cyCheckBoxBTNMin	5
#define cxBullsEyeBTNMin	4
#define cyBullsEyeBTNMin	4

/* BTN.fState flag masks */
#define BTNS_MARQUEED	(1 << 0)
#define BTNS_HILIGHTED	(1 << 2)
#define BTNS_PRESSED	(1 << 3)
#define BTNS_GHOSTED	(1 << 4)

/* BTN.fPref flag masks */
#define BTNP_ODD_GHOST	(1 << 0)

/******************************** T Y P E S *******************************/

typedef enum {
	btnkCheckBox,	/* square with x in middle */
	btnkBullsEye,	/* Circle with dot in middle */
	btnkPush,		/* 3D keyboard key looking button */
} BTNK;	/* Button Kind */

typedef enum {
	bdkBlank,	/* no decal graphic */
	bdkHotText, /* See HotText.H for explanation. */
	bdkGShape, 	/* Gshape. */
	bdkSCIC		/* Scaleable Icon. See SCIC.H for explanation. */
} BDK; /* button decal kind */

typedef enum {
	rlpsRight,
	rlpsLeft,
	rlpsAbove,
	rlpsBelow,
	rlpsOn
} RLPS; /* Relative Position */

typedef enum {
	jstfLeft,
	jstfCenter,
	jstfRight
} JSTF; /* Justification (positional)*/

typedef short (*PFNGAD)(HitInfo *phi);

typedef struct {
	BTNK	 btnk;			/* Button kind */
	short	 x, y;			/* Filled in by DrawBTN */
	short	 cx, cy;		/* Dimensions of button */
	BDK		 bdk;			/* Decal kind */
	void	*pvDecal;		/* Decal data */
	RLPS	 rlpsDecal;		/* Decal position relative to button:
							 *   rlpsRight, rlpsLeft,  
							 *   rlpsAbove, rlpsBelow, rlpsOn 
							 */
	JSTF	 jstfDecal;		/* Decal justification:
							 *    jstfLeft, jstfCenter, jstfRight
							 */
	UBYTE	 fState;		/* BTNS_MARQUEED | BTNS_HILIGHTED | BTNS_PRESSED |
							 *    BTNS_GHOSTED 
							 */
	UBYTE	 fPref;			/* Preference flags:
							 *     BTNP_ODD_GHOST  
							 */
} BTN; /* Button */

typedef struct {
	/* Private */
	void	*pvgad;		/* Pointer to specific gadget structure */
	PFNGAD	 pfngad;	/* Gadget function to call. */
	void	*pvData;	/* Data for gadget function */
	UWORD	 fhaFlags;	/* copy of pha->Flags call handling fields. */
} GHI;	/* Gadget Handler Info */

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/****************** F U N C T I O N   P R O T O T Y P E S *****************/

extern void DrawBTN (BTN *pbtn, short x, short y);
extern void DrawBDK (
	BDK bdk, 
	void *pvDecal, 
	short x, short y, 
	UBYTE fState,
	UBYTE fPref
);

extern void GetDimBDK (BDK bdk, void *pvDecal, short *cx, short *cy);
extern void GetDimBTN (BTN *pbtn, short *cx, short *cy);

#ifdef __cplusplus
}
#endif

#endif /* GADGETS_H */

