/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * HITAREAS.H
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 10/10/90
 *   MODIFIED : 07/26/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Routines to deal with Hit Areas (Gadgets?)
 *
 * HISTORY
 *
*/

#ifndef ECHIDNA_HITAREAS_H
#define ECHIDNA_HITAREAS_H

#include <echidna/platform.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <echidna/etypes.h>
#include <echidna/events.h>

/**************************** C O N S T A N T S ***************************/

/* Satic hit area flags go in 'Flags' field */
#define HTA_TOGGLE		(1 << 0)	/* Toggle HTS_SELECTED with each press	*/
#define HTA_COMPLEMENT	(1 << 1)	/* Complement area when selected		*/
#define HTA_DOWN		(1 << 2)	/* Call function on button first down	*/
#define HTA_REPEAT		(1 << 3)	/* Call function while button pressed	*/
#define HTA_REPORTRIGHT	(1 << 4)	/* Right mouse button valid here		*/
#define HTA_EVENT		(1 << 5)	/* 'Data' points to an event			*/
#define HTA_ESCAPE		(1 << 6)	/* Escape key area.						*/
#define HTA_INSURE_UP	(1 << 7)	/* Call function on button up even if 
									 *   pointer is not over hit area. 
									 *   Default is call only when button up 
									 *   happens over hit area. If you want
									 *	 never to be called on button up then
									 *   set HTA_NO_UP.
									 */
#define HTA_NO_UP		(1 << 8)	/* Never call function on button up. This
									 * flag overrides HTA_INSURE_UP.
									 */
#define HTA_SUBAREAS	(1 << 9)	/* This area has subareas				*/
#define HTA_ENTER_FOCUS	(1 << 10)	/* Can recieve Enter key focus. 
									 * Call func when Enter focus changes.	*/
#define HTA_SPACE_FOCUS	(1 << 11)	/* Not space key focusable. Do not call
									 * function when space focus changes.	*/
#define HTA_DEFAULT_ENTER	(1U << 12)	/* Default Enter key area			*/
#define HTA_KEYS			(1U << 13)	/* Call func with keyboard input.	*/
#define HTA_CURSOR			(1U << 14)	/* Call func with cursor key input.	*/

/* Dynamic hit area flags go in 'fState' field */
#define HA_ENTER			(1 << 0)	/* Current Enter key area 			*/
#define HA_SPACE			(1 << 1)	/* Current Space key area 			*/
#define HA_DISABLED			(1 << 2)	/* Not hitable (Ghosted) 			*/
#define HA_TOGGLE_STATE		(1 << 3)	/* Toggle state if HTA_TOGGLE able	*/
#define HA_WAS_ABLE			 (1 << 4)	/* TRUE if was able before being
										 * disabled by some function which
										 * intends to later restore its 
										 * ability state by consulting this
										 * flag.
										 */

#define HIT_COMPCOLOR	(255 ^ 254)

#define HAL_LAST_AREA	(-1)
#define HTA_LAST_AREA	HAL_LAST_AREA

#define STATIC_END_HIT_AREA	{0,0,HAL_LAST_AREA,0,0,0,NULL, NULL, 0, 0}

/******************************** T Y P E S *******************************/
#if 0
typedef struct {
	short		 X;
	short		 Y;
	short		 Width;
	short		 Height;
	void		*Data;
} SaveAreaType;
#endif

typedef struct HitArea {
	short		 X;
	short		 Y;
	short		 Width;
	short		 Height;
	UWORD		 Flags;		/* Static flags */
	short		 ID;
	void		*Data;		/* function or hitarea array or ... */
	void		*UserData;	
	short		 KeyCode;	/* ASCII code of Key that invokes this area.
							 * 0 means no key is bound to this hit area.
							 */
	UWORD		fState;		/* Dynamic flags: HA_SPACE, HA_ENTER, HA_DISABLED default, 
							 * 		HA_TOGGLE_STATE.
							 */
} HitArea;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/

#define MAKE_END_HITAREA(pha)	((pha)->Width = HAL_LAST_AREA)
#define END_HITAREA(area)	((area)->Width == HAL_LAST_AREA)
#define PREV_HITAREA(area)	(((HitArea *)(area))-1)
#define NEXT_HITAREA(area)	(((HitArea *)(area))+1)

#define fHitAreaIsDisabled(pha)		((pha)->fState & HA_DISABLED)
#define fHitAreaIsEnabled(pha)		(!fHitAreaIsDisabled((pha)))

#define DisableHitArea(pha)	(SetHitAreaAbleState((pha),FALSE))
#define EnableHitArea(pha)	(SetHitAreaAbleState((pha),TRUE))


/****************** F U N C T I O N   P R O T O T Y P E S *****************/

extern HitArea		*CheckHitAreas (HitArea *ha, short xoff, short yoff);
extern HitArea *CheckKeyHitAreas (HitArea *pha, UWORD scancode);
extern void SetHitAreaAbleState (HitArea *pha, short fEnable); 
extern void FocusHitArea (HitArea *rgha, HitArea *phaFocus);
extern void BuryHitArea (HitArea *pha);
extern void UnBuryHitArea (HitArea *pha);

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_HITAREAS_H */

