/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * NAMING.H
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 01/08/92
 *   MODIFIED : 07/30/92
 *       TABS : 04 08
 *
 * DESCRIPTION
 *		This file documents the naming conventions that the
 *		programmer including this file promises to use.  If you are
 *		a programmer, before you include this file in your source
 *		module you must recite the following pledge:
 *
 *		I <your name> promise to abide by the rules of this
 *		notation convention while creating names within the source
 *		file which includes this header file.  Furthermore I
 *		promise to use notation-compliant macros for external
 *		non-complying names (created by some bozo) which I
 *		must use in my source file and which may be decieving (i.e.
 *		any names that are not obviously non-compliant) or at the
 *		very least to comment the use of these evil names.
 * 			
 *			
 *		Also include here are macros for commonly used echidna library types
 *		that do not comply.
 *
 *		See the end of this file for the notation documentaion.
 *		
 *	
 * HISTORY
 *
*/


#ifndef NAMING_H
#define NAMING_H

#include <echidna/platform.h>
#include "switches.h"

#ifdef __cplusplus
extern "C" {
#endif

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/

#define APO	APOType
#define BPI	BPIType
#define CCY	ColorCycleType
#define CCS	ColorCycleSetType
#define CTB	ConversionTableType
#define CTP	ConTablePtrType
#define CRM	ConversionRoomType
#define CTI	CTIType
#define FNT	Font
#define FPF	FPF_InfoType
#define GSH	GShape
#define IRT	InterruptRoutineType
#define LAY	LayerType
#define LST	ListType
#define MEN	MenuType
#define NOD	NodeType
#define PAL	PaletteType
#define PBX	PaintBoxType
#define ROO	RoomType
#define SCO	ScoreType
#define SFX	SndFxType
#define TIM	TimerType
#define TPR	tUMEPackRoomType
#define WIN	Window

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

#ifdef __cplusplus
}
#endif

#endif /* NAMING_H */

/***************************************************************************
Naming Convention Documentation
-------------------------------

[<Prefix>[_]][<Tag>][<Qulaifier>]

Prefix
------
u 	unsigned
p	pointer
i	index
c	count
msk	bit (mask value)
sh	a shift amount (bit number)
mp	an array. This prefix is followed by two types: the index type (domain)
	and element type (range).
rg  an array. This prefix is followed by one type: the element type (range),
	as the index type (domain) is taken to be integer.
 

Common Tags (default is signed word)
------------------------------------
f	flag (0 or 1) may be a bit or byte or word. 
st	pascal string
sz	zero-terminated string
fn	function (usually with 'p' prefix)
b	byte
w	word
l	long
bit	single bit
v	void
x	Cartesian x coordinate
y	Cartesian y coordinate

Common Qualifiers
-----------------
Src		source
Dst		Destination.

Minex		==Minin-1
Minin		==Min
Min		Lowest valid index of an array.
Micex		==Micin-1
Micin		==Mic
Mic		Like Min but for arrays with dynamically changing size.

First		First element in a range to be dealt with.
Crnt		Current element in a range being dealt with.
Last		Last element in a range to be dealt with.
Lim		First invalid upper index in a range to be dealt with.

Macin		==Macex-1
Macex		==Mac. More appropriate as index qualifier than count qualifier.
Mac		First invalid upper index of array with dynamically changing size.
Maxin		==Maxex-1;
Maxex		==Max.  More appropriate as index qualifier than count qualifier.
Max		Maximum number of elements in array.  First invalid upper index 
			of an array (in C, last valid upper index in Pascal).  Use this
			for qualifier on staticly allocated array size constants.

Total		Total number of valid elements in an array.  

			The issue of validity of an array element with regard to data 
			integrity is restricted, but not defined, by memory allocation.
			An array index within the range of allocated storage may or may
			not be valid depending on whether or not it falls in the range
			of initialized elements of the array.  An array element outside
			the range of allocated storage is always invalid, regardless of 
			whether or not it was initialized, because it is not protected
			from interference after initialization.  Max, Min, Maxex, Maxin,
			Minex and Minin define limits of validity based on allocated
			storage.  Mac, Mic Macex, Macin, Minex and Minin define
			limits of validity based on initialization.  


Note: Min<=Mic<=First<=Crnt<=Last<=<Lim<=Mac<=Max
		Minex<Minin=Min
		Micex<Micin=Mic
		Macin<Macex=Mac
		Maxin<Maxex=Max
		Total<=Macex

Note: The "in" and "ex" suffixes used above on Min and Max are short
		for 'included' and 'excluded', in reference to the range of valid
		indexes	of the array.

Wdth	Width 
Hght	Height

Note:
-----
For the width of brush use:
	cx<name>Wdth

For height of brush use:
	cy<name>Hght


***************************************************************************/

