/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * PROC.H
 *
 * PROGRAMMER : Gregg Tavares
 *    VERSION : 00.000
 *    CREATED : 09/24/93
 *   MODIFIED : 09/21/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		
 *
 * HISTORY
 *
*/

#ifndef ECHIDNA_PROC_H
#define ECHIDNA_PROC_H

#include <echidna\platform.h>
#include "switches.h"

#ifdef __cplusplus
extern "C" {
#endif

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/

extern char *CurrentFuncName;

/******************************* M A C R O S ******************************/

/*
 * ******** PROC Declarations
 *
 * General Example:
 *
 * PROC(void,Init,(long size,void *data))
 * RANGE_ARG(size,(size<100000))
 * RANGE_PTR(data)
 * {
 *
 *		**** CODE ****
 *
 * }
 * PROC_END	// Init
 *
 * This example creates a routine called Init, that has no return value and
 * accepts two arguments; a long size, and a pointer to void.
 * If the RANGECHECKING flag is set, the size parameter is checked to make
 * sure it is less than 100000, and the data pointer is checked to point
 * into valid memory.
 *
 * Variations:
 *	PROC_LOCAL()	- for declaring static routines
 *	PROC_GLOBAL()	- for declaring externs
*/

#if (RANGECHECKING != 0) || (DEBUG_PRINTFS != 0)
	#define PROC(r,f,ad)		r f ad \
								{ \
								CurrentFuncName = #f;
	#define PROC_LOCAL(r,f,ad)	static r f ad \
								{ \
								CurrentFuncName = #f;
#else
	#define	PROC(r,f,ad)		r f ad \
								{
	#define PROC_LOCAL(r,f,ad)	static r f ad \
								{
#endif

// --------

#define PROC_GLOBAL(r,f,ad)		extern r f ad

#define	PROC_END				}

// --------

#if RANGECHECKING != 0
	#define	RANGE_CHECK(c,s,v)	if(!(c)) printf(__FILE__"(line %d)-%s() " s ": " #v " = %lx\n",__LINE__,CurrentFuncName,(unsigned long)v);
#else
	#define	RANGE_CHECK(c,s,v)
#endif

// --------

#define	RANGE_PTR(ad)			RANGE_CHECK((!ad),"Pointer is NULL",ad)
#define	RANGE_ARG(ad,r)			RANGE_CHECK(r,"Range check Failed",ad)

// ******** DPRINTF() Declaration

#if DEBUG_PRINTFS != 0
	#define	DPRINTF(s,v)		printf("%s(): "s,CurrentFuncName,v)
#else
	#define	DPRINTF(s,v)
#endif

// These are only needed if you need more than one argument passed into a
// DPRINTF macro
//
// ex.
//	DPRINTF("size %ld, addr. %lx\n",DARG(size) DARG_END(pos));

#define DARG(var)				var,
#define DARG_END(var)			var

// ******** DEBUGONLY() Declaration

#if DEBUG_CODE != 0
	#define	DEBUGONLY(f)		f
#else
	#define	DEBUGONLY(f)
#endif

// ******** FINALONLY() Declaration

#if !DEBUG_CODE
	#define	FINALONLY(f)		f
#else
	#define	FINALONLY(f)
#endif

// ******** MACONLY() Declaration

#if !RUNTIME_CODE
	#define	DEVELONLY(f)		f
#else
	#define	DEVELONLY(f)
#endif

// ******** RUNTIMEONLY() Declaration

#if RUNTIME_CODE != 0
	#define	RUNTIMEONLY(f)		f
#else
	#define	RUNTIMEONLY(f)
#endif

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

#ifdef __cplusplus
}
#endif

#endif /* ECHIDNA_PROC_H */
