/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * READCIM.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 09/25/92
 *   MODIFIED : 06/19/00
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		
 *
 * HISTORY
 *		09/21/94 (dcc) - compiles cleanly with 32-bit Borland C++ 4.02.
 *	
 * TODO
 *		* Make it load NESREF.PAL for palette.
 *	
*/

#include <echidna\platform.h>
#include "switches.h"

#include <echidna\etypes.h>
#include <echidna\listfunc.h>
#include <echidna\eio.h>
#include <echidna\emacros.h>
#include <echidna\eerrors.h>
#include <echidna\xtramem.h>
#include <echidna\strings.h>
#include <echidna\readiff.h>
#include <echidna\bilbm.h>
#include <echidna\btume.h>

#define LOCAL static

/**************************** C O N S T A N T S ***************************/

#define	ROOM_LEVEL				0
#define ROOM_TABLE				1
#define ROOM_PICTURE			2
#define ROOM_PARALLAX			3
#define ROOM_MODE7				4
#define ROOM_FLAT_CONVERSION	5
#define ROOM_LAYERED_CONVERSION	6

#define TILESET_IMAGE		0
#define TILESET_CONTOUR		1
#define TILESET_SPECIAL		2
#define TILESET_OBJECT		3
#define TILESET_2BITS		4
#define	TILESET_8BITS		5
#define TILESET_MODE7		6

#define FLAT_CONVERSION_ROOM	5	// From TPMCKID2
#define SPECIAL_TILES			2	// From TPMCKID2

#define CED_MTYPE_GAMEBOY		0
#define CED_MTYPE_NES			1
#define CED_MTYPE_SFX			2
#define CED_MTYPE_SFXM7			3
#define CED_MTYPE_PCENG			8
#define CED_MTYPE_PCENGSUPER	9
#define CED_MTYPE_GENISIS		16
#define CED_MTYPE_GIGADRIVE		17
#define CED_MTYPE_UNKNOWN		255

#define MAXTILESETS				(256)
#define	LOADFLAG_OLDROOM		(0x0001)

#define SIFF_CHUNK	MakeID('S','I','F','F')

#define CIMH_SIFF	MakeID('C','I','M','H')
#define CIML_SIFF	MakeID('C','I','M','L')

#define HHDR_CHUNK	MakeID('H','H','D','R')
#define LHDR_CHUNK	MakeID('L','H','D','R')
#define CPAL_CHUNK	MakeID('C','P','A','L')
#define SPAL_CHUNK	MakeID('S','P','A','L')
#define CSET_CHUNK	MakeID('C','S','E','T')
#define MSET_CHUNK	MakeID('M','S','E','T')
#define CATR_CHUNK	MakeID('C','A','T','R')
#define MATR_CHUNK	MakeID('M','A','T','R')
#define MSEQ_CHUNK	MakeID('M','S','E','Q')
#define FSEQ_CHUNK	MakeID('F','S','E','Q')
#define FSQ2_CHUNK	MakeID('F','S','Q','2')
#define NTXT_CHUNK	MakeID('N','T','X','T')
#define IFRM_CHUNK	MakeID('I','F','R','M')
#define FOVY_CHUNK	MakeID('F','O','V','Y')

/******************************** T Y P E S *******************************/

typedef struct {
	ListType	 RoomListX;
	ListType	*RoomList;
	ListType	 TilesetListX;
	ListType	*TilesetList;
	ListType	 CycleListX;
	ListType	*CycleList;
	RoomType	*CurrentRoom;
	TilesetType	*CurrentTileset;
	WORD		 CurrentTSetID;
	WORD		 RoomTSetID;
	WORD		 CharTSetID;
	WORD		 MachineType;
	TilesetType	*CharTileset;
	TilesetType	*MatrixTileset;
	UWORD		 TranslateTable[MAXTILESETS];
	MapType		*Map;
} CIMInfo;

/****************************** G L O B A L S *****************************/

static ColorRegister	*CIMNesRefPal;

/******************************* M A C R O S ******************************/

/***************************** R O U T I N E S ****************************/

/*********************************************************************
 *
 * CIMSkipChunk
 *
 * SYNOPSIS
 *		long CIMSkipChunk (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long CIMSkipChunk (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CIMSkipChunk";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	return IFFSkip (iff, iff->Size /* + (iff->Size & 0x01) */);

} /* CIMSkipChunk */

/*********************************************************************
 *
 * CIMCallIDFunc
 *
 * PURPOSE
 *		Call a function based on an IFF ID.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long CIMCallIDFunc (long id, IFFRoutineID *ifffunc, IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CIMCallIDFunc";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	while (ifffunc->ID) {
		if (id == ifffunc->ID) {
			return ifffunc->IFFFunc (iff);
		}
		ifffunc++;
	}

	return CIMSkipChunk (iff);

} /* CIMCallIDFunc */

/*********************************************************************
 *
 * ReadCIMForm
 *
 * SYNOPSIS
 *		long ReadCIMForm (IFFTracker *iff)
 *
 * PURPOSE
 *		Read a IFF FORM.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadCIMForm (IFFTracker *iff)
{
	long		 okay = TRUE;
	long		 size = iff->Size;
	long		 id;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMForm";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (IFFRead (iff, &id, sizeof (long)) != sizeof (long)) {
		GEcatf ("ReadCIMForm:Reading FORM id");
		return FALSE;
	}

	iff->Size = size - 4;
	okay = CIMCallIDFunc (id, iff->FORMTable, iff);
	if (!okay) {
		return FALSE;
	}

	return size;

} /* ReadCIMForm */

/*********************************************************************
 *
 * ReadSIFF
 *
 * PURPOSE
 *		Read an SIFF file
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadSIFF (IFFTracker *iff)
{
	static IFFRoutineID TopChunks[] = {
		{ SIFF_CHUNK,	ReadCIMForm,	},
		{ NULL,			NULL,			},
	};

	ChunkHeader		 chunkhdr;
	long			 okay;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadSIFF";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (!GetChunkHeader (iff, &chunkhdr)) {
		return FALSE;
	}

	iff->Size = chunkhdr.Size;
	okay = CIMCallIDFunc (chunkhdr.ID, TopChunks, iff);
	if (okay) {
		okay += 8;
	}

	return okay;

} /* ReadSIFF */

/*********************************************************************
 *
 * ReadCIMLCPAL
 *
 * SYNOPSIS
 *		long ReadCIMLCPAL (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLCPAL (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLCPAL";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		struct {
			UWORD	NumColors;
			UWORD	NumPal;
			UBYTE	Type;
		} CPALHeader;
		long			 size;
		int				 numColors;
		ColorRegister	*cregs = NULL;
		CIMInfo			*tume  = (CIMInfo *)iff->IFFData.Data;
		RoomType		*room  = tume->CurrentRoom;

		iff->Size = 0;
		size = IFFRead (iff, &CPALHeader, 5);
		if (!size) {
			goto cleanup;
		}

		LilWord2Native(CPALHeader.NumColors);
		LilWord2Native(CPALHeader.NumPal);

		numColors = CPALHeader.NumColors * CPALHeader.NumPal;
		
		cregs =	(ColorRegister *) IFFmalloc (numColors * sizeof (ColorRegister));
		if (!cregs)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCPAL:OOM color/palette");
			goto cleanup;
		}

		switch (CPALHeader.Type)
		{
		case 0:
			size += IFFRead (iff, cregs, numColors * sizeof (ColorRegister));
			break;
		case 1:
			{
				UBYTE	nesColors[17];
				int		i;

				size += IFFRead (iff, nesColors, 17);

				for (i = 0; i < 16; i++)
				{
					cregs[i] = CIMNesRefPal[(nesColors[i] & 0x0F) * 16 + ((nesColors[i] & 0xF0) >> 4)];
				}
			}
			break;
		default:
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadCIMLCPAL:Bad CPAL type");
			goto cleanup;
		}

		room->ColorInfo.NumColors = (UWORD) numColors;
		room->ColorInfo.Colors    = cregs;
		return size;

cleanup:
	if (cregs)	IFFfree (cregs);
	return FALSE;
	}
		
} /* ReadCIMLCPAL */

/*********************************************************************
 *
 * ReadCIMLCSET
 *
 * SYNOPSIS
 *		long ReadCIMLCSET (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLCSET (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLCSET";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		struct {
			UWORD	numChars;
			UBYTE	width;
			UBYTE	height;
			UBYTE	depth;
		} CSETHeader;
		unsigned		 tsize;
		unsigned		 readSize;
		unsigned		 readRowSize;
		unsigned		 tumeRowSize;
		long			 rsize;
		long			 size;
		UBYTE			*tbuf  = NULL;
		CIMInfo			*tume  = (CIMInfo *)iff->IFFData.Data;
		RoomType		*room  = tume->CurrentRoom;
		TilesetType		*ts    = tume->CurrentTileset;
		TMGType			*tmg   = &ts->TMG;
		TileDataGroup	*tdg;

		iff->Size = 0;

		size = IFFRead (iff, &CSETHeader, sizeof (CSETHeader));
		if (!size) {
			return FALSE;
		}

		LilWord2Native (CSETHeader.numChars);

		tmg->TileCount = CSETHeader.numChars;
		tmg->Width     = CSETHeader.width;
		tmg->Height    = CSETHeader.height;
		tmg->Depth     = CSETHeader.depth;

		switch (CSETHeader.depth)
		{
		case 2:
			ts->UserType = TILESET_2BITS;
			break;
		case 8:
			if (tume->MachineType == CED_MTYPE_SFXM7)
			{
				ts->UserType = TILESET_MODE7;
			}
			else
			{
				ts->UserType = TILESET_8BITS;
			}
			break;
		case 4:
		default:
			ts->UserType = TILESET_IMAGE;
			break;
		}

		ts->TileCount  = CSETHeader.numChars;
		ts->Width      = CSETHeader.width;
		ts->Height     = CSETHeader.height;
		ts->Depth      = CSETHeader.depth;

		tume->CharTileset = ts;

		ts->ID = ++tume->Map->NumTilesets;
		tume->RoomTSetID = ts->ID;
		AddTail (tume->TilesetList, ts);

		room->TileWidth  = CSETHeader.width;
		room->TileHeight = CSETHeader.height;

		tumeRowSize = (((tmg->Width + 15) / 16) * 2);
		tsize       = tumeRowSize * tmg->Height * tmg->Depth;
		readRowSize = ((tmg->Width + 7) / 8);
		readSize    = readRowSize * tmg->Height * tmg->Depth;

		tbuf  = (UBYTE *) IFFmalloc (readSize);
		if (!tbuf) {
			return FALSE;
		}

		if (iff->IFFData.Flags & TUME_TMGC) {
			size_t		 maxtilespergrp;
			unsigned	 tcount;
			unsigned	 tilesleft;
			UBYTE		*pdata;
			XTRAPntr	 xtra;
			UBYTE		*data;
			unsigned	 dsize;

			dsize          = tmg->Width * tmg->Height;
			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / dsize);
			tilesleft      = tmg->TileCount;

			/*
			** Allocate Plane Buffer.
			*/
			pdata = (UBYTE *) IFFmalloc (tsize);
			if (!pdata) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadTSETTMGX:OOM TMGX Buffer");
				return FALSE;
			}

			while (tilesleft) {
			
				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * dsize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData (2)");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, tbuf, readSize);
					if (rsize != readSize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGX:Error reading TMGX Tile");
						IFFfree (pdata);
						return FALSE;
					}

					size += rsize;

					{
						int		p;
						int		y;

						UBYTE	*t = pdata;
						UBYTE	*r = tbuf;
						for (p = 0; p < tmg->Depth; p++)
						{
							for (y = 0; y < tmg->Height; y++)
							{
								t[tumeRowSize - 1] = 0;
								memcpy (t, r, readRowSize);
								t += tumeRowSize;
								r += readRowSize;
							}
						}
					}

					if (!ConvertTMGXtoTMGC (pdata, tmg->Width, tmg->Height, tmg->Depth, data)) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGX:Error reading TMGX Tile 2");
						IFFfree (pdata);
						return FALSE;
					}

					data += dsize;
					tcount--;
				}
				UpdateXTRA (xtra);
			}

			IFFfree (pdata);		
		
		} else {
			size_t		 maxtilespergrp;
			unsigned	 tcount;
			unsigned	 tilesleft;
			XTRAPntr	 xtra;
			UBYTE		*data;

			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / tsize);
			tilesleft      = tmg->TileCount;

			while (tilesleft) {

				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * tsize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, tbuf, readSize);
					if (rsize != readSize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGC:Error reading TMGC Tile");
						return FALSE;
					}

					{
						int		p;
						int		y;

						UBYTE	*t = data;
						UBYTE	*r = tbuf;
						for (p = 0; p < tmg->Depth; p++)
						{
							for (y = 0; y < tmg->Height; y++)
							{
								t[tumeRowSize - 1] = 0;
								memcpy (t, r, readRowSize);
								t += tumeRowSize;
								r += readRowSize;
							}
						}
					}

					data += tsize;
					size += rsize;
					tcount--;
				}
				UpdateXTRA (xtra);
			}
		}

		IFFfree (tbuf);

		ts->FileSpec = IFFcalloc (1, EIO_MAXPATH);
		if (!ts->FileSpec) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadTSETDATA:OOM Tileset FileSpec");
			return FALSE;
		}

		EIO_fnmerge (ts->FileSpec, "", iff->Name, ".tst");


		return size;
	}
} /* ReadCIMLCSET */

/*********************************************************************
 *
 * ReadCIMLMSET
 *
 * SYNOPSIS
 *		long ReadCIMLMSET (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLMSET (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLMSET";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		struct {
			UWORD	numMatrices;
			UBYTE	width;
			UBYTE	height;
		} MSETHeader;
		UWORD			*mset  = NULL;
		long			 size;
		CIMInfo			*tume  = (CIMInfo *)iff->IFFData.Data;
		RoomType		*room  = tume->CurrentRoom;
		RoomType		*newrm = NULL;
		TilesetType		*ts    = tume->CurrentTileset;
		TilesetType		*newts = NULL;
		int				 msetSize;
		int				 msetRowSize;
		int				 msetCount;
		int				 l;
		UBYTE			 tsID;

		iff->Size = 0;

		size = IFFRead (iff, &MSETHeader, sizeof (MSETHeader));
		if (!size) {
			return FALSE;
		}

		LilWord2Native (MSETHeader.numMatrices);

		newts = IFFcalloc (1, sizeof (TilesetType));
		newrm = IFFcalloc (1, sizeof (RoomType));

		if (!newts || !newrm) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMSET:OOM comp tileset, comp room");
			return FALSE;
		}

		newts->Width            = (WORD) (ts->Width  * MSETHeader.width);
		newts->Height           = (WORD) (ts->Height * MSETHeader.height);
		newts->TilesAcross      = MSETHeader.width;
		newts->TilesDown        = MSETHeader.height;
		newts->GridedTilesRoom  = newrm;
		newts->TileCount        = MSETHeader.numMatrices;
		newts->Flags           |= TUME_TILESET_COMPOSITE;
		newts->UserType         = ts->UserType;

		tume->MatrixTileset = newts;

		tsID = tume->RoomTSetID;

		newts->ID = ++tume->Map->NumTilesets;
		tume->RoomTSetID = newts->ID;
		AddTail (tume->TilesetList, newts);

		room->TileWidth  = newts->Width;
		room->TileHeight = newts->Height;

		newrm->CompositeTSet = newts;
		newrm->Width  = (WORD) (MSETHeader.width * 16);
		newrm->Height = (WORD) (((MSETHeader.numMatrices + 15) / 16) * MSETHeader.height);
		newrm->TileWidth  = ts->Width;
		newrm->TileHeight = ts->Height;
		newrm->LayerCount = 1;

		newrm->Layers = IFFcalloc (newrm->LayerCount, sizeof (LayerType *));
		if (!newrm->Layers)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMSET:OOM LayerTable");
			goto cleanup;
		}

		tume->Map->NumRooms++;
		AddTail (tume->RoomList, newrm);

		msetSize    = newts->TilesAcross * newts->TilesDown * sizeof (long);
		msetRowSize = msetSize * 16;
		msetCount   = MSETHeader.numMatrices;
		mset = IFFmalloc (msetRowSize);

		for (l = 0; l < newrm->LayerCount; l++) {
			LayerType	*layer;
			unsigned	 row;

			layer = IFFcalloc (1, sizeof (LayerType));
			if (!layer) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadCIMLMSET:OOM LayerType");
				goto cleanup;
			}

			newrm->Layers[l] = layer;
			layer->Plot = IFFcalloc (newrm->Height, sizeof (XTRAPntr));
			if (!layer->Plot) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadCIMLMSET:OOM Layer->Plot");
				goto cleanup;
			}

			for (row = 0; row < newrm->Height; row += newts->TilesDown) {
				int		subrow;

				memset (mset, 0, msetRowSize);
				{
					long	 msize;
					long	 rsize;
					UWORD	*data = mset;
					int		 i;

					msize      = min (16, msetCount);
					msetCount -= (int)msize;
					msize     *= msetSize;

					rsize = IFFRead (iff, mset, msize);
					if (rsize != msize)
					{
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadROOMDATA:Error reading ROOM DATA");
						goto cleanup;
					}
					size += rsize;

					for (i = 0; i < msize / 2; i++)
					{
						LilWord2Native (*data);
						data++;
					}

				}		

				for (subrow = 0; subrow < newts->TilesDown; subrow++)
				{
					XTRAPntr	 plot;
					PlotType	*pt;
					UWORD		*data = mset + subrow * (newts->TilesAcross * 2);
					int			 i;
					int			 x;

					plot = AllocXTRA (16 * newts->TilesAcross * sizeof (PlotType));
					if (!plot) {
						SetGlobalErr (ERR_OUT_OF_MEMORY);
						GEcatf ("\nReadROOMDATA:OOM Room Row");
						goto cleanup;
					}

					layer->Plot[row + subrow] = plot;

					pt = ActivateXTRA (plot);

					for (i = 0; i < 16; i++)
					{
						for (x = 0; x < newts->TilesAcross; x++)
						{
							pt->TilesetID = tsID;
							pt->TileID    = (UWORD) (1 + *data++);
							pt->PlotFlags = (UBYTE)
											(
												0x80 + (*data & 0x0F) +
												((*data & 0x0100U) ? 0x40 : 0) +
	 											((*data & 0x0200U) ? 0x20 : 0)
											);
							data++;
							pt++;
						}
						data += ((msetSize / 2) - (newts->TilesAcross * 2));
					}
					UpdateXTRA (plot);
				}
			}
		}

		newrm->Name = IFFmalloc (EIO_MAXPATH);
		newts->FileSpec = IFFmalloc (EIO_MAXPATH);
		if (!newrm->Name || !newts->FileSpec) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMDATA:OOM ROOM Name");
			goto cleanup;
		}
		EIO_fnmerge (newrm->Name, NULL, iff->Name, ".mtx");
		strcpy (newts->FileSpec, newrm->Name);

		IFFfree (mset);

		return size;
cleanup:
		if (mset) IFFfree (mset);
		return FALSE;
	}
} /* ReadCIMLMSET */

/*********************************************************************
 *
 * ReadCIMLCATR
 *
 * SYNOPSIS
 *		long ReadCIMLCATR (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLCATR (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLCATR";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		UWORD			*attr;
		long			 size;
		long			 asize = iff->Size;
		int				 attrCount;
		CIMInfo			*tume  = (CIMInfo *)iff->IFFData.Data;
		TilesetType		*newts = NULL;
		RoomType		*newrm = NULL;
		LayerType		*layer;
		PlotType		*pt;
		XTRAPntr		 plot;
		int				 x;

		attr = IFFmalloc (asize);
		if (!attr) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM");
			return FALSE;
		}
		attrCount = (int)(asize / 2);

		iff->Size = 0;
		size = IFFRead (iff, attr, asize);

		newts = IFFcalloc (1, sizeof (TilesetType));
		newrm = IFFcalloc (1, sizeof (RoomType));

		if (!newts || !newrm) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM tileset, room");
			return FALSE;
		}

		newts->Width     = tume->CharTileset->Width;
		newts->Height    = tume->CharTileset->Height;
		newts->TileCount = 256;
		newts->UserType	 = TILESET_SPECIAL;

		newts->ID = ++tume->Map->NumTilesets;
		AddTail (tume->TilesetList, newts);

		newrm->Width       = (WORD) attrCount;
		newrm->Height      = 2;
		newrm->TileWidth   = newts->Width;
		newrm->TileHeight  = newts->Height;
		newrm->LayerCount  = 1;
		newrm->UserType    = ROOM_FLAT_CONVERSION;
		newrm->GuideWidth  = 16;
		newrm->GuideHeight = 2;
		
		newrm->Layers = IFFcalloc (newrm->LayerCount, sizeof (LayerType *));
		if (!newrm->Layers)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM LayerTable");
			goto cleanup;
		}

		tume->Map->NumRooms++;
		AddTail (tume->RoomList, newrm);

		layer = IFFcalloc (1, sizeof (LayerType));
		if (!layer) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM LayerType");
			goto cleanup;
		}
		
		newrm->Layers[0] = layer;
		layer->Plot = IFFcalloc (newrm->Height, sizeof (XTRAPntr));
		if (!layer->Plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM Layer->Plot");
			goto cleanup;
		}

		//
		// fill first row with tiles from Matrix tileset
		//
		plot = AllocXTRA (sizeof (PlotType) * newrm->Width);
		if (!plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM Room Row");
			goto cleanup;
		}

		layer->Plot[0] = plot;

		pt = ActivateXTRA (plot);
		for (x = 0; x < newrm->Width; x++)
		{
			pt->TileID    = (UWORD) (x + 1);
			pt->TilesetID = tume->CharTileset->ID;
			pt->PlotFlags = 0;
			pt++;
		}
		UpdateXTRA (plot);

		//
		// fill second row with attributes
		//
		plot = AllocXTRA (sizeof (PlotType) * newrm->Width);
		if (!plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLCATR:OOM Room Row");
			goto cleanup;
		}

		layer->Plot[1] = plot;

		pt = ActivateXTRA (plot);
		for (x = 0; x < newrm->Width; x++)
		{
			pt->TileID    = (UWORD) (attr[x] + 1);
			pt->TilesetID = newts->ID;
			pt->PlotFlags = 0;
			pt++;
		}
		UpdateXTRA (plot);

		newrm->Name = IFFmalloc (EIO_MAXPATH);
		newts->FileSpec = IFFmalloc (EIO_MAXPATH);
		if (!newrm->Name || !newts->FileSpec) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMDATA:OOM ROOM Name");
			goto cleanup;
		}
		EIO_fnmerge (newrm->Name, NULL, iff->Name, ".ccv");
		strcpy (newts->FileSpec, newrm->Name);

		return size;
cleanup:
		return FALSE;
	}
} /* ReadCIMLCATR */

/*********************************************************************
 *
 * ReadCIMLMATR
 *
 * SYNOPSIS
 *		long ReadCIMLMATR (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLMATR (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLMATR";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		UWORD			*attr;
		long			 size;
		long			 asize = iff->Size;
		int				 attrCount;
		CIMInfo			*tume  = (CIMInfo *)iff->IFFData.Data;
		TilesetType		*newts = NULL;
		RoomType		*newrm = NULL;
		LayerType		*layer;
		PlotType		*pt;
		XTRAPntr		 plot;
		int				 x;

		attr = IFFmalloc (asize);
		if (!attr) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM");
			return FALSE;
		}
		attrCount = (int)(asize / 2);

		iff->Size = 0;
		size = IFFRead (iff, attr, asize);

		newts = IFFcalloc (1, sizeof (TilesetType));
		newrm = IFFcalloc (1, sizeof (RoomType));

		if (!newts || !newrm) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM tileset, room");
			return FALSE;
		}

		newts->Width     = tume->MatrixTileset->Width;
		newts->Height    = tume->MatrixTileset->Height;
		newts->TileCount = 256;
		newts->UserType	 = TILESET_SPECIAL;

		newts->ID = ++tume->Map->NumTilesets;
		AddTail (tume->TilesetList, newts);

		newrm->Width       = (WORD) attrCount;
		newrm->Height      = 2;
		newrm->TileWidth   = newts->Width;
		newrm->TileHeight  = newts->Height;
		newrm->LayerCount  = 1;
		newrm->UserType    = ROOM_FLAT_CONVERSION;
		newrm->GuideWidth  = 16;
		newrm->GuideHeight = 2;
		
		newrm->Layers = IFFcalloc (newrm->LayerCount, sizeof (LayerType *));
		if (!newrm->Layers)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM LayerTable");
			goto cleanup;
		}

		tume->Map->NumRooms++;
		AddTail (tume->RoomList, newrm);

		layer = IFFcalloc (1, sizeof (LayerType));
		if (!layer) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM LayerType");
			goto cleanup;
		}
		
		newrm->Layers[0] = layer;
		layer->Plot = IFFcalloc (newrm->Height, sizeof (XTRAPntr));
		if (!layer->Plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM Layer->Plot");
			goto cleanup;
		}

		//
		// fill first row with tiles from Matrix tileset
		//
		plot = AllocXTRA (sizeof (PlotType) * newrm->Width);
		if (!plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM Room Row");
			goto cleanup;
		}

		layer->Plot[0] = plot;

		pt = ActivateXTRA (plot);
		for (x = 0; x < newrm->Width; x++)
		{
			pt->TileID    = (UWORD) (x + 1);
			pt->TilesetID = tume->MatrixTileset->ID;
			pt->PlotFlags = 0;
			pt++;
		}
		UpdateXTRA (plot);

		//
		// fill second row with attributes
		//
		plot = AllocXTRA (sizeof (PlotType) * newrm->Width);
		if (!plot) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLMATR:OOM Room Row");
			goto cleanup;
		}

		layer->Plot[1] = plot;

		pt = ActivateXTRA (plot);
		for (x = 0; x < newrm->Width; x++)
		{
			pt->TileID    = (UWORD) (attr[x] + 1);
			pt->TilesetID = newts->ID;
			pt->PlotFlags = 0;
			pt++;
		}
		UpdateXTRA (plot);

		newrm->Name = IFFmalloc (EIO_MAXPATH);
		newts->FileSpec = IFFmalloc (EIO_MAXPATH);
		if (!newrm->Name || !newts->FileSpec) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMDATA:OOM ROOM Name");
			goto cleanup;
		}
		EIO_fnmerge (newrm->Name, NULL, iff->Name, ".mcv");
		strcpy (newts->FileSpec, newrm->Name);

		return size;
cleanup:
		return FALSE;
	}
} /* ReadCIMLMATR */

/*********************************************************************
 *
 * ReadCIMLIFRM
 *
 * SYNOPSIS
 *		long ReadCIMLIFRM (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMLIFRM (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMLIFRM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		struct {
			UWORD	frameID;
			UWORD	pxlWidth;
			UWORD	pxlHeight;
			UWORD	xOffset;
			UWORD	yOffset;
		} IFRMHeader;
		long		 size;
		unsigned	 rsize;
		UWORD		*data;
		CIMInfo		*tume = (CIMInfo *)iff->IFFData.Data;
		RoomType	*room = tume->CurrentRoom;
		int			 l;
		UBYTE 		 tsID;
		UBYTE		 tileColorMask;

		switch (tume->MachineType)
		{
		case CED_MTYPE_NES:
			tileColorMask = 0x00;
			break;
		default:
			tileColorMask = 0xFF;
			break;
		}

		tsID = tume->RoomTSetID;

		iff->Size = 0;

		size = IFFRead (iff, &IFRMHeader, sizeof (IFRMHeader));
		if (!size) {
			goto cleanup;
		}

		LilWord2Native (IFRMHeader.frameID);
		LilWord2Native (IFRMHeader.pxlWidth);
		LilWord2Native (IFRMHeader.pxlHeight);
		LilWord2Native (IFRMHeader.xOffset);
		LilWord2Native (IFRMHeader.yOffset);

		room->Width  = (WORD) (IFRMHeader.pxlWidth  / room->TileWidth);
		room->Height = (WORD) (IFRMHeader.pxlHeight / room->TileHeight);
		room->LayerCount = 1;
		tume->Map->NumRooms++;
		AddTail (tume->RoomList, room);

		rsize = room->Width * sizeof (PlotType);

		data = IFFmalloc (rsize);

		room->Layers = IFFcalloc (room->LayerCount, sizeof (LayerType *));
		if (!room->Layers)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMLIFRM:OOM LayerTable");
			goto cleanup;
		}

		for (l = 0; l < room->LayerCount; l++) {
			LayerType	*layer;
			unsigned	 row;
			long		 tsize;

			layer = IFFcalloc (1, sizeof (LayerType));
			if (!layer) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadCIMLIFRM:OOM LayerType");
				goto cleanup;
			}

			room->Layers[l] = layer;
			layer->Plot = IFFcalloc (room->Height, sizeof (XTRAPntr));
			if (!layer->Plot) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadCIMLIFRM:OOM Layer->Plot");
				goto cleanup;
			}

			for (row = 0; row < room->Height; row++) {
				XTRAPntr	 plot;
				PlotType	*pt;

				plot = AllocXTRA (rsize);
				if (!plot) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadCIMLIFRM:OOM Room Row");
					goto cleanup;
				}

				layer->Plot[row] = plot;

				pt    = ActivateXTRA (plot);
				tsize = IFFRead (iff, data, rsize);
				if (rsize != tsize) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMDATA:Error reading ROOM DATA");
					goto cleanup;
				}

				size += tsize;

				{
					unsigned	 ndx;
					UWORD		*d;

					d = data;
					for (ndx = 0; ndx < room->Width; ndx++) {
						LilWord2Native (*d);
						pt->TileID    = (UWORD) (1 + *d++);
						pt->TilesetID = tsID;
						LilWord2Native (*d);
						pt->PlotFlags = (UBYTE)
										(
											((0x80 + (*d & 0x0F)) & tileColorMask) +
											((*d & 0x0100U) ? 0x40 : 0) +
	 										((*d & 0x0200U) ? 0x20 : 0)
										);
						d++;
						pt++;
					}
				}
				UpdateXTRA (plot);
			}
		}

		room->Name = IFFcalloc (1, EIO_MAXPATH);
		if (!room->Name) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMDATA:OOM ROOM Name");
			goto cleanup;
		}
		EIO_fnmerge (room->Name, NULL, iff->Name, NULL);

		IFFfree (data);

		return size;
cleanup:
		if (data) IFFfree (data);
		return FALSE;
	}
} /* ReadCIMLIFRM */

/*********************************************************************
 *
 * ReadCIML
 *
 * SYNOPSIS
 *		long ReadCIML (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIML (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIML";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	CIMLTable[] = {
		//	{ LHDR_CHUNK,	ReadCIMLLHDR,	},
			{ CPAL_CHUNK,	ReadCIMLCPAL,	},
			{ CSET_CHUNK,	ReadCIMLCSET,	},
			{ MSET_CHUNK,	ReadCIMLMSET,	},
			{ CATR_CHUNK,	ReadCIMLCATR,	},
			{ MATR_CHUNK,	ReadCIMLMATR,	},
			{ IFRM_CHUNK,	ReadCIMLIFRM,	},
			{ FALSE,		NULL,			},
		};

		long		 size = iff->Size;
		long		 okay = TRUE;
		long		 tsize;
		ChunkHeader	 chunkhdr;
		RoomType	*room;
		TilesetType	*ts;
		CIMInfo		*ti = iff->IFFData.Data;
		
		ts = CreateNode (sizeof (TilesetType), NULL);
		if (!ts) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIML:OOM TilesetType");
			return FALSE;
		}

		ti->CurrentTileset = ts;
		ts->TMG.TileGroupList = &ts->TMG.TileGroupListX;
		InitList (ts->TMG.TileGroupList);

		room = CreateNode (sizeof (RoomType), NULL);
		if (!room) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIML:OOM RoomType");
			return FALSE;
		}

		ti->CycleList = &ti->CycleListX;
		InitList (ti->CycleList);

		ti->CurrentRoom = room;

		iff->Size = 0;
		while (iff->Size < size && okay) {
			if (!GetChunkHeader (iff, &chunkhdr)) {
				return FALSE;
			}

			tsize      = iff->Size;
			iff->Size  = chunkhdr.Size;
			okay       = CIMCallIDFunc (chunkhdr.ID, CIMLTable, iff);
			if (okay != chunkhdr.Size)
			{
				SetGlobalErr (ERR_GENERIC);
				GEcatf1 ("\nReadCIML:'%4.4s' chunk size and data do not match", &chunkhdr.ID);
/**/			goto cleanup;
			}
				
			tsize     += okay;
			iff->Size  = tsize;
		}

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadCIML:CIML form size and data do not match");
/**/		goto cleanup;
		}

		if (room->ColorInfo.NumCycles) {
			room->ColorInfo.Cycles = IFFmalloc (room->ColorInfo.NumCycles * sizeof (CycleInfo *));
			if (!room->ColorInfo.Cycles) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadCIML:OOM ColorInfo.Cycle Array");
				goto cleanup;
			}

			{
				CycleInfo	*cycle;
				CycleInfo	**cycleptr;

				cycleptr = room->ColorInfo.Cycles;
				while ((cycle = RemHead (ti->CycleList)) != NULL) {
					*cycleptr = cycle;
					cycleptr++;
				}
			}
		}

		ti->CurrentRoom = NULL;
		ti->CurrentTileset = NULL;

		return size;

cleanup:
		EmptyList (ti->CycleList);

		return FALSE;

	}
} /* ReadCIML */

/*********************************************************************
 *
 * ReadCIMHHHDR
 *
 * SYNOPSIS
 *		long ReadCIMHHHDR (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadCIMHHHDR (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMHHHDR";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		struct {
			UWORD	 numLayers;
			UBYTE	 layeringType;
			UBYTE	 machineType;
		} HHDRHeader;
		CIMInfo		*tume = (CIMInfo *)iff->IFFData.Data;
		long		 size;
		
		iff->Size = 0;
		size = IFFRead (iff, &HHDRHeader, sizeof (HHDRHeader));
		
		tume->MachineType = HHDRHeader.machineType;

		return size;
	}

} /* ReadCIMHHHDR */

/*********************************************************************
 *
 * ReadCIMH
 *
 * SYNOPSIS
 *		long ReadCIMH (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMH (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMH";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		//
		// NOTE: No relavent chunks in CIMH
		//
		static IFFRoutineID	CIMHTable[] = {
			{ HHDR_CHUNK,	ReadCIMHHHDR,	},
			{ FALSE,		NULL,			},
		};

		long		 size = iff->Size;
		long		 okay = TRUE;
		long		 tsize;
		ChunkHeader	 chunkhdr;

		iff->Size = 0;

		while (iff->Size < size && okay) {
			if (!GetChunkHeader (iff, &chunkhdr)) {
				return FALSE;
			}

			tsize      = iff->Size;
			iff->Size  = chunkhdr.Size;
			okay = CIMCallIDFunc (chunkhdr.ID, CIMHTable, iff);
			if (okay != chunkhdr.Size)
			{
				SetGlobalErr (ERR_GENERIC);
				GEcatf1 ("\nReadCIMH:'%4.4s' chunk size and data do not match", &chunkhdr.ID);
/**/			goto cleanup;
			}
			tsize     += okay;
			iff->Size  = tsize;
		}

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadCIMH:CIMH form size and data do not match");
/**/		goto cleanup;
		}

		return size;

cleanup:
		return FALSE;

	}
} /* ReadCIMH */

/*********************************************************************
 *
 * ReadCIMForms
 *
 * SYNOPSIS
 *		long ReadCIMForms (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
LOCAL long ReadCIMForms (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCIMForms";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	CIMTable[] = {
			{ CIMH_SIFF,	ReadCIMH,	},
			{ CIML_SIFF,	ReadCIML,	},
			{ FALSE,		NULL,		},
		};

		long			 size = iff->Size;
		long			 okay = TRUE;
		long			 tsize;
		IFFRoutineID	*iffrtn;
		CIMInfo			*ti;
		MapType			*map = iff->IFFData.Data;

		iff->Size = 0;

		ti  = IFFcalloc (1, sizeof (CIMInfo));
		if (!map) {
			map = IFFcalloc (1, sizeof (MapType));
		}

		if (!ti || !map) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMForms:OOM CIMInfo/MapType");
			return FALSE;
		}

		iff->IFFData.Data = ti;
		ti->Map           = map;
		ti->RoomList      = &ti->RoomListX;
		ti->TilesetList   = &ti->TilesetListX;
		InitList (ti->RoomList);
		InitList (ti->TilesetList);

		/*
		** Add OLD Rooms and Tileset to CIMInfo
		*/
		{
			unsigned	 i;
			RoomType	**roomptr;
			TilesetType	**tsptr;

			roomptr = map->Rooms + 1;
			for (i = 0; i < map->NumRooms; i++) {
				(*roomptr)->LoadFlags = LOADFLAG_OLDROOM;
				AddTail (ti->RoomList, *roomptr);
				roomptr++;
			}

			tsptr = map->Tilesets + 1;
			for (i = 0; i < map->NumTilesets; i++) {
				AddTail (ti->TilesetList, *tsptr);
				tsptr++;
			}
		}


		iffrtn = iff->FORMTable;
		iff->FORMTable = CIMTable;

		while (iff->Size < size && okay) {
			tsize      = iff->Size;
			okay       = ReadSIFF (iff);
			tsize     += okay;
			iff->Size  = tsize;
		}

		iff->FORMTable = iffrtn;

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadCIMForms:CIM form size and data do not match");
/**/		goto cleanup;
		}

		/*
		**	Renumber New Room Data
		*/
		#if 0
		{
			RoomType	*room;

			room = Head (ti->RoomList);
			while (!EndOfList (room)) {
				if (!(room->LoadFlags & LOADFLAG_OLDROOM)) {
					unsigned	l;

					for (l = 0; l < room->LayerCount; l++) {
						LayerType	*layer;
						unsigned	 row;

						layer = room->Layers[l];
						for (row = 0; row < room->Height; row++) {
							PlotType	*plot;
							unsigned	 ndx;

							plot = ActivateXTRA (layer->Plot[row]);
							for (ndx = 0; ndx < room->Width; ndx++) {
								plot->TilesetID = ti->TranslateTable[plot->TilesetID];
								plot++;
							}
							UpdateXTRA (layer->Plot[row]);
						}
					}
				}
				room = Next (room);
			}
		}
		#endif

		#if 0
		/*
		** Connect Composite Tilesets to their Rooms.
		*/
		{
			TilesetType	*ts;

			ts = Head (ti->TilesetList);
			while (!EndOfList (ts)) {
				if (ts->Flags & TUME_TILESET_COMPOSITE) {
					RoomType	*room;
					RoomType	*nextroom;
					BOOL		 found = FALSE;

					room = Head (ti->RoomList);
					while (!EndOfList (room)) {
						nextroom = Next (room);
						if (!strcmp (room->Name, ts->FileSpec)) {
							if (!found) {
								found = TRUE;
								room->CompositeTSet = ts;
								ts->GridedTilesRoom = room;
								ts->TilesAcross = ts->Width  / room->TileWidth;
								ts->TilesDown   = ts->Height / room->TileHeight;
							} else {
								Remove (room);
								DestroytUMERoom (room);
							}
						}
						room = nextroom;
					}
				}
				ts = Next (ts);
			}
		}
		#endif

		/*
		**	Create Room Array and Tileset Array
		*/
		if (map->Rooms)		IFFfree (map->Rooms);
		if (map->Tilesets)	IFFfree (map->Tilesets);

		map->Rooms    = IFFcalloc (map->NumRooms + 2, sizeof (RoomType *));
		map->Tilesets = IFFcalloc (map->NumTilesets + 2, sizeof (TilesetType *));
		if (!map->Rooms || !map->Tilesets) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadCIMForms:OOM CIMInfo/MapType");
			goto cleanup;
		}

		{
			RoomType	*room;
			RoomType	**roomptr;
			TilesetType	*ts;
			TilesetType	**tsptr;

			roomptr = map->Rooms + 1;
			room    = Head (ti->RoomList);
			while (!EndOfList (room)) {
				*roomptr = room;
				roomptr++;
				room = Next (room);
			}

			tsptr = map->Tilesets + 1;
			ts    = Head (ti->TilesetList);
			while (!EndOfList (ts)) {
				*tsptr = ts;
				tsptr++;
				ts = Next (ts);
			}
		}

		iff->IFFData.Data = map;
		IFFfree (ti);

		return size;

cleanup:
/*		FreeCIMInfo (iff->IFFData.Data); */
		return FALSE;
		
	}
} /* ReadCIMForms */

/*********************************************************************
 *
 * ReadCIM
 *
 * SYNOPSIS
 *		long ReadCIM (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadCIM (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCED";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		#define REF_PALETTE	"npicpal.lbm"
		if (!CIMNesRefPal)
		{
			char		 palname[EIO_MAXPATH];
			ILBMInfo	*refPalILBM;

			if (!EIO_FindFile (REF_PALETTE, REF_PALETTE, palname)) {
				SetGlobalErr (ERR_FILE_NOT_FOUND);
				GEcatf ("\nReadCIM:Couldn't find Reference Palette " REF_PALETTE "\n");
				return 0;
			}
			refPalILBM = CreatePalette (palname);
			if (!refPalILBM) {
				SetGlobalErr (ERR_GENERIC);
				GEcatf1 ("Couldn't load Reference Palette '%s\n", palname);
				return 0;
			}
			if (refPalILBM->NumberOfColors < 256) {
				SetGlobalErr (ERR_GENERIC);
				GEcatf1 ("Less than 256 colors in reference palette '%s'\n", palname);
			}

			CIMNesRefPal = refPalILBM->Colors;
		}

	
		iff->Size = EIO_FileLength (iff->fd);
		return ReadCIMForms (iff);
	}
} /* ReadCIM */

