/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * READILBM.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 05/07/90
 *   MODIFIED : 06/19/00
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Routines to read ILBM and PBM files.
 *
 * HISTORY
 *		12/04/91 (dcc) - WriteXTRA replaces some UpdateXTRA's
 *		01/16/92 (dcc) - Check for alloc's returning NULL
 *						 free temporarily allocated memory
 *		01/18/92 (dcc) - All ActivateXTRA's & UpdateXTRA's removed
 *		09/21/94 (dcc) - Compiles cleanly with 32-bit Borland C++ 4.02.
 *		02/02/95 (dcc) - Add #include <echidna/memory.h> for
 *						 prototype for AllocMemOnPara().
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <echidna/etypes.h>
#include <echidna/eio.h>
#include <echidna/emacros.h>
#include <echidna/eerrors.h>
#include <echidna/memory.h>
#include <echidna/readiff.h>
#include <echidna/bilbm.h>
#include "pcx.h"
#include "szerror.h"

/**************************** C O N S T A N T S ***************************/

#define PBM_DEPTH	8

/******************************** T Y P E S *******************************/

typedef struct {
   ULONG ViewModes;
   } CamgChunk;

typedef struct {
   WORD  direction;  /* 0=don't cycle, 1=forward, -1=backwards */
   UBYTE start;      /* range lower */
   UBYTE end;        /* range upper */
   LONG  seconds;    /* seconds between cycling */
   LONG  microseconds; /* msecs between cycling */
   WORD  pad;        /* future exp - store 0 here */
   } CcrtChunk;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/

/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_FreeBigByteMap                                       */
/*------------------------------------------------------------------------*/


/**************************************************************************
 *
 * FreeBigByteMap
 *
 * PURPOSE
 *		Frees actual memory used by pixel plot array.
 *
 * USAGE
 *
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/

void FreeBigByteMap (MPYPMPXPPIX mpYpMpXpPix, int height)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeBigByteMap";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/


	if (mpYpMpXpPix)
	{
		int iyt;

		for (iyt = 0; iyt < height; iyt++)
		{
			if (mpYpMpXpPix[iyt])
			{
				FreeXTRA(mpYpMpXpPix[iyt]);
			}
		}

		IFFfree(mpYpMpXpPix);
	}

} /* FreeBigByteMap */


/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_AllocateBigByteMap                                   */
/*------------------------------------------------------------------------*/


/**************************************************************************
 *
 * AllocateBigByteMap
 *
 * PURPOSE
 *		Allocates actual memory used by pixel plot array.
 *
 * USAGE
 *
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/

MPYPMPXPPIX AllocateBigByteMap(int width, int height)
{
	MPYPMPXPPIX	mpYpMpXpPix;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AllocateBigByteMap";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/


	if (width != 0 && height != 0)
	{
		int		  iyt;
		UBYTE	 *pPix;

		if ((mpYpMpXpPix = (MPYPMPXPPIX) IFFcalloc(height, sizeof(MPXPPIX))) == NULL)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMPixRow);
			return NULL;
		}
		for (iyt = 0; iyt < height; iyt++)
		{
			if ((mpYpMpXpPix[iyt] = AllocXTRA(width)) == NULL)
			{
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf (szerrOOXMPixel);
/**/			goto ABORT;
			}
		}

		if ((pPix = (UBYTE *) IFFmalloc(width)) == NULL)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMRowBuf);
/**/		goto ABORT;
		}
		memset(pPix, 0, width);

		for (iyt = 0; iyt < height; iyt++)
		{
			WriteXTRA(pPix, width, mpYpMpXpPix[iyt], 0);
		}
		IFFfree(pPix);

		return mpYpMpXpPix;
	}
	else
	{
		return NULL;
	}

/*========================================================================*/
ABORT:

	FreeBigByteMap(mpYpMpXpPix, height);

	return NULL;

} /* AllocateBigByteMap */


/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadBODY                                             */
/*------------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * unpacker.c Convert data from "cmpByteRun1" run compression. 11/15/85
 *
 * By Jerry Morrison and Steve Shaw, Electronic Arts.
 * This software is in the public domain.
 *
 *	control bytes:
 *	 [0..127]   : followed by n+1 bytes of data.
 *	 [-1..-127] : followed by byte to be repeated (-n)+1 times.
 *	 -128       : NOOP.
 *
 * HISTORY
 *		09/21/94 (dcc) - remove 68000 optimizations. Let compiler do it.
 *
 *----------------------------------------------------------------------*/

/*----------- UnPackRow ------------------------------------------------*/
#define MaxPackedSize(rowSize)  ( (rowSize) + ( ((rowSize)+127) >> 7 ) )

#define UGetByte()	(*source++)
#define UPutByte(c)	(*dest++ = (c))

/* Given POINTERS to POINTER variables, unpacks one row, updating the source
 * and destination pointers until it produces dstBytes bytes. */
static BOOL UnPackRow(BYTE **pSource, BYTE **pDest, int srcBytes0, int dstBytes0)
{
    register BYTE *source = *pSource;
    register BYTE *dest   = *pDest;
    register int n;
    register BYTE c;
    register int srcBytes = srcBytes0, dstBytes = dstBytes0;
    BOOL error = TRUE;	/* assume error until we make it through the loop */

    while( dstBytes > 0 )
	{
		if ( (srcBytes -= 1) < 0 )  goto ErrorExit;
    	n = UGetByte();

    	if (n >= 0)
		{
		    n += 1;
		    if ( (srcBytes -= n) < 0 )  goto ErrorExit;
		    if ( (dstBytes -= n) < 0 )  goto ErrorExit;
		    do
			{
				UPutByte(UGetByte());
			} while (--n > 0);
	    }
    	else if (n != -128)
		{
		    n = -n + 1;
		    if ( (srcBytes -= 1) < 0 )  goto ErrorExit;
		    if ( (dstBytes -= n) < 0 )  goto ErrorExit;
		    c = UGetByte();
		    do
			{
				UPutByte(c);
			} while (--n > 0);
	    }
	}
    error = FALSE;	/* success! */

ErrorExit:
    *pSource = source;  *pDest = dest;
    return(error);
}


/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/
/************************************  ************************************/


/*********************************************************************
 *
 * ReadILBMBODY
 *
 * SYNOPSIS
 *		long ReadILBMBODY (IFFTracker *iff, PixelData *pd)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadILBMBODY (IFFTracker *iff, BitMapHeader *bmHdr, PixelData *pd)
{

	int				 srcPlaneCnt = bmHdr->nPlanes;
	UBYTE			 createbmap  = (UBYTE)(iff->Flags & ILBM_CREATEBYTEMAP);
	int				 srcRowBytes = RowBytes(bmHdr->w);
	LONG			 bufRowBytes = MaxPackedSize(srcRowBytes);
	LONG			 okay;
	LONG			 size;
	LONG			 result		 = FALSE;
	LONG			 bufsize;
	int				 nRows       = bmHdr->h;
	Compression		 compression = bmHdr->compression;
	register int	 iPlane, iRow;
	register LONG	 nEmpty;
	register int	 nFilled;
	BYTE			*buf;
	BYTE			*nullDest	 = NULL;
	BYTE			*nullBuf;
	BYTE			**pDest;
	BYTE			*planes[MAXSRCPLANES] = { NULL, };	/* array of ptrs to planes & mask */
	UBYTE			*workp[MAXSRCPLANES];	/* array of ptrs to planes & mask */
	BYTE			*gDest;
	BYTE			*mask		 = NULL;
	BYTE			*buffer		 = NULL;

	UBYTE			*pDataTmp;
	BYTE			*pPlaneTmp;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadILBMBODY";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	size      = iff->Size;
	iff->Size = 0;

	if (compression > cmpByteRun1) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf1 (szerrIUnknownComp, compression);
		return FALSE;
	}

	if (createbmap) {
		pd->BigByteMap.width  = bmHdr->w;
		pd->BigByteMap.height = (WORD) nRows;
		if ((pd->BigByteMap.mpYpMpXpPix = AllocateBigByteMap(bmHdr->w, bmHdr->h)) == NULL)
			return FALSE;
	} else {
		pd->BigBitMap.BytesPerRow = (UWORD) srcRowBytes;
		pd->BigBitMap.Rows        = (UWORD) nRows;
		pd->BigBitMap.Depth       = (UBYTE) srcPlaneCnt;
	}

	bufsize = srcRowBytes * 3;
	if ((buffer  = (BYTE *) IFFmalloc (bufsize)) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrIOOMWorkBuf);
/**/	goto RIBcleanup;
	}

	/* Initialize array "planes" with bitmap ptrs; NULL in empty slots.*/
	for (iPlane = 0; iPlane < bmHdr->nPlanes; iPlane++) {
		if (createbmap) {
			if ((planes[iPlane] = (BYTE *) IFFmalloc (srcRowBytes)) == NULL) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf (szerrOOMPlanes);
/**/			goto RIBcleanup;
			}
		} else {
			if ((pd->BigBitMap.xPlanes[iPlane] = AllocXTRA (nRows * srcRowBytes)) == NULL) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf (szerrOOXMPlanes);
/**/			goto RIBcleanup;
			}
		}
	}
	for (; iPlane < MAXSRCPLANES;  iPlane++) {
		planes[iPlane] = NULL;
		if (!createbmap) {
			pd->BigBitMap.xPlanes[iPlane] = NULL;
		}
		workp[iPlane]  = NULL;
	}

	/* Copy any mask plane ptr into corresponding "planes" slot.*/
	if (bmHdr->masking == mskHasMask) {
		if ((mask = (BYTE *) IFFGmalloc (srcRowBytes * nRows)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMMasks);
/**/		goto RIBcleanup;
		}
		srcPlaneCnt += 1;  /* Include mask plane in count.*/
	}

	/* Setup a sink for dummy destination of rows from unwanted planes.*/
	nullDest = buffer;
	buffer  += srcRowBytes;
	bufsize -= srcRowBytes;

	if (createbmap) {
		if ((pDataTmp = (UBYTE *) IFFmalloc (bmHdr->w)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMPixBuf);
/**/		goto RIBcleanup;
		}
	} else {
		if ((pPlaneTmp = (BYTE *) IFFmalloc (srcRowBytes)) == NULL) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMTempPlane);
/**/		goto RIBcleanup;
		}
	}

	/* Read the BODY contents into client's bitmap.
	 * De-interleave planes and decompress rows.
	 * MODIFIES: Last iteration modifies bufsize.*/

	buf = (BYTE *)((BYTE huge32 *)buffer + bufsize);  /* Buffer is currently empty.*/
	for (iRow = 0; iRow < nRows; iRow++)  {
		for (iPlane = 0; iPlane < srcPlaneCnt; iPlane++)  {
			if (createbmap) {
				workp[iPlane] = (UBYTE *)planes[iPlane];
				gDest = planes[iPlane];
				pDest = &gDest;
			} else {
				if (pd->BigBitMap.xPlanes[iPlane] != NULL) {
//					planes[iPlane] = (BYTE *) ActivateXTRA(pd->BigBitMap.xPlanes[iPlane]) + iRow * srcRowBytes;
					planes[iPlane] = pPlaneTmp;
					pDest = &planes[iPlane];
				} else {
					nullBuf = nullDest;
					pDest = &nullBuf;
				}
			}

			/* Establish a sink for any unwanted plane.*/
			if (*pDest == NULL) {
				nullBuf = nullDest;
				pDest   = &nullBuf;
			}

			/* Read in at least enough bytes to
			   uncompress next row. */

			nEmpty  = (BYTE huge32 *)buf - (BYTE huge32 *)buffer;	/* size of empty part of buffer.*/
			nFilled = (int)(bufsize - nEmpty);	  /* this part has data.*/
			if (nFilled < bufRowBytes) {
				/* Need to read more.*/

				/* Move the existing data to the front of the buffer.*/
				/* Now covers range buffer[0]..buffer[nFilled-1].*/
				IFFmovmem(buf, buffer, nFilled);  /* Could be moving 0 bytes.*/

				if (nEmpty > size - iff->Size) {
					/* There aren't enough bytes left to fill the buffer.*/
					nEmpty  = size - iff->Size;
					bufsize = nFilled + nEmpty;  /* heh-heh */
				}

				/* Append new data to the existing data.*/
				if (nEmpty) {
					okay = IFFRead(iff, &buffer[nFilled], nEmpty);
					if (!okay) {
/**/					goto RIBcleanup;
					}
				}
	
				buf     = buffer;
				nFilled = (int)bufsize;
				nEmpty  = 0;
			}
			/* Copy uncompressed row to destination plane.*/
			if (compression == cmpNone) {
				if (nFilled < srcRowBytes) {
					SetGlobalErr (ERR_INVALID_DATA);
					GEcatf (szerrIShortFORM);
/**/				goto RIBcleanup;
				}
				IFFmovmem(buf, *pDest, srcRowBytes);
				buf    += srcRowBytes;
				*pDest += srcRowBytes;
			} else {
				/* Decompress row to destination plane.*/

				if ( UnPackRow(&buf, pDest, nFilled,  srcRowBytes) ) {
		                    /*  pSource, pDest, srcBytes, dstBytes  */
					SetGlobalErr (ERR_INVALID_DATA);
					GEcatf (szerrIBadPacked);
/**/				goto RIBcleanup;
				}
			}
			if (!createbmap) {
//				UpdateXTRA(pd->BigBitMap.xPlanes[iPlane]);
				if (pd->BigBitMap.xPlanes[iPlane] != NULL)
					WriteXTRA(pPlaneTmp, srcRowBytes, pd->BigBitMap.xPlanes[iPlane], iRow * srcRowBytes);
			}
		}

		if (createbmap) {
			int		 x;
			int		 xend;
			UBYTE	*data;

//			data = (UBYTE *) ActivateXTRA(pd->BigByteMap.mpYpMpXpPix[iRow]);
			data = pDataTmp;

			for (x = 0; x < bmHdr->w; x += 8) {
				xend = min (8, bmHdr->w - x);
				if (xend >= 1) {
					*data++ = (UBYTE)
							  (
								((workp[0] && (*workp[0] & 0x80)) ? 0x01 : 0x00) |
								((workp[1] && (*workp[1] & 0x80)) ? 0x02 : 0x00) |
								((workp[2] && (*workp[2] & 0x80)) ? 0x04 : 0x00) |
								((workp[3] && (*workp[3] & 0x80)) ? 0x08 : 0x00) |
								((workp[4] && (*workp[4] & 0x80)) ? 0x10 : 0x00) |
								((workp[5] && (*workp[5] & 0x80)) ? 0x20 : 0x00) |
								((workp[6] && (*workp[6] & 0x80)) ? 0x40 : 0x00) |
								((workp[7] && (*workp[7] & 0x80)) ? 0x80 : 0x00)
							  );
					if (xend >= 2) {
						*data++ = (UBYTE)
								  (
									((workp[0] && (*workp[0] & 0x40)) ? 0x01 : 0x00) |
									((workp[1] && (*workp[1] & 0x40)) ? 0x02 : 0x00) |
									((workp[2] && (*workp[2] & 0x40)) ? 0x04 : 0x00) |
									((workp[3] && (*workp[3] & 0x40)) ? 0x08 : 0x00) |
									((workp[4] && (*workp[4] & 0x40)) ? 0x10 : 0x00) |
									((workp[5] && (*workp[5] & 0x40)) ? 0x20 : 0x00) |
									((workp[6] && (*workp[6] & 0x40)) ? 0x40 : 0x00) |
									((workp[7] && (*workp[7] & 0x40)) ? 0x80 : 0x00)
								  );
						if (xend >= 3) {
							*data++ = (UBYTE)
									  (
										((workp[0] && (*workp[0] & 0x20)) ? 0x01 : 0x00) |
										((workp[1] && (*workp[1] & 0x20)) ? 0x02 : 0x00) |
										((workp[2] && (*workp[2] & 0x20)) ? 0x04 : 0x00) |
										((workp[3] && (*workp[3] & 0x20)) ? 0x08 : 0x00) |
										((workp[4] && (*workp[4] & 0x20)) ? 0x10 : 0x00) |
										((workp[5] && (*workp[5] & 0x20)) ? 0x20 : 0x00) |
										((workp[6] && (*workp[6] & 0x20)) ? 0x40 : 0x00) |
										((workp[7] && (*workp[7] & 0x20)) ? 0x80 : 0x00)
									  );
							if (xend >= 4) {
								*data++ = (UBYTE)
										  (
											((workp[0] && (*workp[0] & 0x10)) ? 0x01 : 0x00) |
											((workp[1] && (*workp[1] & 0x10)) ? 0x02 : 0x00) |
											((workp[2] && (*workp[2] & 0x10)) ? 0x04 : 0x00) |
											((workp[3] && (*workp[3] & 0x10)) ? 0x08 : 0x00) |
											((workp[4] && (*workp[4] & 0x10)) ? 0x10 : 0x00) |
											((workp[5] && (*workp[5] & 0x10)) ? 0x20 : 0x00) |
											((workp[6] && (*workp[6] & 0x10)) ? 0x40 : 0x00) |
											((workp[7] && (*workp[7] & 0x10)) ? 0x80 : 0x00)
										  );
								if (xend >= 5) {
									*data++ = (UBYTE)
											  (
												((workp[0] && (*workp[0] & 0x08)) ? 0x01 : 0x00) |
												((workp[1] && (*workp[1] & 0x08)) ? 0x02 : 0x00) |
												((workp[2] && (*workp[2] & 0x08)) ? 0x04 : 0x00) |
												((workp[3] && (*workp[3] & 0x08)) ? 0x08 : 0x00) |
												((workp[4] && (*workp[4] & 0x08)) ? 0x10 : 0x00) |
												((workp[5] && (*workp[5] & 0x08)) ? 0x20 : 0x00) |
												((workp[6] && (*workp[6] & 0x08)) ? 0x40 : 0x00) |
												((workp[7] && (*workp[7] & 0x08)) ? 0x80 : 0x00)
											  );
									if (xend >= 6) {
										*data++ = (UBYTE)
												  (
													((workp[0] && (*workp[0] & 0x04)) ? 0x01 : 0x00) |
													((workp[1] && (*workp[1] & 0x04)) ? 0x02 : 0x00) |
													((workp[2] && (*workp[2] & 0x04)) ? 0x04 : 0x00) |
													((workp[3] && (*workp[3] & 0x04)) ? 0x08 : 0x00) |
													((workp[4] && (*workp[4] & 0x04)) ? 0x10 : 0x00) |
													((workp[5] && (*workp[5] & 0x04)) ? 0x20 : 0x00) |
													((workp[6] && (*workp[6] & 0x04)) ? 0x40 : 0x00) |
													((workp[7] && (*workp[7] & 0x04)) ? 0x80 : 0x00)
												  );
										if (xend >= 7) {
											*data++ = (UBYTE)
													  (
														((workp[0] && (*workp[0] & 0x02)) ? 0x01 : 0x00) |
														((workp[1] && (*workp[1] & 0x02)) ? 0x02 : 0x00) |
														((workp[2] && (*workp[2] & 0x02)) ? 0x04 : 0x00) |
														((workp[3] && (*workp[3] & 0x02)) ? 0x08 : 0x00) |
														((workp[4] && (*workp[4] & 0x02)) ? 0x10 : 0x00) |
														((workp[5] && (*workp[5] & 0x02)) ? 0x20 : 0x00) |
														((workp[6] && (*workp[6] & 0x02)) ? 0x40 : 0x00) |
														((workp[7] && (*workp[7] & 0x02)) ? 0x80 : 0x00)
													  );
											if (xend >= 8) {
												*data++ = (UBYTE)
														  (
															((workp[0] && (*workp[0] & 0x01)) ? 0x01 : 0x00) |
															((workp[1] && (*workp[1] & 0x01)) ? 0x02 : 0x00) |
															((workp[2] && (*workp[2] & 0x01)) ? 0x04 : 0x00) |
															((workp[3] && (*workp[3] & 0x01)) ? 0x08 : 0x00) |
															((workp[4] && (*workp[4] & 0x01)) ? 0x10 : 0x00) |
															((workp[5] && (*workp[5] & 0x01)) ? 0x20 : 0x00) |
															((workp[6] && (*workp[6] & 0x01)) ? 0x40 : 0x00) |
															((workp[7] && (*workp[7] & 0x01)) ? 0x80 : 0x00)
														  );
											}
										}
									}
								}
							}
						}
					}
				}
				if (workp[0]) workp[0]++;
				if (workp[1]) workp[1]++;
				if (workp[2]) workp[2]++;
				if (workp[3]) workp[3]++;
				if (workp[4]) workp[4]++;
				if (workp[5]) workp[5]++;
				if (workp[6]) workp[6]++;
				if (workp[7]) workp[7]++;
			}
//			UpdateXTRA(pd->BigByteMap.mpYpMpXpPix[iRow]);
			WriteXTRA(pDataTmp, bmHdr->w, pd->BigByteMap.mpYpMpXpPix[iRow], 0);
		}
	}

	if (createbmap)
		IFFfree(pDataTmp);
	else
		IFFfree(pPlaneTmp);

	result =  iff->Size;

RIBcleanup:
	if (createbmap) {
		for (iPlane = 0; iPlane < MAXSRCPLANES; iPlane++) {
			if (planes[iPlane])
				IFFfree (planes[iPlane]);
		}
	}
	if (nullDest)
		IFFfree (nullDest);

	if (mask)
		IFFGfree (mask);
	return result;

} /* ReadILBMBODY */

/*********************************************************************
 *
 * ReadPBMBODY
 *
 * SYNOPSIS
 *		long ReadPBMBODY (IFFTracker *iff, PixelData *pd)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		09/28/94 (dcc) - set unused bits at end of line in bitplanes to 0
 *
 * SEE ALSO
 *
*/
long ReadPBMBODY (IFFTracker *iff, BitMapHeader *bmHdr, PixelData *pd)
{

	UBYTE			createbmap  = (UBYTE)(iff->Flags & ILBM_CREATEBYTEMAP);
	int				unpRowBytes = bmHdr->w;
	int				srcRowBytes = ((unpRowBytes + 1) / 2) * 2; // & 0xFFFE;
	int				dstRowBytes;
	LONG			bufRowBytes;
	LONG			okay;
	LONG			size;
	LONG			result		= FALSE;
	LONG			bufsize;
	int				nRows       = bmHdr->h;
	Compression		compression = bmHdr->compression;
	register unsigned	iPlane;
	register int	iRow;
	register LONG	nEmpty;
	register int	nFilled;
	UBYTE			*workp[PBM_DEPTH]; /* array of ptrs to planes & mask */
	BYTE			*buf, **pDest;
	BYTE			*gDest;
	UBYTE			*ubuf		= NULL;
	BYTE			*buffer		= NULL;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadPBMBODY";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	size      = iff->Size;
	iff->Size = 0;

	if (compression > cmpByteRun1) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf1 (szerrPUnknownComp, compression);
		return FALSE;
	}

	if (createbmap) {
		dstRowBytes = bmHdr->w;

		pd->BigByteMap.width  = bmHdr->w;
		pd->BigByteMap.height = (WORD) nRows;
		if ((pd->BigByteMap.mpYpMpXpPix = AllocateBigByteMap(bmHdr->w, bmHdr->h)) == NULL) {
/**/		goto RIBcleanup;
		}
	} else {
		dstRowBytes = RowBytes(srcRowBytes);
		for (iPlane = 0; iPlane < PBM_DEPTH; iPlane++) {
			if ((pd->BigBitMap.xPlanes[iPlane] = AllocXTRA (dstRowBytes * nRows)) == NULL) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf (szerrOOXMBitPlanes);
/**/			goto RIBcleanup;
			}
		}
		for (iPlane = 0; iPlane < PBM_DEPTH; iPlane++) {
			if ((workp[iPlane] = (UBYTE *) IFFcalloc(dstRowBytes, 1)) == NULL) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf (szerrOOMTempBitPlanes);
/**/			goto RIBcleanup;
			}
		}
		pd->BigBitMap.BytesPerRow = (UWORD) dstRowBytes;
		pd->BigBitMap.Rows        = (UWORD) nRows;
		pd->BigBitMap.Depth       = PBM_DEPTH;
	}

	bufRowBytes = MaxPackedSize(srcRowBytes);
	bufsize     = srcRowBytes * 3;
	if ((buffer = (BYTE *) IFFcalloc (bufsize, 1)) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrPOOMWorkBuf);
/**/	goto RIBcleanup;
	}

	ubuf     = (UBYTE *)buffer;
	buffer  += srcRowBytes;
	bufsize -= srcRowBytes;

	/*
	 * Read the BODY contents into client's bitmap.
	 * De-interleave planes and decompress rows.
	 * MODIFIES: Last iteration modifies bufsize.
	 */

	buf = (BYTE *)((BYTE huge32 *)buffer + bufsize);  /* Buffer is currently empty.*/
	for (iRow = 0; iRow < nRows; iRow++)  {
		gDest = (BYTE *)ubuf;
		pDest = &gDest;

		/* Read in at least enough bytes to
			uncompress next row. */

		nEmpty  = (BYTE huge32 *)buf - (BYTE huge32 *)buffer;	/* size of empty part of buffer.*/
		nFilled = (int)(bufsize - nEmpty);	  /* this part has data.*/
		if (nFilled < bufRowBytes) {
			/* Need to read more.*/

			/* Move the existing data to the front of the buffer.*/
			/* Now covers range buffer[0]..buffer[nFilled-1].*/
			IFFmovmem(buf, buffer, nFilled);  /* Could be moving 0 bytes.*/

			if (nEmpty > size - iff->Size) {
				/* There aren't enough bytes left to fill the buffer.*/
				nEmpty  = size - iff->Size;
				bufsize = nFilled + nEmpty;  /* heh-heh */
			}

			/* Append new data to the existing data.*/
			if (nEmpty) {
				okay = IFFRead(iff, &buffer[nFilled], nEmpty);
				if (!okay) {
/**/				goto RIBcleanup;
				}
			}

			buf     = buffer;
			nFilled = (int)bufsize;
			nEmpty  = 0;
		}
		/* Copy uncompressed row to destination plane.*/
		if (compression == cmpNone) {
			if (nFilled < srcRowBytes) {
				SetGlobalErr (ERR_INVALID_DATA);
				GEcatf (szerrPShortFORM);
/**/			goto RIBcleanup;
			}
			if (createbmap) {
				WriteXTRA(buf, unpRowBytes, pd->BigByteMap.mpYpMpXpPix[iRow], 0);
			} else {
				IFFmovmem(buf, *pDest, unpRowBytes);
			}
			buf    += srcRowBytes;
			*pDest += dstRowBytes;
		} else {
			/* Decompress row to destination plane.*/

			if ( UnPackRow(&buf, pDest, nFilled, srcRowBytes) ) {
		                /*  pSource, pDest, srcBytes, dstBytes  */
				SetGlobalErr (ERR_INVALID_DATA);
				GEcatf (szerrPBadPacked);
/**/			goto RIBcleanup;
			}
			if (createbmap) {
				WriteXTRA(ubuf, unpRowBytes, pd->BigByteMap.mpYpMpXpPix[iRow], 0);
			}
		}

		/*
		** Make Sure Pad byte for odd sized pictures is ZERO.
		*/
		if (unpRowBytes & 0x01) {
			ubuf[unpRowBytes] = 0x00;
		}

		if (!createbmap)
		{
			int		x;
			UBYTE	temps[8];
			UBYTE	plns[8];
			UBYTE	bit;
			unsigned	p;

			gDest = (BYTE *)ubuf;
			for (x = 0; x < srcRowBytes; x += 8)
			{
				temps[0] = *gDest++;
				temps[1] = *gDest++;
				temps[2] = *gDest++;
				temps[3] = *gDest++;
				temps[4] = *gDest++;
				temps[5] = *gDest++;
				temps[6] = *gDest++;
				temps[7] = *gDest++;

				for (p = srcRowBytes - x; p < 8; p++)
					temps[p] = 0;

				bit      = 0x01;
				for (p = 0; p < 8; p++)
				{
					plns[p] = (UBYTE)
							  (
								((temps[0] & bit) ? 0x80 : 0x00) |
								((temps[1] & bit) ? 0x40 : 0x00) |
								((temps[2] & bit) ? 0x20 : 0x00) |
								((temps[3] & bit) ? 0x10 : 0x00) |
								((temps[4] & bit) ? 0x08 : 0x00) |
								((temps[5] & bit) ? 0x04 : 0x00) |
								((temps[6] & bit) ? 0x02 : 0x00) |
								((temps[7] & bit) ? 0x01 : 0x00)
							  );
					bit <<= 1;
				}
				workp[0][x/8] = plns[0];
				workp[1][x/8] = plns[1];
				workp[2][x/8] = plns[2];
				workp[3][x/8] = plns[3];
				workp[4][x/8] = plns[4];
				workp[5][x/8] = plns[5];
				workp[6][x/8] = plns[6];
				workp[7][x/8] = plns[7];
			}
			for (x = 0; x < PBM_DEPTH; x++)
			{
				WriteXTRA(workp[x], dstRowBytes, pd->BigBitMap.xPlanes[x], dstRowBytes * iRow);
			}
		}
	}

	result = iff->Size;

RIBcleanup:
	if (!createbmap) {
		for (iPlane = 0; iPlane < PBM_DEPTH; iPlane++) {
			if (workp[iPlane]) {
				IFFfree(workp[iPlane]);
			}
		}
	}
	if (ubuf)
		IFFfree (ubuf);
	return result;

} /* ReadPBMBODY */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadCMAP                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadCMAP
 *
 * SYNOPSIS
 *		long ReadCMAP (IFFTracker *iff, UWORD *numcolors, ColorRegister **cr)
 *
 * PURPOSE
 *		Read a CMAP chunk.  Fill in *numcolors with the number of
 *		colors read and allocate memory for the color and assign that
 *		memory to *cr then read the colors.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadCMAP (IFFTracker *iff, UWORD *numcolors, ColorRegister **cr)
{
	UWORD			 ncolors;
	ColorRegister	*cregs;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCMAP";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ncolors = (UWORD)(iff->Size / sizeofColorRegister);
	*numcolors = ncolors;

	if ((cregs = (ColorRegister *) IFFmalloc (iff->Size)) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrOOMPalette);
		return FALSE;
	}

	*cr = cregs;
	return IFFRead (iff, cregs, iff->Size);

} /* ReadCMAP */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ILBM_ReadCMAP                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ILBM_ReadCMAP
 *
 * SYNOPSIS
 *		long ILBM_ReadCMAP (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ILBM_ReadCMAP (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ILBM_ReadCMAP";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	return ReadCMAP (
				iff,
				(&((ILBMInfo *)(iff->IFFData.Data))->NumberOfColors),
				(&((ILBMInfo *)(iff->IFFData.Data))->Colors));

} /* ILBM_ReadCMAP */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ILBM_ReadBODY                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ILBM_ReadBODY
 *
 * SYNOPSIS
 *		long ILBM_ReadBODY (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ILBM_ReadBODY (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ILBM_ReadBODY";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	return ReadILBMBODY (iff, &((ILBMInfo *)(iff->IFFData.Data))->BM, (PixelData *)&((ILBMInfo *)(iff->IFFData.Data))->Pixels);

} /* ILBM_ReadBODY */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_PBM_ReadBODY                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * PBM_ReadBODY
 *
 * SYNOPSIS
 *		long PBM_ReadBODY (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long PBM_ReadBODY (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PBM_ReadBODY";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	return ReadPBMBODY (iff, &((ILBMInfo *)(iff->IFFData.Data))->BM, &((ILBMInfo *)(iff->IFFData.Data))->Pixels);

} /* PBM_ReadBODY */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadCRNG                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadCRNG
 *
 * SYNOPSIS
 *		long ReadCRNG (IFFTracker *iff, CrngChunk *cc)
 *
 * PURPOSE
 *		read a CRNG chunk into *cc;
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadCRNG (IFFTracker *iff, CrngChunk *cc)
{

	long	size;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadCRNG";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		size = IFFRead (iff, cc, sizeof (CrngChunk));
		if (size) {
			BigWord2Native (cc->rate);
			BigWord2Native (cc->active);
		}
		return size;
	}

} /* ReadCRNG */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ILBM_ReadCRNG                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ILBM_ReadCRNG
 *
 * SYNOPSIS
 *		long ILBM_ReadCRNG (IFFTracker *iff)
 *
 * PURPOSE
 *		read a CRNG (Cycle RaNGe)
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ILBM_ReadCRNG (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ILBM_ReadCRNG";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		CrngNode	*cn;
		ILBMInfo	*ilbm;

		ilbm = (ILBMInfo *)iff->IFFData.Data;

		cn = CreateNode (sizeof (CrngNode), NULL);
		if (!cn)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf (szerrOOMCRNG);
			return FALSE;
		}

		AddTail (ilbm->CrngList, cn);

		return ReadCRNG (iff, &cn->Crng);
	}

} /* ILBM_ReadCRNG */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadBMHD                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadBMHD
 *
 * SYNOPSIS
 *		long ReadBMHD (IFFTracker *iff, BitMapHeader *bm)
 *
 * PURPOSE
 *		read a BMHD chunk into *bm;
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadBMHD (IFFTracker *iff, BitMapHeader *bm)
{

	long	size;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadBMHD";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	
	size = IFFRead (iff, bm, sizeof (BitMapHeader));
	if (size) {
		BigWord2Native (bm->w);
		BigWord2Native (bm->h);
		BigWord2Native (bm->x);
		BigWord2Native (bm->y);
		BigWord2Native (bm->transparentColor);
		BigWord2Native (bm->pageWidth);
		BigWord2Native (bm->pageHeight);
	}
	return size;

} /* ReadBMHD */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ILBM_ReadBMHD                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ILBM_ReadBMHD
 *
 * SYNOPSIS
 *		long ILBM_ReadBMHD (IFFTracker *iff)
 *
 * PURPOSE
 *		read a BMHD (BitMap HeaDer)
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ILBM_ReadBMHD (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ILBM_ReadBMHD";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	return ReadBMHD (iff, &((ILBMInfo *)(iff->IFFData.Data))->BM);

} /* ILBM_ReadBMHD */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_FreeILBMInfo                                         */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * FreeILBMInfo
 *
 * SYNOPSIS
 *		void FreeILBMInfo (ILBMInfo *ilbm)
 *
 * PURPOSE
 *		Free stuff accumulated while loading.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		04/04/94 Monday (dcc) - add test for NULL CrngList
 *
 * SEE ALSO
 *
*/
void FreeILBMInfo (ILBMInfo *ilbm)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FreeILBMInfo";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (ilbm)
	{
		{
			ColorRegister	*data;

			data = ilbm->Colors;
			if (data)
			{
				IFFfree (data);
			}
		}

		if (ilbm->CrngList != NULL)
		{
			CrngNode	*nd;

			while ((nd = RemTail (ilbm->CrngList)) != NULL)
			{
				DeleteNode (nd);
			}
		}

		IFFfree (ilbm);
	}

} /* FreeILBMInfo */


/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_AllocaILBMInfo                                       */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * AllocILBMInfo
 *
 * SYNOPSIS
 *		ILBMInfo *AllocILBMInfo (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
ILBMInfo *AllocILBMInfo (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AllocILBMInfo";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		ILBMInfo	*ilbm;

		ilbm = (ILBMInfo *) IFFcalloc (1, sizeof (ILBMInfo));
		if (ilbm)
		{
			ilbm->CrngList = &ilbm->CrngListX;
			InitList (ilbm->CrngList);
		}

		return ilbm;
	}
} /* AllocILBMInfo */


/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadILBM                                             */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadILBM
 *
 * SYNOPSIS
 *		long ReadILBM (IFFTracker *iff)
 *
 * PURPOSE
 *		Read ILBM Form
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadILBM (IFFTracker *iff)
{
	static IFFRoutineID ILBMTable[] = {
		{ BMHD_CHUNK,		ILBM_ReadBMHD,	},
		{ CMAP_CHUNK,		ILBM_ReadCMAP,	},
		{ CRNG_CHUNK,		ILBM_ReadCRNG,	},
/*		{ GRAB_CHUNK,		ILBM_ReadGRAB,	}, */
/*		{ DEST_CHUNK,		ILBM_ReadDEST,	}, */
/*		{ SPRT_CHUNK,		ILBM_ReadSPRT,	}, */
/*		{ CAMG_CHUNK,		ILBM_ReadCAMG,	}, */
		{ BODY_CHUNK,		ILBM_ReadBODY,	},
		{ FALSE,			NULL,			},
	};

	long		 size = iff->Size;
	long		 okay = TRUE;
	long		 tsize;
	ChunkHeader	 chunkhdr;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadILBM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	iff->Size         = 0;
	if ((iff->IFFData.Data = AllocILBMInfo ()) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrIOOMILBMInfo);
		return FALSE;
	}

	while (iff->Size < size && okay) {
		if (!GetChunkHeader (iff, &chunkhdr)) {
			return FALSE;
		}

		tsize      = iff->Size;
		iff->Size  = chunkhdr.Size;
		okay       = CallIDFunc (chunkhdr.ID, ILBMTable, iff);
		if (okay & 0x01) {
			IFFSkip (iff, 1);
			okay++;
		}
		tsize     += okay;
		iff->Size  = tsize;
	}

	if (!okay) {
/**/	goto cleanup;
	}

	if (iff->Size != size) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf (szerrIDataMismatch);
/**/	goto cleanup;
	}

	return size;

cleanup:
	FreeILBMInfo (iff->IFFData.Data);
	return FALSE;

} /* ReadILBM */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_ReadPBM                                              */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadPBM
 *
 * SYNOPSIS
 *		long ReadPBM (IFFTracker *iff)
 *
 * PURPOSE
 *		Read PBM Form
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadPBM (IFFTracker *iff)
{
	static IFFRoutineID PBMTable[] = {
		{ BMHD_CHUNK,		ILBM_ReadBMHD,	},
		{ CMAP_CHUNK,		ILBM_ReadCMAP,	},
		{ CRNG_CHUNK,		ILBM_ReadCRNG,	},
/*		{ GRAB_CHUNK,		ILBM_ReadGRAB,	}, */
/*		{ DEST_CHUNK,		ILBM_ReadDEST,	}, */
/*		{ SPRT_CHUNK,		ILBM_ReadSPRT,	}, */
/*		{ CAMG_CHUNK,		ILBM_ReadCAMG,	}, */
		{ BODY_CHUNK,		PBM_ReadBODY,	},
		{ FALSE,			NULL,			},
	};

	long		 size = iff->Size;
	long		 okay = TRUE;
	long		 tsize;
	ChunkHeader	 chunkhdr;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadPBM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	iff->Size         = 0;
	if ((iff->IFFData.Data = AllocILBMInfo ()) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrPOOMILBMInfo);
		return FALSE;
	}

	while (iff->Size < size && okay) {
		if (!GetChunkHeader (iff, &chunkhdr)) {
			return FALSE;
		}

		tsize      = iff->Size;
		iff->Size  = chunkhdr.Size;
		okay       = CallIDFunc (chunkhdr.ID, PBMTable, iff);
		if (okay & 0x01) {
			IFFSkip (iff, 1);
			okay++;
		}
		tsize     += okay;
		iff->Size  = tsize;
	}

	if (!okay) {
/**/	goto cleanup;
	}

	if (iff->Size != size) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf (szerrPDataMismatch);
/**/	goto cleanup;
	}

	return size;

cleanup:
	FreeILBMInfo (iff->IFFData.Data);
	return FALSE;

} /* ReadPBM */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_DestroyBlockOPixels                                  */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * DestroyBlockOPixels
 *
 * SYNOPSIS
 *		void DestroyBlockOPixels (BlockOPixels *bop)
 *
 * PURPOSE
 *		Free memory used by BlockOPixels
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroyBlockOPixels (BlockOPixels *bop)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroyBlockOPixels";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (bop) {
		FreeBigByteMap (bop->Pixels.BigByteMap.mpYpMpXpPix, bop->BM.h);

		FreeILBMInfo (bop);
	}

} /* DestroyBlockOPixels */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_CreateBlockOPixels                                   */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * CreateBlockOPixels
 *
 * SYNOPSIS
 *		BlockOPixels *CreateBlockOPixels (char *filename)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BlockOPixels *CreateBlockOPixels (char *filename)
{
	static IFFRoutineID ILBMTable[] = {
		{ ILBM_FORM,		ReadILBM,	},
		{ PBM_FORM,			ReadPBM,	},
		{ FALSE,			NULL,		},
	};

	IFFTracker	iff;
	long		okay;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CreateBlockOPixels";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	memset (&iff, 0, sizeof (IFFTracker));

	iff.FORMTable = ILBMTable;
	iff.Flags     = ILBM_CREATEBYTEMAP;
	iff.fd = EIO_ReadOpen (filename);
	if (iff.fd < 0) {
		SetGlobalErr (ERR_FILE_NOT_FOUND);
		GEcatf1 (szerrPCantOpen, filename);
		return NULL;
	}

	if (IsPCXFile (&iff))
	{
		okay = ReadPCX (&iff);
	}
	else
	{
		okay = ReadIFF (&iff);
	}

	EIO_Close(iff.fd);
	if (!okay) {
		return NULL;
	}

	return iff.IFFData.Data;

} /* CreateBlockOPixels */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_DestroyBlockOPixels                                  */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * DestroyPlanesOPixels
 *
 * SYNOPSIS
 *		void DestroyPlanesOPixels (PlanesOPixels *pop)
 *
 * PURPOSE
 *		Free memory used by PlanesOPixels
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroyPlanesOPixels (PlanesOPixels *pop)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroyPlanesOPixels";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (pop) {
		int		 ip;
		void	*data;

		for (ip = 0; ip < pop->BM.nPlanes; ip++) {
			data = pop->Pixels.BigBitMap.xPlanes[ip];
			if (data)
				FreeXTRA (data);
		}

		FreeILBMInfo (pop);
	}

} /* DestroyPlanesOPixels */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_CreatePlanesOPixels                                  */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * CreatePlanesOPixels
 *
 * SYNOPSIS
 *		PlanesOPixels *CreatePlanesOPixels (char *filename)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
PlanesOPixels *CreatePlanesOPixels (char *filename)
{
	static IFFRoutineID ILBMTable[] = {
		{ ILBM_FORM,		ReadILBM,	},
		{ PBM_FORM,			ReadPBM,	},
		{ FALSE,			NULL,		},
	};

	IFFTracker	iff;
	long		okay;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CreatePlanesOPixels";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	memset (&iff, 0, sizeof (IFFTracker));

	iff.FORMTable = ILBMTable;
	iff.fd = EIO_ReadOpen (filename);
	if (iff.fd < 0) {
		SetGlobalErr (ERR_FILE_NOT_FOUND);
		GEcatf1 (szerrICantOpen, filename);
		return NULL;
	}

	if (IsPCXFile (&iff))
	{
		okay = ReadPCX (&iff);
	}
	else
	{
		okay = ReadIFF (&iff);
	}

	EIO_Close(iff.fd);
	if (!okay) {
		return NULL;
	}

	return iff.IFFData.Data;

} /* CreatePlanesOPixels */

/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_CreatePalette                                        */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * ReadILBMCMAP
 *
 * PURPOSE
 *		Read just one CMAP chunk from an ILBM Form
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static long ReadILBMCMAP (IFFTracker *iff)
{
	static IFFRoutineID ILBMTable[] = {
		{ CMAP_CHUNK,		ILBM_ReadCMAP,	},
		{ CRNG_CHUNK,		ILBM_ReadCRNG,	},
		{ FALSE,			NULL,			},
	};

	long		 size = iff->Size;
	long		 okay = TRUE;
	long		 tsize;
	ChunkHeader	 chunkhdr;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadILBMCMAP";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	iff->Size         = 0;
	if ((iff->IFFData.Data = AllocILBMInfo ()) == NULL) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf (szerrCOOMILBMInfo);
		return FALSE;
	}

	while (iff->Size < size && okay) {
		if (!GetChunkHeader (iff, &chunkhdr)) {
			return FALSE;
		}

		tsize      = iff->Size;
		iff->Size  = chunkhdr.Size;
		okay       = CallIDFunc (chunkhdr.ID, ILBMTable, iff);
		if (okay & 0x01) {
			IFFSkip (iff, 1);
			okay++;
		}
		tsize     += okay;
		iff->Size  = tsize;
	}

	if (!okay) {
/**/	goto cleanup;
	}

	if (iff->Size != size) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf (szerrCDataMismatch);
/**/	goto cleanup;
	}

	return size;

cleanup:
	FreeILBMInfo (iff->IFFData.Data);
	return FALSE;

} /* ReadILBMCMAP */

/*********************************************************************
 *
 * CreatePalette
 *
 * SYNOPSIS
 *		ILBMInfo *CreatePalette (char *filename)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
ILBMInfo *CreatePalette (char *filename)
{
	static IFFRoutineID ILBMTable[] = {
		{ ILBM_FORM,		ReadILBMCMAP,	},
		{ PBM_FORM,			ReadILBMCMAP,	},
		{ FALSE,			NULL,			},
	};

	IFFTracker	iff;
	long		okay;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CreatePalette";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	memset (&iff, 0, sizeof (IFFTracker));

	iff.FORMTable = ILBMTable;
	iff.fd = EIO_ReadOpen (filename);
	if (iff.fd < 0) {
		SetGlobalErr (ERR_FILE_NOT_FOUND);
		GEcatf1 (szerrCCantOpen, filename);
		return NULL;
	}

	if (IsPCXFile(&iff))
	{
		SetGlobalErr (ERR_GENERIC);
		GEcatf ("\nCreatePalette:PCX files not supported yet");
		okay = FALSE;
	}
	else
	{
		okay = ReadIFF (&iff);
	}

	EIO_Close(iff.fd);
	if (!okay) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf1 (szerrCCantRead, filename);
		return NULL;
	}

	return iff.IFFData.Data;

} /* CreatePalette */


/*------------------------------------------------------------------------*/
/**# MODULE:READILBM_DestroyPalette                                       */
/*------------------------------------------------------------------------*/

/*********************************************************************
 *
 * DestroyPalette
 *
 * SYNOPSIS
 *		void DestroyPalette (PlanesOPixels *pop)
 *
 * PURPOSE
 *		Free memory used by Palette
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroyPalette (PlanesOPixels *pop)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroyPalette";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	FreeILBMInfo (pop);

} /* DestroyPalette */

