/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * READTUME.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/27/92
 *   MODIFIED : 06/19/00
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Routines to load a tUME map.
 *
 * HISTORY
 *		08/22/94 (dcc) - add support to read 'LAYR' and 'LAY1' chunks
 *		09/21/94 (dcc) - compiles cleanly with 32-bit Borland C++ 4.02.
 *
*/

#include <echidna\platform.h>
#include "switches.h"

#include <echidna\etypes.h>
#include <echidna\listfunc.h>
#include <echidna\eio.h>
#include <echidna\emacros.h>
#include <echidna\eerrors.h>
#include <echidna\xtramem.h>
#include <echidna\strings.h>
#include <echidna\readiff.h>
#include <echidna\bilbm.h>
#include <echidna\btume.h>

/**************************** C O N S T A N T S ***************************/

#if __AMIGAOS__
#define	INPUT_SLASHES	"\\:"
#define OUTPUT_SLASH	'/'
#elif __MSDOS__
#define	INPUT_SLASHES	"/:"
#define OUTPUT_SLASH	'\\'
#elif __MACOS__
#define	INPUT_SLASHES	"/\\"
#define OUTPUT_SLASH	':'
#elif __EGGS_SGI__
#define	INPUT_SLASHES	":\\"
#define OUTPUT_SLASH	'/'
#else
#error Need something here.
#endif

#define	LOADFLAG_OLDROOM		(0x0001)

#define MAXTILESETS				(256)

#define ROOM_DATA_CHUNK_SIZE	(2+2+2+2+2+2+2+2)
#define ROOM_USER_CHUNK_SIZE	(2+2+2+2)
#define	ROOM_CFLG_CHUNK_SIZE	(2)
#define ROOM_GRID_CHUNK_SIZE	(2+2+2+2+1+1)
#define ROOM_GUID_CHUNK_SIZE	(2+2+2+2+1+1)
#define ROOM_ZOOM_CHUNK_SIZE	(2+2+1)
#define ROOM_CINF_CHUNK_SIZE	(2+2+2+2+2)

#define TSET_USER_CHUNK_SIZE	(2+2+2+2)
#define TSET_TMG_CHUNK_SIZE		(2+2+2+2+2)

#define TUME_FORM	MakeID('t','U','M','E')
#define ROOM_FORM	MakeID('R','O','O','M')
#define TSET_FORM	MakeID('T','S','E','T')

#define SIFF_CHUNK	MakeID('S','I','F','F')

#define DATA_CHUNK	MakeID('D','A','T','A')
#if 1
#define NAME_CHUNK	MakeID('N','A','M','E')
#define LAYR_CHUNK	MakeID('L','A','Y','R')
#define LAY1_CHUNK	MakeID('L','A','Y','1')
#endif // 1
#define USER_CHUNK	MakeID('U','S','E','R')
#define CMNT_CHUNK	MakeID('C','M','N','T')
#define CFLG_CHUNK	MakeID('C','F','L','G')
#define CYCL_CHUNK	MakeID('C','Y','C','L')
#define TMGX_CHUNK	MakeID('T','M','G','X')
#define TMGC_CHUNK	MakeID('T','M','G','C')
#define GRID_CHUNK	MakeID('G','R','I','D')
#define GUID_CHUNK	MakeID('G','U','I','D')
#define ZOOM_CHUNK	MakeID('Z','O','O','M')
#define CINF_CHUNK	MakeID('C','I','N','F')
#define HSVP_CHUNK	MakeID('H','S','V','P')

/******************************** T Y P E S *******************************/

typedef struct {
	UWORD		 GWidth;
	UWORD		 GHeight;
	UWORD		 GX;
	UWORD		 GY;
	UBYTE		 GOnOff;
	UBYTE		 pad;
} GChunk;

typedef struct {
	UWORD		 ZoomDstDup;
	UWORD		 ZoomSrcSkip;
	UBYTE		 ZoomOnOff;
} ZOOMChunk;

typedef struct {
	UBYTE	Regs[33];
	UBYTE	Pos[33];
	UWORD	Speed;
	WORD	Direction;
	UWORD	Flags;
} CYCLChunk;

typedef struct {
	ListType	 RoomListX;
	ListType	*RoomList;
	ListType	 TilesetListX;
	ListType	*TilesetList;
	ListType	 CycleListX;
	ListType	*CycleList;
	RoomType	*CurrentRoom;
	TilesetType	*CurrentTileset;
	UWORD		 TranslateTable[MAXTILESETS];
	MapType		*Map;
} TUMEInfo;

/****************************** G L O B A L S *****************************/

extern long ReadCIM (IFFTracker *iff);

/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * DestroytUMETileset
 *
 * SYNOPSIS
 *		void DestroytUMETileset (TilesetType *ts)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroytUMETileset (TilesetType *ts)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroytUMETileset";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	if (ts) {
		IFFfree (ts->FileSpec);

		if (ts->Comments) {
			char	**s;

			s = ts->Comments;
			while (*s) {
				IFFfree (s);
				s++;
			}
			IFFfree (ts->Comments);
		}

		if (ts->TMG.TileCount) {
			TileDataGroup	*tdg;

			while ((tdg = RemTail (ts->TMG.TileGroupList)) != NULL) {
				FreeXTRA (tdg->TileData);
				DeleteNode (tdg);
			}
		}

		DeleteNode (ts);
	}
} /* DestroytUMETileset */


/*********************************************************************
 *
 * DestroytUMERoom
 *
 * SYNOPSIS
 *		void DestroytUMERoom (RoomType *room)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroytUMERoom (RoomType *room)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroytUMERoom";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	if (room) {
		int		l;
		int		y;

		IFFfree (room->Name);
		if (room->Comments) {
			char	**s;

			s = room->Comments;
			while (*s) {
				IFFfree (s);
				s++;
			}
			IFFfree (room->Comments);
		}

		for (l = 0; l < room->LayerCount; l++) {
			for (y = 0; y < room->Height; y++) {
				if (room->Layers[l]->Plot[y]) {
					FreeXTRA (room->Layers[l]->Plot[y]);
				}
			}
		}
		DeleteNode (room);
	}
} /* DestroytUMERoom */


/*********************************************************************
 *
 * DestroytUMEMap
 *
 * SYNOPSIS
 *		void DestroytUMEMap (MapType *map)
 *
 * PURPOSE
 *		Free all memory associated with a tUME Map
 *
 * INPUT
 *		map		: pointer to a MapType
 *
 * EXAMPLE
 *		See AppendtUMEMap
 *
 * RETURN VALUE
 *		None
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		DestroytUMERoom, DestroytUMETileset, AppendtUMEMap
*/
void DestroytUMEMap (MapType *map)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "DestroytUMEMap";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (map) {
		{
			RoomType	**roomptr;

			roomptr = map->Rooms + 1;

			while (*roomptr) {
				DestroytUMERoom (*roomptr++);
			}
		}
		{
			TilesetType	**tsptr;

			tsptr = map->Tilesets + 1;

			while (*tsptr) {
				DestroytUMETileset (*tsptr++);
			}
		}
		IFFfree (map->Rooms);
		IFFfree (map->Tilesets);
		IFFfree (map);
	}
} /* DestroytUMEMap */


/*********************************************************************
 *
 * ReadGChunk
 *
 * SYNOPSIS
 *		long ReadGChunk (IFFTracker *iff, GChunk *g)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadGChunk (IFFTracker *iff, GChunk *g)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadGChunk";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;

		size = IFFRead (iff, g, ROOM_GRID_CHUNK_SIZE);
		if (size) {
			BigWord2Native (g->GWidth);
			BigWord2Native (g->GHeight);
			BigWord2Native (g->GX);
			BigWord2Native (g->GY);
		}
		return size;
	}
} /* ReadGChunk */


/*********************************************************************
 *
 * ReadZOOM
 *
 * SYNOPSIS
 *		long ReadZOOM (IFFTracker *iff, ZoomChunk *zd)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadZOOM (IFFTracker *iff, ZOOMChunk *zd)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadZOOM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		long		 size;

		size = IFFRead (iff, zd, ROOM_ZOOM_CHUNK_SIZE);
		if (size) {
			BigWord2Native (zd->ZoomDstDup);
			BigWord2Native (zd->ZoomSrcSkip);
		}
		return size;
	}
} /* ReadZOOM */


/*********************************************************************
 *
 * ReadTSETZOOM
 *
 * SYNOPSIS
 *		long ReadTSETZOOM (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETZOOM (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETZOOM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		TilesetType	*ts  = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;

		return ReadZOOM (iff, (ZOOMChunk *)&ts->ZoomDstDup);
	}
} /* ReadTSETZOOM */


/*********************************************************************
 *
 * ReadTSETGRID
 *
 * SYNOPSIS
 *		long ReadTSETGRID (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETGRID (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETGRID";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		TilesetType	*ts  = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;

		return ReadGChunk (iff, (GChunk *)&ts->GridWidth);
	}
} /* ReadTSETGRID */


/*********************************************************************
 *
 * ReadTSETGUID
 *
 * SYNOPSIS
 *		long ReadTSETGUID (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETGUID (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETGUID";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		TilesetType	*ts  = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;

		return ReadGChunk (iff, (GChunk *)&ts->GuideWidth);
	}
} /* ReadTSETGUID */


/*********************************************************************
 *
 * ReadTSETDATA
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *		10/01/94 (dcc) - when merging tilesets, make sure that UserType and
 *						 UserNumber are the same in addition to the name
 *						 before merging. (Previously only the name was
 *						 compared.)
 *
 * SEE ALSO
 *
*/
long ReadTSETDATA (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETDATA";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;
		long	 	 origsize = iff->Size;
		TUMEInfo	*tume     = (TUMEInfo *)iff->IFFData.Data;
		TilesetType	*ts       = tume->CurrentTileset;

		iff->Size = 0;

		// ASSUMES Flags immediately follows ID in TilesetType

		size = IFFRead (iff, &ts->ID, 2 * sizeof (UWORD));
		if (size != 2 * sizeof (UWORD)) {
			return FALSE;
		}

		BigWord2Native (ts->ID);
		BigWord2Native (ts->Flags);

		if (ts->Flags & 0x0100) {
			long	xsize;

			// ASSUMES Height immediately follows Width in TilesetType

			xsize = IFFRead (iff, &ts->Width, 2 * sizeof (WORD));
			if (xsize != 2 * sizeof (WORD)) {
				return FALSE;
			}

			BigWord2Native (ts->Width);
			BigWord2Native (ts->Height);

			size += xsize;
		}

		{
			long	 nsize;

			nsize = origsize - size;
			ts->FileSpec = (char *) IFFcalloc (1, nsize + 1);
			if (!ts->FileSpec) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadTSETDATA:OOM Tileset FileSpec");
				return FALSE;
			}

			nsize = IFFRead (iff, ts->FileSpec, nsize);
			size += nsize;

			{
				char *s = ts->FileSpec;

				for ( ; *s != '\0'; s++)
				{
					char *d = INPUT_SLASHES;
					
					while (*d)
					{
						if (*s == *d)
						{
							*s = OUTPUT_SLASH;
							break;
						}
						d++;
					}
				}
			}

			{
				TilesetType	*chkts;
				char		 ts1name[EIO_MAXPATH];
				char		 ts2name[EIO_MAXPATH];

				if (ts->Flags & TUME_TILESET_COMPOSITE) {
					strcpy (ts2name, ts->FileSpec);
				} else {
					strcpy (ts2name, EIO_Filename (ts->FileSpec));
				}

				chkts = Head (tume->TilesetList);
				while (!EndOfList (chkts)) {
					if (ts->Flags & TUME_TILESET_COMPOSITE) {
						strcpy (ts1name, chkts->FileSpec);
					} else {
						strcpy (ts1name, EIO_Filename (chkts->FileSpec));
					}
					if (stricmp (ts1name, ts2name) == 0 &&
						chkts->UserType == ts->UserType &&
						chkts->UserNumber == ts->UserNumber)
					{
						break;
					}
					chkts = Next (chkts);
				}

				if (!EndOfList (chkts)) {
					tume->TranslateTable[ts->ID] = chkts->ID;
					IFFfree (tume->CurrentTileset);
					tume->CurrentTileset = NULL;
				} else {
					unsigned	newID;

					newID = tume->Map->NumTilesets + 1;
					tume->TranslateTable[ts->ID] = (UWORD) newID;
					ts->ID = (WORD) newID;
				}
			}
		}

		return size;

	}
} /* ReadTSETDATA */


/*********************************************************************
 *
 * ReadTSETUSER
 *
 * SYNOPSIS
 *		long ReadTSETUSER (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETUSER (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETUSER";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		long		 size;
		long		 origsize = iff->Size;
		TilesetType	*ts = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;

		iff->Size = 0;

		size = IFFRead (iff, &ts->UserType, TSET_USER_CHUNK_SIZE);
		if (size) {
			BigWord2Native (ts->UserType);
			BigWord2Native (ts->UserNumber);
			BigWord2Native (ts->DisplayRoomID);
			BigWord2Native (ts->TileCount);
		}

		if (origsize - size) {
			static	ListType	 StrListX;
			static	ListType	*StrList = &StrListX;

			int					 strcount = 0;
			UWORD				 strsize;
			long				 okay;
			char				*str;
			NodeType			*nd;

			InitList (StrList);

			while (origsize - size) {

				okay = IFFRead (iff, &strsize, sizeof (UWORD));
				if (okay != sizeof (UWORD)) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadTSETUSER:Error reading comments");
					EmptyList (StrList);
					return FALSE;
				}
				size += okay;

				BigWord2Native (strsize);

				str = (char *) IFFcalloc (1, strsize + 1);
				nd  = CreateNode (sizeof (NodeType), NULL);

				if (!str || !nd) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETUSER:OOM reading comments");
					EmptyList (StrList);
					return FALSE;
				}

				AddTail (StrList, nd);
				NodeName (nd) = str;

				okay = IFFRead (iff, str, strsize);
				if (okay != strsize) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadTSETUSER:Error reading comments (2)");
					EmptyList (StrList);
					return FALSE;
				}

				size += okay;
				strcount++;
			}

			ts->Comments = (char **) IFFcalloc (strcount + 1, sizeof (char *));
			if (!ts->Comments) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadTSETUSER:OOM reading comments (2)");
				EmptyList (StrList);
				return FALSE;
			}
				
			nd = Tail (StrList);
			while (!StartOfList (nd)) {
				strcount--;
				ts->Comments[strcount] = NodeName (nd);
				NodeName(nd) = NULL;
				nd = Prev (nd);
			}

			EmptyList (StrList);
		}

		return size;
	}
} /* ReadTSETUSER */


/*********************************************************************
 *
 * ConvertTMGXtoTMGC
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL ConvertTMGXtoTMGC (
		UBYTE	*src,
		int		 twidth,
		int		 theight,
		int		 tdepth,
		UBYTE	*dst
)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ConvertTMGXtoTMGC";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		UBYTE	*p[8];
		unsigned x;
		unsigned y;
		UBYTE	 b[8];
		UBYTE	 d[8];
		unsigned mod;
		unsigned count;
		int		 width;
		unsigned planesize;
		unsigned rowsize;
		UBYTE	*zerobuffer;

		rowsize   = (((twidth + 15) / 16) * 2);
		planesize = rowsize * theight;

		zerobuffer = (UBYTE *) IFFcalloc (1, planesize);
		if (!zerobuffer) {
			return FALSE;
		}

		mod = ((twidth + 7) / 8) & 0x0001;

		for (x = 0; x < 8; x++) {
			if (x < tdepth) {
				p[x] = src + (planesize * x);
			} else {
				p[x] = zerobuffer;
			}
		}

		for (y = 0; y < theight; y++) {
			width = twidth;
			for (x = 0; x < twidth; x += 8) {
				d[0] = *(p[0]);
				d[1] = *(p[1]);
				d[2] = *(p[2]);
				d[3] = *(p[3]);
				d[4] = *(p[4]);
				d[5] = *(p[5]);
				d[6] = *(p[6]);
				d[7] = *(p[7]);

				p[0]++;
				p[1]++;
				p[2]++;
				p[3]++;
				p[4]++;
				p[5]++;
				p[6]++;
				p[7]++;

				b[0] = (UBYTE)
					   (
						((d[0] & 0x80) ? 0x01 : 0x00) |
						((d[1] & 0x80) ? 0x02 : 0x00) |
						((d[2] & 0x80) ? 0x04 : 0x00) |
						((d[3] & 0x80) ? 0x08 : 0x00) |
						((d[4] & 0x80) ? 0x10 : 0x00) |
						((d[5] & 0x80) ? 0x20 : 0x00) |
						((d[6] & 0x80) ? 0x40 : 0x00) |
						((d[7] & 0x80) ? 0x80 : 0x00)
					   );
				b[1] = (UBYTE)
					   (
						((d[0] & 0x40) ? 0x01 : 0x00) |
						((d[1] & 0x40) ? 0x02 : 0x00) |
						((d[2] & 0x40) ? 0x04 : 0x00) |
						((d[3] & 0x40) ? 0x08 : 0x00) |
						((d[4] & 0x40) ? 0x10 : 0x00) |
						((d[5] & 0x40) ? 0x20 : 0x00) |
						((d[6] & 0x40) ? 0x40 : 0x00) |
						((d[7] & 0x40) ? 0x80 : 0x00)
					   );
				b[2] = (UBYTE)
					   (
						((d[0] & 0x20) ? 0x01 : 0x00) |
						((d[1] & 0x20) ? 0x02 : 0x00) |
						((d[2] & 0x20) ? 0x04 : 0x00) |
						((d[3] & 0x20) ? 0x08 : 0x00) |
						((d[4] & 0x20) ? 0x10 : 0x00) |
						((d[5] & 0x20) ? 0x20 : 0x00) |
						((d[6] & 0x20) ? 0x40 : 0x00) |
						((d[7] & 0x20) ? 0x80 : 0x00)
					   );
				b[3] = (UBYTE)
					   (
						((d[0] & 0x10) ? 0x01 : 0x00) |
						((d[1] & 0x10) ? 0x02 : 0x00) |
						((d[2] & 0x10) ? 0x04 : 0x00) |
						((d[3] & 0x10) ? 0x08 : 0x00) |
						((d[4] & 0x10) ? 0x10 : 0x00) |
						((d[5] & 0x10) ? 0x20 : 0x00) |
						((d[6] & 0x10) ? 0x40 : 0x00) |
						((d[7] & 0x10) ? 0x80 : 0x00)
					   );
				b[4] = (UBYTE)
					   (
						((d[0] & 0x08) ? 0x01 : 0x00) |
						((d[1] & 0x08) ? 0x02 : 0x00) |
						((d[2] & 0x08) ? 0x04 : 0x00) |
						((d[3] & 0x08) ? 0x08 : 0x00) |
						((d[4] & 0x08) ? 0x10 : 0x00) |
						((d[5] & 0x08) ? 0x20 : 0x00) |
						((d[6] & 0x08) ? 0x40 : 0x00) |
						((d[7] & 0x08) ? 0x80 : 0x00)
					   );
				b[5] = (UBYTE)
					   (
						((d[0] & 0x04) ? 0x01 : 0x00) |
						((d[1] & 0x04) ? 0x02 : 0x00) |
						((d[2] & 0x04) ? 0x04 : 0x00) |
						((d[3] & 0x04) ? 0x08 : 0x00) |
						((d[4] & 0x04) ? 0x10 : 0x00) |
						((d[5] & 0x04) ? 0x20 : 0x00) |
						((d[6] & 0x04) ? 0x40 : 0x00) |
						((d[7] & 0x04) ? 0x80 : 0x00)
					   );
				b[6] = (UBYTE)
					   (
						((d[0] & 0x02) ? 0x01 : 0x00) |
						((d[1] & 0x02) ? 0x02 : 0x00) |
						((d[2] & 0x02) ? 0x04 : 0x00) |
						((d[3] & 0x02) ? 0x08 : 0x00) |
						((d[4] & 0x02) ? 0x10 : 0x00) |
						((d[5] & 0x02) ? 0x20 : 0x00) |
						((d[6] & 0x02) ? 0x40 : 0x00) |
						((d[7] & 0x02) ? 0x80 : 0x00)
					   );
				b[7] = (UBYTE)
					   (
						((d[0] & 0x01) ? 0x01 : 0x00) |
						((d[1] & 0x01) ? 0x02 : 0x00) |
						((d[2] & 0x01) ? 0x04 : 0x00) |
						((d[3] & 0x01) ? 0x08 : 0x00) |
						((d[4] & 0x01) ? 0x10 : 0x00) |
						((d[5] & 0x01) ? 0x20 : 0x00) |
						((d[6] & 0x01) ? 0x40 : 0x00) |
						((d[7] & 0x01) ? 0x80 : 0x00)
					   );

				for (count = 0; count < 8 && width > 0; count++, width--) {
					*dst++ = b[count];
				}
			}

			if (mod) {
				p[0]++;
				p[1]++;
				p[2]++;
				p[3]++;
				p[4]++;
				p[5]++;
				p[6]++;
				p[7]++;
			}
		}

		IFFfree (zerobuffer);
		return TRUE;
	}
} /* ConvertTMGXtoTMGC */


/*********************************************************************
 *
 * ReadTSETTMGX
 *
 * SYNOPSIS
 *		long ReadTSETTMGX (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETTMGX (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETTMGX";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long			 size;
		unsigned		 tsize;
		long			 rsize;
		TilesetType		*ts  = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;
		TMGType			*tmg = &ts->TMG;
		TileDataGroup	*tdg;
		
		size = IFFRead (iff, tmg, TSET_TMG_CHUNK_SIZE);
		if (!size) {
			return FALSE;
		}

		BigWord2Native (tmg->TileCount);
		BigWord2Native (tmg->Width);
		BigWord2Native (tmg->Height);
		BigWord2Native (tmg->Depth);
		BigWord2Native (tmg->TransparentColor);

		tsize = (((tmg->Width + 15) / 16) * 2) * tmg->Height * tmg->Depth;

		if (iff->IFFData.Flags & TUME_TMGC) {
			size_t		 maxtilespergrp;
			unsigned	 tcount;
			unsigned	 tilesleft;
			UBYTE		*pdata;
			XTRAPntr	 xtra;
			UBYTE		*data;
			unsigned	 dsize;

			dsize          = tmg->Width * tmg->Height;
			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / dsize);
			tilesleft      = tmg->TileCount;

			/*
			** Allocate Plane Buffer.
			*/
			pdata = (UBYTE *) IFFmalloc (tsize);
			if (!pdata) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadTSETTMGX:OOM TMGX Buffer");
				return FALSE;
			}

			while (tilesleft) {
			
				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * dsize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData (2)");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, pdata, tsize);
					if (rsize != tsize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGX:Error reading TMGX Tile");
						IFFfree (pdata);
						return FALSE;
					}

					size += rsize;

					if (!ConvertTMGXtoTMGC (pdata, tmg->Width, tmg->Height, tmg->Depth, data)) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGX:Error reading TMGX Tile 2");
						IFFfree (pdata);
						return FALSE;
					}

					data += dsize;
					tcount--;
				}
				UpdateXTRA (xtra);
			}

			IFFfree (pdata);		
		
		} else {
			size_t		 maxtilespergrp;
			unsigned	 tcount;
			unsigned	 tilesleft;
			XTRAPntr	 xtra;
			UBYTE		*data;

			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / tsize);
			tilesleft      = tmg->TileCount;

			while (tilesleft) {

				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * tsize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, data, tsize);
					if (tsize != rsize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGC:Error reading TMGC Tile");
						return FALSE;
					}
					data += tsize;
					size += rsize;
					tcount--;
				}
				UpdateXTRA (xtra);
			}
		}

		return size;
	}
} /* ReadTSETTMGX */


/*********************************************************************
 *
 * ConvertTMGCtoTMGX
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void ConvertTMGCtoTMGX (
		UBYTE	*src,
		int		 twidth,
		int		 theight,
		UBYTE	*dst
)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ConvertTMGCtoTMGX";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/


	{
		unsigned i;
		unsigned x;
		unsigned y;
		int		 width;
		UBYTE	 b[16];
		UBYTE	 p[8][2];
		UBYTE	*pptr[8];
		UBYTE	*tdata;
		unsigned planesize;

		planesize = (((twidth + 15) / 16) * 2) * theight;

		tdata = src;
		for (x = 0; x < 8; x++) {
			pptr[x] = dst + (planesize * x);
		}
		for (y = 0; y < theight; y++) {
			width = twidth;
			for (x = 0; x < twidth; x += 16) {
				for (i = 0; i < 16; i++, width--) {
					if (width > 0) {
						b[i] = *tdata++;
					} else {
						b[i] = 0;
					}
				}
				p[0][0] = (UBYTE)
						  (
							((b[ 0] & 0x01) ? 0x80 : 0x00) |
							((b[ 1] & 0x01) ? 0x40 : 0x00) |
							((b[ 2] & 0x01) ? 0x20 : 0x00) |
							((b[ 3] & 0x01) ? 0x10 : 0x00) |
							((b[ 4] & 0x01) ? 0x08 : 0x00) |
							((b[ 5] & 0x01) ? 0x04 : 0x00) |
							((b[ 6] & 0x01) ? 0x02 : 0x00) |
							((b[ 7] & 0x01) ? 0x01 : 0x00)
						  );
				p[0][1] = (UBYTE)
						  (
						  	((b[ 8] & 0x01) ? 0x80 : 0x00) |
							((b[ 9] & 0x01) ? 0x40 : 0x00) |
							((b[10] & 0x01) ? 0x20 : 0x00) |
							((b[11] & 0x01) ? 0x10 : 0x00) |
							((b[12] & 0x01) ? 0x08 : 0x00) |
							((b[13] & 0x01) ? 0x04 : 0x00) |
							((b[14] & 0x01) ? 0x02 : 0x00) |
							((b[15] & 0x01) ? 0x01 : 0x00)
						  );
				p[1][0] = (UBYTE)
						  (
							((b[ 0] & 0x02) ? 0x80 : 0x00) |
							((b[ 1] & 0x02) ? 0x40 : 0x00) |
							((b[ 2] & 0x02) ? 0x20 : 0x00) |
							((b[ 3] & 0x02) ? 0x10 : 0x00) |
							((b[ 4] & 0x02) ? 0x08 : 0x00) |
							((b[ 5] & 0x02) ? 0x04 : 0x00) |
							((b[ 6] & 0x02) ? 0x02 : 0x00) |
							((b[ 7] & 0x02) ? 0x01 : 0x00)
						  );
				p[1][1] = (UBYTE)
						  (
							((b[ 8] & 0x02) ? 0x80 : 0x00) |
							((b[ 9] & 0x02) ? 0x40 : 0x00) |
							((b[10] & 0x02) ? 0x20 : 0x00) |
							((b[11] & 0x02) ? 0x10 : 0x00) |
							((b[12] & 0x02) ? 0x08 : 0x00) |
							((b[13] & 0x02) ? 0x04 : 0x00) |
							((b[14] & 0x02) ? 0x02 : 0x00) |
							((b[15] & 0x02) ? 0x01 : 0x00)
						  );
				p[2][0] = (UBYTE)
						  (
							((b[ 0] & 0x04) ? 0x80 : 0x00) |
							((b[ 1] & 0x04) ? 0x40 : 0x00) |
							((b[ 2] & 0x04) ? 0x20 : 0x00) |
							((b[ 3] & 0x04) ? 0x10 : 0x00) |
							((b[ 4] & 0x04) ? 0x08 : 0x00) |
							((b[ 5] & 0x04) ? 0x04 : 0x00) |
							((b[ 6] & 0x04) ? 0x02 : 0x00) |
							((b[ 7] & 0x04) ? 0x01 : 0x00)
						  );
				p[2][1] = (UBYTE)
						  (
							((b[ 8] & 0x04) ? 0x80 : 0x00) |
							((b[ 9] & 0x04) ? 0x40 : 0x00) |
							((b[10] & 0x04) ? 0x20 : 0x00) |
							((b[11] & 0x04) ? 0x10 : 0x00) |
							((b[12] & 0x04) ? 0x08 : 0x00) |
							((b[13] & 0x04) ? 0x04 : 0x00) |
							((b[14] & 0x04) ? 0x02 : 0x00) |
							((b[15] & 0x04) ? 0x01 : 0x00)
						  );
				p[3][0] = (UBYTE)
						  (
							((b[ 0] & 0x08) ? 0x80 : 0x00) |
							((b[ 1] & 0x08) ? 0x40 : 0x00) |
							((b[ 2] & 0x08) ? 0x20 : 0x00) |
							((b[ 3] & 0x08) ? 0x10 : 0x00) |
							((b[ 4] & 0x08) ? 0x08 : 0x00) |
							((b[ 5] & 0x08) ? 0x04 : 0x00) |
							((b[ 6] & 0x08) ? 0x02 : 0x00) |
							((b[ 7] & 0x08) ? 0x01 : 0x00)
						  );
				p[3][1] = (UBYTE)
						  (
							((b[ 8] & 0x08) ? 0x80 : 0x00) |
							((b[ 9] & 0x08) ? 0x40 : 0x00) |
							((b[10] & 0x08) ? 0x20 : 0x00) |
							((b[11] & 0x08) ? 0x10 : 0x00) |
							((b[12] & 0x08) ? 0x08 : 0x00) |
							((b[13] & 0x08) ? 0x04 : 0x00) |
							((b[14] & 0x08) ? 0x02 : 0x00) |
							((b[15] & 0x08) ? 0x01 : 0x00)
						  );
				p[4][0] = (UBYTE)
						  (
							((b[ 0] & 0x10) ? 0x80 : 0x00) |
							((b[ 1] & 0x10) ? 0x40 : 0x00) |
							((b[ 2] & 0x10) ? 0x20 : 0x00) |
							((b[ 3] & 0x10) ? 0x10 : 0x00) |
							((b[ 4] & 0x10) ? 0x08 : 0x00) |
							((b[ 5] & 0x10) ? 0x04 : 0x00) |
							((b[ 6] & 0x10) ? 0x02 : 0x00) |
							((b[ 7] & 0x10) ? 0x01 : 0x00)
						  );
				p[4][1] = (UBYTE)
						  (
							((b[ 8] & 0x10) ? 0x80 : 0x00) |
							((b[ 9] & 0x10) ? 0x40 : 0x00) |
							((b[10] & 0x10) ? 0x20 : 0x00) |
							((b[11] & 0x10) ? 0x10 : 0x00) |
							((b[12] & 0x10) ? 0x08 : 0x00) |
							((b[13] & 0x10) ? 0x04 : 0x00) |
							((b[14] & 0x10) ? 0x02 : 0x00) |
							((b[15] & 0x10) ? 0x01 : 0x00)
						  );
				p[5][0] = (UBYTE)
						  (
							((b[ 0] & 0x20) ? 0x80 : 0x00) |
							((b[ 1] & 0x20) ? 0x40 : 0x00) |
							((b[ 2] & 0x20) ? 0x20 : 0x00) |
							((b[ 3] & 0x20) ? 0x10 : 0x00) |
							((b[ 4] & 0x20) ? 0x08 : 0x00) |
							((b[ 5] & 0x20) ? 0x04 : 0x00) |
							((b[ 6] & 0x20) ? 0x02 : 0x00) |
							((b[ 7] & 0x20) ? 0x01 : 0x00)
						  );
				p[5][1] = (UBYTE)
						  (
							((b[ 8] & 0x20) ? 0x80 : 0x00) |
							((b[ 9] & 0x20) ? 0x40 : 0x00) |
							((b[10] & 0x20) ? 0x20 : 0x00) |
							((b[11] & 0x20) ? 0x10 : 0x00) |
							((b[12] & 0x20) ? 0x08 : 0x00) |
							((b[13] & 0x20) ? 0x04 : 0x00) |
							((b[14] & 0x20) ? 0x02 : 0x00) |
							((b[15] & 0x20) ? 0x01 : 0x00)
						  );
				p[6][0] = (UBYTE)
						  (
							((b[ 0] & 0x40) ? 0x80 : 0x00) |
							((b[ 1] & 0x40) ? 0x40 : 0x00) |
							((b[ 2] & 0x40) ? 0x20 : 0x00) |
							((b[ 3] & 0x40) ? 0x10 : 0x00) |
							((b[ 4] & 0x40) ? 0x08 : 0x00) |
							((b[ 5] & 0x40) ? 0x04 : 0x00) |
							((b[ 6] & 0x40) ? 0x02 : 0x00) |
							((b[ 7] & 0x40) ? 0x01 : 0x00)
						  );
				p[6][1] = (UBYTE)
						  (
							((b[ 8] & 0x40) ? 0x80 : 0x00) |
							((b[ 9] & 0x40) ? 0x40 : 0x00) |
							((b[10] & 0x40) ? 0x20 : 0x00) |
							((b[11] & 0x40) ? 0x10 : 0x00) |
							((b[12] & 0x40) ? 0x08 : 0x00) |
							((b[13] & 0x40) ? 0x04 : 0x00) |
							((b[14] & 0x40) ? 0x02 : 0x00) |
							((b[15] & 0x40) ? 0x01 : 0x00)
						  );
				p[7][0] = (UBYTE)
						  (
							((b[ 0] & 0x80) ? 0x80 : 0x00) |
							((b[ 1] & 0x80) ? 0x40 : 0x00) |
							((b[ 2] & 0x80) ? 0x20 : 0x00) |
							((b[ 3] & 0x80) ? 0x10 : 0x00) |
							((b[ 4] & 0x80) ? 0x08 : 0x00) |
							((b[ 5] & 0x80) ? 0x04 : 0x00) |
							((b[ 6] & 0x80) ? 0x02 : 0x00) |
							((b[ 7] & 0x80) ? 0x01 : 0x00)
						  );
				p[7][1] = (UBYTE)
						  (
							((b[ 8] & 0x80) ? 0x80 : 0x00) |
							((b[ 9] & 0x80) ? 0x40 : 0x00) |
							((b[10] & 0x80) ? 0x20 : 0x00) |
							((b[11] & 0x80) ? 0x10 : 0x00) |
							((b[12] & 0x80) ? 0x08 : 0x00) |
							((b[13] & 0x80) ? 0x04 : 0x00) |
							((b[14] & 0x80) ? 0x02 : 0x00) |
							((b[15] & 0x80) ? 0x01 : 0x00)
						  );

				*(pptr[0]) = p[0][0];
				pptr[0]++;
				*(pptr[0]) = p[0][1];
				pptr[0]++;

				*(pptr[1]) = p[1][0];
				pptr[1]++;
				*(pptr[1]) = p[1][1];
				pptr[1]++;

				*(pptr[2]) = p[2][0];
				pptr[2]++;
				*(pptr[2]) = p[2][1];
				pptr[2]++;

				*(pptr[3]) = p[3][0];
				pptr[3]++;
				*(pptr[3]) = p[3][1];
				pptr[3]++;

				*(pptr[4]) = p[4][0];
				pptr[4]++;
				*(pptr[4]) = p[4][1];
				pptr[4]++;

				*(pptr[5]) = p[5][0];
				pptr[5]++;
				*(pptr[5]) = p[5][1];
				pptr[5]++;

				*(pptr[6]) = p[6][0];
				pptr[6]++;
				*(pptr[6]) = p[6][1];
				pptr[6]++;

				*(pptr[7]) = p[7][0];
				pptr[7]++;
				*(pptr[7]) = p[7][1];
				pptr[7]++;

			}
		}
	}
} /* ConvertTMGCtoTMGX */


/*********************************************************************
 *
 * ReadTSETTMGC
 *
 * SYNOPSIS
 *		long ReadTSETTMGC (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSETTMGC (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSETTMGC";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long			 size;
		unsigned		 tsize;
		long			 rsize;
		TilesetType		*ts  = ((TUMEInfo *)iff->IFFData.Data)->CurrentTileset;
		TMGType			*tmg = &ts->TMG;
		TileDataGroup	*tdg;
		
		size = IFFRead (iff, tmg, TSET_TMG_CHUNK_SIZE);
		if (!size) {
			return FALSE;
		}

		BigWord2Native (tmg->TileCount);
		BigWord2Native (tmg->Width);
		BigWord2Native (tmg->Height);
		BigWord2Native (tmg->Depth);
		BigWord2Native (tmg->TransparentColor);

		tsize = tmg->Width * tmg->Height;

		if (iff->IFFData.Flags & TUME_TMGC) {

			size_t		 maxtilespergrp;
			unsigned	 tcount;
			unsigned	 tilesleft;
			XTRAPntr	 xtra;
			UBYTE		*data;

			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / tsize);
			tilesleft      = tmg->TileCount;

			while (tilesleft) {

				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * tsize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, data, tsize);
					if (tsize != rsize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGC:Error reading TMGC Tile");
						return FALSE;
					}
					data += tsize;
					size += rsize;
					tcount--;
				}

				UpdateXTRA (xtra);

			}
		} else {
			size_t		 maxtilespergrp;
			UBYTE		*bdata;
			unsigned	 psize;
			unsigned	 planesize;
			unsigned	 tcount;
			unsigned	 tilesleft;
			XTRAPntr	 xtra;
			UBYTE		*data;

			tmg->Depth     = 8;
			planesize      = (((tmg->Width + 15) / 16) * 2) * tmg->Height;
			psize          = planesize * tmg->Depth;
			maxtilespergrp = (size_t)(LARGEST_XTRAMEM / psize);
			tilesleft      = tmg->TileCount;

			/*
			** Allocate temp buffer.
			*/
			bdata = (UBYTE *) IFFmalloc (tsize);
			if (!bdata) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadTSETTMGC:OOM TMGC Tile Buffer");
				return FALSE;
			}

			while (tilesleft) {
			
				tcount     = min (tilesleft, maxtilespergrp);
				tilesleft -= tcount;
				tdg        = CreateNode (sizeof (TileDataGroup), NULL);
				xtra       = AllocXTRA (tcount * psize);

				if (!tdg || !xtra) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadTSETTMGC:OOM TileData (2)");
					return FALSE;
				}

				tdg->TileData = xtra;
				tdg->NumTiles = (UWORD) tcount;
				AddTail (tmg->TileGroupList, tdg);

				data = (UBYTE *) ActivateXTRA (xtra);

				while (tcount) {
					rsize = IFFRead (iff, bdata, tsize);
					if (tsize != rsize) {
						SetGlobalErr (ERR_GENERIC);
						GEcatf ("\nReadTSETTMGC:Error reading TMGC Tile (2)");
						IFFfree (bdata);
						return FALSE;
					}

					size += rsize;

					ConvertTMGCtoTMGX (bdata, tmg->Width, tmg->Height, data);

					data += psize;
					tcount--;
				}

				UpdateXTRA (xtra);


			}

			IFFfree (bdata);
		}

		return size;
	}
} /* ReadTSETTMGC */


/*********************************************************************
 *
 * ReadTSET
 *
 * SYNOPSIS
 *		long ReadTSET (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTSET (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTSET";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	TSETTable[] = {
			{ DATA_CHUNK,	ReadTSETDATA,	},
			{ USER_CHUNK,	ReadTSETUSER,	},
			{ TMGX_CHUNK,	ReadTSETTMGX,	},
			{ TMGC_CHUNK,	ReadTSETTMGC,	},
			{ ZOOM_CHUNK,	ReadTSETZOOM,	},
			{ GUID_CHUNK,	ReadTSETGUID,	},
			{ GRID_CHUNK,	ReadTSETGRID,	},
			{ FALSE,		NULL,			},
		};

		long		 size = iff->Size;
		long		 okay = TRUE;
		long		 tsize;
		ChunkHeader	 chunkhdr;
		TUMEInfo	*ti = iff->IFFData.Data;
		TilesetType	*ts;

		iff->Size = 0;

		ts = CreateNode (sizeof (TilesetType), NULL);
		if (!ts) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadTSET:OOM TilesetType");
			return FALSE;
		}

		ti->CurrentTileset = ts;
		ts->TMG.TileGroupList = &ts->TMG.TileGroupListX;
		InitList (ts->TMG.TileGroupList);

		while (iff->Size < size && okay) {
			if (!GetChunkHeader (iff, &chunkhdr)) {
				return FALSE;
			}

			tsize      = iff->Size;
			iff->Size  = chunkhdr.Size;
			if (ti->CurrentTileset) {
				okay = CallIDFunc (chunkhdr.ID, TSETTable, iff);
				if (okay != chunkhdr.Size)
				{
					SetGlobalErr (ERR_GENERIC);
					GEcatf1 ("\nReadTSET:'%4.4s' chunk size and data do not match", &chunkhdr.ID);
/**/				goto cleanup;
				}

				if (okay & 0x01) {
					IFFSkip (iff, 1);
					okay++;
				}
			} else {
				okay = SkipChunk (iff);
			}
			tsize     += okay;
			iff->Size  = tsize;
		}

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadTSET:TSET form size and data do not match");
/**/		goto cleanup;
		}

		if (ti->CurrentTileset) {
			ti->Map->NumTilesets++;
			AddTail (ti->TilesetList, ti->CurrentTileset);
			ti->CurrentTileset = NULL;
		}

		return size;

cleanup:
		return FALSE;

	}
} /* ReadTSET */


/*********************************************************************
 *
 * ReadROOMDATA
 *
 * SYNOPSIS
 *		long ReadROOMDATA (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMDATA (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMDATA";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 origsize = iff->Size;
		long		 size;
		unsigned	 rsize;
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;
		int			 l;

		iff->Size = 0;

		size = IFFRead (iff, &room->Flags, ROOM_DATA_CHUNK_SIZE);
		if (!size) {
			goto cleanup;
		}

		BigWord2Native (room->Flags);
		BigWord2Native (room->ID);
		BigWord2Native (room->Width);
		BigWord2Native (room->Height);
		BigWord2Native (room->TileWidth);
		BigWord2Native (room->TileHeight);
		BigWord2Native (room->LayerCount);
		BigWord2Native (room->FloorNumber);

		rsize = room->Width * sizeof (PlotType);

		room->Layers = (LayerType **) IFFcalloc (room->LayerCount, sizeof (LayerType *));
		if (!room->Layers)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMDATA:OOM LayerTable");
			goto cleanup;
		}
			
		for (l = 0; l < room->LayerCount; l++) {
			LayerType	*layer;
			unsigned	 row;
			long		 tsize;

			layer = (LayerType *) IFFcalloc (1, sizeof (LayerType));
			if (!layer) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadROOMDATA:OOM LayerType");
				goto cleanup;
			}

			room->Layers[l] = layer;
			layer->Plot = (XTRAPntr *) IFFcalloc (room->Height, sizeof (XTRAPntr));
			if (!layer->Plot) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadROOMDATA:OOM Layer->Plot");
				goto cleanup;
			}

			for (row = 0; row < room->Height; row++) {
				XTRAPntr	 plot;
				PlotType	*pt;

				plot = AllocXTRA (rsize);
				if (!plot) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadROOMDATA:OOM Room Row");
					goto cleanup;
				}

				layer->Plot[row] = plot;

				pt    = (PlotType *) ActivateXTRA (plot);
				tsize = IFFRead (iff, pt, rsize);
				if (rsize != tsize) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMDATA:Error reading ROOM DATA");
					goto cleanup;
				}

				size += tsize;

				{
					unsigned	 ndx;

					for (ndx = 0; ndx < room->Width; ndx++) {
						BigWord2Native (pt->TileID);
						pt++;
					}
				}
				UpdateXTRA (plot);
			}
		}

		{
			long	nsize;

			nsize = origsize - size;

			if (!nsize) {
				room->Name = dupstr ("");
			} else {
				room->Name = (char *) IFFcalloc (1, nsize + 1);
				if (!room->Name) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadROOMDATA:OOM ROOM Name");
					goto cleanup;
				}
				nsize = IFFRead (iff, room->Name, nsize);
				size += nsize;
			}
		}

		return size;
cleanup:
		return FALSE;
	}
} /* ReadROOMDATA */


#if 1
/*********************************************************************
 *
 * ReadROOMNAME
 *
 * PURPOSE
 *		Read a 'ROOM''NAME' chunk.
 *
 * INPUT
 *		
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Number of bytes read, or FALSE if error.
 *
 * HISTORY
 *		08/23/94 (dcc) - created.
 *
*/
typedef struct
{
	UWORD	Flags;
	WORD	R_id;
	UWORD	LayerCount;
	UWORD	FloorNumber;
} RoomBuff1Type;

static long ReadROOMNAME(IFFTracker *iff)
{
	RoomType		*prm = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;
	long			 count, size;
	RoomBuff1Type	 rb1;
	int				 cName0;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMNAME";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	size = iff->Size;
	iff->Size = 0;

	if ((count = IFFRead(iff, &rb1, sizeof (RoomBuff1Type))) == 0)
/**/	goto cleanup;

	BigWord2Native(rb1.Flags);
	BigWord2Native(rb1.R_id);
	BigWord2Native(rb1.LayerCount);
	BigWord2Native(rb1.FloorNumber);

	prm->Flags		= rb1.Flags;
	prm->ID			= rb1.R_id;
	prm->LayerCount	= rb1.LayerCount;
	prm->FloorNumber= rb1.FloorNumber;

	prm->Layers		= (LayerType **) IFFcalloc(prm->LayerCount, sizeof(LayerType *));
	if (!prm->Layers)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf ("\nReadROOMNAME:OOM LayerTable");
/**/	goto cleanup;
	}

	if (count > size)
/**/	goto cleanup;

	cName0 = (int) (size - count);

	if (!cName0)
	{
		prm->Name = dupstr("");
	}
	else
	{
		if ((prm->Name = (char *) IFFcalloc(1, cName0+1)) == NULL)
		{
			SetGlobalErr(ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMNAME:OOM ROOM Name");
/**/		goto cleanup;
		}
		if (IFFRead(iff, prm->Name, cName0) == FALSE)
/**/		goto cleanup;
	}
	return size;

cleanup:
	return FALSE;
} /* ReadROOMNAME */


/*********************************************************************
 *
 * ReadROOMLAYR
 *
 * PURPOSE
 *		Read a 'ROOM''LAYR' chunk.
 *
 * INPUT
 *		
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Number of bytes read, or FALSE if error.
 *
 * HISTORY
 *		08/24/94 (dcc) - created.
 *
*/
typedef struct
{
	UWORD	LayerType;
	UWORD	LayerNumber;
	UWORD	ctilx;
	UWORD	ctily;
	UWORD	cxTile;
	UWORD	cyTile;
} LayerBuffType;

static long ReadROOMLAYR(IFFTracker *iff)
{
	RoomType		*prm = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;
	long			 count, size;
	LayerBuffType	 lyb;
	LayerType		*play;
	unsigned		 cRow;
	unsigned		 itily;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMLAYR";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	size = iff->Size;
	iff->Size = 0;

	/* Read in the fixed size info */

	if ((count = IFFRead(iff, &lyb, sizeof (LayerBuffType))) == 0)
/**/	goto cleanup;

	BigWord2Native(lyb.LayerType);		// ignore for now
	BigWord2Native(lyb.LayerNumber);
	BigWord2Native(lyb.ctilx);
	BigWord2Native(lyb.ctily);
	BigWord2Native(lyb.cxTile);
	BigWord2Native(lyb.cyTile);

	if (lyb.LayerNumber == 1)
	{
		prm->Width		= lyb.ctilx;
		prm->Height		= lyb.ctily;
		prm->TileWidth	= lyb.cxTile;
		prm->TileHeight	= lyb.cyTile;
	}

	play = (LayerType *) IFFcalloc(1, sizeof(LayerType));
	if (!play)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf ("\nReadROOMLAYR:OOM LayerType");
		goto cleanup;
	}
	prm->Layers[lyb.LayerNumber-1] = play;

	play->Plot = (XTRAPntr *) IFFcalloc(lyb.ctily, sizeof(XTRAPntr));
	if (!play->Plot)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf ("\nReadROOMLAYR:OOM Layer->Plot");
/**/	goto cleanup;
	}
	cRow = lyb.ctilx * sizeof(PlotType);

	for (itily = 0; itily < lyb.ctily; itily++)
	{
		XTRAPntr	 xp;
		PlotType	*pplt;

		if ((xp = AllocXTRA(cRow)) == NULL)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMLAYR:OOM Room Row");
/**/		goto cleanup;
		}
		play->Plot[itily] = xp;

		pplt = (PlotType *) ActivateXTRA(xp);
		if (IFFRead(iff, pplt, cRow) != cRow)
		{
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadROOMLAYR:Error reading ROOM LAYR");
/**/		goto cleanup;
		}
		count += cRow;
#if __LITTLEENDIAN__
		{
			unsigned itilx;

			for (itilx = 0; itilx < lyb.ctilx; itilx++)
			{
				BigWord2Native(pplt->TileID);
				pplt++;
			}
		}
#endif // __LITTLEENDIAN__
		UpdateXTRA(xp);
	}
	if (count != size)
/**/	goto cleanup;

	return count;
cleanup:
	return FALSE;
} /* ReadROOMLAYR */
#endif // 1


/*********************************************************************
 *
 * ReadROOMLAY1
 *
 * PURPOSE
 *		Read a 'ROOM''LAY1' chunk.
 *
 * INPUT
 *		
 *
 * ASSUMES
 *		
 *
 * RETURN VALUE
 *		Number of bytes read, or FALSE if error.
 *
 * HISTORY
 *		08/24/94 (dcc) - created.
 *
*/
typedef struct
{
	UWORD	dtilx;
	UWORD	dtily;
	PlotType plt;
} XYTileBuffType;

static long ReadROOMLAY1(IFFTracker *iff)
{
	RoomType		*prm = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;
	long			 count, size;
	LayerBuffType	 lyb;
	LayerType		*play;
	unsigned		 cRow;
	unsigned		 itily;
	ULONG			 ulTiles;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMLAY1";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	size = iff->Size;
	iff->Size = 0;

	/* Read in the fixed size info */

	if ((count = IFFRead(iff, &lyb, sizeof (LayerBuffType))) == 0)
/**/	goto cleanup;

	BigWord2Native(lyb.LayerType);		// ignore for now
	BigWord2Native(lyb.LayerNumber);
	BigWord2Native(lyb.ctilx);
	BigWord2Native(lyb.ctily);
	BigWord2Native(lyb.cxTile);
	BigWord2Native(lyb.cyTile);

	if (lyb.LayerNumber == 1)
	{
		prm->Width		= lyb.ctilx;
		prm->Height		= lyb.ctily;
		prm->TileWidth	= lyb.cxTile;
		prm->TileHeight	= lyb.cyTile;
	}

	play = (LayerType *) IFFcalloc(1, sizeof(LayerType));
	if (!play)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf ("\nReadROOMLAY1:OOM LayerType");
		goto cleanup;
	}
	prm->Layers[lyb.LayerNumber-1] = play;

	play->Plot = (XTRAPntr *) IFFcalloc(lyb.ctily, sizeof(XTRAPntr));
	if (!play->Plot)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEcatf ("\nReadROOMLAY1:OOM Layer->Plot");
/**/	goto cleanup;
	}
	cRow = lyb.ctilx * sizeof(PlotType);

	for (itily = 0; itily < lyb.ctily; itily++)
	{
		XTRAPntr	 xp;

		if ((xp = AllocXTRA(cRow)) == NULL)
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMLAY1:OOM Room Row");
/**/		goto cleanup;
		}
		play->Plot[itily] = xp;
	}

	if (IFFRead (iff, &ulTiles, sizeof(ULONG)) != sizeof(ULONG))
/**/	goto cleanup;

	BigLong2Native(ulTiles);

	count += sizeof(ULONG);

	while (ulTiles--)
	{
		long cBytes;
		XYTileBuffType xyt;

		if ((cBytes = IFFRead (iff, &xyt, sizeof(XYTileBuffType))) != sizeof(XYTileBuffType))
/**/		goto cleanup;

		BigWord2Native(xyt.dtilx);
		BigWord2Native(xyt.dtily);
		BigWord2Native(xyt.plt.TileID);

		count += cBytes;

		WriteXTRA(&xyt.plt, sizeof(PlotType),
					play->Plot[xyt.dtily], xyt.dtilx * sizeof(PlotType));
	}
	if (count != size)
/**/	goto cleanup;

	return count;
cleanup:
	return FALSE;
} /* ReadROOMLAY1 */


/*********************************************************************
 *
 * ReadROOMUSER
 *
 * SYNOPSIS
 *		long ReadROOMUSER (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMUSER (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMUSER";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		size = IFFRead (iff, &room->UserType, ROOM_USER_CHUNK_SIZE);
		if (size) {
			BigWord2Native (room->UserType);
			BigWord2Native (room->UserNumber);
			BigWord2Native (room->pad0);
			BigWord2Native (room->pad1);
		}
		return size;
	}
} /* ReadROOMUSER */

/*********************************************************************
 *
 * ReadROOMCMNT
 *
 * SYNOPSIS
 *		long ReadROOMCMNT (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMCMNT (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMCMNT";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		long		 origsize = iff->Size;
		long		 size     = 0;
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		iff->Size = 0;

		if (origsize) {
			static	ListType	 StrListX;
			static	ListType	*StrList = &StrListX;

			int					 strcount = 0;
			UWORD				 strsize;
			long				 okay;
			char				*str;
			NodeType			*nd;

			InitList (StrList);

			while (origsize - size) {

				okay = IFFRead (iff, &strsize, sizeof (UWORD));
				if (okay != sizeof (UWORD)) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMCMNT:Error reading comments");
					EmptyList (StrList);
					return FALSE;
				}
				size += okay;

				BigWord2Native (strsize);

				str = (char *) IFFcalloc (1, strsize + 1);
				nd  = CreateNode (sizeof (NodeType), NULL);

				if (!str || !nd) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadROOMCMNT:OOM reading comments");
					EmptyList (StrList);
					return FALSE;
				}

				AddTail (StrList, nd);
				NodeName (nd) = str;

				okay = IFFRead (iff, str, strsize);
				if (okay != strsize) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMCMNT:Error reading comments (2)");
					EmptyList (StrList);
					return FALSE;
				}

				size += okay;
				strcount++;
			}

			room->Comments = (char **) IFFcalloc (strcount + 1, sizeof (char *));
			if (!room->Comments) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadROOMCMNT:OOM reading comments (2)");
				EmptyList (StrList);
				return FALSE;
			}
				
			nd = Tail (StrList);
			while (!StartOfList (nd)) {
				strcount--;
				room->Comments[strcount] = NodeName (nd);
				NodeName(nd) = NULL;
				nd = Prev (nd);
			}

			EmptyList (StrList);
		}

		return size;
	}
} /* ReadROOMCMNT */

/*********************************************************************
 *
 * ReadROOMCMAP
 *
 * SYNOPSIS
 *		long ReadROOMCMAP (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMCMAP (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMCMAP";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		return ReadCMAP (iff, &room->ColorInfo.NumColors, &room->ColorInfo.Colors);
	}
} /* ReadROOMCMAP */

/*********************************************************************
 *
 * ReadROOMCFLG
 *
 * SYNOPSIS
 *		long ReadROOMCFLG (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMCFLG (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMCFLG";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		size = IFFRead (iff, &room->ColorInfo.CycleFlag, ROOM_CFLG_CHUNK_SIZE);
		if (size) {
			BigWord2Native (room->ColorInfo.CycleFlag);
		}
		return size;
	}
} /* ReadROOMCFLG */

/*********************************************************************
 *
 * ReadROOMGRID
 *
 * SYNOPSIS
 *		long ReadROOMGRID (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMGRID (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMGRID";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		return ReadGChunk (iff, (GChunk *)&room->GridWidth);
	}
} /* ReadROOMGRID */

/*********************************************************************
 *
 * ReadROOMGUID
 *
 * SYNOPSIS
 *		long ReadROOMGUID (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMGUID (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMGUID";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		return ReadGChunk (iff, (GChunk *)&room->GuideWidth);
	}
} /* ReadROOMGUID */

/*********************************************************************
 *
 * ReadROOMZOOM
 *
 * SYNOPSIS
 *		long ReadROOMZOOM (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMZOOM (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMZOOM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;

		return ReadZOOM (iff, (ZOOMChunk *)&room->ZoomDstDup);
	}
} /* ReadROOMZOOM */

/*********************************************************************
 *
 * ReadROOMHSVP
 *
 * SYNOPSIS
 *		long ReadROOMHSVP (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMHSVP (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMHSVP";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;
		RoomType	*room = ((TUMEInfo *)iff->IFFData.Data)->CurrentRoom;
		HSVType		*phsv;
		unsigned	 ncolors;

		ncolors = (unsigned)(iff->Size / sizeof (HSVType));

		phsv = (HSVType *) IFFmalloc (iff->Size);
		if (!phsv) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMHSVP:OOM hsv palette");
			return FALSE;
		}

		room->ColorInfo.HSVs = phsv;

		size = IFFRead (iff, phsv, iff->Size);
		if (size) {
			while (ncolors) {
				BigWord2Native (phsv->Hue);
				phsv++;
				ncolors--;
			}
		}

		return size;
	}
} /* ReadROOMHSVP */

/*********************************************************************
 *
 * ReadROOMCINF
 *
 * SYNOPSIS
 *		long ReadROOMCINF (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOMCINF (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOMCINF";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		long		 size;
		TUMEInfo	*ti   = iff->IFFData.Data;
		RoomType	*room = ti->CurrentRoom;
		CycleInfo	*cycle;

		cycle = (CycleInfo *) IFFcalloc (sizeof (CycleInfo), 1);
		if (!cycle) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOMCINF:OOM cycleinfo");
			return FALSE;
		}

		size = IFFRead (iff, &cycle->Speed, ROOM_CINF_CHUNK_SIZE);
		if (size) {
			ColorRegister	*pcr;
			HSVType			*phsv;
			UWORD			*preg;
			unsigned		 count;
			long			 tsize;

			BigWord2Native (cycle->Speed);
			BigWord2Native (cycle->Direction);
			BigWord2Native (cycle->Flags);
			BigWord2Native (cycle->NumColors);
			BigWord2Native (cycle->NumRegs);

			if (cycle->NumColors) {
				pcr  = (ColorRegister *) IFFmalloc (sizeofColorRegister * cycle->NumColors);
				phsv = (HSVType *) IFFmalloc (sizeof (HSVType) * cycle->NumColors);
				if (!pcr || !phsv) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadROOMCINF:OOM cycleinfo 2");
					return FALSE;
				}
				cycle->Colors    = pcr;
				cycle->HSVs      = phsv;
			}
			if (cycle->NumRegs) {
				preg = (UWORD *) IFFmalloc (sizeof (UWORD) * cycle->NumRegs);
				if (!preg) {
					SetGlobalErr (ERR_OUT_OF_MEMORY);
					GEcatf ("\nReadROOMCINF:OOM cycleinfo 2");
					return FALSE;
				}
				cycle->Registers = preg;
			}

			count = cycle->NumColors;
			while (count) {

				tsize = IFFRead (iff, pcr, sizeofColorRegister);
				if (tsize != sizeofColorRegister) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMCINF:Error reading ROOM CINF");
					goto cleanup;
				}
				size += tsize;

				tsize = IFFRead (iff, phsv, sizeof (HSVType));
				if (tsize != sizeof (HSVType)) {
					SetGlobalErr (ERR_GENERIC);
					GEcatf ("\nReadROOMCINF:Error reading ROOM CINF HSV");
					goto cleanup;
				}
				size += tsize;

				BigWord2Native (phsv->Hue);

				pcr++;
				phsv++;
				count--;
			}

			tsize = IFFRead (iff, preg, sizeof (UWORD) * cycle->NumRegs);
			if (tsize != sizeof (UWORD) * cycle->NumRegs) {
				SetGlobalErr (ERR_GENERIC);
				GEcatf ("\nReadROOMCINF:Error reading ROOM CINF REGS");
				goto cleanup;
			}

			size += tsize;

			count = cycle->NumRegs;
			while (count) {
			 	BigWord2Native (*preg);
				preg++;
				count--;
			}
		}

		AddTail (ti->CycleList, cycle);
		room->ColorInfo.NumCycles++;

		return size;
cleanup:
		return FALSE;
	}
} /* ReadROOMCINF */

/*********************************************************************
 *
 * ReadROOM
 *
 * SYNOPSIS
 *		long ReadROOM (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadROOM (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadROOM";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	ROOMTable[] = {
			{ DATA_CHUNK,	ReadROOMDATA,	},
#if 1
			{ NAME_CHUNK,	ReadROOMNAME,	},
			{ LAYR_CHUNK,	ReadROOMLAYR,	},
			{ LAY1_CHUNK,	ReadROOMLAY1,	},
#endif // 1
			{ USER_CHUNK,	ReadROOMUSER,	},
			{ CMNT_CHUNK,	ReadROOMCMNT,	},
			{ CMAP_CHUNK,	ReadROOMCMAP,	},
			{ CFLG_CHUNK,	ReadROOMCFLG,	},
		/*	{ CYCL_CHUNK,	ReadROOMCYCL,	}, */
			{ GRID_CHUNK,	ReadROOMGRID,	},
			{ GUID_CHUNK,	ReadROOMGUID,	},
			{ ZOOM_CHUNK,	ReadROOMZOOM,	},
			{ CINF_CHUNK,	ReadROOMCINF,	},
			{ HSVP_CHUNK,	ReadROOMHSVP,	},
			{ FALSE,		NULL,			},
		};

		long		 size = iff->Size;
		long		 okay = TRUE;
		long		 tsize;
		ChunkHeader	 chunkhdr;
		RoomType	*room;
		TUMEInfo	*ti = iff->IFFData.Data;
		
		room = CreateNode (sizeof (RoomType), NULL);
		if (!room) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadROOM:OOM RoomType");
			return FALSE;
		}

		ti->CycleList = &ti->CycleListX;
		InitList (ti->CycleList);

		ti->CurrentRoom = room;

		iff->Size = 0;
		while (iff->Size < size && okay) {
			if (!GetChunkHeader (iff, &chunkhdr)) {
				return FALSE;
			}

			tsize      = iff->Size;
			iff->Size  = chunkhdr.Size;
			okay       = CallIDFunc (chunkhdr.ID, ROOMTable, iff);
			if (okay != chunkhdr.Size)
			{
				SetGlobalErr (ERR_GENERIC);
				GEcatf1 ("\nReadROOM:'%4.4s' chunk size and data do not match", &chunkhdr.ID);
/**/			goto cleanup;
			}

			if (okay & 0x01) {
				IFFSkip (iff, 1);
				okay++;
			}
			tsize     += okay;
			iff->Size  = tsize;
		}

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadROOM:ROOM form size and data do not match");
/**/		goto cleanup;
		}

		if (room->ColorInfo.NumCycles) {
			room->ColorInfo.Cycles = (CycleInfo **) IFFmalloc (room->ColorInfo.NumCycles * sizeof (CycleInfo *));
			if (!room->ColorInfo.Cycles) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEcatf ("\nReadROOM:OOM ColorInfo.Cycle Array");
				goto cleanup;
			}

			{
				CycleInfo	*cycle;
				CycleInfo	**cycleptr;

				cycleptr = room->ColorInfo.Cycles;
				while ((cycle = RemHead (ti->CycleList)) != NULL) {
					*cycleptr = cycle;
					cycleptr++;
				}
			}
		}

		ti->Map->NumRooms++;
		AddTail (ti->RoomList, room);
		ti->CurrentRoom = NULL;

		return size;

cleanup:
		EmptyList (ti->CycleList);

		return FALSE;

	}
} /* ReadROOM */

/*********************************************************************
 *
 * ReadTUME
 *
 * SYNOPSIS
 *		long ReadTUME (IFFTracker *iff)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
long ReadTUME (IFFTracker *iff)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ReadTUME";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	TUMETable[] = {
			{ ROOM_FORM,	ReadROOM,	},
			{ TSET_FORM,	ReadTSET,	},
			{ FALSE,		NULL,		},
		};

		long			 size = iff->Size;
		long			 okay = TRUE;
		long			 tsize;
		IFFRoutineID	*iffrtn;
		TUMEInfo		*ti;
		MapType			*map = iff->IFFData.Data;

		iff->Size = 0;

		ti  = (TUMEInfo *) IFFcalloc (1, sizeof (TUMEInfo));
		if (!map) {
			map = (MapType *) IFFcalloc (1, sizeof (MapType));
		}

		if (!ti || !map) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadTUME:OOM TUMEInfo/MapType");
			return FALSE;
		}

		iff->IFFData.Data = ti;
		ti->Map           = map;
		ti->RoomList      = &ti->RoomListX;
		ti->TilesetList   = &ti->TilesetListX;
		InitList (ti->RoomList);
		InitList (ti->TilesetList);

		/*
		** Add OLD Rooms and Tileset to TUMEInfo
		*/
		{
			unsigned	 i;
			RoomType	**roomptr;
			TilesetType	**tsptr;

			roomptr = map->Rooms + 1;
			for (i = 0; i < map->NumRooms; i++) {
				(*roomptr)->LoadFlags = LOADFLAG_OLDROOM;
				AddTail (ti->RoomList, *roomptr);
				roomptr++;
			}

			tsptr = map->Tilesets + 1;
			for (i = 0; i < map->NumTilesets; i++) {
				AddTail (ti->TilesetList, *tsptr);
				tsptr++;
			}
		}


		iffrtn = iff->FORMTable;
		iff->FORMTable = TUMETable;

		while (iff->Size < size && okay) {
			tsize      = iff->Size;
			okay       = ReadIFF (iff);
			tsize     += okay;
			iff->Size  = tsize;
		}

		iff->FORMTable = iffrtn;

		if (!okay) {
/**/		goto cleanup;
		}

		if (iff->Size != size) {
			SetGlobalErr (ERR_GENERIC);
			GEcatf ("\nReadTUME:TUME form size and data do not match");
/**/		goto cleanup;
		}

		/*
		**	Renumber New Room Data
		*/
		{
			RoomType	*room;

			room = Head (ti->RoomList);
			while (!EndOfList (room)) {
				if (!(room->LoadFlags & LOADFLAG_OLDROOM)) {
					unsigned	l;

					for (l = 0; l < room->LayerCount; l++) {
						LayerType	*layer;
						unsigned	 row;

						layer = room->Layers[l];
						for (row = 0; row < room->Height; row++) {
							PlotType	*plot;
							unsigned	 ndx;

							plot = (PlotType *) ActivateXTRA (layer->Plot[row]);
							for (ndx = 0; ndx < room->Width; ndx++) {
								plot->TilesetID = ti->TranslateTable[plot->TilesetID];
								plot++;
							}
							UpdateXTRA (layer->Plot[row]);
						}
					}
				}
				room = Next (room);
			}
		}

		/*
		** Connect Composite Tilesets to their Rooms.
		*/
		{
			TilesetType	*ts;

			ts = Head (ti->TilesetList);
			while (!EndOfList (ts)) {
				if (ts->Flags & TUME_TILESET_COMPOSITE) {
					RoomType	*room;
					RoomType	*nextroom;
					BOOL		 found = FALSE;

					room = Head (ti->RoomList);
					while (!EndOfList (room)) {
						nextroom = Next (room);
						if (!strcmp (room->Name, ts->FileSpec)) {
							if (!found) {
								found = TRUE;
								room->CompositeTSet = ts;
								ts->GridedTilesRoom = room;
								ts->TilesAcross = (WORD) (ts->Width  / room->TileWidth);
								ts->TilesDown   = (WORD) (ts->Height / room->TileHeight);
							} else {
								Remove (room);
								DestroytUMERoom (room);
							}
						}
						room = nextroom;
					}
				}
				ts = Next (ts);
			}
		}

		/*
		**	Create Room Array and Tileset Array
		*/
		if (map->Rooms)		IFFfree (map->Rooms);
		if (map->Tilesets)	IFFfree (map->Tilesets);

		map->Rooms    = (RoomType **) IFFcalloc (map->NumRooms + 2, sizeof (RoomType *));
		map->Tilesets = (TilesetType **) IFFcalloc (map->NumTilesets + 2, sizeof (TilesetType *));
		if (!map->Rooms || !map->Tilesets) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEcatf ("\nReadTUME:OOM TUMEInfo/MapType");
			goto cleanup;
		}

		{
			RoomType	*room;
			RoomType	**roomptr;
			TilesetType	*ts;
			TilesetType	**tsptr;

			roomptr = map->Rooms + 1;
			room    = Head (ti->RoomList);
			while (!EndOfList (room)) {
				*roomptr = room;
				roomptr++;
				room = Next (room);
			}

			tsptr = map->Tilesets + 1;
			ts    = Head (ti->TilesetList);
			while (!EndOfList (ts)) {
				*tsptr = ts;
				tsptr++;
				ts = Next (ts);
			}
		}

		iff->IFFData.Data = map;
		IFFfree (ti);

		return size;

cleanup:
/*		FreeTUMEInfo (iff->IFFData.Data); */
		return FALSE;
		
	}
} /* ReadTUME */

/*********************************************************************
 *
 * AppendtUMEMap
 *
 * SYNOPSIS
 *		MapType *AppendtUMEMap (MapType *map, char *filename)
 *
 * PURPOSE
 *		Load or Append a tUME map info memory from a tUME map file.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
MapType *AppendtUMEMap (MapType *map, char *filename, UWORD flags)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "AppendtUMEMap";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	{
		static IFFRoutineID	TUMETable[] = {
			{ TUME_FORM,	ReadTUME,	},
			{ FALSE,		NULL,		},
		};

		IFFTracker	iff = { 0, 0, TUMETable, };
		long		okay;
		long		id = 0;

		iff.IFFData.Flags = flags;
		iff.IFFData.Data  = map;

		iff.fd = EIO_ReadOpen (filename);
		if (iff.fd < 0) {
			SetGlobalErr (ERR_FILE_NOT_FOUND);
			GEcatf1 ("\nAppendtUMEMap:Couldn't open '%s'", filename);
			return NULL;
		}

		strcpy (iff.Name, EIO_Name(filename));
		EIO_Read (iff.fd, &id, sizeof (id));
		EIO_Seek (iff.fd, 0, EIO_SEEK_BEGINNING);
		if (id == SIFF_CHUNK)
		{
			okay = ReadCIM (&iff);
		}
		else
		{
			okay = ReadIFF (&iff);
		}
		EIO_Close(iff.fd);
		if (!okay) {
			return NULL;
		}

		return iff.IFFData.Data;

	}
} /* AppendtUMEMap */

