/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * KEYEVENT.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 08/08/89 
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *    Routines to "Handle" keys.
 *		
 *
 * HISTORY
 *		03/19/93 Friday (dcc) - fix spelling of "non-existent".
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/etypes.h>
#include <ctype.h>
#include <stdio.h>
#include <echidna/listfunc.h>
#include <echidna/argparse.h>
#include <echidna/eerrors.h>
#include <echidna/events.h>
#include <echidna/readini.h>
#include <echidna/keyboard.h>
#include <echidna/strings.h>
#include <echidna/menus2.h>
#include <echidna/eui.h>

#define DEBUGGING	0

#if DEBUGGING
#include <echidna/fonts.h>
#include <echidna/graphics.h>
#endif


/**************************** C O N S T A N T S ***************************/

#define LOCAL	static

#define KEY_EVENT	1

/******************************** T Y P E S *******************************/

typedef struct {
	NodeType	Node;
	KeyValue	Key;
	USHORT		Qualifier;
	short		Command;
	void		*CommandData;
	char		*Label;
} KeyType;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/

/*
 * Global Key Routine variables.
 */
ListType	 KeyListX;
ListType	*KeyList = &KeyListX;

#if __AMIGAOS__
/*
 * DeadKeyConvert Global Variables
 */
struct	ConsoleDevice	*ConsoleDevice = NULL;
struct	IOStdReq		 ConsoleIoReq;
#endif

#define QUAL(q)		(((q) & (IEQUALIFIER_LSHIFT  | \
				 IEQUALIFIER_CONTROL | \
				 IEQUALIFIER_LALT    | \
				 IEQUALIFIER_NUMERICPAD)) | \
			(((q) & IEQUALIFIER_RSHIFT) >> 1) | \
			(((q) & IEQUALIFIER_RALT) >> 1 ))

typedef struct {
	char	*name;
	USHORT	bits;
} Qualifier;

LOCAL Qualifier Qualifiers[] = {
{	"Alt-",		QUAL_ALT,	},
{	"Shift-",	QUAL_SHIFT, },
{	"Ctrl-",	QUAL_CTRL,	},
#if __AMIGAOS__/*dcc*/
{	"KeyPad-",	QUAL_KEYPAD, },
#endif
{	NULL, NULL,		},
};

/*
** Error Routines *********************************************************
*/
LOCAL	short	 LineNo = 0;

LOCAL void StartError (void)
{
	printf ("Error at line #%d: ", LineNo);
}

LOCAL void ParseError1 (
	char	*message,
	char	*arg1
) {
	StartError ();
	printf (message, arg1);
}

LOCAL void KeyError (
	char *mess
) {
	printf ("%s\n", mess);
}

/*
** Get / Strip Qualifier.
*/
LOCAL USHORT GetQualifier (
	char	*keystring
) {
	Qualifier	*qu  = Qualifiers;
	USHORT		qual = NULL;
	int		len;

	while (qu->name) {
		len = strlen (qu->name);
		if (!strnicmp (keystring, qu->name, len)) {
			strcpy (keystring, keystring + len);
			qual = qu->bits | GetQualifier (keystring);
			break;
		}
		qu++;
	}

	return qual;
}


#if __AMIGAOS__
/*
** KeyCode type.
*/
typedef struct {
	char		*KeyName;
	char		KeyCode[10];
	USHORT		Qualifier;
} KeyCode;

#define CopyKeyCodes(dest,src)	\
	strcpy ((dest), (src));

#define SetKeyCode(kc,key)		\
	(kc)->KeyCode[0] = (key);	\
	(kc)->KeyCode[1] = '\0';

#define CmpKeyCodes(one,two)	\
	(strcmp(one,two))

/*
** Key Code Data.
*/
LOCAL KeyCode KeyCodes[] = {
	{ "F1",		{ 0x9B, '0', '~', 0 },	NULL },
	{ "F2",		{ 0x9B, '1', '~', 0 },	NULL },
	{ "F3",		{ 0x9B, '2', '~', 0 },	NULL },
	{ "F4",		{ 0x9B, '3', '~', 0 },	NULL },
	{ "F5",		{ 0x9B, '4', '~', 0 },	NULL },
	{ "F6",		{ 0x9B, '5', '~', 0 },	NULL },
	{ "F7",		{ 0x9B, '6', '~', 0 },	NULL },
	{ "F8",		{ 0x9B, '7', '~', 0 },	NULL },
	{ "F9",		{ 0x9B, '8', '~', 0 },	NULL },
	{ "F10",	{ 0x9B, '9', '~', 0 },	NULL },
	{ "Help",	{ 0x9B, '?', '~', 0 },	NULL },

	{ "Shift-F1",	{ 0x9B, '1', '0', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F2",	{ 0x9B, '1', '1', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F3",	{ 0x9B, '1', '2', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F4",	{ 0x9B, '1', '3', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F5",	{ 0x9B, '1', '4', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F6",	{ 0x9B, '1', '5', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F7",	{ 0x9B, '1', '6', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F8",	{ 0x9B, '1', '7', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F9",	{ 0x9B, '1', '8', '~', 0 },	QUAL_SHIFT },
	{ "Shift-F10",	{ 0x9B, '1', '9', '~', 0 },	QUAL_SHIFT },

	{ "Up",		{ 0x9B, 'A', 0 },	NULL },
	{ "Down",	{ 0x9B, 'B', 0 },	NULL },
	{ "Left",	{ 0x9B, 'D', 0 },	NULL },
	{ "Right",	{ 0x9B, 'C', 0 },	NULL },

	{ "Shift-Up",	{ 0x9B, 'T', 0 },	QUAL_SHIFT },
	{ "Shift-Down",	{ 0x9B, 'S', 0 },	QUAL_SHIFT },
	{ "Shift-Left",	{ 0x9B, 'A', ' ', 0 },	QUAL_SHIFT },
	{ "Shift-Right",{ 0x9B, '@', ' ', 0 },	QUAL_SHIFT },

	{ "BkSp",	{ 0x08, 0 },	NULL },
	{ "Esc",	{ 0x1b, 0 },	NULL },
	{ "Tab",	{ 0x09, 0 },	NULL },
	{ "Space",	{ 0x20, 0 },	NULL },
	{ "Return",	{ 0x0D, 0 },	NULL },
	{ "Enter",	{ 0x0D, 0 },	NULL },
	{ "Del", 	{ 0x7F, 0 },	NULL },

	{ "^a",		{ 0x01, 0 },	QUAL_CTRL },
	{ "^b",		{ 0x02, 0 },	QUAL_CTRL },
	{ "^c",		{ 0x03, 0 },	QUAL_CTRL },
	{ "^d",		{ 0x04, 0 },	QUAL_CTRL },
	{ "^e",		{ 0x05, 0 },	QUAL_CTRL },
	{ "^f",		{ 0x06, 0 },	QUAL_CTRL },
	{ "^g",		{ 0x07, 0 },	QUAL_CTRL },
	{ "^h",		{ 0x08, 0 },	QUAL_CTRL },
	{ "^i",		{ 0x09, 0 },	QUAL_CTRL },
	{ "^j",		{ 0x0A, 0 },	QUAL_CTRL },
	{ "^k",		{ 0x0B, 0 },	QUAL_CTRL },
	{ "^l",		{ 0x0C, 0 },	QUAL_CTRL },
	{ "^m",		{ 0x0D, 0 },	QUAL_CTRL },
	{ "^n",		{ 0x0E, 0 },	QUAL_CTRL },
	{ "^o",		{ 0x0F, 0 },	QUAL_CTRL },
	{ "^p",		{ 0x10, 0 },	QUAL_CTRL },
	{ "^q",		{ 0x11, 0 },	QUAL_CTRL },
	{ "^r",		{ 0x12, 0 },	QUAL_CTRL },
	{ "^s",		{ 0x13, 0 },	QUAL_CTRL },
	{ "^t",		{ 0x14, 0 },	QUAL_CTRL },
	{ "^u",		{ 0x15, 0 },	QUAL_CTRL },
	{ "^v",		{ 0x16, 0 },	QUAL_CTRL },
	{ "^w",		{ 0x17, 0 },	QUAL_CTRL },
	{ "^x",		{ 0x18, 0 },	QUAL_CTRL },
	{ "^y",		{ 0x19, 0 },	QUAL_CTRL },
	{ "^z",		{ 0x1A, 0 },	QUAL_CTRL },

	{ "Shift-a",	{ 0x41, 0 },	QUAL_SHIFT },
	{ "Shift-b",	{ 0x42, 0 },	QUAL_SHIFT },
	{ "Shift-c",	{ 0x43, 0 },	QUAL_SHIFT },
	{ "Shift-d",	{ 0x44, 0 },	QUAL_SHIFT },
	{ "Shift-e",	{ 0x45, 0 },	QUAL_SHIFT },
	{ "Shift-f",	{ 0x46, 0 },	QUAL_SHIFT },
	{ "Shift-g",	{ 0x47, 0 },	QUAL_SHIFT },
	{ "Shift-h",	{ 0x48, 0 },	QUAL_SHIFT },
	{ "Shift-i",	{ 0x49, 0 },	QUAL_SHIFT },
	{ "Shift-j",	{ 0x4A, 0 },	QUAL_SHIFT },
	{ "Shift-k",	{ 0x4B, 0 },	QUAL_SHIFT },
	{ "Shift-l",	{ 0x4C, 0 },	QUAL_SHIFT },
	{ "Shift-m",	{ 0x4D, 0 },	QUAL_SHIFT },
	{ "Shift-n",	{ 0x4E, 0 },	QUAL_SHIFT },
	{ "Shift-o",	{ 0x4F, 0 },	QUAL_SHIFT },
	{ "Shift-p",	{ 0x50, 0 },	QUAL_SHIFT },
	{ "Shift-q",	{ 0x51, 0 },	QUAL_SHIFT },
	{ "Shift-r",	{ 0x52, 0 },	QUAL_SHIFT },
	{ "Shift-s",	{ 0x53, 0 },	QUAL_SHIFT },
	{ "Shift-t",	{ 0x54, 0 },	QUAL_SHIFT },
	{ "Shift-u",	{ 0x55, 0 },	QUAL_SHIFT },
	{ "Shift-v",	{ 0x56, 0 },	QUAL_SHIFT },
	{ "Shift-w",	{ 0x57, 0 },	QUAL_SHIFT },
	{ "Shift-x",	{ 0x58, 0 },	QUAL_SHIFT },
	{ "Shift-y",	{ 0x59, 0 },	QUAL_SHIFT },
	{ "Shift-z",	{ 0x5A, 0 },	QUAL_SHIFT },

	{ NULL, },
};
#elif __MSDOS__
/*
** KeyCode type.
*/
typedef struct {
	char		*KeyName;
	USHORT		KeyCode;
	USHORT		Qualifier;
} KeyCode;

#define GREY	0x80

#define CopyKeyCodes(dest,src)	\
	(dest) = (src);

#define SetKeyCode(kc,key)		\
	(kc)->KeyCode = (key);

#define CmpKeyCodes(one,two)	\
	((one) - (two))

/*
** Key Code Data.
*/
LOCAL KeyCode KeyCodes[] = {
	{ "F1",		59,	NULL },
	{ "F2",		60,	NULL },
	{ "F3",		61,	NULL },
	{ "F4",		62,	NULL },
	{ "F5",		63,	NULL },
	{ "F6",		64,	NULL },
	{ "F7",		65,	NULL },
	{ "F8",		66,	NULL },
	{ "F9",		67,	NULL },
	{ "F10",	68,	NULL },
	{ "F11",	11 + GREY,	NULL },
	{ "F12",	12 + GREY,	NULL },

	{ "Up",		72 + GREY,	NULL },
	{ "Down",	80 + GREY,	NULL },
	{ "Left",	75 + GREY,	NULL },
	{ "Right",	77 + GREY,	NULL },

	{ "BkSp",	14,			NULL },
	{ "Esc",	 1,			NULL },
	{ "Tab",	15,			NULL },
	{ "Space",	57,			NULL },
	{ "Return",	28,			NULL },
	{ "Enter",	28 + GREY,	NULL },
	{ "Del", 	83 + GREY,	NULL },
	{ "Delete",	83 + GREY,	NULL },
	{ "Ins",	82 + GREY,	NULL },
	{ "Insert",	82 + GREY,	NULL },
	{ "Home",	71 + GREY,	NULL },
	{ "End",	79 + GREY,	NULL },
	{ "PgUp",	73 + GREY,	NULL },
	{ "PgDn",	81 + GREY,  NULL },

	{ "Grey-+",	78,			NULL },
	{ "Grey--", 74,			NULL },
	{ "Grey-*", 55,			NULL },
	{ "Grey-/", 53 + GREY,	NULL },

	{ "Keypad-+",	78,			NULL },
	{ "Keypad--",	74,			NULL },
	{ "Keypad-*",	55,			NULL },
	{ "Keypad-/",	53 + GREY,	NULL },

	{ "Keypad-0",	82,		NULL },
	{ "Keypad-1",	79,		NULL },
	{ "Keypad-2",	80,		NULL },
	{ "Keypad-3",	81,		NULL },
	{ "Keypad-4",	75,		NULL },
	{ "Keypad-5",	76,		NULL },
	{ "Keypad-6",	77,		NULL },
	{ "Keypad-7",	71,		NULL },
	{ "Keypad-8",	72,		NULL },
	{ "Keypad-9",	73,		NULL },
	{ "Keypad-.",	83,		NULL },

	{ "`",			41,		NULL },
	{ "1",			2,		NULL },
	{ "2",			3,		NULL },
	{ "3",			4,		NULL },
	{ "4",			5,		NULL },
	{ "5",			6,		NULL },
	{ "6",			7,		NULL },
	{ "7",			8,		NULL },
	{ "8",			9,		NULL },
	{ "9",			10,		NULL },
	{ "0",			11,		NULL },
	{ "-",			12,		NULL },
	{ "=",			13,		NULL },
	{ "\\",			43,		NULL },

	{ "q",			16,		NULL },
	{ "w",			17,		NULL },
	{ "e",			18,		NULL },
	{ "r",			19,		NULL },
	{ "t",			20,		NULL },
	{ "y",			21,		NULL },
	{ "u",			22,		NULL },
	{ "i",			23,		NULL },
	{ "o",			24,		NULL },
	{ "p",			25,		NULL },
	{ "[",			26,		NULL },
	{ "]",			27,		NULL },

	{ "a",			30,		NULL },
	{ "s",			31,		NULL },
	{ "d",			32,		NULL },
	{ "f",			33,		NULL },
	{ "g",			34,		NULL },
	{ "h",			35,		NULL },
	{ "j",			36,		NULL },
	{ "k",			37,		NULL },
	{ "l",			38,		NULL },
	{ ";",			39,		NULL },
	{ "'",			40,		NULL },

	{ "z",			44,		NULL },
	{ "x",			45,		NULL },
	{ "c",			46,		NULL },
	{ "v",			47,		NULL },
	{ "b",			48,		NULL },
	{ "n",			49,		NULL },
	{ "m",			50,		NULL },
	{ ",",			51,		NULL },
	{ ".",			52,		NULL },
	{ "/",			53,		NULL },

/************************************  ************************************/
	
	{ "~",			41,		QUAL_SHIFT },
	{ "!",			2,		QUAL_SHIFT },
	{ "@",			3,		QUAL_SHIFT },
	{ "#",			4,		QUAL_SHIFT },
	{ "$",			5,		QUAL_SHIFT },
	{ "%",			6,		QUAL_SHIFT },
	{ "^",			7,		QUAL_SHIFT },
	{ "&",			8,		QUAL_SHIFT },
	{ "*",			9,		QUAL_SHIFT },
	{ "(",			10,		QUAL_SHIFT },
	{ ")",			11,		QUAL_SHIFT },
	{ "_",			12,		QUAL_SHIFT },
	{ "+",			13,		QUAL_SHIFT },
	{ "|",			43,		QUAL_SHIFT },

	{ "Q",			16,		QUAL_SHIFT },
	{ "W",			17,		QUAL_SHIFT },
	{ "E",			18,		QUAL_SHIFT },
	{ "R",			19,		QUAL_SHIFT },
	{ "T",			20,		QUAL_SHIFT },
	{ "Y",			21,		QUAL_SHIFT },
	{ "U",			22,		QUAL_SHIFT },
	{ "I",			23,		QUAL_SHIFT },
	{ "O",			24,		QUAL_SHIFT },
	{ "P",			25,		QUAL_SHIFT },
	{ "{",			26,		QUAL_SHIFT },
	{ "}",			27,		QUAL_SHIFT },

	{ "A",			30,		QUAL_SHIFT },
	{ "S",			31,		QUAL_SHIFT },
	{ "D",			32,		QUAL_SHIFT },
	{ "F",			33,		QUAL_SHIFT },
	{ "G",			34,		QUAL_SHIFT },
	{ "H",			35,		QUAL_SHIFT },
	{ "J",			36,		QUAL_SHIFT },
	{ "K",			37,		QUAL_SHIFT },
	{ "L",			38,		QUAL_SHIFT },
	{ ":",			39,		QUAL_SHIFT },
	{ "\"",			40,		QUAL_SHIFT },

	{ "Z",			44,		QUAL_SHIFT },
	{ "X",			45,		QUAL_SHIFT },
	{ "C",			46,		QUAL_SHIFT },
	{ "V",			47,		QUAL_SHIFT },
	{ "B",			48,		QUAL_SHIFT },
	{ "N",			49,		QUAL_SHIFT },
	{ "M",			50,		QUAL_SHIFT },
	{ "<",			51,		QUAL_SHIFT },
	{ ">",			52,		QUAL_SHIFT },
	{ "?",			53,		QUAL_SHIFT },

	{ NULL, },
};
#else
#error Need Keyboard support for this platform
#endif

/*
** FindKeyCode *******************************************************************
*/
LOCAL KeyCode *FindKeyCode (
	char *key
) {
	register KeyCode *k = KeyCodes;

	while (k->KeyName) {
		if (!stricmp (k->KeyName, key)) {
			return k;
		}
		k++;
	}

	return NULL;
}

/*
** GetKeyCode ********************************************************************
*/
LOCAL KeyCode *GetKeyCode (
	char *key
) {
	KeyCode	*k;

	if ((k = FindKeyCode (key)) != NULL) {
		return k;
	}
#if __AMIGAOS__
	if (!strnicmp ("Shift-", key, 6)) {
		key += 6;
		if ((k = FindKeyCode (key))) {
			return k;
		}
	}

	if (strlen (key) != 1) {
		KeyError ("Bad Key Code\n");
		return NULL;
	}

	{
		static KeyCode JunkKC = { NULL, };

		SetKeyCode (&JunkKC,*key);
		JunkKC.Qualifier  = NULL;
		if (isalpha(*key) && isupper (*key)) {
			JunkKC.Qualifier = QUAL_SHIFT;
		}
		return &JunkKC;
	}
#else
	return NULL;
#endif
}

/**************************************************************************
 *
 * AddKey
 *
 * SYNOPSIS
 *   int AddKey (int action, void *data, char *key);
 *
 * FUNCTION
 *   Adds a specific key specified by 'key' string as in "SHIFT-A" or
 *   "CTRL-ALT-UP" or "KEYPAD-9" and associates that key with a
 *   particular command.  Valid commands are specifed in 'KEYCOMMANDS.h'.
 *   The most common command is KEY_EVENT which binds a key to an 'EVENT'
 *   (the same events that menus use.)
 *
 *   Note: key string will be modified.
 *
 * INPUTS
 *   action = ACTION Value as defined in 'KEYCOMMANDS.h'
 *   data   = pointer to data specific to type of command.  (if KEY_EVENT)
 *            then data should be a pointer to the 'EventState' for that
 *            event.
 *   key    = pointer to a NULL terminated string specifing the key.
 *
 * RESULTS
 *   TRUE if successful.
 *   FALSE if unsuccessful.
 *
*/
short AddKey (
	short	 action,
	void	*data,
	char	*key
) {
	register KeyType *k;
	register KeyCode *ks;
	register KeyType *kt;
	char		hold[50];

	strcpy (hold, key);

	if ((k = (KeyType *)CreateNode (sizeof (KeyType), key)) == NULL) {
		KeyError ("OOM: KeyType\n");
		return FALSE;
	}

	/* Check Quotes. */
	if ((*key == '\'') && (key[strlen (key) - 1] == '\'')) {
		/* Strip Quotes */
		key[strlen (key) - 1] = '\0';
		key++;
	}

	if ((*key == '<') && (key[strlen (key) - 1] == '>')) {
		key[strlen (key) - 1] = '\0';
		key++;
	}

	k->Command	= action;
	k->CommandData	= data;

	if (strlen (key) > 2) {
		k->Qualifier = GetQualifier (key);
	}

#if __AMIGAOS__
	if (k->Qualifier & QUAL_SHIFT) {
		char		work[50];

		strcpy (work, "Shift-");
		strcat (work, key);
		key = work;
	}
#endif

	if ((ks = GetKeyCode (key)) == NULL) {
		ParseError1 ("Bad Key Code\n", NULL);
		DeleteNode (k);
		return FALSE;
	}

#if __AMIGAOS__
	if (strlen(ks->KeyCode) > 9) {
		KeyError ("Key Code too long");
		DeleteNode (k);
		return FALSE;
	}
#endif

	CopyKeyCodes (k->Key, ks->KeyCode);
	k->Qualifier |= ks->Qualifier;
	k->Label      = ks->KeyName;

	kt = (KeyType *)KeyList->Head;
	while (!IsEOList(kt)) {
		if ((!CmpKeyCodes (kt->Key, k->Key)) &&
		    (kt->Qualifier == k->Qualifier)) {
			ParseError1 ("Duplicate key '%s'\n", hold);
			DeleteNode (k);
			return FALSE;
		}
		kt = (KeyType *)Next(kt);
	}

	AddTail (KeyList, k);

#if DEBUGGING2
				{
					char *kc;
					int   cnt;

					printf ("Keyspec - '%s'\n", hold);
					PrintQual (k->Qualifier);
					kc = k->Key;
					while (*kc) {
						printf ("0x%02x, ", *kc++);
					}
					printf ("\n");
			}
#endif

	return TRUE;
}

/**************************************************************************
 *
 * AddKeyEvent
 *
 * SYNOPSIS
 *   int AddKeyEvent (char *key, char *event)
 *
 * FUNCTION
 *   Binds 'key' to 'event'.  Key is specified in the same way as AddKey().
 *   'event' is the ASCII name of the event.
 *
 * INPUTS
 *   key   = pointer to NULL terminated string of ASCII description of key
 *   event = pointer to NULL terminated string of ASCII name of event.
 *
 * RESULTS
 *   TRUE  = success
 *   FALSE = failure
*/
short AddKeyEvent (
	char	*key,
	char	*event
) {
	EventState	*es;

	if ((es = FindEvent (event)) == NULL) {
		ParseError1 ("Nonexistent event '%s'\n", event);
		return FALSE;
	}
	return AddKey (KEY_EVENT, es, key);
}

#if 0
static UBYTE NumLockTable[] = {
	0x08	{	0x37,	EASCII_HOME,	},	//	KEYPAD HOME
	0x09	{	0x38,	EASCII_UP,		},	//	KEYPAD UP	/* 48 */
	0x0A	{	0x39,	EASCII_PGUP,	},	//	KEYPAD PGUP
	0x4A	{	0x2D,	0x2D,			},	//	KEYPAD GREY -
{	0x34,	EASCII_LEFT,	},	//	KEYPAD LEFT
{	0x35,	EASCII_CENTER,	},	//	KEYPAD CENTER
{	0x36,	EASCII_RIGHT,	},	//	KEYPAD RIGHT
{	0x2B,	0x2B,			},	//	KEYPAD GREY +
{	0x31,	EASCII_END,		},	//	KEYPAD END
{	0x32,	EASCII_DOWN,	},	//	KEYPAD DOWN	/* 50 */
{	0x33,	EASCII_PGDN,	},	//	KEYPAD PGDN
{	0x30,	EASCII_INSERT,	},	//	KEYPAD INS
{	0x2E,	EASCII_DELETE,	},	//	KEYPAD DEL
}
#endif

short DoKeyCommand (
	USHORT	  keyvalue
) {
	register KeyType		*kt;
	KeyValue				 key;
	USHORT					 qual;

	key      = keyvalue & 0xFF;
	qual     = (keyvalue >> 8) & 0x0F;
	if (qual & 0x02) {
		qual = (qual & ~0x02) | 0x01;
	}

	kt = (KeyType *)KeyList->Head;

	while (!IsEOList(kt)) {
		if (kt->Qualifier == qual && !CmpKeyCodes (kt->Key, key)) {
			switch (kt->Command) {
			case KEY_EVENT:
				TriggerEvent (kt->CommandData, ETYPE_KEY, NULL);
				break;
			default:
				puts ("Nonexistent command\n");
				break;
			}
			break;
		}
		kt = (KeyType *)Next(kt);
	}

	if (IsEOList(kt)) {
		return FALSE;
	}

	return TRUE;
}

#if __AMIGAOS__
/**************************************************************************
 *
 * RawKey
 *
 * SYNOPSIS
 *    int RawKey (struct IntuiMessage *msg);
 *
 * FUNCTION
 *    Excepts Intuition messages of Class RAWKEY and issues and command
 *    bound to that key.  NOTE: RawKey will ReplyMsg() the message passed
 *    to it.  DO NOT pass it a message that you have already replyed to.
 *
 * INPUTS
 *    msg = pointer to an IntuiMessage structure as returned from
 *          GetMsg ();
 *
 * RESULTS
 *    TRUE  = continue
 *    FALSE = program should cleanup and exit.
 *
*/
int RawKey (msg)
	struct	IntuiMessage 		*msg;
{
	UBYTE				KeyBuffer[100];
	LONG				Keys;
	USHORT				qual;
	int				status = TRUE;

	if (!(msg->Code & IECODE_UP_PREFIX)) {
		qual = msg->Qualifier;
		Keys = DeadKeyConvert (msg,
				KeyBuffer,
				sizeof (KeyBuffer),
				NULL);
	}
	ReplyMsg (msg);
	if (Keys > 0 && !(msg->Code & IECODE_UP_PREFIX)) {
#if DEBUGGING2
				{
					char *k;
					int   cnt;
					
					PrintQual (qual);
					k = KeyBuffer;
					for (cnt = 0; cnt <Keys; cnt++) {
						printf ("0x%02x, ", *k++);
					}
					printf ("\n");
				}
#endif
		KeyBuffer[Keys] = '\0';
		status = DoKeyCommand (KeyBuffer, QUAL(qual));
	}

	return status;
}
#endif


/*
 * ==================================================================== *
 *    Debugging Routines.
 * ==================================================================== *
 */

#if TEST
PrintKeys ()
{
	register struct	IntuiMessage 	*msg;
	UBYTE				KeyBuffer[100];
	LONG				Keys;
	int				quit = 0;
	USHORT				qual;
	register int			cnt;
	register UBYTE			*k;

	while (!quit) {
		WaitPort (MainWindow->UserPort);
		while (!quit && (msg = GetMsg (MainWindow->UserPort))) {
			if (msg->Class == RAWKEY) {
				qual = msg->Qualifier;
				Keys = DeadKeyConvert (msg,
							KeyBuffer,
							sizeof (KeyBuffer),
							NULL);
				ReplyMsg (msg);
				if (Keys > 0) {
					if (KeyBuffer[0] == 'q') {
						quit = TRUE;
					}
					PrintQual (qual);
					k = KeyBuffer;
					for (cnt = 0; cnt <Keys; cnt++) {
						printf ("0x%02x, ", *k++);
					}
					printf ("\n");
				}
			} else {
				printf ("Bad IDCMP class 0x%08x\n", msg->Class);
				ReplyMsg (msg);
			}
		}
	}
}
#endif

#if DEBUGGING2
void PrintQual (qual)
	USHORT	qual;
{
	USHORT	q;

	q = QUAL (qual);
	printf ("%s\t%s\t%s\t%s\t:",
		(q & QUAL_SHIFT)  ? "Shift"  : "",
		(q & QUAL_CTRL)   ? "Ctrl"   : "",
		(q & QUAL_ALT)    ? "Alt"    : "",
		(q & QUAL_KEYPAD) ? "KeyPad" : "");
}
#endif

/*
** ParseKeyEvent ***********************************************************
**
** KeyEvent EV KEY
**    EV = Event Label
**   KEY = Key to Set Special Frame.
*/

#define KE_EVENT	1
#define KE_KEY		2

LOCAL short ParseKeyEvent (argc, argv)
	short	 argc;
	char	**argv;
{
	argc = argc;
	return AddKeyEvent (argv[KE_EVENT], argv[KE_KEY]);
}

/* -----------------------------------------------------------------------*/

typedef struct {
	char	*name;
	short	(*func)(short argc, char **argv);
} Command;

LOCAL Command Commands[] = {
	{ "keyevent",	ParseKeyEvent, },
	{ NULL, NULL, },
};


#define MAX_ARGS	24

LOCAL	short	 p_argc;
LOCAL	char	*p_argv[MAX_ARGS];  /* 4 more than template? */

#define LINESIZE	256
LOCAL	char	line[LINESIZE];

short ParseKeys (
	IniList	*inilist
) {
	short			 goodscan = TRUE;
	Command			*cmd;
	SectionTracker	 secx;
	SectionTracker	*sec = &secx;
	ConfigLine		*cl;

	sec = FindSection (sec, inilist, "[KEYS]");
	if (!sec) {
		SetGlobalErr (ERR_GENERIC);
		GEcatf ("\nNo [KEYS] section found in .ini file");
		goodscan = FALSE;
	} else {
		while ((cl = GetNextINILine (sec)) != NULL) {
			LineNo = GetConfigLineNo (cl);
			strcpy (line, NodeName (cl));
 			StripComment (line);
 			p_argc = argify (line, MAX_ARGS, p_argv);
 			cmd = Commands;
 			if (p_argc > 0) {
 				while (cmd->name && p_argc > 0) {
 					if (!stricmp (cmd->name, p_argv[0])) {
 						if (!(*cmd->func)(p_argc, p_argv)) {
 							goodscan = FALSE;
 						}
 						p_argc = -1;
 					} else {
 						cmd++;
 					}
 				}
 				if (!cmd->name) {
 					ParseError1 ("ERROR: Unknown command '%s'\n",
 						p_argv[0]);
 					goodscan = FALSE;
 				}
 			}
		}
	}

	return goodscan;
}


void UninitKeys (void)
{
	if (KeyList)		EmptyList (KeyList);
#if __AMIGAOS__
	if (ConsoleDevice)	CloseDevice (&ConsoleIoReq);
#endif
}

short InitKeys (IniList	*inilist)
{
#if __AMIGAOS__
	/* Set up DeadKeyConvert */
	if (OpenDevice ("console.device", -1L, &ConsoleIoReq, 0L))
		return FALSE;

	ConsoleDevice = (struct ConsoleDevice *)ConsoleIoReq.io_Device;
#endif

	InitList (KeyList);
	
	if (!ParseKeys (inilist)) {
		UninitKeys ();
		return FALSE;
	}

	return TRUE;
}

/*********************************************************************
 *
 * HandleKeys
 *
 * SYNOPSIS
 *		short HandleKeys (Window *win)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short HandleKeys (Window *win)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "HandleKeys";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	win = win;

	{
		UWORD		 key;

		do {
			key = GetKeyNoWait ();
			if (key) {
				if (!DoKeyCommand (key)) {
					CheckMenuKeys (key);
				}
			}
 		} while (key);
	}
	return TRUE;

} /* HandleKeys */

/*********************************************************************
 *
 * FindFirstEventKey
 *
 * SYNOPSIS
 *		char *FindFirstEventKey (
 *					char	*eventname,
 *					USHORT	*qualifiers);
 *
 * PURPOSE
 *		Given an Event Name, returns the first key assigned to the specified
 *		event along with its qualifiers.
 *
 * INPUT
 *
 *
 * EFFECTS
 *		return a pointer to the first keylabel for this event.
 *		NOTE: do NOT modify the string.  (make your own copy if you
 *		need to modify it.)  USHORT pointed to by qualifiers
 *		will have the key's qualifier bits.
 *
 * RETURN VALUE
 *		Returns NULL if no such event or no keys assigned to 
 *		specified event.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
char *FindFirstEventKey (
			char	*eventname,
			USHORT	*qualifier)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FindFirstEventKey";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		KeyType		*kt;
		EventState	*es;

		kt = Head (KeyList);
		while (!EndOfList (kt)) {
			if (kt->Command == KEY_EVENT) {
				es = kt->CommandData;
				if (!stricmp (NodeName (es), eventname)) {
					*qualifier = kt->Qualifier;
					return kt->Label;
				}
			}
			kt = Next (kt);
		}

		return NULL;
	}

} /* FindFirstEventKey */

