/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * APALETTE.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 11/08/89
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oO///_
 *	-----w/-w-----
 *	 E C H i D N A
 *	--------------
 *
 * DESCRIPTION
 *		Routines for setting the Amiga/CDTV palette.
 *
 * HISTORY
 *	06/19/91 Wednesday (RGM) - Taken from MPalette.C.
 *	06/27/91 Thursday (RGM) - Support for program syncronized palette
 *		loading (eg for use in JMA's anti-alias scrolling).
 *
*/
#include <echidna/platform.h>
#include "switches.h"

#include <string.h>
#include <echidna/etypes.h>
#include <echidna/eerrors.h>
#include <echidna/memory.h>
#include <echidna/graphics.h>
#include <echidna/timer.h>

#if __AMIGAOS__
#include <intuition/screens.h>
#if LATTICE
#include <proto/graphics.h>
#include <stdio.h>
#endif
#endif


/**************************** C O N S T A N T S ***************************/

/* number of different components in color model (in this case R,G,B = 3) */
#define	COLOR_COMPONENTS		(3)

#if __MSDOS__
#define HI_RES_SHIFT			(8)
#elif __AMIGAOS__
#define HI_RES_SHIFT			(10)
#elif
#error You need to define a shift value here.
#endif

#define DONT_FAKE_IT			(1)
#define GFX_COLORWRITE_VBLANK	(0)

#define MAX_COPY_RANGES					32
#define MAX_COLORS_COPIED_PER_FRAME		32

#define MAX_ALLOWED_AMIGA_COLORS (32)

/******************************** T Y P E S *******************************/

typedef struct {
    UWORD	 NumColors; 
    void	*ColorPtr;  
    UWORD	 ColorReg;  
} CopyRangeType;

static struct Cyc {
	short 		 MaxColors;			/* Max cycle colors routines handle */
	short		 FadeFrames;		/* Fade duration count-down timer.
							 		 * Non-zero if fade in progress */
	ColorCycleSetType	*UserSet;	/* User Color Cycling Set */
	ColorType			*RGB;		/* Private copy of cycle color data */
	short		 ToBlack;			/* Flag: True if fading to black */
	WORD		*HiResColors;	/* Used for smooth color fades */
	WORD		*HiResDeltas;	/* Used for smooth color fades */
	size_t		 CBytes;		/* Number of bytes in color data */
	size_t		 FadeBufferSize;/* Size of buffer allocated for fading */			
	short		 HasTopBuffer;	/* True if has top temp buffer; Only valid
								 *  if HiResColors is NON-NULL */
	struct {
		BYTE Change;
		BYTE Fade;
		BYTE Activate;
		short Hz60FadeFrames;		/* 60 Hz count-down timer.*/
		ColorCycleSetType *CycleSet;
	} Request;
};

static struct Pal {
	short		 MaxColors;		/* Max palette colors routines handle */
	short		 FadeFrames;/* Fade duration count-down timer.
								 * Non-zero if fade in progress */
	short		 SrcOffset;		/* Offset in src pal of start of fade range*/
	short		 DstOffset;		/* Offset in Dst pal of start of fade range*/
	short		 NumColors;		/* NumColors to fade */
	PaletteType	*UserPalette;	/* User palette structure */
	ColorType	*RGB;			/* Private copy of palette color data */
	WORD		*HiResColors;	/* Used for smooth color fades */
	WORD		*HiResDeltas;	/* Used for smooth color fades */
	size_t		 CBytes;		/* Number of bytes in color data */
	size_t		 FadeBufferSize;/* Size of buffer allocated for fading */			
	short		 HasTopBuffer;	/* True if has top temp buffer; Only valid
								 *  if HiResColors is NON-NULL */
	struct {
		BYTE Change;
		BYTE Fade;
		short SrcOffset;
		short DstOffset;
		short NumColors;
		short Hz60FadeFrames;	/* 60 Hz count-down timer. */
		PaletteType *Palette;
	} Request;
};

static struct PaintCtrlStruct {
	int				 InterruptHandle;
	int				 VBlankInterruptHandle;
	short			 PHertz;			/* Highest palette change rate */
	short			 Hz60ToPHertz;	
	short			 PrevNotDone;	/* True if not done from prev change */
	UBYTE			*RegsChanged;	/* Flags what palette registers need updating */
	struct Cyc		 Cyc;			/* Info for color cycling */
	struct Pal		 Pal;			/* Info for palette changes */
};

/****************************** E X T E R N S *****************************/

extern struct Screen	*AmigaScreen;
extern void far			MCGA_ASM_WriteColors (void);
/* extern short			ActivationFlag; */
extern CopyRangeType	*CopyRanges;

extern void __regargs AmigaPaletteToHardware (
	UWORD	*colors,
	UWORD	count
);

#if 0
extern VOID LoadRGBBoth (
	UWORD	start_reg,
	UWORD	*colors,
	UWORD	count
);
extern VOID LoadRGB (
	UWORD	start_reg,
	UWORD	*colors,
	UWORD	count
);
#endif

/****************************** G L O B A L S *****************************/

static struct PaintCtrlStruct	D;
short							ActivationFlag	= FALSE;
CopyRangeType					*CopyRanges		= NULL;

/******************************* L O C A L S ******************************/


/******************************* M A C R O S ******************************/

#define ColorsDifferent(c1,c2)			\
		((c1)->Red   != (c2)->Red ||	\
		 (c1)->Green != (c2)->Green ||	\
		 (c1)->Blue  != (c2)->Blue)

#if __AMIGAOS__

#define	X_RED(w)	((BYTE) (((w) & 0x0F00) >> 8))
#define	X_GREEN(w)	((BYTE) (((w) & 0x00F0) >> 4))
#define	X_BLUE(w)	((BYTE) ((w) & 0x000F))

#define I_RED(c,b)		((c) |= (((UWORD) (b)) << 8))
#define I_GREEN(c,b)	((c) |= (((UWORD) (b)) << 4))
#define I_BLUE(c,b)		((c) |= ((UWORD) (b)))

#define I_RGB(c,r,g,b) \
	((c) = ((((UWORD) (r)) << 8) | (((UWORD) (g)) << 4) | ((UWORD) (b))))

#endif

/***************************** R O U T I N E S ****************************/

/*********************************************************************
 *
 * CreateCopyRanges
 *
 * SYNOPSIS
 *		static void CreateCopyRanges (void)
 *
 * PURPOSE
 *		To set up the CopyRanges array according to the flags in 
 *		D.RegsChanged, so that only the registers that are changed
 *		will get copied to the hardware palette.
 *	
 *		IMPORTANT: Only call if at least one register has changed.
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void CreateCopyRanges (void)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CreateCopyRanges";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT

#if 1
	/* we want to stuff the colors into the Amiga palette */

	if (AmigaScreen) {
		short	max_color;

		max_color	= min (D.Pal.MaxColors, MAX_ALLOWED_AMIGA_COLORS);

		/* LoadRGB4 (&AmigaScreen->ViewPort, (void *) D.Pal.RGB, max_color); */
		/* LoadRGBBoth (0, (void *) D.Pal.RGB, max_color); */
		/* LoadRGB (0, (void *) D.Pal.RGB, max_color); */

		AmigaPaletteToHardware ((UWORD *) D.Pal.RGB, (UWORD) max_color);
	}

	D.PrevNotDone	= FALSE;
	ActivationFlag	= FALSE;

#else
	int register i;
	short totalcolors = 1;
	short numranges = 0;
	UBYTE *regschanged = D.RegsChanged;
	CopyRangeType *copyrange = CopyRanges;

	D.PrevNotDone = FALSE;
	D.RegsChanged[D.Pal.MaxColors] = FALSE;	/* elephant at end */

	i = 0;
	while (i <= D.Pal.MaxColors) {
		/* Find start of changed range */
		if (*regschanged) {
			short start = i++;
			*regschanged++ = FALSE;
			while (i <= D.Pal.MaxColors) {
				/* Find end of changed range */
				if (!(*regschanged) || 
					totalcolors == MAX_COLORS_COPIED_PER_FRAME) 
				{
					/* Set up copy range */
					copyrange->NumColors = i - start;
 					copyrange->ColorPtr = &D.Pal.RGB[start];
					copyrange->ColorReg = start;
					++copyrange;
					++numranges;
					if (numranges == MAX_COPY_RANGES ||
						totalcolors >= MAX_COLORS_COPIED_PER_FRAME) 
					{
						/* Can't stuff anymore colors this frame; quit */
						/* Check if more left to do still */
						while (i < D.Pal.MaxColors) {
							if (*regschanged++) {
								D.PrevNotDone = TRUE;
								break;
							}
							++i;
						}
						/* remaining colors will get stuffed next frame */
						/* break out of both loops */
						i = D.Pal.MaxColors + 1;
					}
					break;
				} else {
					*regschanged = FALSE;
					++totalcolors;
				}
				++regschanged;
				++i;
			}
		}
		++regschanged;
		++i;
	}
	/* Mark end of copy ranges list */
	copyrange->NumColors = 0;

	/* Activate interrupt driven color stuffer */
	ActivationFlag = TRUE;
#endif

#endif
}
} /* CreateCopyRanges */

/*********************************************************************
 *
 * Paint
 *
 * SYNOPSIS
 *		InterruptRoutineType Paint (void)
 *
 * PURPOSE
 *		To Make changes requested by user.  This routine is called by
 *		an interrupt that is installed by OpenPainter().
 *
 * INPUT
 *		
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
InterruptRoutineType Paint (void) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "Paint";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT

	short changed = FALSE;

	/* If didn't finish work last time, then try to finish now */
	if (D.PrevNotDone) {
		CreateCopyRanges ();
		return;
	}

	/********** P R O C E S S   F A D I N G   P A L E T T E  **********/
	if (D.Pal.FadeFrames) {
		register int i;

#if __MSDOS__
		WORD *delta		= D.Pal.HiResDeltas; 
		WORD *tmp		= D.Pal.HiResColors;
		UBYTE *color		= (UBYTE *)(D.Pal.RGB + D.Pal.SrcOffset);
		UBYTE *regschanged	= D.RegsChanged + D.Pal.SrcOffset;

		/* Fade HiResColors */
		for (i = D.Pal.CBytes; i; i--) {
			*tmp++ += *delta++;
		}
		/* Copy HiResColors to Private Palette */
		tmp = D.Pal.HiResColors;
		for (i = D.Pal.CBytes; i; i--) {
			*color++ = (UBYTE)(*tmp++ >> 8);
		}
#else
		WORD	*delta	= (WORD *) D.Pal.HiResDeltas; 
		WORD	*tmp	= (WORD *) D.Pal.HiResColors;
		UWORD	*color	= (UWORD *)(D.Pal.RGB + D.Pal.SrcOffset);
		UBYTE	*regschanged	= D.RegsChanged + D.Pal.SrcOffset;

		/* Fade HiResColors & Copy HiResColors to Private Palette */

		for (i = D.Pal.NumColors; i; i--) {
			UWORD	color_val;

			color_val	= 0;

			I_RED(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));
			I_GREEN(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));
			I_BLUE(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));

			*color++ = color_val;
		}
#endif

		/* Set Flags in RegsChanged Array */
		for (i = D.Pal.NumColors; i; i--) *regschanged++ = TRUE;

		/* Is Fade done? */
		if (!(--D.Pal.FadeFrames)) {
			/* Try to release fade buffers */
			if (D.Pal.HasTopBuffer) {
				DestroyTempBuffer (D.Pal.HiResColors, D.Pal.FadeBufferSize);
				D.Pal.HiResColors = NULL;
				D.Cyc.HasTopBuffer = TRUE;
				/* Free Cycle fade buffers if not used */ 
				if (D.Cyc.HiResColors && !D.Cyc.FadeFrames) {
					DestroyTempBuffer (D.Cyc.HiResColors, 
						D.Cyc.FadeBufferSize);
					D.Cyc.HiResColors = NULL;
				}
			}
		}
		changed = TRUE;
	}
	
	/********** P R O C E S S   F A D I N G   C Y C L E  **********/
	if (D.Cyc.FadeFrames) {
#if __MSDOS__
		register int i;
		WORD *delta	= D.Cyc.HiResDeltas; 
		WORD *tmp		= D.Cyc.HiResColors;
		UBYTE *color	= (UBYTE *)D.Cyc.RGB;

		/* Fade HiResColors */
		for (i = D.Cyc.CBytes; i; i--) {
			*tmp++ += *delta++;
		}

		/* Copy HiResColors to Private Palette */
		tmp = D.Cyc.HiResColors;
		color = (UBYTE *)D.Cyc.RGB;
		for (i = D.Cyc.CBytes; i; i--) {
			*color++ = (UBYTE)(*tmp++ >> 8);
		}

#else
		register int	i;
		WORD			*delta	= D.Cyc.HiResDeltas; 
		WORD			*tmp	= D.Cyc.HiResColors;
		UWORD			*color	= (UWORD *) D.Cyc.RGB;

		/* Fade HiResColors & Copy HiResColors to Private Palette */

		{
			UWORD	color_count;

			color_count = D.Cyc.CBytes / COLOR_COMPONENTS;

			color = (UWORD *) D.Cyc.RGB;

			for (i = color_count; i; i--) {
				UWORD	color_val;

				color_val	= 0;

				I_RED(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));
				I_GREEN(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));
				I_BLUE(color_val,((*tmp++ += *delta++) >> HI_RES_SHIFT));

				*color++ = color_val;
			}
		}
#endif

		/* Is Fade done? */
		if (!(--D.Cyc.FadeFrames)) {
			/* Try to release fade buffers */
			if (D.Cyc.HasTopBuffer) {
				DestroyTempBuffer (D.Cyc.HiResColors, D.Cyc.FadeBufferSize);
				D.Cyc.HiResColors = NULL;
				D.Pal.HasTopBuffer = TRUE;
				/* Free Palette fade buffers if not used */ 
				if (D.Pal.HiResColors && !D.Pal.FadeFrames) {
					DestroyTempBuffer (D.Pal.HiResColors, 
						D.Pal.FadeBufferSize);
					D.Pal.HiResColors = NULL;
				}
			}
			/* If fade was to black then install NULL cycle set */
			if (D.Cyc.ToBlack) {
				D.Cyc.UserSet = NULL;
			}
		}
		changed = TRUE;
	} 

	/**********  P R O C E S S   C O L O R   C Y C L I N G  *********/
	if (ColorCyclingActive && D.Cyc.UserSet && D.Cyc.UserSet->Active) {
		ColorCycleType	*cycle;
		ColorType		*rgbptr;
		short i;

		cycle = D.Cyc.UserSet->Cycles;
		rgbptr = D.Cyc.RGB;
		for (i = 0; i < D.Cyc.UserSet->NumCycles; i++) {
			UBYTE *reg;
			WORD icolor;
			ColorType *newcolor;
			short i;
			WORD hertz = cycle->Hertz;

			if (cycle->Active) {

				if (hertz < 0) {			/* NEW for anti-alias scroll */
					cycle->Timer	= 0;
					cycle->Active	= FALSE;
				}

				/* Update Cycle Timer etc. if it has a speed */
				if (hertz) {
					/* Reset Timer and Cycle Colors if necessary */
					if (cycle->Timer <= 0) {
						cycle->Timer = D.PHertz / hertz;

						/* Cycle Colors */
						if (cycle->Direction) {
							if (++cycle->CurrentColor >= cycle->NumColors) {
								cycle->CurrentColor = 0; 
							}
						} else {
							if (--cycle->CurrentColor < 0) {
								cycle->CurrentColor = cycle->NumColors - 1;
							}
						}
					} 
					--cycle->Timer;
				}

				/* Copy Cycle Colors into Shadow Palette if they Differ */
				reg = cycle->Registers;
				icolor = cycle->CurrentColor;
				newcolor = &rgbptr[icolor];
				for (i = cycle->NumRegisters; i; i--) {
					ColorType *crntcolor = &D.Pal.RGB[*reg];
#if __AMIGAOS__
					if ( ((*((UWORD *)crntcolor)) & 0x0FFF)
										!= ((*((UWORD *)newcolor)) & 0x0FFF) ) {
#else
					if (ColorsDifferent (crntcolor, newcolor)) {
#endif
						*crntcolor = *newcolor;
						D.RegsChanged[*reg] = changed = TRUE;
					}
					++newcolor; ++icolor;
					if (icolor >= cycle->NumColors) {
						icolor = 0;
						newcolor = rgbptr;
					}
					++reg;
				}

			}
			rgbptr += cycle->NumColors;
			++cycle;
		}
	}

	/****** P R O C E S S   P A L E T T E   R E Q U E S T S ********/
	if (D.Pal.Request.Change) {
		/* Is palette currently fading? */
		if (D.Pal.FadeFrames) { /* Terminate the fade */
			D.Pal.FadeFrames = 0;
			/* Free fade buffer if possible */
			if (D.Pal.HasTopBuffer) {
				DestroyTempBuffer (D.Pal.HiResColors, D.Pal.FadeBufferSize);
				D.Pal.HiResColors = NULL;
				D.Cyc.HasTopBuffer = TRUE;
			}
		}
		/* Initiate palette change */
		{
			register int 		 i; 
			ColorType	*src_color;
			ColorType	*dst_color;
			UBYTE		*regschanged;
			UBYTE		toblack;
			static ColorType black_color = {0};

			toblack = (D.Pal.UserPalette = D.Pal.Request.Palette) == NULL;
			if (!toblack) {
				src_color = D.Pal.UserPalette->Data.RGB	+ 
					D.Pal.Request.SrcOffset;
			}
			dst_color = D.Pal.RGB					+ D.Pal.Request.DstOffset;
			regschanged = D.RegsChanged				+ D.Pal.Request.DstOffset;

			for (i = D.Pal.Request.NumColors; i; i--) {
				*dst_color++ = (toblack) ? black_color : *src_color++;	
				*regschanged++ = TRUE;
			}
		}
		/* Turn off change request */
		D.Pal.Request.Change = FALSE;
		changed = TRUE;

	} else if (D.Pal.Request.Fade) { /****** PALETTE FADE REQUEST ****/
		short ok_to_start;

		/* Is palette currently fading? */
		if (D.Pal.FadeFrames) {	/* Terminate old fade */
			D.Pal.FadeFrames = 0;
		}

		/** Make sure I have a fade buffer **/

		ok_to_start = TRUE;	/* Assume I do until proven otherwise */
		/* Do I already have a buffer? */
		if (D.Pal.HiResColors) { 
			/* Is it too small for the current request? */
			if (D.Pal.Request.NumColors > D.Pal.NumColors) {/* Too small */
				/*Can I free it to get bigger one? */
				if (D.Pal.HasTopBuffer) { /* Free it */
					DestroyTempBuffer (D.Pal.HiResColors, 
						D.Pal.FadeBufferSize);
					D.Pal.HiResColors = NULL;
					D.Cyc.HasTopBuffer = TRUE;
				} else { /* buffer too small and can't free it for new one */
					ok_to_start = FALSE;				
				}
			}
		} /* else ok to allocate new buffer */

		if (ok_to_start) {
			size_t hi_res_pal_size;
			UBYTE toblack;

			/* Turn off request for fade */
			D.Pal.Request.Fade = FALSE;

			/* Set up new palette fade */
			toblack = (D.Pal.UserPalette = D.Pal.Request.Palette) == NULL;
			D.Pal.SrcOffset		= D.Pal.Request.SrcOffset;
			D.Pal.DstOffset		= D.Pal.Request.DstOffset;
			D.Pal.NumColors		= D.Pal.Request.NumColors;
			D.Pal.FadeFrames	= (D.Pal.Request.Hz60FadeFrames + 
				D.Hz60ToPHertz - 1) / D.Hz60ToPHertz;

			if (!D.Pal.HiResColors) {
				/* Allocate temp buffer big enough for 2 "hi-res" palettes
			 	 * ("hi-res" meaning a WORD per color component instead of
			 	 * the normal UBYTE per color component)
			 	 */
				hi_res_pal_size =
					D.Pal.NumColors * COLOR_COMPONENTS * sizeof (WORD);

				D.Pal.FadeBufferSize = 2 * hi_res_pal_size;
				if (D.Pal.HiResColors = 
					CreateTempBuffer (D.Pal.FadeBufferSize))
				{
					/* Put delta values in same buffer after HiResColors part */
					D.Pal.HiResDeltas = (WORD *)(((BYTE *)D.Pal.HiResColors)+ 
						hi_res_pal_size);
				}

				/* If I succeeded in allocating a temp buffer, then
				 * I will now have the top most buffer on the temp buffer
				 * stack and the cycle color structure will not.
				 * So set up the flags to reflect the current state.
			 	 */
				D.Cyc.HasTopBuffer = !(D.Pal.HasTopBuffer = 
					(D.Pal.HiResColors != NULL));
			}

			/* I now either have a buffer or ran out of memory */
			if (D.Pal.HiResColors) {
#if __MSDOS__
				register int i;
				WORD *delta, *tmp;
				BYTE *src_color, *save_sc, *dst_color;
				WORD  diff;

				/* Compute increment */
				D.Pal.CBytes = D.Pal.NumColors * COLOR_COMPONENTS;
				delta = D.Pal.HiResDeltas;
				if (!toblack) {
					dst_color = (BYTE *)(D.Pal.UserPalette->Data.RGB + 
						D.Pal.Request.DstOffset);
				}
				src_color = save_sc = (BYTE *)(D.Pal.RGB + 
					D.Pal.Request.SrcOffset);
 				
				for (i = D.Pal.CBytes; i; i--) {
					diff = ((toblack) ? 0 : *dst_color++) - 
						*src_color++;
					*delta = (diff << 8) / D.Pal.FadeFrames;
					if (*delta > 0) {
						++(*delta);
					}
					++delta;
				}
			
				/* Make temporary palette */
				tmp = D.Pal.HiResColors;
				src_color = save_sc;
				for (i = D.Pal.CBytes; i; i--) {
					*tmp++ = ((WORD)*src_color++) << 8;
				}
#else
				register int	i;
				WORD			*tmp;
				WORD			*delta;
				UWORD			*src_color, *save_sc, *dst_color; /* xRGB */
				WORD			diff;
				UWORD			source_color;
				UWORD			dest_color;

				/* Compute increment & Make temporary palette */

				D.Pal.CBytes	= D.Pal.NumColors * COLOR_COMPONENTS;
				delta			= D.Pal.HiResDeltas;

				src_color	= save_sc =
					(UWORD *)(D.Pal.RGB + D.Pal.Request.SrcOffset);
				if (! toblack) {
					dst_color	=
						(UWORD *)
						(D.Pal.UserPalette->Data.RGB+D.Pal.Request.DstOffset);
				}

				tmp			= (WORD *) D.Pal.HiResColors;

				for (i = D.Pal.NumColors ; i; i--) {
					WORD	temp_color;

					source_color	= *src_color++;
					dest_color		= (toblack) ? 0 : *dst_color++;


					/*** do Red ***/

					diff	= X_RED(dest_color) - (temp_color = X_RED(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Pal.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;


					/*** do Green ***/

					diff	= X_GREEN(dest_color) - (temp_color = X_GREEN(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Pal.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;


					/*** do Blue ***/

					diff	= X_BLUE(dest_color) - (temp_color = X_BLUE(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Pal.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;
				}

#endif
			} else { /* Out of memory for fade buffer */
				/* Substitute Change-Palette-Request for failed Fade-Request */
				D.Pal.Request.Change = TRUE;
				if (D.Cyc.Request.Fade) {
					D.Cyc.Request.Fade = !(D.Cyc.Request.Change = TRUE);
				}
			}
		}
	}

	/****** P R O C E S S   C Y C L E   R E Q U E S T S ********/
	/* Process Cycle Requests (change fade) */
	if (D.Cyc.Request.Change) {
		/* Is cycle currently fading? */
		if (D.Cyc.FadeFrames) { /* Terminate the fade */
			D.Cyc.FadeFrames = 0;
			/* Free fade buffer if possible */
			if (D.Cyc.HasTopBuffer) {
				DestroyTempBuffer (D.Cyc.HiResColors, D.Cyc.FadeBufferSize);
				D.Cyc.HiResColors = NULL;
				D.Pal.HasTopBuffer = TRUE;
			}

		}
		/* Install new user cycle set */
		if (D.Cyc.UserSet = D.Cyc.Request.CycleSet) {
			/* Make private copy of RGB data */
			memcpy (D.Cyc.RGB, D.Cyc.UserSet->Colors, 
				D.Cyc.UserSet->TotalColors * sizeof (ColorType));
		} else { /* Set colors to black */
			memset (D.Cyc.RGB, 0, D.Cyc.MaxColors * sizeof (ColorType));
		}

		if (D.Cyc.Request.Activate) ColorCyclingActive = TRUE;

		/* Turn off change request */
		D.Cyc.Request.Change = FALSE;
		
	} else if (D.Cyc.Request.Fade) { /***** CYCLE FADE REQUEST *****/
		short ok_to_start;
		short numcolors;

		numcolors = (D.Cyc.Request.CycleSet) ?
			D.Cyc.Request.CycleSet->TotalColors :
			D.Cyc.UserSet->TotalColors;

		/* Is cycle currently fading? */
		if (D.Cyc.FadeFrames) { /* Terminate old fade */
			D.Cyc.FadeFrames = 0;
		}

		/** Make sure I have a fade buffer **/

		ok_to_start = TRUE;	/* Assume I do until proven otherwise */
		/* Do I already have a buffer? */
		if (D.Cyc.HiResColors) { 
			/* Is it too small for the current request? */
			if (numcolors > D.Cyc.UserSet->TotalColors) 
			{/* Too small */
				/*Can I free it to get bigger one? */
				if (D.Cyc.HasTopBuffer) { /* Free it */
					DestroyTempBuffer (D.Cyc.HiResColors, 
						D.Cyc.FadeBufferSize);
					D.Cyc.HiResColors = NULL;
					D.Pal.HasTopBuffer = TRUE;
				} else { /* buffer too small and can't free it for new one */
					ok_to_start = FALSE;				
				}
			}
		} /* else ok to allocate new buffer */

		if (ok_to_start) {
			size_t hi_res_pal_size;

			/* Turn off request for fade */
			D.Cyc.Request.Fade = FALSE;

			/* Set up new cycle fade */
			if (!(D.Cyc.ToBlack = !D.Cyc.Request.CycleSet)) {
				D.Cyc.UserSet = D.Cyc.Request.CycleSet;
			}
			ColorCyclingActive = TRUE;
			D.Cyc.FadeFrames	= (D.Cyc.Request.Hz60FadeFrames + 
				D.Hz60ToPHertz - 1) / D.Hz60ToPHertz;

			if (!D.Cyc.HiResColors) {
				/* Allocate temp buffer big enough for 2 "hi-res" palettes
			 	 * ("hi-res" meaning a WORD per color component instead of
			 	 * the normal UBYTE per color component)
			 	 */
				hi_res_pal_size =
					numcolors * COLOR_COMPONENTS * sizeof(WORD);

				D.Cyc.FadeBufferSize = 2 * hi_res_pal_size;
				if (D.Cyc.HiResColors = 
					CreateTempBuffer (D.Cyc.FadeBufferSize))
				{
					/* Put delta values in same buffer after HiResColors part */
					D.Cyc.HiResDeltas = (WORD *)(((BYTE *)D.Cyc.HiResColors)
						+ hi_res_pal_size);
				}

				/* If I succeeded in allocating a temp buffer, then
				 * I will now have the top most buffer on the temp buffer
				 * stack and the cycle color structure will not.
				 * So set up the flags to reflect the current state.
			 	 */
				D.Pal.HasTopBuffer = !(D.Cyc.HasTopBuffer = 
					(D.Cyc.HiResColors != NULL));
			}

			/* I now either have a buffer or ran out of memory */
			if (D.Cyc.HiResColors) {
#if __MSDOS__
				register int i;
				WORD *delta, *tmp;
				BYTE *src_color, *dst_color;
				WORD  diff;

				/* Compute increment */
				D.Cyc.CBytes = numcolors * COLOR_COMPONENTS;
				delta = D.Cyc.HiResDeltas;
				if (!D.Cyc.ToBlack) {
					dst_color = (BYTE *)D.Cyc.UserSet->Colors;
				}
				src_color = (BYTE *)D.Cyc.RGB;

				for (i = D.Cyc.CBytes; i; i--) {
					diff = ((D.Cyc.ToBlack) ? 0 : *dst_color++) - 
						*src_color++;
					*delta = (diff << 8) / D.Cyc.FadeFrames;
					if (*delta > 0) {
						++(*delta);
					}
					++delta;
				}


				/* Make temporary palette */
				tmp = D.Cyc.HiResColors;
				src_color = (BYTE *)D.Cyc.RGB;
				for (i = D.Cyc.CBytes; i; i--) {
					*tmp++ = ((WORD)*src_color++) << 8;
				}
#else
				register int	i;
				WORD			*tmp;
				WORD			*delta;
				UWORD			*src_color, *dst_color; /* xRGB */
				WORD			diff;
				UWORD			source_color;
				UWORD			dest_color;

				/* Compute increment & Make temporary palette */

				D.Cyc.CBytes	= numcolors * COLOR_COMPONENTS;

				delta		= D.Cyc.HiResDeltas;

				src_color	= (UWORD *) D.Cyc.RGB;
				if (!D.Cyc.ToBlack) {
					dst_color	= (UWORD *) D.Cyc.UserSet->Colors;
				}

				tmp			= (WORD *) D.Cyc.HiResColors;

				for (i = numcolors; i; i--) {
					WORD	temp_color;

					source_color	= *src_color++;
					dest_color		= (D.Cyc.ToBlack) ? 0 : *dst_color++;

					/*** do Red ***/

					diff	= X_RED(dest_color) - (temp_color = X_RED(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Cyc.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;


					/*** do Green ***/

					diff	= X_GREEN(dest_color) - (temp_color = X_GREEN(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Cyc.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;


					/*** do Blue ***/

					diff	= X_BLUE(dest_color) - (temp_color = X_BLUE(source_color));
					*delta	= (diff << HI_RES_SHIFT) / D.Cyc.FadeFrames;

					if (*delta > 0) {
						++(*delta);
					}
					++delta;

					*tmp++ = temp_color << HI_RES_SHIFT;
				}
#endif

			} else { /* Out of memory for fade buffer */
				/* Substitute Change-Cycle-Request for failed Fade-Request */
				D.Cyc.Request.Change = TRUE;
				if (D.Pal.Request.Fade) {
					D.Pal.Request.Fade = !(D.Pal.Request.Change = TRUE);
				}
			}
		}
	}

	/* Maintain public flags */
	PaletteIsFading = D.Pal.FadeFrames;
	CycleSetIsFading = D.Cyc.FadeFrames;

	if (changed) {
		CreateCopyRanges ();
	}
#endif
}
} /* Paint */

/*********************************************************************
 *
 * ClosePainter
 *
 * SYNOPSIS
 *		void ClosePainter (void) {
 *
 * PURPOSE
 *		To clean up stuff done in OpenPainter()
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void ClosePainter (void) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "ClosePainter";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT

	if (D.VBlankInterruptHandle > 0) RemoveVBlank (D.VBlankInterruptHandle);
	if (D.InterruptHandle > 0) RemoveInterrupt (D.InterruptHandle);

	if (D.RegsChanged) free (D.RegsChanged);
#if 0
	if (CopyRanges) free (CopyRanges);
#endif
	if (D.Pal.RGB) free (D.Pal.RGB);
	if (D.Cyc.RGB) free (D.Cyc.RGB);

#endif
}
} /* ClosePainter */

/*********************************************************************
 *
 * OpenPainter
 *
 * SYNOPSIS
 *		short OpenPainter (	
 *			short max_colors, 
 *			short max_cycle_colors, 
 *			short hertz
 *		)
 *
 * PURPOSE
 *		To initialize and allocate structures for controlling palette.
 *
 * INPUT
 *		max_colors: # of colors in largest palette or cycle range used.
 *		max_cycle_colors: Total # of colors in largest cycle set used.
 *				For example if you use only one cycle set, and it has
 *				a cycle range with four colors and another with 3 colors,
 *				then mac_cycle_colors == (4 + 3) == 7.
 * 		hertz:	Maximum number of times per second that any color cycle
 *				will change. Or if no cycle colors are used then the rate
 *				at which you would like the the palette handling interrupt
 *				routine to check for palette change requests.
 *				Must be <= 60.
 *
 *		IMPORTANT:	OpenTimers() must be called before this routine is 
 *				called, becuase it tries to install interrupts.
 *	
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		TRUE on success.
 *		FALSE on failure.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short OpenPainter (	
	short max_colors, 
	short max_cycle_colors, 
	short hertz
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "OpenPainter";
#endif
#if REQUIRE
	if (max_colors < 2) {
		SetGlobalErr ( ERR_REQUIREMENTS);
		GEprintf ("Error: max_colors < 2 in OpenPainter()\n");
		return FALSE;
	}
#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	size_t size;

	/* Clear Palette Control Structure */
	memset (&D, 0, sizeof (D));

	D.Pal.MaxColors = PaletteMaxColors = max_colors;
	D.Cyc.MaxColors = max_cycle_colors;
	D.PHertz = hertz;
	D.Hz60ToPHertz = 60 / D.PHertz;
	D.InterruptHandle = -1;
	PaletteIsFading = CycleSetIsFading = FALSE;

	if (D.Cyc.MaxColors) {
		/* Allocate Space for Private color cycle colors */
		if (!(D.Cyc.RGB = (ColorType *)calloc 
			(D.Cyc.MaxColors, sizeof (ColorType))))
		{
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf ("OpenPainter:OOM D.Cyc.RGB\n");
			goto CLOSE_FAIL;
		}
	}

	/* Allocate Space for Private RGB Data */
	if (!(D.Pal.RGB = (ColorType *)calloc 
		(D.Pal.MaxColors, sizeof (ColorType))))
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OpenPainter:OOM D.Pal.RGB\n");
		goto CLOSE_FAIL;
	}

	/** Allocate CopyRanges array */
	size = (MAX_COPY_RANGES + 1) * sizeof (CopyRangeType);
	if (!(CopyRanges = (CopyRangeType *)malloc (size))) 
	{
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf ("OpenPainter:OOM CopyRanges\n");
		goto CLOSE_FAIL;
	}

	/** Allocate RegsChanged array */
	if (!(D.RegsChanged = (UBYTE *)malloc (D.Pal.MaxColors + 1))) 
	{
		SetGlobalErr ( ERR_OUT_OF_MEMORY);
		GEprintf ("OpenPainter:OOM D.RegsChanged\n");
		goto CLOSE_FAIL;
	}

	/* Set Palette to Black (black because D.Pal.RGB allocated with calloc()) */
	memset (D.RegsChanged, 1, D.Pal.MaxColors * sizeof (UBYTE));

	CreateCopyRanges ();

#if 1
	/* Install Palette handling interrupt routine */
	if (!(D.InterruptHandle = InstallInterrupt ((void __far *) Paint, D.PHertz))) {
	 	SetGlobalErr (ERR_GENERIC);
		GEprintf ("OpenPainter: couldn't install Paint()\n");
		goto CLOSE_FAIL;
	}
#else
	/* Install Palette handling interrupt routine */
	if (!(D.InterruptHandle = InstallInterrupt ((void __far *) Paint, 1))) {
	 	SetGlobalErr (ERR_GENERIC);
		GEprintf ("OpenPainter: couldn't install Paint()\n");
		goto CLOSE_FAIL;
	}
#endif

#if GFX_COLORWRITE_VBLANK
	/* Install vertical blank color stuffer */
	if (!(D.VBlankInterruptHandle = InstallVBlank (ASM_WriteColors))) {
	 	SetGlobalErr (ERR_GENERIC);
		GEprintf ("OpenPainter: couldn't install VBlank()\n");
		goto CLOSE_FAIL;
	}
#endif


#if 1
	/* Wait for palette */
	while (D.PrevNotDone);
#endif

	return TRUE;
CLOSE_FAIL:

	ClosePainter();
	return FALSE;
#endif
}
} /* OpenPainter */

/*********************************************************************
 *
 * UsePaletteN
 *
 * SYNOPSIS
 *		void	UsePaletteN  (
 *			PaletteType	*palette,
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors
 *		) 
 *
 * PURPOSE
 *		To set a range of colors in the hardware palette from the
 *		corresponding range in a user palette, or to black if 'palette'
 *		is NULL.
 *
 * INPUT
 *		palette:
 *					Installed 
 *		palette		palette		Action
 *		---------	---------	------------
 *		valid		valid		Installs your palette.
 *		valid		NULL		Installs your palette.
 *		NULL		valid		Sets palette to black and 
 *									installs NULL palette.
 *		NULL		NULL		No action.
 *	
 *		src_offset:	First color of range in palette to copy from.
 *		dst_offset:	First color of range in palette to write to.
 *		numcolors:	Number of colors to copy.
 *
 * AFFECTS
 *		Sets up flags to tell interrupt driven Palette updating routine to
 *		copy the color range to the Private (shadow) palette then to hardware
 *		palette. Waits until palette is set before returning.
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void UsePaletteN  (
	PaletteType	*palette,
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors
) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "UsePaletteN";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	short tmp;

	if (!palette && !D.Pal.UserPalette) return;

	/* Do nothing if offset is out of palette range */
	if (dst_offset >= D.Pal.MaxColors) return;

	/* Reduce numcolors if it is out of palette range */
	if ((tmp = (numcolors + dst_offset)) > D.Pal.MaxColors) {
		numcolors -= (tmp - D.Pal.MaxColors);
	}

	D.Pal.Request.SrcOffset = src_offset;
	D.Pal.Request.DstOffset = dst_offset;
	D.Pal.Request.NumColors = numcolors;
	D.Pal.Request.Palette = palette;
	D.Pal.Request.Change = TRUE;

	/* Wait til request recieved by interrupt driven Paint() routine */
	while (D.Pal.Request.Change);

	/* Wait for palette to be all set */
	while (D.PrevNotDone || ActivationFlag);
#endif
}
} /* UsePaletteN */

/*********************************************************************
 *
 * UseColorCycleSet
 *
 * SYNOPSIS
 *		void UseColorCycleSet (
 *			ColorCycleSetType	*color_cycle_set,
 *			short				 activate
 *		)
 *
 * PURPOSE
 *		To install a new cycle set and optionally activate it.
 *
 * INPUT
 *		color_cycle_set:
 *					Installed
 *		cycle set	CycleSet	Action
 *		---------	---------	----------
 *		valid		valid		Installs your cycle set.
 *		valid		NULL		Installs your cycle set.
 *		NULL		valid		Installs Null cycle set.
 *		NULL		NULL		No action.
 *	
 *		activate:			TRUE if cycling should be activated.
 *
 * AFFECTS
 *		Puts in request to change the cycle-set and waits til request is
 *		handled before returning.
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void UseColorCycleSet (
	ColorCycleSetType	*color_cycle_set,
	short				 activate
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "UseColorCycleSet";
#endif
#if REQUIRE
	if (color_cycle_set && color_cycle_set->TotalColors > D.Cyc.MaxColors) {
	 	SetGlobalErr (ERR_GENERIC);
		GEprintf1 ("Error: UseColorCycleSet(): TotalColors > (MaxColors=%d)\n",
			D.Cyc.MaxColors);
		return;
	}
#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	if (!color_cycle_set && !D.Cyc.UserSet) return;

	D.Cyc.Request.CycleSet = color_cycle_set;
	D.Cyc.Request.Activate = activate;
	D.Cyc.Request.Change = TRUE;

	/* Wait til request recieved by interrupt driven Paint() routine */
	while (D.Cyc.Request.Change);
#endif
}
} /* UseColorCycleSet */

/*********************************************************************
 *
 * UsePaintBoxN
 *
 * SYNOPSIS
 *		void UsePaintBoxN (
 *			PaintBoxType	*paintbox,
 *			short			 src_offset,
 *			short			 dst_offset,
 *			short		 	 numcolors,
 *			short			 activate_color_cycling
 *		)
 *
 * PURPOSE
 *		To install new palette, and install new color cycling set and
 *		optionally activate it.
 *
 * INPUT
 *					paintbox	Installed 
 *		paintbox	->Palette	Palette		Action
 *		--------	---------	---------	------------
 *		valid		valid		valid		Installs your palette.
 *		valid		valid		NULL		Installs your palette.
 *		valid		NULL		valid		Sets palette to black and 
 *												installs NULL palette.
 *		valid		NULL		NULL		No action.
 *		NULL					valid		Sets palette to black and
 *												installs NULL palette.
 *		NULL					NULL		No action.
 *
 * 
 *					paintbox	Installed 
 *		paintbox	->CycleSet	CycleSet	Action
 *		--------	----------	--------	----------
 *		valid		valid		valid		Installs your cycle set.
 *		valid		valid		NULL		Installs your cycle set.
 *		valid		NULL		valid		Installs Null cycle set.
 *		valid		NULL		NULL		No action.
 *		NULL					valid		Installs Null cycle set.
 *		NULL					NULL		No action.
 *	
 *		src_offset:	First color of range in palette to copy from.
 *		dst_offset:	First color of range in palette to write to.
 *		numcolors:	Number of colors in palette range.
 *		activate_color_cyclinge:	non-zero if you want color cycling
 *					to be activated after cycle set is installed.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void UsePaintBoxN (
	PaintBoxType	*paintbox,
	short			 src_offset,
	short			 dst_offset,
	short			 numcolors,
	short			 activate_color_cycling
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "UsePaintBox";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	PaletteType			*palette = NULL;
	ColorCycleSetType	*cycleset = NULL;
	short tmp;
	short do_cyc;

	if (paintbox) {
		palette = paintbox->Palette;
		cycleset = paintbox->CycleSet;
	}

	/* Do nothing if offset is out of palette range */
	if (dst_offset >= D.Pal.MaxColors) return;

	/* Reduce numcolors if it is out of palette range */
	if ((tmp = (numcolors + dst_offset)) > D.Pal.MaxColors) {
		numcolors -= (tmp - D.Pal.MaxColors);
	}

	D.Pal.Request.SrcOffset = src_offset;
	D.Pal.Request.DstOffset = dst_offset;
	D.Pal.Request.NumColors = numcolors;
	D.Pal.Request.Palette = palette;

	D.Cyc.Request.CycleSet = cycleset;
	D.Cyc.Request.Activate = activate_color_cycling;

	do_cyc = (cycleset || D.Cyc.UserSet);
	D.Pal.Request.Change = (palette || D.Pal.UserPalette);
	D.Cyc.Request.Change = do_cyc;


	/* Wait til request recieved by interrupt driven Paint() routine */
	while (D.Pal.Request.Change || D.Cyc.Request.Change); 

	/* Wait for palette to be all set */
	while (D.PrevNotDone || ActivationFlag);
#endif
}
} /* UsePaintBoxN */

/*********************************************************************
 *
 * FadeToPaletteN
 *
 * SYNOPSIS
 *		void FadeToPaletteN (
 *			PaletteType *palette, 
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors,
 *			short		 hz60frames
 *		) 
 *
 * PURPOSE
 *		To fade from the current palette colors to those in 'palette'.
 *	
 * INPUT
 *		palette:	
 *					Installed
 *		palette		Palette		Action
 *		---------	---------	----------
 *		valid		valid		Fades from installed palette to 
 *									your palette.
 *		valid		NULL		Fades from black to your palette.
 *		NULL		valid		Fades from installed palette 
 *									to black then installs NULL palette.
 *		NULL		NULL		No action.
 *	
 *		src_offset:	First color of range in palette to fade from.
 *		dst_offset:	First color of range in palette to fade to.
 *		numcolors:	Number of colors to fade.
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FadeToPaletteN (
	PaletteType *palette, 
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors,
	short		 hz60frames
) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FadeToPaletteN";
#endif
#if REQUIRE
	if (palette && palette->NumColors > D.Pal.MaxColors) {
		SetGlobalErr (ERR_GENERIC);
		GEprintf1 ("Error: FadeToPaletteN(): NumColors > (MaxColors=%d)\n",
			D.Pal.MaxColors);
		return;
	}
#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	short tmp;

	if (!palette && !D.Pal.UserPalette) return;

	/* Do nothing if offset is out of palette range */
	if (dst_offset >= D.Pal.MaxColors) return;

	/* Reduce numcolors if it is out of palette range */
	if ((tmp = (numcolors + dst_offset)) > D.Pal.MaxColors) {
		numcolors -= (tmp - D.Pal.MaxColors);
	}

	D.Pal.Request.SrcOffset = src_offset;
	D.Pal.Request.DstOffset = dst_offset;
	D.Pal.Request.NumColors = numcolors;
	D.Pal.Request.Hz60FadeFrames = hz60frames;
	D.Pal.Request.Palette = palette;
	D.Pal.Request.Fade = TRUE;

	/* Wait til fade has started before exiting */
	while (D.Pal.Request.Fade); 
#endif
}
} /* FadeToPaletteN */

/*********************************************************************
 *
 * FadeToColorCycleSet
 *
 * SYNOPSIS
 *		void FadeToColorCycleSet (
 *			ColorCycleSetType	*color_cycle_set,
 *			short				 hz60frames
 *		)
 *
 * PURPOSE
 *		To fade from the current colors in the current CycleSet to 
 *		those in 'color_cycle_set'.		
 *
 * INPUT
 *		color_cycle_set: 
 *						Installed 
 *		color_cycle_set	CycleSet	Action
 *		---------------	--------	---------
 *		valid			valid		Installs your cycle set but fades
 *										from old installed cycleset
 *										colors to your cycleset colors.
 *		valid			NULL		Installs your cycle set and fades 
 *										from black cycle set colors 
 *										to your cycleset colors.
 *		NULL			valid		Fades installed cycle set colors 
 *										to black then installs NULL
 *										cycle set.
 *		NULL			NULL		No action.
 *
 *
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FadeToColorCycleSet (
	ColorCycleSetType	*color_cycle_set,
	short				 hz60frames
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FadeToColorCycleSet";
#endif
#if REQUIRE
	if (color_cycle_set && color_cycle_set->TotalColors > D.Cyc.MaxColors) {
	 	SetGlobalErr (ERR_GENERIC);
		GEprintf1 ("Error: MCGA_FadeToColorCycleSetSet(): TotalColors > (MaxColors=%d)\n",
			D.Cyc.MaxColors);
		return;
	}

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	if (color_cycle_set || D.Cyc.UserSet) {
		D.Cyc.Request.Hz60FadeFrames = hz60frames;
		D.Cyc.Request.CycleSet = color_cycle_set;
		D.Cyc.Request.Fade = TRUE;

		/* Wait til fade has started before exiting */
		while (D.Cyc.Request.Fade); 
	}
#endif
}
} /* FadeToColorCycleSet */

/*********************************************************************
 *
 * FadeToPaintBoxN
 *
 * SYNOPSIS
 *		void FadeToPaintBoxN (
 *			PaintBoxType *paintbox,
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors,
 *			short		 hz60frames
 *		)
 *
 * PURPOSE
 *		To fade from the colors in the current paintbox to those in
 *		'paintbox'.
 *
 * INPUT
 *		paintbox:	
 * 
 *					paintbox	Installed 
 *		paintbox	->Palette	Palette		Action
 *		--------	---------	-------		----------
 *		valid		valid		valid		Fades from installed palette to 
 *												your palette.
 *		valid		valid		NULL		Fades from black to your palette.
 *		valid		NULL		valid		Fades from installed palette 
 *												to black then installs 
 *												NULL palette.
 *		valid		NULL		NULL		No action.
 *		NULL					valid		Fades from installed palette 
 *												to Black then installs 
 *												NULL palette.
 *		NULL					NULL		No Action.
 *
 * 
 *					paintbox	Installed 
 *		paintbox	->CycleSet	CycleSet	Action
 *		--------	----------	--------	----------
 *		valid		valid		valid		Installs your cycle set but fades
 *												from old installed cycleset
 *												colors to your cycleset colors.
 *		valid		valid		NULL		Installs your cycle set and fades 
 *												from black cycle set colors 
 *												to your cycleset colors.
 *		valid		NULL		valid		Fades installed cycle set colors
 *												to black then installs 
 *												NULL cycle set.
 *		valid		NULL		NULL		No action.
 *		NULL					valid		Fades installed cycle set colors 
 *												to black then installs
 *												NULL cycle set.
 *		NULL					NULL		No Action.                           
 *
 *	
 *		src_offset:	First color of range in palette to fade from.
 *		dst_offset:	First color of range in palette to fade to.
 *		numcolors:	Number of colors in palette range.
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FadeToPaintBoxN (
	PaintBoxType *paintbox,
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors,
	short		 hz60frames
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "FadeToPaintBoxN";
#endif
#if REQUIRE
	if (paintbox) {
		if (paintbox->CycleSet && 
			paintbox->CycleSet->TotalColors > D.Cyc.MaxColors) 
		{
		 	SetGlobalErr (ERR_GENERIC);
			GEprintf1 ("Error: FadePaintBoxN(): TotalColors > (MaxColors=%d)\n",
				D.Cyc.MaxColors);
			return;
		}
		if (paintbox->Palette && paintbox->Palette->NumColors >
			D.Pal.MaxColors)
		{
		 	SetGlobalErr (ERR_GENERIC);
			GEprintf1 ("Error: FadePaintBoxN(): NumColors > (MaxColors=%d)\n",
				D.Pal.MaxColors);
			return;
		}
	}

#endif
/*-------------------------------------------------------------------------*/
{
#if DONT_FAKE_IT
	PaletteType			*palette = NULL;
	ColorCycleSetType	*cycleset = NULL;
	short tmp;
	short do_cyc;

	if (paintbox) {
		palette = paintbox->Palette;
		cycleset = paintbox->CycleSet;
	}

	/* Do nothing if offset is out of palette range */
	if (dst_offset >= D.Pal.MaxColors) return;

	/* Reduce numcolors if it is out of palette range */
	if ((tmp = (numcolors + dst_offset)) > D.Pal.MaxColors) {
		numcolors -= (tmp - D.Pal.MaxColors);
	}

	D.Pal.Request.Palette = palette;
	D.Pal.Request.SrcOffset = src_offset;
	D.Pal.Request.DstOffset = dst_offset;
	D.Pal.Request.NumColors = numcolors;

	D.Pal.Request.Hz60FadeFrames = D.Cyc.Request.Hz60FadeFrames = hz60frames;

	D.Cyc.Request.CycleSet = cycleset;

	do_cyc = (cycleset || D.Cyc.UserSet);
	D.Pal.Request.Fade = (palette || D.Pal.UserPalette);
	D.Cyc.Request.Fade = do_cyc;

	/* Wait til fade has started before exiting */
	while (D.Pal.Request.Fade || D.Cyc.Request.Fade);
#endif
}
} /* FadeToPaintBoxN */

