/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ASUP.c
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/27/90
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		C routine support for Amiga mode graphics
 *
 * HISTORY
 *	06/19/91 Wednesday (RGM) - Moved WaitVTimer, CheckVTimer, and
 *		ResetVTimer to GRAPHICS.C.
 *	10/10/93 Sunday (dcc) - changed short's to int's. Did not verify
 *		operation (no Amiga available).
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <echidna/etypes.h>
#if LATTICE
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#else
#include <functions.h>
#endif
#include <exec/memory.h>
#include <hardware/blit.h>
#include <hardware/dmabits.h>
#include <hardware/custom.h>
#include <graphics/gfx.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <echidna/emacros.h>
#include <echidna/graphics.h>
#include <echidna/eerrors.h>
#include <echidna/memory.h>
#include <echidna/sound.h>
#include "easydbuf.h"

#define PRINTDATA 	0
#define CREATEMASK	0

/*************************** EXTERNAL VARIABLES ***************************/
extern struct IntuitionBase *IntuitionBase;

/********************************* MACROS *********************************/
#define WAITFORBLITTER	while(((struct Custom *)(0x0DFF000L))->dmaconr & DMAF_BLTDONE);

/**************************** C O N S T A N T S ***************************/
#define GHEADERSIZE	(GSHAPESIZE + AMIGASDATASIZE)

/**************************** GLOBAL VARAIABLES ***************************/

/* LATTICE BLINK says duplicate symbol _VBTimer, */
/* so we have the allocation in easydbuf.c       */
extern ULONG		 VBTimer;
UBYTE		 PenColor		= 1;
UBYTE		 PaperColor		= 0;
GPage		 GrSeg;
GPage		 WorkBuffer		= NULL;
GPage		 BackBuffer		= NULL;
GPage		 DisplayBuffer	= NULL;

int			 ClipLeft		= 0;
int			 ClipTop		= 0;
int			 ClipWidth		= 320;
int			 ClipHeight		= 200;

struct Screen	*AmigaScreen = NULL;
struct Window	*MainWindow = NULL;

PauseFunc	 BeforePause = NULL;
PauseFunc	 AfterPause  = NULL;

/************************************  ************************************/

#if PRINTDATA
/**************************************************************************
 *
 * PrintPlane 
 *
 * SYNOPSIS
 *		void PrintPlane (Shape *shape, UBYTE *planedata)
 *
 * PURPOSE
 *		Print a bitplane diagram to stdout.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void PrintPlane (UWORD *planedata, short w, short h)
{
	short	y;
	short	x;
	short	b;
	UWORD	data;

	for (y = 0; y < h; y++) {
		for (x = 0; x < w; x++) {
			printf (" ");
			data = *planedata++;
			for (b = 0; b < 16; b++) {
				if (data & 0x8000) {
					printf ("*");
				} else {
					printf ("");
				}
				data <<= 1;
			}
		}
		printf ("\n");
	}

} /* PrintPlane  */

void PrintGShape (GShape *gs)
{
	UWORD	 size;
	UWORD	*data;
	UWORD	*mask;
	UWORD	 w;
	UWORD	 h;
	UWORD	 p;
	
	w    = (gs->Data.Amiga.DWidth + 15) >> 4;
	h    = gs->Data.Amiga.DHeight;
	size = w * h;

	/* copy data */
	mask = (UWORD *)(((ULONG)gs) + GHEADERSIZE);
	data = mask + size;

	printf ("gs->Size   == %u\n", gs->Size);
	printf ("gs->Width  == %u\n", gs->Width);
	printf ("gs->Height == %u\n", gs->Height);
	printf ("gs->TColor == %u\n", gs->TColor);
	printf ("gs->Data.Amiga.XOffset == %u\n", gs->Data.Amiga.XOffset);
	printf ("gs->Data.Amiga.DWidth  == %u\n", gs->Data.Amiga.DWidth);
	printf ("gs->Data.Amiga.YOffset == %u\n", gs->Data.Amiga.YOffset);
	printf ("gs->Data.Amiga.DHeight == %u\n", gs->Data.Amiga.DHeight);

	if (size) {
		printf ("Mask\n");
		PrintPlane (mask, w, h);

		for (p = 0; p < 8; p++ ) {
			if (gs->Data.Amiga.PlanePick & (1 << p)) {
				printf ("Plane %d\n", p);
				PrintPlane (data, w, h);
			} else {
				if (gs->Data.Amiga.PlaneOnOff & (1 << p)) {
					printf ("Plane %d Same as Mask\n", p);
				} else {
					printf ("Plane %d Clear\n", p);
				}
			}
		}
	} else {
		printf ("Size = 0...No Data\n");
	}
}
#endif

/************************************  ************************************/
#if 0
/************************************  ************************************/
/*
 *	Synopsis
 *
 *	void *malloc(size_t size);
 *
 *
 *	Description
 *
 *		The malloc function allocates space for an object whose size is
 *	specified by size and whose value is indeterminate.
 *
 *
 *	Returns
 *
 *		The malloc function returns either a null pointer or a pointer to the
 *	allocated space.
 */

#if AZTEC_C
struct mem {
	struct mem *next;
	long size;
};

extern struct mem *_Free;

void *mallocAmigaMem(size_t size, ULONG type)
{
	static UBYTE init = FALSE;
	register struct mem *ptr;
	extern void (*_cln)(void);

	if (size == 0)
		return(0);
	if (!init) {
		init = TRUE;
		malloc (1L);
	}
	if ((ptr = AllocMem(size+sizeof(struct mem), type)) == 0)
		return((void *)0);
	ptr->next = _Free;
	ptr->size = size;
	_Free = ptr;
/*
printf ("allocating %015ld bytes, %015ld free\n", size, AvailMem (NULL));
*/
	return((char *)ptr + sizeof(struct mem));
}

void *callocAmigaMem(size_t nelem, size_t size, ULONG type)
{
	register size_t i = nelem*size;
	register void *cp;

	if ((cp = mallocAmigaMem(i, type)) != NULL)
		memset(cp, 0, i);
	return(cp);
}

#endif
#endif
/************************************  ************************************/
/************************************  ************************************/

#undef DEBUG

#if DEBUG
typedef struct {
	UWORD bltddat;
	UWORD bltcon0;
	UWORD bltcon1;
	UWORD bltafwm;
	UWORD bltalwm;
	APTR bltcpt;
	APTR bltbpt;
	APTR bltapt;
	APTR bltdpt;
	UWORD bltsize;
	UWORD bltcmod;
	UWORD bltbmod;
	UWORD bltamod;
	UWORD bltdmod;
	UWORD bltcdat;
	UWORD bltbdat;
	UWORD bltadat;
} BLITTER;
#else
typedef struct Custom BLITTER;
#endif

#if DEBUG
void dorealblit (BLITTER *hardware)
{
	register struct Custom *cs = (struct Custom *)0x0DFF000;
	OwnBlitter ();

	WAITFORBLITTER

	cs->bltddat = hardware->bltddat;
	cs->bltcon0 = hardware->bltcon0;
	cs->bltcon1 = hardware->bltcon1;
	cs->bltafwm = hardware->bltafwm;
	cs->bltalwm = hardware->bltalwm;
	cs->bltcpt  = hardware->bltcpt;
	cs->bltbpt  = hardware->bltbpt;
	cs->bltapt  = hardware->bltapt;
	cs->bltdpt  = hardware->bltdpt;
	cs->bltcmod = hardware->bltcmod;
	cs->bltbmod = hardware->bltbmod;
	cs->bltamod = hardware->bltamod;
	cs->bltdmod = hardware->bltdmod;
	cs->bltcdat = hardware->bltcdat;
	cs->bltbdat = hardware->bltbdat;
	cs->bltadat = hardware->bltadat;
	cs->bltsize = hardware->bltsize;

	DisownBlitter ();
}
#endif

/**************************************************************************
 *
 * DrawGShape
 *
 * SYNOPSIS
 *    void DrawGShape (
 *				GShape *shape,
 *				int  toleft,
 *				int  totop
 *             );
 *
 * FUNCTION
 *    Draw a shape with the blitter.
 *
 * INPUTS
 *
 * BUGS
 *
 * HISTORY
 *
*/

void DrawGShape (GShape *shape, int toleft, int totop)
{
	register ULONG		 fromoffset;
	register ULONG		 tooffset;
#if DEBUG
		 BLITTER	 h;
#endif
	register BLITTER	*hardware = 
#if DEBUG
						&h;
#else
						(struct Custom *)0x0DFF000;
#endif
	UBYTE		*mask;
	UBYTE		*sdata;
	ULONG		ssize;
	UWORD		sbpr;
	UWORD		fsword;
	UWORD		tsword;
	UWORD		newwidth;
	UWORD		fend;
	UWORD		tend;
	UWORD		fsbit;
	UWORD		tsbit;
	UWORD		febit;
	UWORD		tebit;
	UWORD		frommod;
	UWORD		tomod;
	UWORD		shift;
	UWORD		bltshift;
	UWORD		fbwidth;
	UWORD		tbwidth;
	UWORD		bltwidth;
	ULONG		toptr;
	UWORD		size;
	UWORD		firstmask;
	UWORD		lastmask;
	UWORD		ndx;
	UWORD		depth;
	UBYTE		planepick;
	UBYTE		planeonoff;

/* --------------------------------------------------------------------- */
	int fromleft = 0;
	int fromtop  = 0;
	int inwidth;
	int inheight;
	int width  = inwidth  = shape->Data.Amiga.DWidth;
	int height = inheight = shape->Data.Amiga.DHeight;
	toleft      += shape->Data.Amiga.XOffset;
	totop       += shape->Data.Amiga.YOffset;

	if (toleft < ClipLeft) {
		width	  = inwidth - (ClipLeft - toleft);
		toleft	  = ClipLeft;
		fromleft += inwidth - width;
	} else if (toleft > ClipLeft + ClipWidth - inwidth) {
		width	  = (ClipLeft + ClipWidth) - toleft;
	}

	if (totop < ClipTop) {
		height	 = inheight - (ClipTop - totop);
		totop	 = ClipTop;
		fromtop += inheight - height;
	} else  if (totop > ClipTop + ClipHeight - inheight) {
		height	 = (ClipTop + ClipHeight) - totop;
	}

	if (width <= 0 || height <= 0) {
		return;
	}
/* --------------------------------------------------------------------- */

#ifndef DEBUG
	OwnBlitter ();
#endif

	sbpr  = ((shape->Data.Amiga.DWidth + 15) >> 4) << 1;
	ssize = sbpr * shape->Data.Amiga.DHeight;
	mask  = (UBYTE *)((ULONG)shape + GHEADERSIZE);

	depth = GrSeg->Depth;

	fsword = fromleft >> 4;
	tsword = toleft   >> 4;

	newwidth = width - 1;
	fend   = fromleft + newwidth;
	tend   = toleft   + newwidth;

	fromoffset = fromtop * sbpr                + (fsword << 1);
	tooffset   = totop   * GrSeg->BytesPerRow  + (tsword << 1);

	fsbit	= fromleft & 0x0F;
	tsbit	= toleft   & 0x0F;

	febit	= fend & 0x0F;
	tebit	= tend & 0x0F;

	fbwidth = (fend >> 4) - fsword + 1;
	tbwidth = (tend >> 4) - tsword + 1;

	bltwidth = fbwidth;
	shift    = (tsbit - fsbit);
	if (shift > 15) {
		tbwidth  ++;
		shift    &= 0x0F;
		tooffset -= 2;
	}

	firstmask = 0xFFFF >> fsbit;

	if (tbwidth > fbwidth) {
		lastmask = 0x0000;
		bltwidth = tbwidth;
	} else {
		lastmask  = 0xFFFF << (15 - febit);
	}

	frommod = sbpr - (bltwidth << 1);
	tomod   = GrSeg->BytesPerRow   - (bltwidth << 1);

	bltshift = shift << 12;
	size = (height << 6) | bltwidth;

	mask      += fromoffset;
    sdata      = mask + ssize;
	planepick  = shape->Data.Amiga.PlanePick;
	planeonoff = shape->Data.Amiga.PlaneOnOff;

#ifndef DEBUG
	WAITFORBLITTER
#endif

	hardware->bltafwm = firstmask;
	hardware->bltalwm = lastmask;
	hardware->bltamod = frommod;
	hardware->bltbmod = frommod;
	hardware->bltcmod = tomod;
	hardware->bltdmod = tomod;

	for (ndx = 0; ndx < depth; ndx++) {
#ifndef DEBUG
		WAITFORBLITTER
#endif
		if (planepick & 0x01) {
			hardware->bltcon0 = bltshift | 0x0FCA;
			hardware->bltcon1 = bltshift;
			hardware->bltbpt  = (APTR)sdata;
			sdata += ssize;
		} else {
			hardware->bltcon0 = bltshift | 0x0BCA;
			hardware->bltcon1 = 0x0000;
			if (planeonoff & 0x01) {
				hardware->bltbdat = 0xFFFF;
			} else {
				hardware->bltbdat = 0x0000;
			}
		}
		toptr = (ULONG)GrSeg->Planes[ndx] + (ULONG)tooffset;

		hardware->bltapt  = (APTR)mask;
		hardware->bltcpt  = (APTR)toptr;
		hardware->bltdpt  = (APTR)toptr;
		hardware->bltsize = size;

#if DEBUG
		dorealblit (hardware);
#endif

		planepick  >>= 1;
		planeonoff >>= 1;
	}

#ifndef DEBUG
	DisownBlitter ();
#endif

}

/*********************************************************************
 *
 * DrawRect
 *
 * SYNOPSIS
 *		DrawRect (int x, int y, int width, int height)
 *
 * PURPOSE
 *      Draw a Filled Rectangle in the currect color using the blitter
 *		for speed.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawRect (int toleft, int totop, int inwidth, int inheight)
{
	int width  = inwidth;
	int height = inheight;
	register ULONG		 tooffset;
#if DEBUG
		 BLITTER	 h;
#endif
	register BLITTER	*hardware = 
#if DEBUG
						&h;
#else
						(struct Custom *)0x0DFF000;
#endif
	UWORD		tsword;
	UWORD		tend;
	UWORD		tsbit;
	UWORD		tebit;
	UWORD		tomod;
	UWORD		tbwidth;
	ULONG		toptr;
	UWORD		size;
	UWORD		ndx;
	UWORD		depth;
	UBYTE		color;

/* --------------------------------------------------------------------- */
	if (toleft < 0) {
		width = inwidth - (ClipWidth - toleft);
		toleft = ClipLeft;
	} else if (toleft > ClipLeft + ClipWidth - inwidth) {
		width = (ClipLeft + ClipWidth) - toleft;
	}

	if (totop < ClipTop) {
		height = inheight - (ClipTop - totop);
		totop = ClipTop;
	} else if (totop > ClipTop + ClipHeight - inheight) {
		height = (ClipTop + ClipHeight) - totop;
	}

	if (width <= 0 || height <= 0) {
		return;
	}
/* --------------------------------------------------------------------- */
	depth = GrSeg->Depth;

	tsword = toleft >> 4;
	tend   = toleft + width - 1;

	tooffset   = totop   * GrSeg->BytesPerRow   + (tsword << 1);

	tsbit	= toleft & 0x0F;
	tebit	= tend   & 0x0F;

	tbwidth = (tend >> 4) - tsword + 1;

	if (!tsbit && tebit == 15) {
		size = (height << 6) | tbwidth;

#ifndef DEBUG
		OwnBlitter ();
		WAITFORBLITTER
#endif

		hardware->bltcon1 = 0x0000;
		hardware->bltafwm = 0xFFFF;
		hardware->bltalwm = 0xFFFF;
		hardware->bltdmod = GrSeg->BytesPerRow - (tbwidth << 1);
		color             = PenColor;

		for (ndx = 0; ndx < depth; ndx++) {
#ifndef DEBUG
			WAITFORBLITTER
#endif
			if (color & 0x01) {
				hardware->bltcon0 = 0x01FF;
			} else {
				hardware->bltcon0 = 0x0100;
			}
			toptr = (ULONG)GrSeg->Planes[ndx] + (ULONG)tooffset;

			hardware->bltdpt  = (APTR)toptr;
			hardware->bltsize = size;

#if DEBUG
	dorealblit (hardware);
#endif

			color >>= 1;
		}
#ifndef DEBUG
		DisownBlitter ();
#endif

	} else {

		tomod   = GrSeg->BytesPerRow   - (tbwidth << 1);

		size = (height << 6) | tbwidth;

#ifndef DEBUG
		OwnBlitter ();
		WAITFORBLITTER
#endif

		hardware->bltcon0 = 0x03CA;
		hardware->bltcon1 = 0x0000;

		hardware->bltafwm = 0xFFFF >> tsbit;
		hardware->bltalwm = 0xFFFF << (15 - tebit);
		hardware->bltadat = 0xFFFF;
		hardware->bltcmod = tomod;
		hardware->bltdmod = tomod;

		color             = PenColor;

		for (ndx = 0; ndx < depth; ndx++) {
#ifndef DEBUG
			WAITFORBLITTER
#endif
			if (color & 0x01) {
				hardware->bltbdat = 0xFFFF;
			} else {
				hardware->bltbdat = 0x0000;
			}
				
			toptr = (ULONG)GrSeg->Planes[ndx] + (ULONG)tooffset;

			hardware->bltcpt  = (APTR)toptr;
			hardware->bltdpt  = (APTR)toptr;
			hardware->bltsize = size;

#if DEBUG
	dorealblit (hardware);
#endif

			color >>= 1;
		}
#ifndef DEBUG
		DisownBlitter ();
#endif
	}
} /* DrawRect */

#if CREATEMASK
/**************************************************************************
 *
 * CreateMask 
 *
 * SYNOPSIS
 *		CreateMask (GShape *shape)
 *
 * PURPOSE
 *		'OR' data planes of new shape to create mask.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void CreateMask (GShape *shape)
{

#if DEBUG
		 BLITTER	 h;
#endif
	register BLITTER	*hardware = 
#if DEBUG
						&h;
#else
						(struct Custom *)0x0DFF000;
#endif
	UBYTE		*mask;
	UBYTE		*sdata;
	ULONG		ssize;
	UWORD		sbpr;
	UWORD		bltwidth;
	UWORD		size;
	UWORD		ndx;
	UBYTE		planepick;
	UBYTE		planeonoff;
	UBYTE		first;

	bltwidth = (shape->Data.Amiga.DWidth + 15) >> 4;
	sbpr  = bltwidth << 1;
	ssize = sbpr * shape->Data.Amiga.DHeight;
	mask  = ((UBYTE *)(&shape->Data)) + sizeof (shape->Data.Amiga);

	size = (shape->Data.Amiga.DHeight << 6) | bltwidth;

    sdata      = mask + ssize;
	planepick  = shape->Data.Amiga.PlanePick;
	planeonoff = shape->Data.Amiga.PlaneOnOff;
	first      = FALSE;

#ifndef DEBUG
	OwnBlitter ();
	WAITFORBLITTER
#endif

	hardware->bltcon1 = 0;
	hardware->bltafwm = 0xFFFF;
	hardware->bltalwm = 0xFFFF;
	hardware->bltamod = 0;
	hardware->bltbmod = 0;
	hardware->bltdmod = 0;

	for (ndx = 0; ndx < 8; ndx++) {
#ifndef DEBUG
		WAITFORBLITTER
#endif
		if (planepick & 0x01) {
			if (!first) {
				first = TRUE;
				hardware->bltcon0 = 0x09F0;
			} else {
				hardware->bltcon0 = 0x0DFC;
			}
			hardware->bltbpt  = (APTR)mask;
			hardware->bltapt  = (APTR)sdata;
			hardware->bltdpt  = (APTR)mask;
			hardware->bltsize = size;
			sdata += ssize;
		} else if (!first) {
			first = TRUE;
			hardware->bltcon0 = 0x0100;
			hardware->bltdpt  = (APTR)mask;
			hardware->bltsize = size;
		}

#if DEBUG
		dorealblit (hardware);
#endif
		planepick  >>= 1;
		planeonoff >>= 1;
	}

#ifndef DEBUG
	DisownBlitter ();
#endif

} /* CreateMask  */
#endif

static fliptable[] = {
	  0,	/*   0 00000000 -> 00000000 */
	128,	/*   1 00000001 -> 10000000 */
	 64,	/*   2 00000010 -> 01000000 */
	192,	/*   3 00000011 -> 11000000 */
	 32,	/*   4 00000100 -> 00100000 */
	160,	/*   5 00000101 -> 10100000 */
	 96,	/*   6 00000110 -> 01100000 */
	224,	/*   7 00000111 -> 11100000 */
	 16,	/*   8 00001000 -> 00010000 */
	144,	/*   9 00001001 -> 10010000 */
	 80,	/*  10 00001010 -> 01010000 */
	208,	/*  11 00001011 -> 11010000 */
	 48,	/*  12 00001100 -> 00110000 */
	176,	/*  13 00001101 -> 10110000 */
	112,	/*  14 00001110 -> 01110000 */
	240,	/*  15 00001111 -> 11110000 */
	  8,	/*  16 00010000 -> 00001000 */
	136,	/*  17 00010001 -> 10001000 */
	 72,	/*  18 00010010 -> 01001000 */
	200,	/*  19 00010011 -> 11001000 */
	 40,	/*  20 00010100 -> 00101000 */
	168,	/*  21 00010101 -> 10101000 */
	104,	/*  22 00010110 -> 01101000 */
	232,	/*  23 00010111 -> 11101000 */
	 24,	/*  24 00011000 -> 00011000 */
	152,	/*  25 00011001 -> 10011000 */
	 88,	/*  26 00011010 -> 01011000 */
	216,	/*  27 00011011 -> 11011000 */
	 56,	/*  28 00011100 -> 00111000 */
	184,	/*  29 00011101 -> 10111000 */
	120,	/*  30 00011110 -> 01111000 */
	248,	/*  31 00011111 -> 11111000 */
	  4,	/*  32 00100000 -> 00000100 */
	132,	/*  33 00100001 -> 10000100 */
	 68,	/*  34 00100010 -> 01000100 */
	196,	/*  35 00100011 -> 11000100 */
	 36,	/*  36 00100100 -> 00100100 */
	164,	/*  37 00100101 -> 10100100 */
	100,	/*  38 00100110 -> 01100100 */
	228,	/*  39 00100111 -> 11100100 */
	 20,	/*  40 00101000 -> 00010100 */
	148,	/*  41 00101001 -> 10010100 */
	 84,	/*  42 00101010 -> 01010100 */
	212,	/*  43 00101011 -> 11010100 */
	 52,	/*  44 00101100 -> 00110100 */
	180,	/*  45 00101101 -> 10110100 */
	116,	/*  46 00101110 -> 01110100 */
	244,	/*  47 00101111 -> 11110100 */
	 12,	/*  48 00110000 -> 00001100 */
	140,	/*  49 00110001 -> 10001100 */
	 76,	/*  50 00110010 -> 01001100 */
	204,	/*  51 00110011 -> 11001100 */
	 44,	/*  52 00110100 -> 00101100 */
	172,	/*  53 00110101 -> 10101100 */
	108,	/*  54 00110110 -> 01101100 */
	236,	/*  55 00110111 -> 11101100 */
	 28,	/*  56 00111000 -> 00011100 */
	156,	/*  57 00111001 -> 10011100 */
	 92,	/*  58 00111010 -> 01011100 */
	220,	/*  59 00111011 -> 11011100 */
	 60,	/*  60 00111100 -> 00111100 */
	188,	/*  61 00111101 -> 10111100 */
	124,	/*  62 00111110 -> 01111100 */
	252,	/*  63 00111111 -> 11111100 */
	  2,	/*  64 01000000 -> 00000010 */
	130,	/*  65 01000001 -> 10000010 */
	 66,	/*  66 01000010 -> 01000010 */
	194,	/*  67 01000011 -> 11000010 */
	 34,	/*  68 01000100 -> 00100010 */
	162,	/*  69 01000101 -> 10100010 */
	 98,	/*  70 01000110 -> 01100010 */
	226,	/*  71 01000111 -> 11100010 */
	 18,	/*  72 01001000 -> 00010010 */
	146,	/*  73 01001001 -> 10010010 */
	 82,	/*  74 01001010 -> 01010010 */
	210,	/*  75 01001011 -> 11010010 */
	 50,	/*  76 01001100 -> 00110010 */
	178,	/*  77 01001101 -> 10110010 */
	114,	/*  78 01001110 -> 01110010 */
	242,	/*  79 01001111 -> 11110010 */
	 10,	/*  80 01010000 -> 00001010 */
	138,	/*  81 01010001 -> 10001010 */
	 74,	/*  82 01010010 -> 01001010 */
	202,	/*  83 01010011 -> 11001010 */
	 42,	/*  84 01010100 -> 00101010 */
	170,	/*  85 01010101 -> 10101010 */
	106,	/*  86 01010110 -> 01101010 */
	234,	/*  87 01010111 -> 11101010 */
	 26,	/*  88 01011000 -> 00011010 */
	154,	/*  89 01011001 -> 10011010 */
	 90,	/*  90 01011010 -> 01011010 */
	218,	/*  91 01011011 -> 11011010 */
	 58,	/*  92 01011100 -> 00111010 */
	186,	/*  93 01011101 -> 10111010 */
	122,	/*  94 01011110 -> 01111010 */
	250,	/*  95 01011111 -> 11111010 */
	  6,	/*  96 01100000 -> 00000110 */
	134,	/*  97 01100001 -> 10000110 */
	 70,	/*  98 01100010 -> 01000110 */
	198,	/*  99 01100011 -> 11000110 */
	 38,	/* 100 01100100 -> 00100110 */
	166,	/* 101 01100101 -> 10100110 */
	102,	/* 102 01100110 -> 01100110 */
	230,	/* 103 01100111 -> 11100110 */
	 22,	/* 104 01101000 -> 00010110 */
	150,	/* 105 01101001 -> 10010110 */
	 86,	/* 106 01101010 -> 01010110 */
	214,	/* 107 01101011 -> 11010110 */
	 54,	/* 108 01101100 -> 00110110 */
	182,	/* 109 01101101 -> 10110110 */
	118,	/* 110 01101110 -> 01110110 */
	246,	/* 111 01101111 -> 11110110 */
	 14,	/* 112 01110000 -> 00001110 */
	142,	/* 113 01110001 -> 10001110 */
	 78,	/* 114 01110010 -> 01001110 */
	206,	/* 115 01110011 -> 11001110 */
	 46,	/* 116 01110100 -> 00101110 */
	174,	/* 117 01110101 -> 10101110 */
	110,	/* 118 01110110 -> 01101110 */
	238,	/* 119 01110111 -> 11101110 */
	 30,	/* 120 01111000 -> 00011110 */
	158,	/* 121 01111001 -> 10011110 */
	 94,	/* 122 01111010 -> 01011110 */
	222,	/* 123 01111011 -> 11011110 */
	 62,	/* 124 01111100 -> 00111110 */
	190,	/* 125 01111101 -> 10111110 */
	126,	/* 126 01111110 -> 01111110 */
	254,	/* 127 01111111 -> 11111110 */
	  1,	/* 128 10000000 -> 00000001 */
	129,	/* 129 10000001 -> 10000001 */
	 65,	/* 130 10000010 -> 01000001 */
	193,	/* 131 10000011 -> 11000001 */
	 33,	/* 132 10000100 -> 00100001 */
	161,	/* 133 10000101 -> 10100001 */
	 97,	/* 134 10000110 -> 01100001 */
	225,	/* 135 10000111 -> 11100001 */
	 17,	/* 136 10001000 -> 00010001 */
	145,	/* 137 10001001 -> 10010001 */
	 81,	/* 138 10001010 -> 01010001 */
	209,	/* 139 10001011 -> 11010001 */
	 49,	/* 140 10001100 -> 00110001 */
	177,	/* 141 10001101 -> 10110001 */
	113,	/* 142 10001110 -> 01110001 */
	241,	/* 143 10001111 -> 11110001 */
	  9,	/* 144 10010000 -> 00001001 */
	137,	/* 145 10010001 -> 10001001 */
	 73,	/* 146 10010010 -> 01001001 */
	201,	/* 147 10010011 -> 11001001 */
	 41,	/* 148 10010100 -> 00101001 */
	169,	/* 149 10010101 -> 10101001 */
	105,	/* 150 10010110 -> 01101001 */
	233,	/* 151 10010111 -> 11101001 */
	 25,	/* 152 10011000 -> 00011001 */
	153,	/* 153 10011001 -> 10011001 */
	 89,	/* 154 10011010 -> 01011001 */
	217,	/* 155 10011011 -> 11011001 */
	 57,	/* 156 10011100 -> 00111001 */
	185,	/* 157 10011101 -> 10111001 */
	121,	/* 158 10011110 -> 01111001 */
	249,	/* 159 10011111 -> 11111001 */
	  5,	/* 160 10100000 -> 00000101 */
	133,	/* 161 10100001 -> 10000101 */
	 69,	/* 162 10100010 -> 01000101 */
	197,	/* 163 10100011 -> 11000101 */
	 37,	/* 164 10100100 -> 00100101 */
	165,	/* 165 10100101 -> 10100101 */
	101,	/* 166 10100110 -> 01100101 */
	229,	/* 167 10100111 -> 11100101 */
	 21,	/* 168 10101000 -> 00010101 */
	149,	/* 169 10101001 -> 10010101 */
	 85,	/* 170 10101010 -> 01010101 */
	213,	/* 171 10101011 -> 11010101 */
	 53,	/* 172 10101100 -> 00110101 */
	181,	/* 173 10101101 -> 10110101 */
	117,	/* 174 10101110 -> 01110101 */
	245,	/* 175 10101111 -> 11110101 */
	 13,	/* 176 10110000 -> 00001101 */
	141,	/* 177 10110001 -> 10001101 */
	 77,	/* 178 10110010 -> 01001101 */
	205,	/* 179 10110011 -> 11001101 */
	 45,	/* 180 10110100 -> 00101101 */
	173,	/* 181 10110101 -> 10101101 */
	109,	/* 182 10110110 -> 01101101 */
	237,	/* 183 10110111 -> 11101101 */
	 29,	/* 184 10111000 -> 00011101 */
	157,	/* 185 10111001 -> 10011101 */
	 93,	/* 186 10111010 -> 01011101 */
	221,	/* 187 10111011 -> 11011101 */
	 61,	/* 188 10111100 -> 00111101 */
	189,	/* 189 10111101 -> 10111101 */
	125,	/* 190 10111110 -> 01111101 */
	253,	/* 191 10111111 -> 11111101 */
	  3,	/* 192 11000000 -> 00000011 */
	131,	/* 193 11000001 -> 10000011 */
	 67,	/* 194 11000010 -> 01000011 */
	195,	/* 195 11000011 -> 11000011 */
	 35,	/* 196 11000100 -> 00100011 */
	163,	/* 197 11000101 -> 10100011 */
	 99,	/* 198 11000110 -> 01100011 */
	227,	/* 199 11000111 -> 11100011 */
	 19,	/* 200 11001000 -> 00010011 */
	147,	/* 201 11001001 -> 10010011 */
	 83,	/* 202 11001010 -> 01010011 */
	211,	/* 203 11001011 -> 11010011 */
	 51,	/* 204 11001100 -> 00110011 */
	179,	/* 205 11001101 -> 10110011 */
	115,	/* 206 11001110 -> 01110011 */
	243,	/* 207 11001111 -> 11110011 */
	 11,	/* 208 11010000 -> 00001011 */
	139,	/* 209 11010001 -> 10001011 */
	 75,	/* 210 11010010 -> 01001011 */
	203,	/* 211 11010011 -> 11001011 */
	 43,	/* 212 11010100 -> 00101011 */
	171,	/* 213 11010101 -> 10101011 */
	107,	/* 214 11010110 -> 01101011 */
	235,	/* 215 11010111 -> 11101011 */
	 27,	/* 216 11011000 -> 00011011 */
	155,	/* 217 11011001 -> 10011011 */
	 91,	/* 218 11011010 -> 01011011 */
	219,	/* 219 11011011 -> 11011011 */
	 59,	/* 220 11011100 -> 00111011 */
	187,	/* 221 11011101 -> 10111011 */
	123,	/* 222 11011110 -> 01111011 */
	251,	/* 223 11011111 -> 11111011 */
	  7,	/* 224 11100000 -> 00000111 */
	135,	/* 225 11100001 -> 10000111 */
	 71,	/* 226 11100010 -> 01000111 */
	199,	/* 227 11100011 -> 11000111 */
	 39,	/* 228 11100100 -> 00100111 */
	167,	/* 229 11100101 -> 10100111 */
	103,	/* 230 11100110 -> 01100111 */
	231,	/* 231 11100111 -> 11100111 */
	 23,	/* 232 11101000 -> 00010111 */
	151,	/* 233 11101001 -> 10010111 */
	 87,	/* 234 11101010 -> 01010111 */
	215,	/* 235 11101011 -> 11010111 */
	 55,	/* 236 11101100 -> 00110111 */
	183,	/* 237 11101101 -> 10110111 */
	119,	/* 238 11101110 -> 01110111 */
	247,	/* 239 11101111 -> 11110111 */
	 15,	/* 240 11110000 -> 00001111 */
	143,	/* 241 11110001 -> 10001111 */
	 79,	/* 242 11110010 -> 01001111 */
	207,	/* 243 11110011 -> 11001111 */
	 47,	/* 244 11110100 -> 00101111 */
	175,	/* 245 11110101 -> 10101111 */
	111,	/* 246 11110110 -> 01101111 */
	239,	/* 247 11110111 -> 11101111 */
	 31,	/* 248 11111000 -> 00011111 */
	159,	/* 249 11111001 -> 10011111 */
	 95,	/* 250 11111010 -> 01011111 */
	223,	/* 251 11111011 -> 11011111 */
	 63,	/* 252 11111100 -> 00111111 */
	191,	/* 253 11111101 -> 10111111 */
	127,	/* 254 11111110 -> 01111111 */
	255,	/* 255 11111111 -> 11111111 */
};

/**************************************************************************
 *
 * BltMem
 *
 * SYNOPSIS
 *		static void BltMem (
 *			UBYTE	*sptr,
 *			UBYTE	*dptr,
 *			UBYTE	 minterm,
 *			ULONG	 size
 *		)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void BltMem (
	UBYTE	*sptr,
	UBYTE	*dptr,
	UBYTE	 minterm,
	ULONG	 size
)
{
#if DEBUG
		 BLITTER	 h;
#endif
	register BLITTER	*hardware = 
#if DEBUG
						&h;
#else
						(struct Custom *)0x0DFF000;
#endif

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "BltMem";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

#ifndef DEBUG
	OwnBlitter ();
	WAITFORBLITTER
#endif

	if (sptr) {
		hardware->bltcon0 = 0x0700 | minterm;
		hardware->bltbpt  = (APTR)sptr;
		hardware->bltbmod = 0;
	} else {
		hardware->bltcon0 = 0x0300 | minterm;
		hardware->bltbdat = 0x0000;
	}
	hardware->bltcon1 = 0;
	hardware->bltafwm = 0xFFFF;
	hardware->bltalwm = 0xFFFF;
	hardware->bltadat = 0xFFFF;
	hardware->bltcmod = 0;
	hardware->bltdmod = 0;

	hardware->bltcpt  = (APTR)dptr;
	hardware->bltdpt  = (APTR)dptr;
	hardware->bltsize = 0x0040 | (size >> 1);

#if DEBUG
	dorealblit (hardware);
#endif

#ifndef DEBUG
	DisownBlitter ();
#endif

} /* BltMem */

/**************************************************************************
 *
 * RemapColor
 *
 * SYNOPSIS
 *		static void RemapColor (
 *			GShape	*origshape,
 *			GShape	*newshape,
 *			UBYTE	 fromcolor,
 *			UBYTE	 tocolor
 *		)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void RemapColor (
	GShape	*origshape,
	GShape	*newshape,
	UBYTE	 fromcolor,
	UBYTE	 tocolor
)
{
	UBYTE	*orig;
	UBYTE	*new;
	UBYTE	*mask;
	UBYTE	*cmask;
	UBYTE	*optr;
	UBYTE	*nptr;
	ULONG	 size;
	short	 width;
	short	 height;
	UWORD	 bpr;
	UBYTE	 temp;
	UBYTE	 minterm;
	UBYTE	 planepick;
	UBYTE	 planeonoff;
	UBYTE	 p;
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "RemapColor";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	width   = origshape->Data.Amiga.DWidth;
	height  = origshape->Data.Amiga.DHeight;
	bpr     = (((width + 15) >> 4) << 1);
	size    = bpr * height;

	orig	= (UBYTE *)origshape + GHEADERSIZE + size;
	mask	= (UBYTE *)newshape  + GHEADERSIZE;
	new		= mask + size;
	cmask	= new  + size * MAX_DEPTH;

	/*** Fill work with 1's ***/
	BltMem (NULL, cmask, 0xFF, size);

	/*** Extract Color Mask ***/
	optr = orig;
	temp = fromcolor;
	planepick  = origshape->Data.Amiga.PlanePick;
	planeonoff = origshape->Data.Amiga.PlaneOnOff;
	for (p = 0; p < MAX_DEPTH; p++) {
		if (temp & 0x1) {
			minterm = 0x80;
		} else {
			minterm = 0x20;
		}
		if (planepick & 0x01) {
			BltMem (optr, cmask, minterm, size);
			optr += size;
		} else {
			if (planeonoff & 0x01) {
				BltMem (mask, cmask, minterm, size);
			} else {
				BltMem (NULL, cmask, minterm, size);
			}
		}
		temp       >>= 1;
		planepick  >>= 1;
		planeonoff >>= 1;
	}

	/*** Draw new color into new shape ***/
	nptr = new;
	temp = tocolor;
	for (p = 0; p < MAX_DEPTH; p++) {
		if (temp & 0x1) {
			minterm = 0xE0;
		} else {
			minterm = 0x20;
		}
		BltMem (cmask, nptr, minterm, size);
		nptr  += size;
		temp >>= 1;
	}

} /* RemapColor */


/*********************************************************************
 *
 * ConvertGShape
 *
 * SYNOPSIS
 *		GShape	*ConvertGShape (
 *			GShape	*genericshape
 *		)
 *
 * PURPOSE
 *		To shift the color of each pixel in a generic MCGA Shape up the 
 *		palette by one.
 *
 * INPUTS
 *		genericshape:	far pointer to shape data.
 *		datasize:		size of 'genericshape' data in bytes.
 *
*/
GShape	*ConvertGShape (
	ConvertSType	*cs
)
{
	ULONG	 size;
	ULONG	 origsize;
	short	 width;
	short	 height;
	UWORD	 bpr;
	GShape	*gshape;
	GShape	*newshape;
	UBYTE	*ps_ColorMap;
	UBYTE	*planeptr;
	UBYTE	*data;
	UWORD	 x;
	UWORD	 y;
	UWORD	 p;
	UBYTE	 planepick;
	UBYTE	 temp;

	if (!cs) {
		return NULL;
	}

	gshape 		= cs->ShapeData; 
	ps_ColorMap	= cs->ColorMap;
	width       = gshape->Data.Amiga.DWidth;
	height      = gshape->Data.Amiga.DHeight;
	bpr         = (((width + 15) >> 4) << 1);
	size        = bpr * height;

	/* Allocate new shape */
	origsize	= GHEADERSIZE + size * (MAX_DEPTH + 2);
	if (!(newshape = (GShape *)mallocChipMem (origsize))) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf (ERRINFO("Couldn't allocate shape"));
		return NULL;
	}

	/* copy header */
	memcpy (newshape, gshape, GHEADERSIZE);

	if (size) {
		if (!cs->Remap) {
			memcpy ((UBYTE *)newshape + GHEADERSIZE,
					(UBYTE *)gshape + GHEADERSIZE,
					gshape->Size - GHEADERSIZE);
			freesome (origsize - gshape->Size);
		} else {
			UBYTE	i;

			memcpy ((UBYTE *)newshape + GHEADERSIZE,
					(UBYTE *)gshape   + GHEADERSIZE,
					size);
			
			newshape->Data.Amiga.PlanePick  = 0xFF;
			newshape->Data.Amiga.PlaneOnOff = 0x00;
			for (i=0; i < MAX_COLORS; i++) {
				RemapColor (gshape, newshape, i, ps_ColorMap[i]);
			}
			freesome (size);
		}
		if (cs->VFlip) {
			UBYTE	*topdata;
			UBYTE	*botdata;
			UWORD	 lines;

			newshape->Data.Amiga.YOffset = newshape->Height -
				height - newshape->Data.Amiga.YOffset;

			planepick = newshape->Data.Amiga.PlanePick;
			planeptr  = (UBYTE *)newshape + GHEADERSIZE;
			lines     = (height + 1) >> 1;
			for (p = 0; p < 9; p++) {
				if ((!p) || (planepick & 0x01)) {
					data = planeptr;
					for (y = 0; y < lines; y++) {
						topdata = data;
						botdata = data + bpr * (height - 1 - y - y);
						for (x = 0; x < bpr; x++) {
							temp       = *topdata;
							*topdata++ = *botdata;
							*botdata++ = temp;
						}
						data += bpr;
					}
					planeptr += size;
				}
				if (p) {
					planepick >>= 1;
				}
			}
		}
		if (cs->HFlip) {
			UBYTE	*leftdata;
			UBYTE	*ritedata;
			UWORD	 bytes;
			short	 extra;

			extra = 15 - ((width - 1) & 0x0F);
			newshape->Data.Amiga.XOffset = (newshape->Width -
				width - newshape->Data.Amiga.XOffset) - extra;
			newshape->Data.Amiga.DWidth  = width + extra;

			planepick = newshape->Data.Amiga.PlanePick;
			planeptr  = (UBYTE *)newshape + GHEADERSIZE;
			bytes     = (bpr + 1) >> 1;
			for (p = 0; p < 9; p++) {
				if ((!p) || (planepick & 0x01)) {
					data = planeptr;
					for (y = 0; y < height; y++) {
						leftdata = data;
						ritedata = data + bpr - 1;
						for (x = 0; x < bytes; x++) {
							temp        = *leftdata;
							*leftdata++ = fliptable[*ritedata];
							*ritedata-- = fliptable[temp];
						}
						data += bpr;
					}
					planeptr += size;
				}
				if (p) {
					planepick >>= 1;
				}
			}
		}
	}

	return newshape;

} /* ConvertGShape */

/*********************************************************************
 *
 * FreeGShape
 *
 * SYNOPSIS
 *		void	FreeGShape (
 *			GShape	*shapeptr
 *		)
 *
 * PURPOSE
 *		To Free the memory allocated for shape by ConvertGShape.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FreeGShape(GShape *shape)
{
	free(shape);
} /* FreeShape */

struct MyRast {
	struct RastPort RastPort;
	struct BitMap	BitMap;
};

/***************************** DELETERASTPORT *****************************/
void DeleteRastPort (struct RastPort *mr)
{
#define MR	((struct MyRast *)mr)

	if (mr) {
		if (MR->BitMap.Planes[0]) {
			free (MR->BitMap.Planes[0]);
		}
		free (mr);
	}
}

/***************************** CREATERASTPORT *****************************/
struct RastPort *CreateRastPort (short depth, short width, short height)
{
	struct MyRast	*mr;
	ULONG			 psize;
	short			 i;

	if (!( mr = calloc (1, sizeof (struct MyRast)))) {
		return (NULL);
	}

	InitRastPort (&mr->RastPort);

	mr->RastPort.BitMap = &mr->BitMap;

	InitBitMap (&mr->BitMap, (long)depth, (long)width, (long)height);

	psize = RASSIZE(width,height);
	if (!(mr->BitMap.Planes[0] = mallocChipMem (psize * depth))) {
		DeleteRastPort (&mr->RastPort);
		return (NULL);
	}
	for (i=1; i < depth; i++) {
		mr->BitMap.Planes[i] = (void *)(((UBYTE *)(mr->BitMap.Planes[i - 1])) + psize);
	}

	SetRast (&mr->RastPort, 0L);

	return ((struct RastPort *)mr);
}

/****************************** POINTERSTUFF ******************************/
void AClearPointer (void)
{
	ClearPointer (MainWindow);
}

void ARestorePointer (void)
{
	static short	 init = TRUE;
	static short	*data;

	if (init) {
		init = FALSE;
		data = callocChipMem (12, 1);
	}

	SetPointer (MainWindow, (void *)data, 1, 16, 0, 0);
}
/******************************** NewWindow *******************************/

#define WMODES1		BORDERLESS|ACTIVATE|SIMPLE_REFRESH|NOCAREREFRESH
#define WMODES2		BACKDROP|RMBTRAP
#define WMODES		WMODES1|WMODES2

#define IMODES		ACTIVEWINDOW|MOUSEBUTTONS

struct	NewWindow	NewWindow = {
	0, 0,				/* Left, Top Edge			*/
	320, 200,	 		/* Width, Height			*/
	0, 1,				/* DetailPen, BlockPen		*/
	IMODES,				/* IDCMP Flags				*/
	WMODES, 			/* WINDOWFLAGS				*/
	NULL,				/* User Gadgets				*/
	NULL,				/* Check Mark				*/
	NULL,				/* Title					*/
	NULL,				/* Screen					*/
	NULL,				/* SuperBitMap				*/
	0,0,0,0,			/* Min, Max Size			*/
	CUSTOMSCREEN		/* Screen Type				*/
};

/*********************************************************************
 *
 * C_CloseGraphics
 *
 * SYNOPSIS
 *		void C_CloseGraphics (void)
 *
 * PURPOSE
 *		Free resources used by MCGA graphics routines.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static struct RastPort *BgRp;
void C_CloseGraphics (void)
{
	if (MainWindow)	{
		AClearPointer ();
		CloseWindow (MainWindow);
	}
	if (BgRp)		DeleteRastPort (BgRp);
	if (AmigaScreen) DeleteDBufScreen (AmigaScreen);

} /* C_CloseGraphics */

/*********************************************************************
 *
 * OpenGraphics
 *
 * SYNOPSIS
 *		short OpenGraphics(void)
 *
 * PURPOSE
 *		Init Amiga graphics routines.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short C_OpenGraphics(void)
{
	if ((AmigaScreen = CreateDBufScreen ())) {
		DisplayBuffer = FrontRp->BitMap;
		WorkBuffer    = BackRp->BitMap;
		if ((BgRp = CreateRastPort (5, 320, 200))) {
			BackBuffer = BgRp->BitMap;
			SetWorkPage (WorkBuffer);
			NewWindow.Screen = AmigaScreen;
			if ((MainWindow = OpenWindow (&NewWindow))) {
				ARestorePointer ();
				CloseWorkBench ();
				return 1;
			}
		}
		C_CloseGraphics ();
	}
	return 0;

} /* C_OpenGraphics */

/*********************************************************************
 *
 * DrawGShapeNoTrans
 *
 * SYNOPSIS
 *		DrawGShapeNoTrans (GShape *shape, short x, short y)
 *
 * PURPOSE
 *		Draw a Shape without transparency.  Use the Blitter!
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DrawGShapeNoTrans (GShape *shape, int x, int y)
{
	SetPenColor (shape->TColor);
	DrawRect (x, y, shape->Width, shape->Height);
	DrawGShape (shape, x, y);

} /* DrawGShapeNoTrans */

/**************************************************************************
 *
 * ClearWorkPage 
 *
 * SYNOPSIS
 *		void ClearWorkPage (void)
 *
 * PURPOSE
 *		Clear the current work page to the currnet Paper color.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void ClearWorkPage (void)
{
	UBYTE	tempcolor;

	tempcolor = PenColor;
	PenColor  = PaperColor;
	DrawRect (0, 0, 320, 200);
	PenColor  = tempcolor;

} /* ClearWorkPage  */


void DrawPixel (int x, int y)
{
	static UBYTE bits[] = {
		0x80,0x40,0x20,0x10,
		0x08,0x04,0x02,0x01,
	};

	ULONG	 offset;
	UBYTE	**plane;
	UBYTE	 color;
	UBYTE	 bit;
	UBYTE	 xbit;
	UBYTE	 i;

	offset = (x >> 3) + y * GrSeg->BytesPerRow;

	color = PenColor;
	plane = (UBYTE **)&GrSeg->Planes[0];
	bit   = bits[x & 0x07];
	xbit  = ~bit;
	for (i = 0; i < MAX_DEPTH; i++) {
		if (color & 0x01) {
			*(*plane + offset) |= bit;
		} else {
			*(*plane + offset) &= xbit;
		}
		color >>= 1;
		plane++;
	}
}

/*********************************************************************
 *
 * SetClipArea
 *
 * SYNOPSIS
 *		void SetClipArea (short xmin, short ymin, short xmax, short ymax)
 *
 * PURPOSE
 *		To set the shape clipping variables.
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void SetClipArea (short xmin, short ymin, short xmax, short ymax)
{
	ClipLeft		= xmin;
	ClipTop			= ymin;
	ClipWidth		= xmax - xmin + 1;
	ClipHeight		= ymax - ymin + 1;

} /* SetClipArea */

