/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * CLIP.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/04/91
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Stuff for clipping graphics!?!?!
 *
 * HISTORY
 *		10/10/93 Sunday (dcc) - changed short's to int's to support 32-bit procs
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/eerrors.h>
#include <echidna/listfunc.h>
#include <echidna/grafx.h>

#define LOCAL static

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/

typedef struct {
	NodeType	 Node;
	int			 X;
	int			 Y;
	int			 Width;
	int			 Height;
} ClipNode;

typedef struct {
	NodeType		 Node;
	int				 X;
	int				 Y;
	int				 Width;
	int				 Height;
	SaveAreaType	*Area;
} ScreenArea;

/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/

LOCAL ListType	 FreeListX;
LOCAL ListType	*FreeList = &FreeListX;
LOCAL ListType	 ClipListX;
LOCAL ListType	*ClipList = &ClipListX;

/*********************************************************************
 *
 * PushClipValues
 *
 * PURPOSE
 *		save current clip values and use new ones.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL PushClipValues (int x, int y, int w, int h)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PushClipValues";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	
	{
		static int	 flag = FALSE;
		ClipNode		*cn;

		if (!flag) {
			flag = TRUE;
			InitList (ClipList);
			InitList (FreeList);
		}

		cn = RemTail (FreeList);
		if (!cn) {
			cn = CreateNode (sizeof (ClipNode), NULL);
			if (!cn) {
				SetGlobalErr (ERR_OUT_OF_MEMORY);
				GEprintf ("OOM: ClipNode");
				return FALSE;
			}
		}

		cn->X      = GetClipLeft ();
		cn->Y      = GetClipTop ();
		cn->Width  = GetClipWidth ();
		cn->Height = GetClipHeight ();

		SetClipDimensions (x,y,w,h);

		AddTail (ClipList, cn);
		return TRUE;
	}

} /* PushClipValues */


/*********************************************************************
 *
 * PopClipValues
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void PopClipValues (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PopClipValues";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	{
		ClipNode	*cn;

		cn = RemTail (ClipList);

		SetClipDimensions (cn->X,cn->Y,cn->Width,cn->Height);

		AddTail (FreeList, cn);
	}

} /* PopClipValues */

LOCAL ListType			 SaveAreaListX;
LOCAL ListType			*SaveAreaList = &SaveAreaListX;

/*********************************************************************
 *
 * PopSaveArea
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void PopSaveArea (void)
{
	ScreenArea	*ma;
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PopSaveArea";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	ma = RemTail (SaveAreaList);

	RestoreArea (ma->Area);
	DeleteNode (ma);

} /* PopSaveArea */


/*********************************************************************
 *
 * PushSaveArea
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
BOOL PushSaveArea (int x, int y, int w, int h)
{
	static int	 smflag = FALSE;
	ScreenArea		*ma;
	SaveAreaType	*sa;

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "PushSaveArea";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	if (!smflag) {
		InitList (SaveAreaList);
		smflag = TRUE;
	}

	ma = CreateNode (sizeof (ScreenArea), NULL);
	if (!ma) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: ScreenArea");
		return FALSE;
	}

	sa = SaveArea (x, y, w, h);
	if (!sa) {
		DeleteNode (ma);
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: MenuSaveArea");
		return FALSE;
	}

	AddTail (SaveAreaList, ma);
	ma->X      = x;
	ma->Y      = y;
	ma->Width  = w;
	ma->Height = h;
	ma->Area   = sa;

	return TRUE;

} /* PushSaveArea */

