/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * CPalette.C
 *
 * PROGRAMMER : Gregg A. Tavares and Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 03/05/90
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 * DESCRIPTION
 *		Routines for setting the CGA palette.
 *
 * HISTORY
 *
*/
#include <echidna/platform.h>
#include "switches.h"

#include <dos.h>
#include <time.h>
#include <stdlib.h>
#include <echidna/etypes.h>
#include <echidna/graphics.h>

/******************************** CONSTANTS *******************************/

#define MAX_CGA_COLORS	4

/* Default Echidna palette code: black background/border, cyan, red, white */
#define CGA_DEFAULT_PALETTE_CODE	2 

#define VIDEO_DISPLAY_DATA_AREA_SEGMENT	0x0040
#define CRT_MODE_SET_OFFSET	0x65
#define CRT_PALETTE_OFFSET		0x66

/* Bit positions of flags in Echidna CGA palette code */
#define E_INTENSITY_BIT				2
#define E_COLOR_BURST_ENABLE_BIT	1
#define E_ADD_BLUE_BIT				0
#define E_BACKGROUND_LOW_BIT		8

/* Masks for getting flags out of Echidna CGA palette code */
#define E_INTENSITY_MASK			(1 << E_INTENSITY_BIT)
#define E_COLOR_BURST_ENABLE_MASK	(1 << E_COLOR_BURST_ENABLE_BIT)
#define E_ADD_BLUE_MASK				(1 << E_ADD_BLUE_BIT)

/* Bit positions of flags in CGA Mode Control Register */
#define COLOR_BURST_ENABLE_BIT	2
#define VIDEO_ENABLE_BIT		3

/* Masks for flags in CGA Mode Control Register */
#define COLOR_BURST_ENABLE_MASK	(1 << COLOR_BURST_ENABLE_BIT)
#define VIDEO_ENABLE_MASK		(1 << VIDEO_ENABLE_BIT)

/* Bit positions of flags in CGA Color Select Register */
#define INTENSITY_BIT	4
#define ADD_BLUE_BIT	5

/* Masks for flags in CGA Color Select Register */
#define INTENSITY_MASK	(1 << INTENSITY_BIT)
#define ADD_BLUE_MASK	(1 << ADD_BLUE_BIT)


/************************** L O C A L   S T U F F *************************/
static union REGS inregs;
static union REGS outregs;
/***
 *struct WORDREGS {
 *	unsigned short	ax, bx, cx, dx, si, di, cflag, flags;
 *};
 *
 *struct BYTEREGS {
 *	unsigned char	al, ah, bl, bh, cl, ch, dl, dh;
 *};
 *
 *union	REGS	{
 *	struct	WORDREGS x;
 *	struct	BYTEREGS h;
 *};
*/

#if 0
/*********************************************************************
 *
 * CGA_Low_SetPalette
 *
 * SYNOPSIS
 *		void	CGA_Low_SetPalette (
 *			ColorType	*colortable
 *		)
 *
 * USAGE
 *
 *
 * PURPOSE
 *		Set palette with tume color map.
 *
 * INPUTS
 *		colortable:	48 BYTE array of 16 Colors (3 BYTES per color: red,
 *						green, blue). 
 *
 *
 * RESULTS
 *
 *
 * BUGS
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	CGA_Low_SetPalette (
	UWORD	color,
	UBYTE	on
)
{
	static UBYTE defcolors1[] = {
		 0, 63, 63,
		63,  0,  0,
		63, 63, 63,
	};

	static UBYTE defcolors2[] = {
		 0, 63, 0,
		63,  0, 0,
		63, 63, 0,
	};

	UBYTE		*rgb;
	UBYTE		 c;
	UBYTE		 d;
	UBYTE		cr;
	UBYTE far	*cs;

	cs  = MK_FP(0x0040,0x0000);

	cr  = cs[0x65];
	cr &= 0xF3;						/* remove bit 2 and 3	*/
	cr |= ((color & 0x02) << 1) |	/* add new bit 2 and 3	*/
		  ((on) ? 0x08 : 0x00);
	outportb (0x3D8, cr);
	cs[0x65] = cr;

	cr  = cs[0x66];
	cr &= 0xCF;
	cr |= ((color & 0x01) << 5) |
		  ((color & 0x04) << 2);
	outportb (0x3D9, cr);
	cs[0x66] = cr;
	
	inregs.h.ah = 0x0B;
	inregs.h.bh = 0x00;
	if (on) {
		inregs.h.bl  = (color >> 8);
		inregs.h.bl += ((color & 0x04) ? 16 : 0);
	} else {
		inregs.h.bl = 0;
	}
	int86 (0x10, &inregs, &outregs);

	if (on && ((color & 0x02) || !(color & 0x03))) {
		rgb = (color & 0x02) ? defcolors1 : defcolors2;
		d   = (color & 0x04) ? 0 : 1;
		for (c = 1; c < 4; c++) {
			inregs.h.ah = 0x10;
			inregs.h.al = 0x10;
			inregs.x.bx = c * 2 + ((d) ? 0 : 16);
			inregs.h.dh = (*rgb++) >> d;
			inregs.h.ch = (*rgb++) >> d;
			inregs.h.cl = (*rgb++) >> d;
			int86 (0x10, &inregs, &outregs);
		}
	}

} /* CGA_Low_SetPalette */
#else
/*********************************************************************
 *
 * CGA_Low_SetPalette
 *
 * SYNOPSIS
 *		void	CGA_Low_SetPalette (
 *			UWORD	palcode,
 *			UBYTE	on
 *		)
 *
 * PURPOSE
 *		palcode: Echidna code for CGA palette:
 *                  High Byte = Background Color (0..15).
 *                  Low Byte (bit 2 is high-intensity flag):
 *                      0 (000):    Green, Red, Yellow.
 *                      1 (001):    Cyan, Violet, White.
 *                      2 (010):    Cyan, Red, White.
 *                      4 (100):    High-Intensity Green, Red, Yellow.
 *                      5 (101):    High-Intensity Cyan, Violet, White.
 *                      6 (110):    High-Intensity Cyan, Red, White.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void	CGA_Low_SetPalette (
	UWORD	palcode,
	UBYTE	on
)
{
	/* RGB values for Cyan-Red-White palette */
	static UBYTE defcolors1[] = {
		 0, 63, 63,
		63,  0,  0,
		63, 63, 63,
	};

	/* RGB values for Green-Red-Yellow palette */
	static UBYTE defcolors2[] = {
		 0, 63, 0,
		63,  0, 0,
		63, 63, 0,
	};

	UBYTE		*rgb;
	UBYTE		 c;
	UBYTE		 d;
	UBYTE		cr;
	UBYTE far	*cs;	/* Pointer to Bios Video Display Data Area */

	/* Get ptr to base address of BIOS Video Display Data Area */
	cs  = MK_FP(VIDEO_DISPLAY_DATA_AREA_SEGMENT,0x0000);

	/* Get value of CRT Mode Control Register */
	cr  = cs[CRT_MODE_SET_OFFSET];

	/* Clear Color Burst Enable bit and Video Enable bit */
	cr &= ~(COLOR_BURST_ENABLE_MASK | VIDEO_ENABLE_MASK);

	/* Or in new Color Burst Enable bit from palette code.
	 * A 1 here selects cyan, red, white palette. A 0 here allows
	 * bit 5 of Color Select Register to choose between the green,
	 * red, yellow palette (0) or the cyan, violet, white palette (1)
	 */
	cr |= ((palcode & E_COLOR_BURST_ENABLE_MASK) << 
			(COLOR_BURST_ENABLE_BIT - E_COLOR_BURST_ENABLE_BIT));

	/* Enable or disable video depending on value of 'on'*/
	cr |= ((on) ? VIDEO_ENABLE_MASK : 0x00);

	/* Write new value to Mode Control Register */
	outportb (0x3D8, cr);
	/* Write new value to BIOS shadow register */
	cs[CRT_MODE_SET_OFFSET] = cr;

	/* Get value of Color Select Register */
	cr  = cs[CRT_PALETTE_OFFSET];

	/* Clear bit 5 (add blue flag) and Intensity bit */
	cr &= ~(ADD_BLUE_MASK | INTENSITY_MASK);

	/* Or in new Add-Blue bit. A 1 here selects the cyan, violet, white 
	 * palette.  A 0 here selects the green, red, yellow palette.
	 */
	cr |= ((palcode & E_ADD_BLUE_MASK) << 
		(ADD_BLUE_BIT - E_ADD_BLUE_BIT));
	
	/* Or in new Intensity bit. A 1 here selects high intensity.
	 * A 0 here selects low intensity.
	 */
	cr |= (palcode & E_INTENSITY_MASK) << 
		(INTENSITY_BIT - E_INTENSITY_BIT);

	/* Write new value to Color Select Register */
	outportb (0x3D9, cr);
	/* Write new value to BIOS shadow register */
	cs[CRT_PALETTE_OFFSET] = cr;

	/* The background is getting set here and since the bios routine being
	 * called also sets the intensity, once again the intensity must be 
	 * put in. Why didn't Gregg set these values when he did the out to
	 * the color select register above?  Maybe this also sets the 
	 * background in EGA & VGA cards that are emulating CGA? Yeah that's it.
	 */
	inregs.h.ah = 0x0B;
	inregs.h.bh = 0x00;
	if (on) {
		inregs.h.bl  = (palcode >> E_BACKGROUND_LOW_BIT);
		inregs.h.bl += ((palcode & E_INTENSITY_MASK) ? INTENSITY_MASK : 0);
	} else {
		inregs.h.bl = 0;
	}
	int86 (0x10, &inregs, &outregs);

	/* Set the DAC color registers for VGA cards or Attribute Controller
	 * palette registers for EGA cards emulating CGA 
	 */
	/* If 'on' and palette is Cyan-Red-White or Green-Red-Yellow */
	if (on && ((palcode & E_COLOR_BURST_ENABLE_MASK) || 
		!(palcode & E_ADD_BLUE_MASK))) 
	{
		rgb = (palcode & E_COLOR_BURST_ENABLE_MASK) ? defcolors1 : defcolors2;
		/* d = # of times to >> the color values to reduce intensity:
		 *  0 for high-intensity, 1 for low-intensity. 
		 */
		d   = (palcode & E_INTENSITY_MASK) ? 0 : 1;
		for (c = 1; c < 4; c++) {
			inregs.h.ah = 0x10;
			inregs.h.al = 0x10;
			inregs.x.bx = c * 2 + ((d) ? 0 : 16);
			inregs.h.dh = (*rgb++) >> d;
			inregs.h.ch = (*rgb++) >> d;
			inregs.h.cl = (*rgb++) >> d;
			int86 (0x10, &inregs, &outregs);
		}
	}

} /* CGA_Low_SetPalette */

#endif

/*********************************************************************
 *
 * CGA_ClosePainter
 *
 * SYNOPSIS
 *		void CGA_ClosePainter (void) {
 *
 * PURPOSE
 *		To clean up stuff done in CGA_OpenPainter()
 *
 * INPUT
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_ClosePainter (void) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_ClosePainter";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{

}
} /* CGA_ClosePainter */

/******************************* P R A G M A ******************************/
/* Get rid of warning: Paremeter 'X' is not used in function. */
#pragma warn -par  
/******************************* P R A G M A ******************************/

/*********************************************************************
 *
 * CGA_OpenPainter
 *
 * SYNOPSIS
 *		short CGA_OpenPainter (	
 *			short max_colors, 
 *			short max_cycle_colors, 
 *			short hertz
 *		)
 *
 * PURPOSE
 *		To initialize and allocate structures for controlling palette.
 *
 * INPUT
 *		max_colors: # of colors in largest palette or cycle range used.
 *		max_cycle_colors: Total # of colors in largest cycle set used.
 *				For example if you use only one cycle set, and it has
 *				a cycle range with four colors and another with 3 colors,
 *				then mac_cycle_colors == (4 + 3) == 7.
 * 		hertz:	Maximum number of times per second that any color cycle
 *				will change. Or if no cycle colors are used then the rate
 *				at which you would like the the palette handling interrupt
 *				routine to check for palette change requests.
 *				Must be <= 60.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *		TRUE on success.
 *		FALSE on failure.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short CGA_OpenPainter (	
	short max_colors, 
	short max_cycle_colors, 
	short hertz
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_OpenPainter";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{

	PaletteMaxColors = MAX_CGA_COLORS;
	PaletteIsFading = CycleSetIsFading = FALSE;

	/* Set Palette to Black */
	
	CGA_Low_SetPalette (CGA_DEFAULT_PALETTE_CODE, 0);
	return TRUE;
}
} /* CGA_OpenPainter */

/*********************************************************************
 *
 * CGA_UsePaletteN
 *
 * SYNOPSIS
 *		void	CGA_UsePaletteN  (
 *			PaletteType	*palette,
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors
 *		) 
 *
 * PURPOSE
 *		To set a range of colors in the hardware palette from the
 *		corresponding range in a user palette, or to black if 'palette'
 *		is NULL.
 *
 * INPUT
 *		palette:	User palette to get colors from, or NULL if you want
 *					to set colors to black.
 *		src_offset:	First color of range in palette to copy from.
 *		dst_offset:	First color of range in palette to write to.
 *		numcolors:	Number of colors to set.
 *
 * AFFECTS
 *		Sets up flags to tell interrupt driven Palette updating routine to
 *		copy the color range to the Private (shadow) palette then to hardware
 *		palette. Waits until palette is set before returning.
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_UsePaletteN  (
	PaletteType	*palette,
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors
) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_UsePaletteN";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	if (palette) {
		CGA_Low_SetPalette (palette->Data.CGA, TRUE);
	} else {
		CGA_Low_SetPalette (CGA_DEFAULT_PALETTE_CODE, FALSE);
	}
}
} /* CGA_UsePaletteN */

/*********************************************************************
 *
 * CGA_UseColorCycleSet
 *
 * SYNOPSIS
 *		void CGA_UseColorCycleSet (
 *			ColorCycleSetType	*color_cycle_set,
 *			short				 activate
 *		)
 *
 * PURPOSE
 *		To install a new cycle set and optionally activate it.
 *
 * INPUT
 *		color_cycle_set:	User cycle set to install. Must be non-NULL.
 *		activate:			TRUE if cycling should be activated.
 *
 * AFFECTS
 *		Puts in request to change the cycle-set and waits til request is
 *		handled before returning.
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_UseColorCycleSet (
	ColorCycleSetType	*color_cycle_set,
	short				 activate
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_UseColorCycleSet";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{

}
} /* CGA_UseColorCycleSet */

/*********************************************************************
 *
 * CGA_UsePaintBoxN
 *
 * SYNOPSIS
 *		void CGA_UsePaintBoxN (
 *			PaintBoxType	*paintbox,
 *			short			 src_offset,
 *			short			 dst_offset,
 *			short		 	 numcolors,
 *			short			 activate_color_cycling
 *		)
 *
 * PURPOSE
 *		To install new palette, and install new color cycling set and
 *		optionally activate it.
 *
 * INPUT
 *		paintbox:	ptr to PaintBoxType struct. CycleSet in paintbox must
 *					be non-NULL.  Palette in paintbox may be NULL, and if
 *					so the color range specified will be set to black.
 *					paintbox may be NULL, and if so the color range will
 *					be set to black and nothing will happen to the CycleSet.
 *		src_offset:	First color of range in palette to copy from.
 *		dst_offset:	First color of range in palette to write to.
 *		numcolors:	Number of colors in palette range.
 *		activate_color_cyclinge:	non-zero if you want color cycling
 *					to be activated after cycle set is installed.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_UsePaintBoxN (
	PaintBoxType	*paintbox,
	short			 src_offset,
	short			 dst_offset,
	short			 numcolors,
	short			 activate_color_cycling
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "UsePaintBox";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{

	if (paintbox && paintbox->Palette) {
		CGA_Low_SetPalette (paintbox->Palette->Data.CGA, TRUE);
	} else {
		CGA_Low_SetPalette (CGA_DEFAULT_PALETTE_CODE, FALSE);
	}
}

} /* CGA_UsePaintBoxN */

/*********************************************************************
 *
 * CGA_FadeToPaletteN
 *
 * SYNOPSIS
 *		void CGA_FadeToPaletteN (
 *			PaletteType *palette, 
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors,
 *			short		 hz60frames
 *		) 
 *
 * PURPOSE
 *		To fade from the current palette colors to those in 'palette'.
 *	
 * INPUT
 *		palette:	User palette to fade to, or NULL if you want
 *					to fade to black.
 *		src_offset:	First color of range in palette to fade from.
 *		dst_offset:	First color of range in palette to fade to.
 *		numcolors:	Number of colors to fade.
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_FadeToPaletteN (
	PaletteType *palette, 
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors,
	short		 hz60frames
) 
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_FadeToPaletteN";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{

	if (palette) {
		CGA_Low_SetPalette (palette->Data.CGA, TRUE);
	} else {
		CGA_Low_SetPalette (CGA_DEFAULT_PALETTE_CODE, FALSE);
	}

}
} /* CGA_FadeToPaletteN */

/*********************************************************************
 *
 * CGA_FadeToColorCycleSet
 *
 * SYNOPSIS
 *		void CGA_FadeToColorCycleSet (
 *			ColorCycleSetType	*color_cycle_set,
 *			short				 hz60frames
 *		)
 *
 * PURPOSE
 *		To fade from the current colors in the current CycleSet to 
 *		those in 'color_cycle_set'.		
 *
 * INPUT
 *		color_cycle_set: ColorCycleSet to fade to, or NULL if you
 *				want to fade to black.
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_FadeToColorCycleSet (
	ColorCycleSetType	*color_cycle_set,
	short				 hz60frames
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_FadeToColorCycleSet";
#endif
#if REQUIRE
#endif
/*-------------------------------------------------------------------------*/
{

}
} /* CGA_FadeToColorCycleSet */

/*********************************************************************
 *
 * CGA_FadeToPaintBoxN
 *
 * SYNOPSIS
 *		void CGA_FadeToPaintBoxN (
 *			PaintBoxType *paintbox,
 *			short		 src_offset,
 *			short		 dst_offset,
 *			short		 numcolors,
 *			short		 hz60frames
 *		)
 *
 * PURPOSE
 *		To fade from the colors in the current paintbox to those in
 *		'paintbox'.
 *
 * INPUT
 *		paintbox:	Pointer to PaintBoxType.  CycleSet in paintbox may
 *			be NULL, in which case the cycle colors will fade to black.
 *			Palette in paintbox may be null, in which case the palette
 *			colors in specified range will fade to black. paintbox may
 *			be NULL, in which case the cycle colors and color range will
 *			fade to black.
 *		src_offset:	First color of range in palette to fade from.
 *		dst_offset:	First color of range in palette to fade to.
 *		numcolors:	Number of colors in palette range.
 *		hz60frames:	Duration of fade in 1/60 second frames.
 *
 *
 * AFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CGA_FadeToPaintBoxN (
	PaintBoxType *paintbox,
	short		 src_offset,
	short		 dst_offset,
	short		 numcolors,
	short		 hz60frames
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "CGA_FadeToPaintBoxN";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{

	if (paintbox && paintbox->Palette) {
		CGA_Low_SetPalette (paintbox->Palette->Data.CGA, TRUE);
	} else {
		CGA_Low_SetPalette (CGA_DEFAULT_PALETTE_CODE, FALSE);
	}
}
} /* CGA_FadeToPaintBoxN */

