/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * EDRAW.c
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 11/05/89
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 * DESCRIPTION
 *		Routines to follow draw commands and show a dialog box.
 *
 * HISTORY
 *		06/01/91 Saturday (RGM) - SAS/Lattice Amiga additions.
 *		10/13/93 Wednesday (dcc) - updated for MS-DOS Watcom C32.
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#define __ARCADE__ 1

#if __MSDOS__
#include <stdio.h>
#else
#include <echidna/etypes.h>
#include <graphics/rastport.h>
#include <graphics/text.h>
#if LATTICE
#include <proto/graphics.h>
#endif
#endif

#include <echidna/graphics.h>
#include <echidna/edraw.h>

#if __TURBOC__
#pragma warn -par
#endif

#define LOCAL

typedef EPic *(*epicfuncptr)(struct RastPort *rp, EPic *epic, short xpos, short ypos);

extern void DrawEPic (struct RastPort *rp, EPic *epic, short xpos, short ypos);

LOCAL struct TextFont	***EFonts;
LOCAL char				***EStrings;
LOCAL EPic				**EEPics;
LOCAL GShape			***EEShapes;
LOCAL GShape			**EShapes;
LOCAL short				 EXPos;
LOCAL short				 EYPos;
LOCAL UBYTE				*EEColors;

/**************************************************************************
 *
 * EDFn_SetFgColor
 *
 * SYNOPSIS
 *   EPic *EDFn_SetFgColor (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Foreground Pen Color.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_SetFgColor (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
#if __WATCOMC__
	rp = rp;	xpos = xpos;	ypos = ypos;
#endif
#if __ARCADE__
	SetPenColor (*epic++);
#else
	SetAPen (rp, *epic++);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_SetFgColorTbl
 *
 * SYNOPSIS
 *   EPic *EDFn_SetFgColorTbl (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Foreground Pen Color.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_SetFgColorTbl (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	UBYTE color;

#if __WATCOMC__
	rp = rp;	xpos = xpos;	ypos = ypos;
#endif
	color = EEColors[*epic++];
#if __ARCADE__
	SetPenColor (color);
#else
	SetAPen (rp, color);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_SetBgColorTbl
 *
 * SYNOPSIS
 *   EPic *EDFn_SetBgColorTbl (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Foreground Pen Color.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_SetBgColorTbl (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	UBYTE color;

#if __WATCOMC__
	rp = rp;	xpos = xpos;	ypos = ypos;
#endif
	color = EEColors[*epic++];
#if __ARCADE__
	SetPaperColor (color);
#else
	SetBPen (rp, color);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_SetBgColor
 *
 * SYNOPSIS
 *   EPic *EDFn_SetBgColor (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Background Pen Color.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_SetBgColor (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
#if __WATCOMC__
	rp = rp;	xpos = xpos;	ypos = ypos;
#endif
#if __ARCADE__
	SetPaperColor (*epic++);
#else
	SetBPen (rp, *epic++);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_SetDrawColor
 *
 * SYNOPSIS
 *   EPic *EDFn_SetDrawColor (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Drawing Mode.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_SetDrawMode (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	SetDrMd (rp, *epic++);
	return epic;
}

/**************************************************************************
 *
 * EDFn_Plot
 *
 * SYNOPSIS
 *   EPic *EDFn_Plot (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Plot a dot.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Plot (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		x;
	short		y;

	x = xpos + *epic++;
	y = ypos + *epic++;
#if __ARCADE__
	DrawPixel (x, y);
	rp->cp_x = x;
	rp->cp_y = y;
#else
	WritePixel (rp, x, y);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_Draw
 *
 * SYNOPSIS
 *   EPic *EDFn_Draw (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Draw a line from last position to new one.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Draw (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		x;
	short		y;

	x = xpos + *epic++;
	y = ypos + *epic++;
#if __ARCADE__
	rp->cp_x = x;
	rp->cp_y = y;
#else
	Draw (rp, x, y);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_Move
 *
 * SYNOPSIS
 *   EPic *EDFn_Move (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Move "Drawing Cursor" to a new position.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Move (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		 x;
	short		 y;

	x = xpos + *epic++;
	y = ypos + *epic++;
#if __ARCADE__
	rp->cp_x = x;
	rp->cp_y = y;
#else
	Move (rp, x, y);
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_Rect
 *
 * SYNOPSIS
 *   EPic *EDFn_Rect (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Draw a rectangle.
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Rect (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		x1, y1;
	short		w , h;

	x1 = xpos + *epic++;
	y1 = ypos + *epic++;
	w  = *epic++;
	h  = *epic++;

	if (w < 0) {
		x1 = x1 + w + 1;
		w  = -w;
	}

	if (h < 0) {
		y1 = y1 + h + 1;
		h  = -h;
	}

#if __ARCADE__
	if (rp->DrawMode == JAM2) {
		DrawRect (x1, y1, w, h);
	} else {
		DrawBox (x1, y1, w, h);
	}
	rp->cp_x = x1;
	rp->cp_y = y1;
#else
	{
		short		x2, y2;

		x2 = x1 + w - 1;
		y2 = y1 + h - 1;
		if (rp->DrawMode & JAM2) {
			RectFill (rp, x1, y1, x2, y2);
		} else {
			Move (rp, x1, y1);
			Draw (rp, x2, y1);
			Draw (rp, x2, y2);
			Draw (rp, x1, y2);
			Draw (rp, x1, y1);
		}
	}
#endif
	return epic;
}

/**************************************************************************
 *
 * EDFn_Font
 *
 * SYNOPSIS
 *   EPic *EDFn_Font (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set font
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Font (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	struct TextFont	**tf;

#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	tf = EFonts[*epic++];
	SetFont (rp, *tf);

	return epic;
}

/**************************************************************************
 *
 * EDFn_Text
 *
 * SYNOPSIS
 *   EPic *EDFn_Text (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Text
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Text (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	char **tx;

#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	tx = EStrings[*epic++];

#if __ARCADE__
	rp->cp_x = DrawString ((Font *)rp->Font, rp->cp_x, rp->cp_y, *tx);
#else
	Move (rp, rp->cp_x, rp->cp_y + rp->TxBaseline);
	Text (rp, *tx, strlen (*tx));
	Move (rp, rp->cp_x, rp->cp_y - rp->TxBaseline);
#endif

	return epic;
}

/**************************************************************************
 *
 * EDFn_Gosub
 *
 * SYNOPSIS
 *   EPic *EDFn_Gosub (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Text
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_Gosub (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	EPic		*newep;

#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	newep = EEPics[*epic++];

	DrawEPic (rp, newep, rp->cp_x, rp->cp_y);

	return epic;
}

/**************************************************************************
 *
 * EDFn_ForX
 *
 * SYNOPSIS
 *   EPic *EDFn_ForX (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Text
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_ForX (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		 i;
	short		 istep;
	short		 ilimit;
	short		 xstep;
	EPic	*ep;

	short		 x;
	short		 y;

#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	i      = *epic++;
	ilimit = *epic++;
	istep  = *epic++;
	xstep  = *epic++;
	ep     = EEPics[*epic++];

	x = rp->cp_x;
	y = rp->cp_y;

	if (istep > 0) {
		for (; i <= ilimit; i += istep, x += xstep) {
			DrawEPic (rp, ep, x, y);
		}
	} else {
		for (; i >= ilimit; i += istep, x += xstep) {
			DrawEPic (rp, ep, x, y);
		}
	}

	return epic;
}

/**************************************************************************
 *
 * EDFn_ForY
 *
 * SYNOPSIS
 *   EPic *EDFn_ForY (struct RastPort *rp, EPic *epic);
 *
 * FUNCTION
 *   Set Text
 *
 * INPUTS
 *
 *
*/
LOCAL EPic *EDFn_ForY (
	struct RastPort	*rp,
	EPic			*epic,
	short			 xpos,
	short			 ypos
)
{
	short		 i;
	short		 istep;
	short		 ilimit;
	short		 ystep;
	EPic	*ep;

	short		 x;
	short		 y;

#if __WATCOMC__
	xpos = xpos;	ypos = ypos;
#endif
	i      = *epic++;
	ilimit = *epic++;
	istep  = *epic++;
	ystep  = *epic++;
	ep     = EEPics[*epic++];

	x = rp->cp_x;
	y = rp->cp_y;

	if (istep > 0) {
		for (; i <= ilimit; i += istep, y += ystep) {
			DrawEPic (rp, ep, x, y);
		}
	} else {
		for (; i >= ilimit; i += istep, y += ystep) {
			DrawEPic (rp, ep, x, y);
		}
	}

	return epic;
}

/*********************************************************************
 *
 * EDFn_DrawGShape
 *
 * SYNOPSIS
 *		EPic *EDFn_DrawGShape (rp, epic, x, y)
 *
 * PURPOSE
 *		Draw a shape
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
EPic *EDFn_DrawGShape (
	struct RastPort	*rp,
	EPic			*epic,
	short			 x,
	short			 y
)
{
#if __WATCOMC__
	x = x;	y = y;
#endif
#if __ARCADE__
	DrawGShape (EShapes[*epic++], rp->cp_x, rp->cp_y);
#else
	epic++;
#endif
	return epic;

} /* EDFn_DrawGShape */

/*********************************************************************
 *
 * EDFn_SetGShapeTable
 *
 * SYNOPSIS
 *		EPic *EDFn_SetGShapeTable (rp, epic, x, y)
 *
 * PURPOSE
 *		Set Shape table
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
EPic *EDFn_SetGShapeTable (
	struct RastPort	*rp,
	EPic			*epic,
	short			 x,
	short			 y
)
{
#if __WATCOMC__
	rp = rp;	x = x;	y = y;
#endif
	EShapes = EEShapes[*epic++];

	return epic;

} /* EDFn_SetGShapeTable */

/*********************************************************************
 *
 * EDFn_MoveRel
 *
 * SYNOPSIS
 *		EPic *EDFn_MoveRel (rp, epic, x, y)
 *
 * PURPOSE
 *		Move relative?
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
EPic *EDFn_MoveRel (
	struct RastPort	*rp,
	EPic			*epic,
	short			 x,
	short			 y
)
{
#if __WATCOMC__
	x = x;	y = y;
#endif
	rp->cp_x += *epic++;
	rp->cp_y += *epic++;

	return epic;

} /* EDFn_MoveRel */

LOCAL epicfuncptr EDrawFuncs[] = {
	NULL,				/*   0 */
	EDFn_SetFgColor,
	EDFn_SetBgColor,
	EDFn_SetDrawMode,
	EDFn_Plot,
	EDFn_Draw,			/*   5 */
	EDFn_Move,
	EDFn_Rect,
	EDFn_Font,
	EDFn_Text,
	EDFn_Gosub,			/*  10 */
	EDFn_ForX,
	EDFn_ForY,
	EDFn_DrawGShape,
	EDFn_SetGShapeTable,
	EDFn_MoveRel,		/*  15 */
	EDFn_SetFgColorTbl,
	EDFn_SetBgColorTbl,
};

/**************************************************************************
 *
 * DrawEPic
 *
 * SYNOPSIS
 *   void DrawEPic (struct RastPort *rp, EPic *epic, short x, short y);
 *
 * FUNCTION
 *   Draw a picture from opcodes.
 *
 * INPUTS
 *
 *
*/
void DrawEPic (
	struct RastPort	*rp,
	EPic		*epic,
	short		 x,
	short		 y
)
{
	short		 command;

	rp->cp_x = x;
	rp->cp_y = y;
	while ((command = *epic++) != 0) {
		epic = (*EDrawFuncs[command])(rp, epic, x, y);
	}
}

/**************************************************************************
 *
 * StartEPic
 *
 * SYNOPSIS
 *   void StartEPic (
 *		struct RastPort	*rp,
 *
 * FUNCTION
 *
 *
 * INPUTS
 *
 *
*/

void StartEPic (
	struct RastPort	*rp,
	EPic			*epic,
	short			 x,
	short			 y,
	struct TextFont ***fonts,
	char			***strings,
	EPic			**epics,
	GShape			***shapes
)
{
	EFonts	 = fonts;
	EStrings = strings;
	EEPics	 = epics;
	EEShapes = shapes;

	BeforeGraphics ();
	DrawEPic (rp, epic, x, y);
	AfterGraphics ();
}
 

extern void StartEPicture (
	struct RastPort	*rp,
	EPicture		*ep,
	short			 x,
	short			 y
) {
	ep->rp = rp;

	EEColors = ep->Colors;

	StartEPic (rp, ep->EPic, x, y, ep->Fonts, ep->Strings, ep->EPics, ep->Shapes);
}

#if __TURBOC__
#pragma warn .par
#endif

