/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ESUP.c
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 03/05/90
 *   MODIFIED : 10/08/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		C routine support for EGA mode
 *
 * HISTORY
 *		10/08/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <alloc.h>
#include <mem.h>
#include <dos.h>
#include <echidna/alloc.h>
#include <echidna/eerrors.h>
#include <echidna/graphics.h>
#include <echidna/pixlscan.h>
#include <echidna/memory.h>

/***************************** E X T E R N A L ****************************/
extern far EGA_ASMOpenGraphics (void);
extern far EGA_ASMCloseGraphics (void);
extern far EGA_DrawRectAsm (short x, short y, short width, short height);
extern GfxRoutinePtr EGARoutineTable[];
extern GfxDataTbl	 EGADataTable;

/**************************** C O N S T A N T S ***************************/


/******************************* L O C A L S ******************************/
static UWORD	EGALoadBuffer = NULL;

/************************************  ************************************/

/*********************************************************************
 *
 * EGA_GetGShapeLoadBuf
 *
 * SYNOPSIS
 *		void far *EGA_GetGShapeLoadBuf (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far *EGA_GetGShapeLoadBuf (void)
{
	return MK_FP(EGALoadBuffer, 0x0000);

} /* EGA_GetGShapeLoadBuf */

/*********************************************************************
 *
 * EGA_DrawRect
 *
 * SYNOPSIS
 *		EGA_DrawRect (short x, short y, short width, short height)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void EGA_DrawRect (short x, short y, short inwidth, short inheight)
{
	short width  = inwidth;
	short height = inheight;

	if (x < 0) {
		width = inwidth - (0 - x);
		x     = 0;
	} else if (x > 0 + 320 - inwidth) {
		width = (0 + 320) - x;
	}

	if (y < 0) {
		height = inheight - (0 - y);
		y      = 0;
	} else if (y > 0 + 200 - inheight) {
		height = (0 + 200) - y;
	}

	if (width > 0 && height > 0) {
		EGA_DrawRectAsm (x, y, width, height);
	}
} /* EGA_DrawRect */

/*********************************************************************
 *
 * GetEGARoutineTable
 *
 * SYNOPSIS
 *		GfxRoutinePtr *GetEGARoutineTable (void)
 *
 * PURPOSE
 *		Return pointer to EGARoutineTable.  Calling this routine
 *		should force loading of the EGA routine overlay.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void GetEGARoutineTable (
	GfxRoutinePtr	**grptr,
	GfxDataTbl		**gdptr
) {
	*grptr = EGARoutineTable;
	*gdptr = &EGADataTable;
} /* GetEGARoutineTable */


/*********************************************************************
 *
 * EGA_FreeShape
 *
 * SYNOPSIS
 *		void	EGA_FreeShape (
 *			GShape	*shapeptr
 *		)
 *
 * PURPOSE
 *		To Free the memory allocated for shape by To_MCG_Shape.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void EGA_FreeShape(GShape *shape)
{
	MEM_free (shape);

} /* EGA_FreeShape */

typedef UBYTE *parray[4];

static UBYTE	 cmasks[] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01, };
static UBYTE	*ps_ColorMap;
static UWORD	 ps_Size;
static short	 ps_DWidth;
static UBYTE	*ps_EPatternMap;
static UBYTE	*ps_OPatternMap;
static parray	*ps_PatMap;
static short	 ps_DeltaCol;
static short	 ps_DeltaRow;
static UBYTE	*ps_PixlPtr;
static short	 ps_ColCount;
static short	 ps_TrueWidth;
static short	 ps_TrueWidthHold;
static UBYTE	 ps_tskip;
static UBYTE	 ps_tskiphold;
static UBYTE	 ps_tpad;
static UBYTE	 ps_tpadhold;
static UBYTE	 ps_odd;
static UBYTE	 pixel;
static UBYTE	 cdata[4];
static UBYTE	 cmask;
static UBYTE 	 tcolor;
static UBYTE	 color;
static parray	 ps_EOPatMap = {NULL, };
static parray	 ps_OOPatMap = {NULL, };

static void getpixl (void)
{
	UBYTE	b;

	/* Get two pixels and their mask */
	cdata[0] = 0;
	cdata[1] = 0;
	cdata[2] = 0;
	cdata[3] = 0;
	cmask    = 0;

	if (ps_Size) {
		for (pixel = 0; pixel < 8; pixel++) {
			if (!ps_TrueWidth) {
				ps_TrueWidth = ps_TrueWidthHold;
				ps_tskip     = ps_tskiphold;
				ps_tpad      = ps_tpadhold;
				ps_ColCount  = 0;
				ps_PixlPtr  += ps_DeltaRow;
				ps_odd = !ps_odd;
				if (ps_odd) {
					ps_PatMap = &ps_EOPatMap;
				} else {
					ps_PatMap = &ps_OOPatMap;
				}
				return;
			}
			ps_TrueWidth--;
			if (ps_tskip) {
				ps_tskip--;
				color = tcolor;
			} else if (ps_ColCount < ps_DWidth) {
				ps_Size--;
				ps_ColCount++;
				color = ps_ColorMap[*ps_PixlPtr];
				ps_PixlPtr += ps_DeltaCol;
			} else if (ps_tpad) {
				ps_tpad--;
				color = tcolor;
			}

			for (b = 0; b < 4; b++) {
				cdata[b] |= (*ps_PatMap)[b][color] & ((color == tcolor) ? 0 : cmasks[pixel]);
			}
			cmask |= ((color == tcolor) ? 0 : cmasks[pixel]);
		}
	}
}

/*********************************************************************
 *
 * EGA_ConvertShape
 *
 * SYNOPSIS
 *		GShape	*EGA_ConvertShape (
 *			GShape			*gshape, 
 *			short				 Hflip,
 *			short				 Vflip,
 *			UBYTE			 pcc,
 *			GShape			*outshape
 *		)
 *
 * PURPOSE
 *		To Create a shape usable by the EGA shape routines.
 *
 * INPUT
 *		gshape:		Raw unpacked MCGA shape.
 *		attrib:		Value from BrushPak that tells me if I need to flip
 *					and/or remap the shape.
 *
 * EFFECTS
 *		Makes Converted copy of 'gshape' in 'outshape' and then allocates 
 *		space for and copies 'outshape'
 *
 * RETURN VALUE
 *		Returns pointer to allocated EGA shape if succesful.
 *		Returns NULL if unable to allocate space.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
GShape	*EGA_ConvertShape (
	ConvertSType	*cs
)
{
	GShape		*gshape;
	short		 Hflip;
	short		 Vflip;
	short		 newxoff;
	GShape		*outshape;
	UBYTE 		*outbuf;
	UWORD 		 origsize;
	UWORD 		 packedsize;
	short		 ps_2xDWidth;
	UWORD		 s;
	UWORD		 b;

	if (!cs) {
		if (ps_EOPatMap[0])	DestroyTempBuffer (ps_EOPatMap[0], 2048);
		ps_EOPatMap[0] = NULL;
		return NULL;
	}

	if (!ps_EOPatMap[0]) {
		ps_EOPatMap[0] = CreateTempBuffer (2048);
		if (!ps_EOPatMap[0]) {
			SetGlobalErr (ERR_OUT_OF_MEMORY);
			GEprintf ("EGA_ConvertShape:OOM allocating odd pattern maps");
			return NULL;
		}
		for (s = 0; s < 4; s++) {
			ps_EOPatMap[s] = ps_EOPatMap[0] + 256 * s;
			ps_OOPatMap[s] = ps_EOPatMap[0] + 1024 + 256 * s;
		}
	}

	gshape 			= cs->ShapeData; 
	ps_ColorMap		= cs->ColorMap;
	ps_EPatternMap	= cs->PatternMap->Even;
	ps_OPatternMap	= cs->PatternMap->Odd;
	Hflip			= cs->HFlip;
	Vflip			= cs->VFlip;

	/* if new pattern create odd pattern set */
	if (cs->ExpandPatterns) {
		UBYTE	*oop;
		UBYTE	*oep;
		UBYTE	 bemask;
		UBYTE	 bomask;

		oep = ps_EPatternMap;
		oop = ps_OPatternMap;
		
		for (s = 0; s < 256; s++) {
			bemask = 0x10;
			bomask = 0x01;
			for (b = 0; b < 4; b++) {
				ps_EOPatMap[b][s] = ((*oep & bemask) ? 0xAA : 0x00) |
									((*oep & bomask) ? 0x55 : 0x00);
				ps_OOPatMap[b][s] = ((*oop & bemask) ? 0xAA : 0x00) |
									((*oop & bomask) ? 0x55 : 0x00);
				bemask <<= 1;
				bomask <<= 1;
			}
			oep++;
			oop++;
		}
	}

	ResetPixlScan (gshape, Hflip, Vflip);

	if (Hflip) {
		newxoff = gshape->Width - ps_DWidth - gshape->Data.MCGA.XOffset;
	} else {
		newxoff = gshape->Data.MCGA.XOffset;
	}

	origsize = ((((gshape->Data.MCGA.DWidth + newxoff - 1) >> 3) -
				(newxoff >> 3) + 1) + 1) *
				gshape->Data.MCGA.DHeight * 5 + EGASDATASIZE + GSHAPESIZE + 5;
	/* Allocate Space for Worst Case Packing */
	if ( (outshape = (GShape *)MEM_calloc (1, origsize)) == NULL)
	{
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: in EGA_ConvertShape");
		return NULL;
	}

	/* Copy Shape Header info to outshape */
	memcpy (outshape, gshape, GSHAPESIZE);
	tcolor = outshape->TColor;
#if !USE_SHAPE_TPAT
	outshape->TColor = ps_EPatternMap[tcolor] + (ps_OPatternMap[tcolor] << 8);
#else
	outshape->Data.EGA.TPatEven = ps_EPatternMap[tcolor];
	outshape->Data.EGA.TPatOdd = ps_OPatternMap[tcolor];
#endif

	/* Pack Shape */
	outbuf = (UBYTE *)outshape + GSHAPESIZE + EGASDATASIZE;
	packedsize = 0;

	outshape->Data.EGA.XOffset = newxoff & 0xFFF8;

	if (Vflip) {
		outshape->Data.EGA.YOffset = gshape->Height - 
			gshape->Data.MCGA.DHeight - gshape->Data.MCGA.YOffset;
	} else {
		outshape->Data.EGA.YOffset = gshape->Data.MCGA.YOffset;
	}

	outshape->Data.EGA.DWidth  = (((ps_DWidth + (newxoff & 0x07)) + 7) >> 3) + 1;
	outshape->Data.EGA.DHeight = gshape->Data.MCGA.DHeight;

	ps_tskiphold     = newxoff & 0x07;
	ps_tpadhold      = 7 - ((newxoff + ps_DWidth - 1) & 0x07);
	ps_TrueWidthHold = ps_tskiphold + ps_DWidth + ps_tpadhold;

	ps_TrueWidth	 = ps_TrueWidthHold;
	ps_tskip     	 = ps_tskiphold;
	ps_tpad      	 = ps_tpadhold;

	ps_odd = outshape->Data.EGA.YOffset & 0x01;
	if (ps_odd) {
		ps_PatMap = &ps_EOPatMap;
	} else {
		ps_PatMap = &ps_OOPatMap;
	}

	if (ps_Size) {
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		packedsize += 5;
		while (ps_Size || ps_tpad) {
			getpixl ();
			*outbuf++   = cdata[0];
			*outbuf++   = cdata[1];
			*outbuf++   = cdata[2];
			*outbuf++   = cdata[3];
			*outbuf++   = cmask;
			packedsize += 5;
		}
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		*outbuf++   = 0;
		packedsize += 5;
	}
	return outshape;

EGAcleanup:
	EGA_FreeShape (outshape);
	return NULL;
} /* EGA_ConvertShape */

/*********************************************************************
 *
 * EGA_OpenGraphics
 *
 * SYNOPSIS
 *		short EGA_OpenGraphics(void)
 *
 * PURPOSE
 *		Init EGA graphics routines.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short EGA_OpenGraphics(size_t size)
{
	/** This really SUCKS! **/
	EGALoadBuffer = FP_SEG(AllocMemOnPara (max (size, 32000U)));	/* work space too :-) */
	if (!EGALoadBuffer) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: allocating EGA loadbuffer");
		return 0;
	}

	EGA_ASMOpenGraphics ();

	return 1;

} /* EGA_OpenGraphics */

/*********************************************************************
 *
 * EGA_CloseGraphics
 *
 * SYNOPSIS
 *		void EGA_CloseGraphics (void)
 *
 * PURPOSE
 *		Free resources used by EGA graphics routines.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void EGA_CloseGraphics (void)
{
	EGA_ASMCloseGraphics ();

	if (EGALoadBuffer)	FreeMemOnPara (MK_FP(EGALoadBuffer, 0));

} /* EGA_CloseGraphics */

/*********************************************************************
 *
 * EGA_DrawShapeNoTrans
 *
 * SYNOPSIS
 *		EGA_DrawShapeNoTrans (GShape *shape, short x, short y)
 *
 * PURPOSE
 *		DrawShape without transparency.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void EGA_DrawShapeNoTrans (GShape *shape, short x, short y)
{
	SetPenColor (255);
#if !USE_SHAPE_TPAT
	EvenPatternTbl[255] = shape->TColor;
	OddPatternTbl[255]  = shape->TColor >> 8;
#else
	EvenPatternTbl[255] = shape->Data.EGA.TPatEven;
	OddPatternTbl[255]  = shape->Data.EGA.TPatOdd;
#endif
	DrawRect (x, y, shape->Width, shape->Height);
	DrawGShape (shape, x, y);

} /* EGA_DrawShapeNoTrans */

