/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * MDBUF.C
 *
 * PROGRAMMER : Gregg A. Tavares
 *    VERSION : 00.000
 *    CREATED : 01/12/90
 *   MODIFIED : 10/08/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		MCGA specific DBUF routines.
 *
 * HISTORY
 *		10/13/93 Wednesday (dcc) - updated for MS-DOS Watcom C32.
 *		10/08/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/alloc.h>
#include <echidna/dbuf.h>
#include <echidna/memory.h>
#include <echidna/eerrors.h>
#include "dbufarea.h"

/**************************** C O N S T A N T S ***************************/
#define GHEADERSIZE	(GSHAPESIZE + MCGASDATASIZE)

/*********************************************************************
 *
 * MCGA_CloseDBufGraphics
 *
 * SYNOPSIS
 *		void MCGA_CloseDBufGraphics (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_CloseDBufGraphics (void)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "MCGA_CloseDBufGraphics";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

#if __MSDOS32X__
	if (*BackBuffer)	MEM_free (*BackBuffer);
	if (*WorkBuffer)	MEM_free (*WorkBuffer);
#else
	if (*BackBuffer)	FreeMemOnPara (MK_FP(*BackBuffer, 0));
	if (*WorkBuffer)	FreeMemOnPara (MK_FP(*WorkBuffer, 0));
#endif

} /* MCGA_CloseDBufGraphics */

/*********************************************************************
 *
 * MCGA_OpenDBufGraphics
 *
 * SYNOPSIS
 *		short MCGA_OpenDBufGraphics (short mode, size_t size)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short MCGA_OpenDBufGraphics (short mode, size_t size)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "MCGA_OpenDBufGraphics";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	mode = mode;

	/** This really SUCKS! **/
#if __MSDOS32X__
	*WorkBuffer = MEM_malloc(64000U);
#else
	*WorkBuffer = FP_SEG(AllocMemOnPara (64000U));
#endif
	if (!*WorkBuffer) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: allocating MCGA workbuffer");
		return FALSE;
	}
#if __MSDOS32X__
	*BackBuffer = MEM_malloc(max(size, 64000U));
#else
	*BackBuffer = FP_SEG(AllocMemOnPara (max(size,64000U)));
#endif
	if (!*BackBuffer) {
		SetGlobalErr (ERR_OUT_OF_MEMORY);
		GEprintf ("OOM: allocating MCGA backbuffer");
		return FALSE;
	}
	return TRUE;

} /* MCGA_OpenDBufGraphics */

/*********************************************************************
 *
 * MCGA_DBufPeepRect
 *
 * SYNOPSIS
 *		void MCGA_DBufPeepRect (
 *			GShape	*shapeptr,
 *			short	 x,
 *			short	 y,
 *			short	 width,
 *			short	 height,
 *		)
 *
 * PURPOSE
 *		To draw a rect in shape's size ^ onto the background buffer, and work buffer and
 *		and enter the area into the shape area list so it will be copied
 *		to the display buffer on the next call to DBufFlipScreen().
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufPeepRect (
	short	 x,
	short	 y,
	short	 width,
	short	 height
)
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x;
	*(saptr++)	= width;
	*(saptr++) 	= y;
	*(saptr++) 	= height;
  	*saptr		= PEEP_FLAG_MASK;	/* Erase next frame */    

	DrawRect (x, y, width, height);

} /* MCGA_DBufPeepRect */

/*********************************************************************
 *
 * MCGA_DBufPeepShape
 *
 * SYNOPSIS
 *		void MCGA_DBufPeepShape (
 *			GShape	*shapeptr,
 *			short x,
 *			short y
 *		) 
 *
 * PURPOSE
 *		To record the rectangular area where the shape will be drawn in the
 *		shape area draw list.  
 *		Then to draw the shape via call to 'DrawGShape'. This shape will
 *		stay on the display screen for only one frame.
 *
 * INPUT
 *		shapeptr: 	Far pointer to shape data.
 *		x, y:		Screen coordinates for drawing shape.	
 *		SAdrawPtr:	(GLOBAL) Pointer to first free record in Shape Area Draw
 *					List.  
 *  
 * EFFECTS
 *		SAdrawPtr:	(GLOBAL) Incremented to point to next free record.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufPeepShape (
	GShape	*shapeptr,
	short x,
	short y
) 
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;
	WORD	*dataptr	= (WORD *)((UBYTE *)shapeptr + GSHAPESIZE);

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x + *dataptr++;	/* Area->x = x + shape->xoffset */
	*(saptr++)	= *dataptr++;		/* Area->width = shape->width   */
	*(saptr++) 	= y + *dataptr++;	/* Area->y = y + shape->yoffset */
	*(saptr++) 	= *dataptr;			/* Area->height = shape->height */
  	*saptr		= PEEP_FLAG_MASK;	/* Erase next frame */

	DrawGShape (shapeptr, x, y);

} /* MCGA_DBufPeepShape */

/*********************************************************************
 *
 * MCGA_DBufStampShape
 *
 * SYNOPSIS
 *		void MCGA_DBufStampShape (
 *			GShape	*shapeptr,
 *			short	 x,
 *			short	 y
 *		)
 *
 * PURPOSE
 *		To draw a shape onto the background buffer, and work buffer and
 *		and enter the area into the shape area list so it will be copied
 *		to the display buffer on the next call to DBufFlipScreen().
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufStampShape (
	GShape	*shapeptr,
	short	 x,
	short	 y
)
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;
	WORD	*dataptr	= (WORD *)((UBYTE *)shapeptr + GSHAPESIZE);

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x + *dataptr++;	/* Area->x = x + shape->xoffset */
	*(saptr++)	= *dataptr++;		/* Area->width = shape->width   */
	*(saptr++) 	= y + *dataptr++;	/* Area->y = y + shape->yoffset */
	*(saptr++) 	= *dataptr;			/* Area->height = shape->height */
  	*saptr		= STAMP_FLAG_MASK;	/* Do not Erase next frame */    

	SetWorkPage (BackBuffer);
	DrawGShape (shapeptr, x, y);	/* in back */
	
	SetWorkPage (WorkBuffer);
	DrawGShape (shapeptr, x, y);	/* in work */

} /* MCGA_DBufStampShape */

/*********************************************************************
 *
 * MCGA_DBufStampShapeNoTrans
 *
 * SYNOPSIS
 *		void MCGA_DBufStampShapeNoTrans (
 *			GShape	*shapeptr,
 *			short	 x,
 *			short	 y
 *		)
 *
 * PURPOSE
 *                   without transparency.
 *		To draw a shape ^ onto the background buffer, and work buffer and
 *		and enter the area into the shape area list so it will be copied
 *		to the display buffer on the next call to DBufFlipScreen().
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufStampShapeNoTrans (
	GShape	*shapeptr,
	short	 x,
	short	 y
)
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;
	UWORD	 width		= shapeptr->Width;
	UWORD	 height		= shapeptr->Height;

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x;
	*(saptr++)	= width;			/* Area->width = shape->Width   */
	*(saptr++) 	= y;
	*(saptr++) 	= height;			/* Area->height = shape->Height */
  	*saptr		= STAMP_FLAG_MASK;	/* Do not Erase next frame */    

	SetWorkPage (BackBuffer);
	
	DrawGShapeNoTrans (shapeptr, x, y);	/* in back */
	
	SetWorkPage (WorkBuffer);
	
	DrawGShapeNoTrans (shapeptr, x, y);	/* in back */

} /* MCGA_DBufStampShapeNoTrans */

/*********************************************************************
 *
 * MCGA_DBufStampClearShape
 *
 * SYNOPSIS
 *		void MCGA_DBufStampClearShape (
 *			GShape	*shapeptr,
 *			short	 x,
 *			short	 y
 *		)
 *
 * PURPOSE
 *		To draw a rect in shape's size ^ onto the background buffer, and work buffer and
 *		and enter the area into the shape area list so it will be copied
 *		to the display buffer on the next call to DBufFlipScreen().
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufStampClearShape (
	GShape	*shapeptr,
	short	 x,
	short	 y
)
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;
	WORD	width		= shapeptr->Width, 
			height		= shapeptr->Height;

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x;
	*(saptr++)	= width;			/* Area->width = shape->width   */
	*(saptr++) 	= y;
	*(saptr++) 	= height;			/* Area->height = shape->height */
  	*saptr		= STAMP_FLAG_MASK;	/* Do not Erase next frame */    

	SetWorkPage (BackBuffer);
	
	SetPaperColor (shapeptr->TColor);
	DrawRect (x, y, shapeptr->Width, shapeptr->Height);
	
	SetWorkPage (WorkBuffer);
	
	DrawRect (x, y, shapeptr->Width, shapeptr->Height);

} /* MCGA_DBufStampClearShape */

/*********************************************************************
 *
 * MCGA_DBufStampRect
 *
 * SYNOPSIS
 *		void MCGA_DBufStampRect (
 *			GShape	*shapeptr,
 *			short	 x,
 *			short	 y,
 *			short	 width,
 *			short	 height,
 *		)
 *
 * PURPOSE
 *		To draw a rect in shape's size ^ onto the background buffer, and work buffer and
 *		and enter the area into the shape area list so it will be copied
 *		to the display buffer on the next call to DBufFlipScreen().
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void MCGA_DBufStampRect (
	short	 x,
	short	 y,
	short	 width,
	short	 height
)
{
	/* saptr:			Points to the succesive fields 
	 *					of the current 'AreaRecType' struct pointed to
	 *					by 'SAdrawPtr'. The fields are: 'x', 'width',  
	 *					'y', 'height' and 'flags'.
	*/	
	/* dataptr:			Data Pointer. Points to the succesive fields of the
	 *					shape data header: <x offset>, <width>,
	 *					<y offset>, <height>. 
	*/

	WORD	*saptr 		= (WORD *)SAdrawPtr++;

	/** Record in draw list the area rectangle where shape will be drawn **/
	*(saptr++) 	= x;
	*(saptr++)	= width;
	*(saptr++) 	= y;
	*(saptr++) 	= height;
  	*saptr		= STAMP_FLAG_MASK;	/* Do not Erase next frame */    

	SetWorkPage (BackBuffer);

	DrawRect (x, y, width, height);

	SetWorkPage (WorkBuffer);

	DrawRect (x, y, width, height);

} /* MCGA_DBufStampRect */

/*********************************************************************
 *
 * MergeShowAreas
 *
 * SYNOPSIS
 *		static void MergeShowAreas (
 *			AreaRecType	*alist,
 *			AreaRecType *blist
 *		)
 *
 * PURPOSE
 *		To merge show areas from 'alist' with overlapping areas of
 *		same dimension in 'blist'.  This is to avoid flickering and
 *		to get better speed.
 *
 * INPUT
 *		alist:	Array of AreaRecType structures with end of list marked
 *					by entry with height field of 0.
 *		blist:	Array of AreaRecType structures with end of list marked
 *					by entry with height field of 0.
 *
 * AFFECTS
 *		When two areas meet the test, the area in 'alist' is expanded to
 *		cover both areas and the area from 'blist' is marked so it won't
 *		be shown.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
static void MergeShowAreas (
	AreaRecType	*alist,
	AreaRecType *blist
)
{
	AreaRecType *bl, *blstart = blist;
	WORD a_height, a_width;
	WORD b_height;

	for ( ; (a_height = alist->height) != 0; alist++) {
		/* Clear skip flag from previous call to MergeShowAreas() */
		alist->flags &= ~(UWORD)(SKIP_FLAG_MASK);
		for (bl = blstart; (b_height = bl->height) != 0; bl++) {
			/* Only merge items of same dimensions since these are the
			 * areas most likely to have been generated by the same object
			 * moved over from the previous frame.
			 */
			 /* Checks b_height > 0 to make sure area not already merged */
			if	(!(bl->flags & SKIP_FLAG_MASK) && 
					a_height == b_height &&
					(a_width = alist->width) == bl->width
				) 
			{
				WORD a_x, a_y;
				WORD b_x, b_y;
				WORD xdist, ydist;
				WORD xmin, ymin;

				if ((xdist = (a_x = alist->x) - (b_x = bl->x)) < 0) {
					xdist = -xdist;
					xmin = a_x;
				} else {
					xmin = b_x;
				}

				if ((ydist = (a_y = alist->y) - (b_y = bl->y)) < 0) {
					ydist = -ydist;
					ymin = a_y;
				} else {
					ymin = b_y;
				}

				if (xdist <= a_width && ydist <= a_height)
				{ /* squares overlap */
					alist->x = xmin;
					alist->y = ymin;
					alist->width += xdist;
					alist->height += ydist;
					/* Flag area in other list as already taken care of */
					bl->flags |= SKIP_FLAG_MASK;
					if (bl == blstart) {
						/* Move start of list to skip this one next time */
						++blstart;
					}
					/* One match is all I need. Skip rest of list */
					break;
				}
			}
		}
	}
} /* MergeShowAreas */

/*********************************************************************
 *
 * DisplayWorkAreas
 *
 * SYNOPSIS
 *		void DisplayWorkAreas (
 *			AreaRecType	*arealist
 *		)
 *
 * PURPOSE
 *		To copy the rectangular areas in 'arealist' from the Work screen to
 *		the Display screen via call to 'ShowRect'.  
 *
 * INPUT
 *		arealist:	Array of AreaRecType structures with end of list marked
 *					by entry with height field of 0.
 *		
 * AFFECTS
 *		Copies areas from work buffer to display buffer.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		ShowRect()
 *
*/
static void DisplayWorkAreas (
	AreaRecType	*arealist
)
{
	WORD al_height;

	for ( ; (al_height = arealist->height) != 0; arealist++) {
		/* Do not show if area eliminated by MergeShowAreas() */
		if (!(arealist->flags & SKIP_FLAG_MASK)) {
			ShowRect (arealist->x, arealist->y, 
						arealist->width, al_height);		
		}
	}
} /* DisplayWorkAreas */

/*********************************************************************
 *
 * RestoreBackAreas
 *
 * SYNOPSIS
 *		void RestoreBackAreas (
 *			AreaRecType	*arealist
 *		)
 *
 * PURPOSE
 *		To copy the rectangular areas in 'arealist' from the Background
 *		screen to the Work screen via call to 'Restore'.  
 *
 * INPUT
 *		arealist:	Array of AreaRecType structures with end of list marked
 *					by entry with height field of 0.
 *
 * AFFECTS
 *		Copies areas from background buffer to work buffer. 
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		Restore()
 *
*/
static void RestoreBackAreas (
	AreaRecType	*arealist
)
{
	WORD al_height;

	for ( ; (al_height = arealist->height) != 0; arealist++) {
		if (arealist->flags & PEEP_FLAG_MASK) {
			Restore (arealist->x, arealist->y, 
						arealist->width, al_height);		
		}
	}

} /* RestoreBackAreas */

/*********************************************************************
 *
 * MCGA_DBufFlipScreen
 *
 * SYNOPSIS
 *		void MCGA_DBufFlipScreen (void)
 *
 * PURPOSE
 *		To ShapeFlip double buffering screens and update the display.
 *
 * INPUT
 *		(GLOBALS)
 *		SAList1, SAList2:	Shape Area lists (arrays).
 *		SAeraseList:		List from which to restore areas of Display  
 *							screen where old shapes are.
 *		SAdrawList:			List from which to update areas of Display
 *							screen where new shapes are. List from which
 *							to restore areas of Work screen where new shapes
 *							were drawn.
 *		SAerasePtr:			Pointer to end of SAerasList.
 *		SAdrawPtr:			Pointer to end of SAdrawList.
 *
 * EFFECTS
 *		(GLOBALS)
 *		'SAeraseList' and 'SAdrawList' are swapped.
 *		'SAerasePtr' is set to final value of 'SAdrawPtr'.
 *		'SAdrawPtr' is set to start of new 'SAdrawList'.
 *
 * RETURN VALUE
 *		None.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *		DisplayWorkAreas (), RestoreBackAreas (),
 *		SWAP_RESET_SALISTS ()
 *
*/
void MCGA_DBufFlipScreen (void)
{

	/* Mark end of area lists */
		SAerasePtr->height = 0;
		SAdrawPtr->height = 0; 

	/* Optimize area lists to do least amount of overlapping copies
	 * from work to display (which also reduces flicker).
	 */
		MergeShowAreas (SAeraseList, SAdrawList);

	/* Copy Old Shape areas from Work to Display */
		DisplayWorkAreas (SAeraseList); 

  	/* Copy New Shape areas from Work to Display */
		DisplayWorkAreas (SAdrawList); 

	/* Copy New Shape areas from Background to Work */
		RestoreBackAreas (SAdrawList);

	/* Reset shape save areas */
		SWAP_RESET_SALISTS ();

} /* MCGA_DBufFlipScreen */

typedef void (far32 *RoutinePtr)();

static RoutinePtr MCGADBufRoutineTable[] = {
	(RoutinePtr)MCGA_OpenDBufGraphics,
	(RoutinePtr)MCGA_CloseDBufGraphics,
	(RoutinePtr)MCGA_DBufPeepShape,
	(RoutinePtr)MCGA_DBufStampShape,
	(RoutinePtr)MCGA_DBufStampShapeNoTrans,
	(RoutinePtr)MCGA_DBufStampClearShape, 
	(RoutinePtr)MCGA_DBufStampRect,
	(RoutinePtr)MCGA_DBufPeepRect,
	(RoutinePtr)MCGA_DBufFlipScreen,
};


/*********************************************************************
 *
 * GetMCGADBufRoutineTable
 *
 * SYNOPSIS
 *		RoutinePtr *GetMCGADBufRoutineTable (void)
 *
 * PURPOSE
 *		Returns pointer to MCGADBufRoutineTable.  Calling this
 *		routine should force loading if MCGA_DBufRoutines as an
 *		overlay.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
RoutinePtr *GetMCGADBufRoutineTable (void)
{
	return MCGADBufRoutineTable;

} /* GetMCGADBufRoutineTable */

