;* 
;* The contents of this file are subject to the Mozilla Public 
;* License Version 1.1 (the "License"); you may not use this file 
;* except in compliance with the License. You may obtain a copy of 
;* the License at http://www.mozilla.org/MPL/ 
;* 
;* Software distributed under the License is distributed on an "AS 
;* IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
;* implied. See the License for the specific language governing 
;* rights and limitations under the License. 
;* 
;* The Original Code is tUME: the Universal Map Editor, released 
;* June 18, 2000. 
;* 
;* The Initial Developer of the Original Code is Echidna. 
;* Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
;* All Rights Reserved. 
;* 
;* Contributor(s): 
;* 

;**
;** MRESTORE.ASM
;**
;** PROGRAMMER : Daniel C. Chang
;**    VERSION : 00.000
;**    CREATED : 07/23/89
;**   MODIFIED : 10/13/93
;**       TABS : 10 18 40 44
;**
;**	     \|///-_
;**	     \oo///_
;**	-----w/-w------
;**	 E C H I D N A
;**	---------------
;**
;** DESCRIPTION
;**	ShowRect/Restore for MCGA.
;**
;** HISTORY
;**	09/30/93 Thursday (dcc) - fix MCGA_Restore so it copies from Back to Work
;**	01/16/92 Thursday (DCC) - add support for very large shapes
;**	05/14/91 Tuesday (DCC) - fixed so source & dest offset may be != 0
;**
	IDEAL

	if USE_GRAFX

; CopyRect - copy rectangle starting at source x, y from screen
; buffer pointed to by DS: to destination x, y screen buffer pointed
; to by ES: starting at x, y of xsize, ysize.
; Known bugs: doesn't do any clipping.
; by Daniel Chang, 11/25/89.
;
; INPUT:
;
; DS:SI - points to source bytemap
; ES:DI - points to destination bytemap
; AX - source x position of left edge of rectangle to copy from
; BX - source y position of top edge of rectangle to copy from
; CX - destination x position of left edge of rectangle to copy to
; DX - destination y position of top edge of rectangle to copy to
; [cs:sw] - width of source rectangle to copy
; [cs:sh] - height of source rectangle to copy
; [cs:sRowByte] - bytes to add to get to next row in source screen
; [cs:dRowByte] - bytes to add to get to next row in destination screen
;
; REGISTERS CHANGED: AX, BX, CX, DX, SI, DI, BP, ES, DS

	STRUC	ByteMap
Width	DW	?
Height	DW	?
gfxData	DD	?
TColor	DW	?
	ENDS

	EVEN

sRowByte dw	?	;source bytemap bytes per row
sHeight	dw	?	;source bytemap number of rows
dRowByte dw	?	;destination bytemap bytes per row
sw	dw	?
sh	dw	?
dskip	dw	?	;bytes to skip to start of next row

	MPROC	MCGA_CopyRect

	cld

	mov	bp, [cs:sw]	;save # of bytes to copy per row
	or	bp, bp
	jle	@@done	;=> width <= 0, no work
	cmp	[cs:sh], 0
	jle	@@done	;=> height <= 0, no work

	push	dx	;save dstY
	push	ax	;save srcX

	mov	ax, bx
	mul	[cs:sRowByte]	;AX = srcY * [cs:sRowByte]
	pop	bx
	add	ax, bx	;AX = srcY * [cs:sRowByte] + srcX
	add	si, ax
	jnc	@@doneMUL

	mov	ax, ds
	add	ax, 1000h
	mov	ds, ax
@@doneMUL:
	call	NormalizeDSSI

	pop	ax	;restore dstY
	mul	[cs:dRowByte]	;AX = dstY * [cs:dRowByte]
	add	ax, cx	;AX = dstY * [cs:dRowByte] + dstX
	add	di, ax
	jnc	@@donMUL

	mov	ax, es
	add	ax, 1000h
	mov	es, ax
@@donMUL:
	call	NormalizeESDI

	mov	dx, [cs:sRowByte]
	sub	dx, bp

	mov	ax, [cs:dRowByte]
	sub	ax, bp
	mov	[cs:dskip], ax	;bytes to skip to next dest row

	mov	bx, [cs:sh]	;# of rows to copy
	shr	bp, 1	;# bytes per row / 2
	jnc	@@even	;=> even number of bytes to copy

	EVEN
@@yloop:
	mov	cx, bp	;get number of bytes to copy
	movsb
	rep	movsw	;copy row

	add	si, dx	;move to next row in source
	js	@@fixSrc0	;=> normalize number
@@SrcFixed0:
	add	di, [cs:dskip]	;move to next row in destination
	js	@@fixDst0	;=> normalize number
@@DstFixed0:
	dec	bx
	jg	@@yloop	;=> more rows to copy
@@done:
	ret

@@fixSrc0:
	call	NormalizeDSSI
	jmp	SHORT @@SrcFixed0
@@fixDst0:
	call	NormalizeESDI
	jmp	SHORT @@DstFixed0

	EVEN
@@even:
	mov	cx, bp	;get number of bytes to copy
	rep	movsw	;copy row

	add	si, dx	;move to next row in source
	js	@@fixSrc1	;=> normalize number
@@SrcFixed1:
	add	di, [cs:dskip]	;move to next row in destination
	js	@@fixDst1	;=> normalize number
@@DstFixed1:
	dec	bx
	jg	@@even	;=> more rows to copy

	ret

@@fixSrc1:
	call	NormalizeDSSI
	jmp	SHORT @@SrcFixed1
@@fixDst1:
	call	NormalizeESDI
	jmp	SHORT @@DstFixed1

	ENDP	MCGA_CopyRect

NormalizeDSSI:

	; Normalize DS:SI. Destroys AX register. Preserves all others.

	push	bx
	mov	ax, si
	shr	ax, 1
	shr	ax, 1
	shr	ax, 1
	shr	ax, 1
	mov	bx, ds
	add	ax, bx
	mov	ds, ax
	and	si, 0Fh
	pop	bx
	ret

NormalizeESDI:

	; Normalize ES:DI. Destroys AX register. Preserves all others.

	push	bx
	mov	ax, di
	shr	ax, 1
	shr	ax, 1
	shr	ax, 1
	shr	ax, 1
	mov	bx, es
	add	ax, bx
	mov	es, ax
	and	di, 0Fh
	pop	bx
	ret

;=====================================================================
;**
;** _CopyRect
;**
;** PURPOSE
;**	C glue for copy rect as above.  C Prototype is
;**
;**	CopyRect (
;**		ByteMap	*src,
;**		short	 fromx,
;**		short	 fromy,
;**		ByteMap	*dest,
;**		short	 tox,
;**		short	 toy,
;**		short	 width,
;**		short	 height
;**		)
;**
;** INPUT
;**	
;**
;** REGISTER USED
;**	
;**
;** EFFECTS
;**	
;**
;** RETURN VALUE
;**	
;**
;** LIMITS/BUGS
;**	
;**
;** HISTORY
;**	
;**

	EVEN

	PUBLIC	_MCGA_CopyRect
	MPROC	_MCGA_CopyRect
	ARG	sBMoff:WORD,sBMseg:WORD,FromX:WORD,FromY:WORD,dBMoff:WORD,dBMseg:WORD,ToX:WORD,ToY:WORD,GWidth:WORD,Height:WORD

	push	bp
	mov	bp, sp

	push	ds
	push	di
	push	si

	BeforeGraphics

	mov	ax, [GWidth]
	mov	[cs:sw], ax
	mov	ax, [Height]
	mov	[cs:sh], ax

	lds	si,[dword ptr dBMoff]	; Get Pointer to Dest ByteMap
	mov	ax,[(ByteMap ptr ds:si).Width]
  	mov	[cs:dRowByte],ax	; Set width of Dest ByteMap
	les	di, [(ByteMap ptr ds:si).gfxData]	;DCC Get Off, Seg of Dest ByteMap

	lds	si,[dword ptr sBMoff]	; Get Pointer to Src ByteMap
	mov	ax,[(ByteMap ptr ds:si).Width]
  	mov	[cs:sRowByte],ax	; Set width of Src ByteMap
	lds	si, [(ByteMap ptr ds:si).gfxData]	;DCC Get Off, Seg of Src ByteMap

	mov	ax, [FromX]
	mov	bx, [FromY]
	mov	cx, [ToX]
	mov	dx, [ToY]

	call	MCGA_CopyRect

	AfterGraphics

	pop	si
	pop	di
	pop	ds

	pop	bp
	ret
	ENDP	_MCGA_CopyRect


;
; C PROTOTYPE:
;
; void far MCGA_Restore(int xmin, int ymin, int xmax, int ymax);
;
; _Restore - restores from background buffer to the current work page.
; segment of background buffer is in [cs:Pages+4]
; segment of current work page is in [cs:GrPage]
; by Daniel Chang, 07/23/89.

	EVEN

	MPROC	_MCGA_Restore
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp

	push	ds
	push	di
	push	si

	BeforeGraphics

	mov	ax, 320
	mov	[cs:sRowByte],ax
	mov	[cs:dRowByte],ax

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	ax, [rwidth]
	mov	[cs:sw], ax
	mov	ax, [height]
	mov	[cs:sh], ax
	mov	ax, cx
	mov	bx, dx

	sub	si, si
	sub	di, di
	mov	ds, [cs:MBackBuffer]	;source segment
	mov	es, [cs:MWorkBuffer]	;destination segment

	call	MCGA_CopyRect

	AfterGraphics

	pop	si
	pop	di
	pop	ds

	pop	bp
	ret
	ENDP	_MCGA_Restore

;
; C PROTOTYPE:
;
; void far MCGA_UnRestore(int xmin, int ymin, int xmax, int ymax);
;
; _UnRestore - 'unrestores' from the current work page to background buffer.
; segment of background buffer is in [cs:Pages+4]
; segment of current work page is in [cs:GrPage]
; by Daniel Chang, 07/23/89.

	EVEN

	MPROC	_MCGA_UnRestore
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp

	push	ds
	push	di
	push	si

	BeforeGraphics

	mov	ax, 320
	mov	[cs:sRowByte],ax
	mov	[cs:dRowByte],ax

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	ax, [rwidth]
	mov	[cs:sw], ax
	mov	ax, [height]
	mov	[cs:sh], ax
	mov	ax, cx
	mov	bx, dx

	sub	si, si
	sub	di, di
	mov	ds, [cs:MWorkBuffer]	;source segment
	mov	es, [cs:MBackBuffer]	;destination segment

	call	MCGA_CopyRect

	AfterGraphics

	pop	si
	pop	di
	pop	ds

	pop	bp
	ret
	ENDP	_MCGA_UnRestore

;
; C PROTOTYPE:
;
; void far MCGA_ShowRect(int xmin, int ymin, int width, int height);
;
; _Restore - copies from work page to the display page.
; segment of current work page is in [cs:GrPage]
; segment of display buffer is in [cs:Pages]
; by Daniel Chang, 07/23/89.

	EVEN

	MPROC	_MCGA_ShowRect
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp

	push	ds
	push	di
	push	si

	BeforeGraphics

	mov	ax, 320
	mov	[cs:sRowByte],ax
	mov	[cs:dRowByte],ax

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	ax, [rwidth]
	mov	[cs:sw], ax
	mov	ax, [height]
	mov	[cs:sh], ax
	mov	ax, cx
	mov	bx, dx

	sub	si, si
	sub	di, di
	mov	ds, [cs:MWorkBuffer]	;source segment
	mov	es, [cs:MDisplayBuffer]	;destination segment

	call	MCGA_CopyRect

	AfterGraphics

	pop	si
	pop	di
	pop	ds

	pop	bp
	ret
	ENDP	_MCGA_ShowRect

	else

;************************************  *************************************
;************************************  *************************************
;************************************  *************************************
;************************************  *************************************
;************************************  *************************************
;************************************  *************************************


	IDEAL

; CopyRect - copy rectangle from screen buffer pointed to by DS: to
; screen buffer pointed to by ES: starting at x, y of xsize, ysize.
; Known bugs: doesn't clip to source rectangle (does clip to dest).
; by Daniel Chang, 07/23/89.
;
; INPUT:
;
; AX - source x position of left edge of rectangle to copy from
; BX - source y position of top edge of rectangle to copy from
; CX - destination x position of left edge of rectangle to copy to
; DX - destination y position of top edge of rectangle to copy to
; SI - x width of source rectangle to copy
; DI - y height of source rectangle to copy
; DS - segment of source MCGA screen
; ES - segment of destination MCGA screen
;
; [cs:MClipLeft] - left edge of destination clip window
; [cs:MClipRight] - right edge of destination clip window
; [cs:MClipTop] - top edge of destination clip window
; [cs:MClipBottom] - bottom edge of destination clip window
;
; [cs:sRowByte] - bytes to add to get to next row in source screen
; [cs:dRowByte] - bytes to add to get to next row in destination screen
;
; REGISTERS CHANGED: AX, BX, CX, DX, SI, DI

sRowByte dw	320	;source bytes per row
dRowByte dw	320	;destination bytes per row

	MPROC	MCGA_CopyRect

	push	bp

	; Check for clipping on destination left edge.
	; If we clip, adjust width, srcX, and dstY.

	cld

	mov	bp, cx
	sub	bp, [cs:MClipLeft]	;BP = dstX-MClipLeft
	jge	@@dxOK	;=> no clipping

	add	si, bp	;width = width + (dstX-MClipLeft)
	sub	ax, bp	;srcX = srcX - (dstX-MClipLeft)
	mov	cx, [cs:MClipLeft]	;dstX = MClipLeft
@@dxOK:
	; Check for clipping on destination top edge.
	; If we clip, adjust height, srcY, and dstY.

	mov	bp, dx
	sub	bp, [cs:MClipTop]	;BP = dstY-MClipTop
	jge	@@dyOK	;=> no clipping

	add	di, bp	;height = height + (dstY-MClipTop)
	sub	bx, bp	;srcY = srcY - (dstY-MClipTop)
	mov	dx, [cs:MClipTop]	;dstY = MClipTop
@@dyOK:
	; Check for clipping on destination right edge.
	; If we clip, adjust width.

	mov	bp, [cs:MClipRight]
	inc	bp
	sub	bp, cx
	cmp	si, bp	;width <= MClipRight+1 - dstX
	jle	@@wOK	;=> yes, no clip
	mov	si, bp	;width = MClipRight+1 - dstX
@@wOK:
	; Check for clipping on destination bottom edge.
	; If we clip, adjust height.

	mov	bp, [cs:MClipBottom]
	inc	bp
	sub	bp, dx
	cmp	di, bp	;height <= MClipBottom+1 - dstY
	jle	@@hOK	;=> yes, no clip
	mov	di, bp	;height = MClipBottom+1 - dstY
@@hOK:
	cmp	si, 0
	jle	@@done	;=> width <= 0, no work
	cmp	di, 0
	jle	@@done	;=> height <= 0, no work

	mov	bp, si	;save # of bytes to copy per row

	push	dx	;save dstY
	push	ax	;save srcX

	mov	ax, bx
	mul	[cs:sRowByte]	;AX = srcY * [cs:sRowByte]
	pop	bx	;restore srcX
	add	ax, bx	;AX = srcY * [cs:sRowByte] + srcX
	mov	si, ax

	mov	bx, di	;save # of rows to copy

	pop	ax	;restore dstY
	mul	[cs:dRowByte]	;AX = dstY * [cs:dRowByte]
	add	ax, cx	;AX = dstY * [cs:dRowByte] + dstX
	mov	di, ax
@@yloop:
	mov	cx, bp	;get number of bytes to copy
	rep	movsb	;copy row

	add	si, [cs:sRowByte]
	sub	si, bp	;move to next row in source
	add	di, [cs:dRowByte]
	sub	di, bp	;move to next row in destination
	dec	bx
	jg	@@yloop	;=> more rows to copy
@@done:
	pop	bp
	ret
	ENDP	MCGA_CopyRect

;
; C PROTOTYPE:
;
; void far MCGA_Restore(int xmin, int ymin, int xmax, int ymax);
;
; _Restore - restores from background buffer to the current work page.
; segment of background buffer is in [cs:Pages+4]
; segment of current work page is in [cs:GrPage]
; by Daniel Chang, 07/23/89.

	MPROC	_MCGA_Restore
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp
	sub	sp, AUTO_SIZE

	push	ds
	push	es
	push	di
	push	si

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	si, [rwidth]
	mov	di, [height]
	mov	ax, cx
	mov	bx, dx

	mov	ds, [cs:MBackBuffer]	;source segment
	mov	es, [cs:MWorkBuffer]	;destination segment

	call	MCGA_CopyRect

	pop	si
	pop	di
	pop	es
	pop	ds

	add	sp, AUTO_SIZE
	pop	bp
	ret
	ENDP	_MCGA_Restore

;
; C PROTOTYPE:
;
; void far MCGA_UnRestore(int xmin, int ymin, int xmax, int ymax);
;
; _UnRestore - 'unrestores' from the current work page to background buffer.
; segment of background buffer is in [cs:Pages+4]
; segment of current work page is in [cs:GrPage]
; by Daniel Chang, 07/23/89.
; modified from MCGA_Restore 7/08/91 (RGM).

	MPROC	_MCGA_UnRestore
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp
	sub	sp, AUTO_SIZE

	push	ds
	push	es
	push	di
	push	si

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	si, [rwidth]
	mov	di, [height]
	mov	ax, cx
	mov	bx, dx

	mov	ds, [cs:MWorkBuffer]	;source segment
	mov	es, [cs:MBackBuffer]	;destination segment

	call	MCGA_CopyRect

	pop	si
	pop	di
	pop	es
	pop	ds

	add	sp, AUTO_SIZE
	pop	bp
	ret
	ENDP	_MCGA_UnRestore


;
; C PROTOTYPE:
;
; void far MCGA_ShowRect(int xmin, int ymin, int width, int height);
;
; _Restore - copies from work page to the display page.
; segment of current work page is in [cs:GrPage]
; segment of display buffer is in [cs:Pages]
; by Daniel Chang, 07/23/89.

	MPROC	_MCGA_ShowRect
	ARG	xmin:WORD, ymin:WORD, rwidth:WORD, height:WORD=AUTO_SIZE

	push	bp
	mov	bp, sp
	sub	sp, AUTO_SIZE

	push	ds
	push	es
	push	di
	push	si

	mov	cx, [xmin]
	mov	dx, [ymin]
	mov	si, [rwidth]
	mov	di, [height]
	mov	ax, cx
	mov	bx, dx

	mov	ds, [cs:MWorkBuffer]	;source segment
	mov	es, [cs:MDisplayBuffer]	;destination segment

	call	MCGA_CopyRect

	pop	si
	pop	di
	pop	es
	pop	ds

	add	sp, AUTO_SIZE
	pop	bp
	ret
	ENDP	_MCGA_ShowRect

 
	endif

