;* 
;* The contents of this file are subject to the Mozilla Public 
;* License Version 1.1 (the "License"); you may not use this file 
;* except in compliance with the License. You may obtain a copy of 
;* the License at http://www.mozilla.org/MPL/ 
;* 
;* Software distributed under the License is distributed on an "AS 
;* IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
;* implied. See the License for the specific language governing 
;* rights and limitations under the License. 
;* 
;* The Original Code is tUME: the Universal Map Editor, released 
;* June 18, 2000. 
;* 
;* The Initial Developer of the Original Code is Echidna. 
;* Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
;* All Rights Reserved. 
;* 
;* Contributor(s): 
;* 

;**
;** TRECT.asm
;**
;** PROGRAMMER : Gregg A. Tavares
;**    VERSION : 00.000
;**    CREATED : 03/05/90
;**   MODIFIED : 09/14/90
;**       TABS : 10 18 40 44
;**
;**	     \|///-_
;**	     \oo///_
;**	-----w/-w------
;**	 E C H I D N A
;**	---------------
;**
;** DESCRIPTION
;**	Routine to draw a rectangle in current color on TGA screen.
;**
;**
;** HISTORY
;**
;**

	IDEAL

;=====================================================================
;**
;** SmallRect
;**
;** PURPOSE
;**	Draw Even or Odd lines of a rectangle
;**
;** INPUT
;**	ES:DI	First Byte of rectangle
;**	DL	Mask for start of line
;**	DH	Mask for end of line
;**	SI	Width in bytes
;**	AL	Even pattern for lines being drawn
;**	AH	Odd pattern for lines being drawn
;**	Modulo	Modulo for rect
;**	Height	Height in lines (half?)
;**	Odd	Starting line odd = 1, even = 0
;**
;** 	NOTE: DS must = CS
;**
;** REGISTER USED
;**	ALL of them
;**
;** EFFECTS
;**	
;**
;** RETURN VALUE
;**	
;**
;** LIMITS/BUGS
;**	
;**
;** HISTORY
;**	
;**
Modulo	dw	0
RHeight	dw	0
Line	db	0
Odd	db	0
EvenPat	db	0
OddPat	db	0

SMasks	DB	000h
	DB	0F0h
EMasks	DB	00Fh
	DB	000h


	MPROC SmallRect NEAR

	cld		; Set Store Direction.

	mov	[EvenPat],al	; Save pattern
	mov	[OddPat],ah

	cmp	[Odd],0	; Init pattern for odd or even lines
	jz	@@even

	mov	al,ah
	jmp	short @@drawnextrow

@@even:
	mov	ah,al

@@drawnextrow:
	mov	cx,si	; restore width
	cmp	cx,1
	jne	@@normalrect

;*
;* both first and last bytes are the same so do them here.
;*
	mov	cl,dl
	xor	cl,dh
	mov	bl,[es:di]	; get original data
	xor	bl,al	; put in pixels
	and	bl,cl
	xor	bl,al
	mov	[es:di],bl
	jmp	short @@nolstbytehere
@@stone1:
	jmp	@@drawnextrow
@@normalrect:
	cmp	dl,000h	; do first byte if necessary.
	je	@@no1stbyte
	mov	bl,[es:di]	; get original data
	xor	bl,al	; put in pixels
	and	bl,dl
	xor	bl,al
	mov	[es:di],bl
	inc	di
	dec	cx

@@no1stbyte:
	cmp	dh,000h	; is there a last byte?
	je	@@nolstbyte
	dec	cx
	jz	@@not_odd

@@nolstbyte:
	push	cx	; save low cl
	shr	cx,1	; Get Width in words.

	rep	stosw

	pop	cx
	test	cl,1	; if width is odd plot one more pixel
	jz	@@not_odd

	stosb
@@not_odd:
	cmp	dh,000h
	je	@@nolstbytestl	; is there a last byte?
	mov	bl,[es:di]	; get original data
	xor	bl,ah	; put in pixels
	and	bl,dh
	xor	bl,ah
	mov	[es:di],bl
@@nolstbytehere:
	inc	di
	dec	cx

@@nolstbytestl:
	cmp	[Odd],0	; did we just do an odd line?
	je	@@doodd

	mov	al,[EvenPat]	; Get Even Pattern
	mov	ah,al
	jmp	short @@nextline

@@doodd:
	mov	al,[OddPat]	; Get odd pattern
	mov	ah,al

@@nextline:
	cmp	[Line],3
	je	@@next0line

	inc	[Line]
	sub	di,si	; move to next line
	add	di,2000h
	jmp	short @@realnext

@@next0line:
	mov	[Line],0
	sub	di,6000h	; move to next line.
	add	di,[Modulo]

@@realnext:
	xor	[Odd],01h
	dec	[RHeight]
	jg	@@stone1
	ret

	ENDP

;=====================================================================
;**
;** TGA_DrawRect
;**
;** PURPOSE
;**	Draws a rectangle on the specified rastport in the current color
;**
;** INPUT
;**	DS:SI RastPort			:Not Implemented
;**	AX = starting x pos
;**	BX = starting y pos
;**	CX = width
;**	DX = height
;**
;** EFFECTS
;**	AX,BX,CX,DX,ES,DI,SI,DS
;**
;** RETURN VALUE
;**	NONE
;**
;** LIMITS/BUGS
;**	
;**
;** HISTORY
;**	
;**

	MPROC xxTGA_DrawRect
tga_no_rect:
	ret

	MPROC TGA_DrawRect

	cmp	dx,0	; Skip if height = 0
	je	tga_no_rect
	cmp	cx,0	; Skip if width = 0
	je	tga_no_rect

	mov	[cs:RHeight],dx	; Save Height

	push	ds
	mov	dx,cs
	mov	ds,dx

	test	bx,01h	; Set up Odd flag.
	jnz	@@startodd
	mov	[Odd],0
	jmp	short @@setstart
@@startodd:
	mov	[Odd],1
@@setstart:
	push	bx
	and	bl,03h
	mov	[Line],bl	; set up line status (0-3)
	pop	bx

	shl	bx,1
	mov	di,[LineSt+BX]	; Get Starting address of y row
	mov	bx,ax
	shr	bx,1
	add	di,bx	; add x column.
	push	bx	; save start byte.

	mov	si,ax	; save xpos

	and	al,01h	; get start pixel num (0-1)
	xor	bh,bh
	mov	bl,al
	mov	dl,[SMasks+bx]	; get start mask
	mov	bx,si	; restore xpos.
	add	bx,cx	; add width
	dec	bx	; comp last pixel
	mov	cx,bx	; save end pixel
	and	bl,01h	; get end pixel num (0-3)
	xor	bh,bh
	mov	dh,[EMasks+bx]	; get end mask

	pop	ax	; get start byte
	shr	cx,1	; comp last byte
	sub	cx,ax	; comp width in bytes
	inc	cx

	mov	ax,320/2	; compute modulo of rect.
	sub	ax,cx
	mov	[Modulo],ax

	push	dx
	mov	bx,[CS:TGrSeg]	; Point Hardware to current page
	mov	ax,[CS:_TDAcsTbl+bx]	;
	or	al,[CS:_TDispVal]	; Or in current display
	mov	dx,03DFh
	out	dx,al
	mov	bx,[CS:_TDMemTbl+bx]	; Get current page's segment
	mov	es,bx
	pop	dx

	mov	si,cx	; save byte width.

	xor	bh,bh
	mov	bl,[TPenColor]	; Get Pen Color.
	mov	ah,[_TGAOddPatTbl+bx]	; Get Odd pattern image
	mov	al,[_TGAEvenPatTbl+bx]; Get Even pattern image

	call	SmallRect

	pop	ds

	ret

	ENDP TGA_DrawRect
	ENDP xxTGA_DrawRect

;=====================================================================
;**
;** _TGA_DrawRect(x, y, width, height)
;**
;** PURPOSE
;**	Draws a rectangle on the current workpage in the current color
;**
;** INPUT
;**	From stack using C conventions.
;**
;** RETURN VALUE
;**	NONE
;**
;** LIMITS/BUGS
;**	
;**
;** HISTORY
;**	
;**
	PUBLIC _TGA_DrawRectAsm
	MPROC	_TGA_DrawRectAsm
	ARG	_LEFT:WORD, _TOP:WORD, _WIDTH:WORD, _HEIGHT:WORD

	push	bp
	mov	bp,sp

	push	di	; Save SI and DI
	push	si

	mov	dx,[_HEIGHT]	; Get C Parms
	mov	cx,[_WIDTH]
	mov	bx,[_TOP]
	mov	ax,[_LEFT]

	call	TGA_DrawRect	; Draw the Rect

	pop	si	; Restore SI and DI
	pop	di

	pop	bp
	ret

	ENDP _TGA_DrawRectAsm

