/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * MEMORY.C
 *
 * PROGRAMMER : Gregg A. Tavares (with suggestions by Juan M. Alavardo)
 *    VERSION : 00.000
 *    CREATED : 03/15/90
 *   MODIFIED : 10/14/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Memory allocation routines for (games?)
 *
 * HISTORY
 *		10/06/94 (dcc) - updated to use <echidna/alloc.h>
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#define PMESS 0

#include <echidna/etypes.h>
#include <echidna/alloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#if __TURBOC__
#include <alloc.h>
#include <dos.h>
#endif
#if __AMIGAOS__
#include <exec/memory.h>
#include <echidna/listfunc.h>
#if AZTEC_C
#include <functions.h>
#endif
#endif
#include <echidna/eerrors.h>
#include <echidna/memory.h>

/******************************** CONSTANTS *******************************/

/********************************** TYPES *********************************/

#if __AMIGAOS__
typedef struct {
	NodeType	 Node;
	void		*Mem;
	ULONG		 Size;
} MemTracker;
#endif

#if __MSDOS__
typedef struct {
	void	*OrigMem;
} ParaMemTracker;
#endif

/****************************** TRACKING VARS *****************************/

#if __AMIGAOS__
#if AZTEC_C
typedef struct mem {
	struct mem *next;
	long size;
} mem;

struct mem *_Free;
#endif
#endif

/******************************** ROUTINES ********************************/

#if !__MSDOS32X__
void freesome (size_t size) { size = size; }


/*********************************************************************
 *
 * OpenMemory
 *
 * SYNOPSIS
 *		short OpenMemory (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short OpenMemory (void)
{
	return TRUE;

} /* OpenMemory */


/*********************************************************************
 *
 * CloseMemory
 *
 * SYNOPSIS
 *		void CloseMemory (void)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *		ExtraMemFunc
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void CloseMemory (void)
{
} /* CloseMemory */


#if			DEBUG

/*********************************************************************
 *
 * __CreateTempBuffer
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far32 *__CreateTempBuffer (size_t size, char *pFile, int iLine)
{
  	return __malloc(size, pFile, iLine, TRUE);

} /* __CreateTempBuffer */


/*********************************************************************
 *
 * __DestroyTempBuffer
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void __DestroyTempBuffer(void far *mem, size_t size, char *pFile, int iLine)
{
	size = size;
	__free(mem, pFile, iLine, TRUE);

} /* __DestroyTempBuffer */

#else	/* !DEBUG */

/*********************************************************************
 *
 * CreateTempBuffer
 *
 * SYNOPSIS
 *		void far32 *CreateTempBuffer (size_t size)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far32 *CreateTempBuffer (size_t size)
{
  	return MEM_malloc (size);

} /* CreateTempBuffer */


/*********************************************************************
 *
 * DestroyTempBuffer
 *
 * SYNOPSIS
 *		void DestroyTempBuffer (size_t size)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void DestroyTempBuffer (void *mem, size_t size)
{
	size = size;
	MEM_free (mem);

} /* DestroyTempBuffer */

#endif	/* !DEBUG */

#endif	/* !__MSDOS32X__ */


/************************************  ************************************/
/********************************** MSDOS *********************************/
/************************************  ************************************/


#if (__MSDOS__ && !__MSDOS32X__)

#if			DEBUG

/*********************************************************************
 *
 * __FreeMemOnPara
 *
 * PURPOSE
 *		Free memory allocated by AllocMemOnPara()
 *
 * INPUT
 *		seg = SEGMENT of memory as returned by AllocMemOnPara()
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void __FreeMemOnPara(void *mem, char *pFile, int iLine)
{
	UBYTE huge32	*d   = (UBYTE huge32 *)mem;

	d -= 16;
	__free (((ParaMemTracker *)d)->OrigMem, pFile, iLine, TRUE);

} /* __FreeMemOnPara */


/*********************************************************************
 *
 * __AllocMemOnPara
 *
 * PURPOSE
 *		Allocate memory for graphics on paragraph boundary because
 *		Turbo C is poo!
 *		(But not as much as some other compilers...)
 *
 * INPUT
 *		Size of Mem you need in bytes
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Segment of Mem or 0
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far32 *__AllocMemOnPara (ULONG size, char *pFile, int iLine)
{
	UBYTE 			*mem;
	UBYTE huge32	*d;

#if __TURBOC__
	mem = __farmalloc ((unsigned long) size+34, pFile, iLine, TRUE);
#else
#error Better find something to replace farmalloc( > 64K ) !
#endif

	if (!mem) {
		return NULL;
	}

	d = mem;
	while (FP_OFF(d)) {
		d++;
	}
	((ParaMemTracker *)d)->OrigMem = mem;
	d += 16;

	return (void far32 *)d;

} /* AllocMemOnPara */

#else	/* !DEBUG */

/*********************************************************************
 *
 * FreeMemOnPara
 *
 * SYNOPSIS
 *		void FreeMemOnPara (UWORD seg)
 *
 * PURPOSE
 *		Free memory allocated by AllocMemOnPara()
 *
 * INPUT
 *		seg = SEGMENT of memory as returned by AllocMemOnPara()
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void FreeMemOnPara (void *mem)
{
	UBYTE huge32	*d   = (UBYTE huge32 *)mem;

	d -= 16;
	MEM_free (((ParaMemTracker *)d)->OrigMem);

} /* FreeMemOnPara */


/*********************************************************************
 *
 * AllocMemOnPara
 *
 * PURPOSE
 *		Allocate memory for graphics on paragraph boundary because
 *		Turbo C is poo!
 *		(But not as much as some other compilers...)
 *
 * INPUT
 *		Size of Mem you need in bytes
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *		Segment of Mem or 0
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void far32 *AllocMemOnPara (ULONG size)
{
	UBYTE 			*mem;
	UBYTE huge32	*d;

#if __TURBOC__
	mem = MEM_farmalloc ((unsigned long) size+34);
#else
#error Better find something to replace farmalloc( > 64K ) !
#endif

	if (!mem) {
		return NULL;
	}

	d = mem;
	while (FP_OFF(d)) {
		d++;
	}
	((ParaMemTracker *)d)->OrigMem = mem;
	d += 16;

	return (void far32 *)d;

} /* AllocMemOnPara */

#endif	/* !DEBUG */


/*********************************************************************
 *
 * SetExtraMemFuncs
 *
 * SYNOPSIS
 *		void SetExtraMemFuncs (AllocMemFuncPtr memfunc)
 *
 * PURPOSE
 *		
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void SetExtraMemFuncs (AllocMemFuncPtr memfunc, FreeSomeFuncPtr fsfunc)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "SetExtraMemFunc";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
	memfunc = memfunc;
	fsfunc  = fsfunc;

} /* SetExtraMemFuncs */


/*********************************************************************
 *
 * UseExtraMemory
 *
 * SYNOPSIS
 *		void UseExtraMemory (void *mem, size_t size)
 *
 * PURPOSE
 *		Specify a 'simple' piece of memory for use as extra memory.
 *
 * INPUT
 *
 *
 * EFFECTS
 *
 *
 * RETURN VALUE
 *
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
void UseExtraMemory (void *mem, size_t size)
{

/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "UseExtraMemory";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/

	mem  = mem;
	size = size;

} /* UseExtraMemory */
#endif


/************************************  ************************************/
/********************************* AMIGAOS ********************************/
/************************************  ************************************/


#if __AMIGAOS__

static void cleanup(void)
{
	register mem *mp, *xp;

	for (mp=_Free;mp;mp=xp) {
		xp = mp->next;
		FreeMem(mp, mp->size+sizeof(struct mem));
	}
	_Free = 0;
}

static void *getmem (size_t size, ULONG type)
{
	register mem *ptr;
	extern void (*_cln)(void);

	if (size == 0) {
		return(0);
	}
	if ((ptr = AllocMem(size+sizeof(mem), type)) == 0) {
		return((void *)0);
	}
	_cln      = cleanup;
	ptr->next = _Free;
	ptr->size = size;
	_Free     = ptr;
	return((char *)ptr + sizeof(mem));
}

void *malloc(size_t size)
{
	return getmem (size, NULL);
}

void *calloc(size_t num, size_t size)
{
	return getmem (size * num, MEMF_CLEAR);
}

void *mallocChipMem (size_t size)
{
	return getmem (size, MEMF_CHIP);
}

void *callocChipMem (size_t num, size_t size)
{
	return getmem (size, MEMF_CHIP | MEMF_CLEAR);
}

#endif

