/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * ATIMER.H
 *
 * PROGRAMMER : R.G. Marquez 
 *    VERSION : 00.000
 *    CREATED : 06/23/91 
 *   MODIFIED : 07/27/94
 *       TABS : 04 09
 *
 *	     \|///-_
 *	     \oO///_
 *	-----w/-w-----
 *	 E C H i D N A
 *	--------------
 *
 * DESCRIPTION
 *		Stuff for AMiga timer library.
 *
 * HISTORY
 *	06/23/91 Sunday (RGM) - Created.
 *
*/
#ifndef ATIMER_H
#define ATIMER_H

#include <echidna/platform.h>
#include "switches.h"

#include <exec/types.h>

#include "atimlist.h"

/**************************** C O N S T A N T S ***************************/

#define ATIME_FREE		(0)
#define ATIME_VBLANK	(1)
#define ATIME_TIMER		(2)
#define ATIME_INTERRUPT	(3)

/******************************** T Y P E S *******************************/

typedef struct ATimer_EventNode {
	ATimer_MinNode			ate_Node;
	ULONG					ate_IdleTimer;
	ULONG					ate_WaitPeriod;
	UBYTE					ate_Type;		/* ATIME_VBLANK,   *
										 	 * ATIME_TIMER,    *
										 	 * ATIME_INTERRUPT */
	UBYTE					ate_Handle;
	BYTE					ate_Direction;	/* > 0 is inc, otherwise dec */
	UBYTE					ate_DataSize;	/* 1, 2, 4 */
	void			__far	*ate_DataPtr;
} ATimer_EventNode;		/* 24 bytes */

typedef struct {
	ATimer_MinList	atel_List;
	UWORD			atel_Lock;
} ATimer_EventList;

/****************************** E X T E R N S *****************************/


/****************************** G L O B A L S *****************************/


/******************************* L O C A L S ******************************/


/******************************* M A C R O S ******************************/

#define ate_IsLocked(ate_list_ptr)\
	((ate_list_ptr)->atel_Lock)

#define ate_IsNotLocked(ate_list_ptr)\
	(! ((ate_list_ptr)->atel_Lock))

#define ate_Lock(ate_list_ptr)\
	(ate_list_ptr)->atel_Lock = 1;

#define ate_Unlock(ate_list_ptr)\
	(ate_list_ptr)->atel_Lock = 0;

/****************** F U N C T I O N   P R O T O T Y P E S *****************/

#endif /* ATIMER_H */

