/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * BPMasks.H
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 02/05/90
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Constants for bit positions and masks used to by BrushPak to make
 *		BPI data files.
 *
 * HISTORY
 *
*/

#ifndef BPMASKS_H
#define BPMASKS_H

/**************************** C O N S T A N T S ***************************/
/* Brush Attribute flags Bit positions */
#define ATT_EGA_PAT_LO_BIT		24
#define ATT_CGA_PAT_LO_BIT		16
#define ATT_NULL_BRUSH_BIT		15
#define ATT_COPIED_BIT			14
#define ATT_CLONED_BIT			13
#define ATT_COPY_BIT			12
#define ATT_CLONE_BIT			11
#define ATT_REMAP_BIT			10
#define ATT_HFLIP_BIT			9
#define ATT_VFLIP_BIT			8
#define ATT_PCC_LO_BIT			4

#define INF_REMAP_INDEX_LO_BIT	24

/* Brush Attribute flags Masks */
#define ATT_NULL_FLAG		(1L << ATT_NULL_BRUSH_BIT)
#define ATT_COPIED_FLAG		(1L << ATT_COPIED_BIT)
#define ATT_CLONED_FLAG		(1L << ATT_CLONED_BIT)
#define ATT_COPY_FLAG		(1L << ATT_COPY_BIT)
#define ATT_CLONE_FLAG		(1L << ATT_CLONE_BIT)
#define ATT_REMAP_FLAG		(1L << ATT_REMAP_BIT)
#define ATT_HFLIP_FLAG		(1L << ATT_HFLIP_BIT)
#define ATT_VFLIP_FLAG		(1L << ATT_VFLIP_BIT)
#define ATT_PCC_MASK		(0x0FL << ATT_PCC_LO_BIT)
#define ATT_USER_CODE_MASK	(0x0FL)
#define ATT_EGA_PAT_MASK	(0xFFL << ATT_EGA_PAT_LO_BIT)
#define ATT_CGA_PAT_MASK	(0xFFL << ATT_CGA_PAT_LO_BIT)

#define ATT_PROCESSED_COPY_FLAG		ATT_COPIED_FLAG
#define ATT_PROCESSED_KLONE_FLAG	ATT_CLONED_FLAG

#define INF_REMAP_INDEX_MASK	(0xFFL << INF_REMAP_INDEX_LO_BIT)
#define INF_BRUSH_SIZE_MASK		(0x00FFFFFFL)

/******************************* M A C R O S ******************************/

#define IsNull(a)			((a) & ATT_NULL_FLAG)

#define	IsKlone(a)			((a) & ATT_CLONE_FLAG)
#define IsKloned(a)			((a) & ATT_CLONED_FLAG)
#define IsCopy(a)			((a) & ATT_COPY_FLAG)
#define IsCopied(a)			((a) & ATT_COPIED_FLAG)

#define KloneSource(info)	((info) & INF_BRUSH_SIZE_MASK)
#define CopySource(info)	((info) & INF_BRUSH_SIZE_MASK)
#define ShapeSize(info)		((info) & INF_BRUSH_SIZE_MASK)
#define RemapIndex(info)	(((info) & INF_REMAP_INDEX_MASK) >>	\
								INF_REMAP_INDEX_LO_BIT)

#define IsRemapped(a)		((a) & ATT_REMAP_FLAG)
#define IsHflipped(a)		((a) & ATT_HFLIP_FLAG)
#define IsVflipped(a)		((a) & ATT_VFLIP_FLAG)

#define UserCode(a)			((a) & ATT_USER_CODE_MASK)
#define PCC(a)				(((a) & ATT_PCC_MASK) >> ATT_PCC_LO_BIT)
#define EGAPatmapIndex(a)	(((a) & ATT_EGA_PAT_MASK) >> ATT_EGA_PAT_LO_BIT)
#define CGAPatmapIndex(a)	(((a) & ATT_CGA_PAT_MASK) >> ATT_CGA_PAT_LO_BIT)

#define IsProcessedCopy(a)	(IsCopied((a)))
#define IsProcessedKlone(a)	(IsKloned((a)))

#endif /* BPMASKS_H */

