/* 
 * The contents of this file are subject to the Mozilla Public 
 * License Version 1.1 (the "License"); you may not use this file 
 * except in compliance with the License. You may obtain a copy of 
 * the License at http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS 
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing 
 * rights and limitations under the License. 
 * 
 * The Original Code is tUME: the Universal Map Editor, released 
 * June 18, 2000. 
 * 
 * The Initial Developer of the Original Code is Echidna. 
 * Portions created by Echidna are Copyright (c) 1989-2000 Echidna. 
 * All Rights Reserved. 
 * 
 * Contributor(s): 
 */

/*
 * LOADBACK.C
 *
 * PROGRAMMER : Juan M. Alvarado
 *    VERSION : 00.000
 *    CREATED : 03/21/91
 *   MODIFIED : 07/27/94
 *       TABS : 05 09
 *
 *	     \|///-_
 *	     \oo///_
 *	-----w/-w------
 *	 E C H I D N A
 *	---------------
 *
 * DESCRIPTION
 *		Routine for loading a background screen into BackBuffer and
 *		not keeping any other data around.
 *
 * HISTORY
 *	07/08/91 Monday (RGM) - Changed LoadBackground to LoadScreen.  Set
 *		the work page with SetWorkPage (<BUFFER>) before calling.
 *
*/

#include <echidna/platform.h>
#include "switches.h"

#include <echidna/etypes.h>
#include <echidna/stdload.h>
#include <echidna/memory.h>
#include <echidna/eerrors.h>
#include <echidna/graphics.h>

/**************************** C O N S T A N T S ***************************/


/******************************** T Y P E S *******************************/


/****************************** G L O B A L S *****************************/


/******************************* M A C R O S ******************************/


/***************************** R O U T I N E S ****************************/


/*********************************************************************
 *
 * LoadScreen
 *
 * SYNOPSIS
 *		short	LoadScreen (
 *			char	*filename,
 *			short	 selector
 *		)
 *
 * PURPOSE
 *		To load a screen onto the workbuffer.  It does this by
 *		loading a brushpack file (file number 'selector' in the FPF
 *		file 'filename') and drawing all the brushes in the BPI to
 *		the workbuffer at position (0,0).  
 *
 *		The brushes in the bpi must all be smaller than 40,000
 *		bytes, because on the Amiga that is the biggest BPI brush
 *		that can be loaded into the translation buffer.  So what
 *		you do is cut out your screen in 4 (or more) pieces,
 *		starting each cut at position (0,0) of your dpaint screen. 
 *		Before this routine draws any brushes, it clears the
 *		workscreen to the transparent color of the first brush in
 *		the BPI.  It also uses the global pattern from the BPI so
 *		that it will work in EGA, CGA etc.  
 *		
 *
 * INPUT
 *
 *
 * AFFECTS
 *		Overwrites Global Patterns.  So if you set the global pattern 
 *		before calling this routine, you will have to reset it after
 *		calling the routine (unless you know the pattern in the screen 
 *		bpi to be the same as the one you set it to before).
 *
 * RETURN VALUE
 *		Non-zero on success.
 *		FALSE on Failure.
 *
 * HISTORY
 *
 *
 * SEE ALSO
 *
*/
short	LoadScreen (
	char	*filename,
	short	 selector
)
{
/*-------------------------------------------------------------------------*/
#if FUNC_NAMES
 CurrentFuncName = "LoadScreen";
#endif
#if REQUIRE

#endif
/*-------------------------------------------------------------------------*/
{
	BPIType *tempbpi;
	short result;

	if ((result = OpenMemory ()) != 0) {
		if ((tempbpi = LoadBPI (filename, selector)) != NULL)
		{
			short i;

			SetGlobalPattern (&tempbpi->GlobalPattern);
			SetPaperColor (tempbpi->Shapes[0]->TColor);
			ClearWorkPage ();

			for (i = 0; i < tempbpi->NumShapes; i++) {
				DrawGShape (tempbpi->Shapes[i], 0, 0);
			}

		} else {
			GEcatf ("LoadScreen():");
			result = FALSE;
		}
		CloseMemory ();
	}

	return result;
}
} /* LoadScreen */

